#ifndef __HELPER_H__
#define __HELPER_H__

//#include "cblas.h"
extern "C" {
#include "cblas.h"
#include "atlas_cblascalias.h"
}

inline void xerbla(char* func_name, int info){
  printf(" On entry to %s parameter number %d had an illegal value\n", func_name, info);
}


inline void idamax(int n, double* dx, int incx, int* index, double* pivot){
  
  //  *index = cblas_iamax(n, dx, incx);
  //  *pivot = dx[*index*incx];
  //};
  
   double dmax, xmag;
   int i, ii, xindex;
   xindex = 0;
   if ( n <= 0 )
     return;
   if (incx <= 0 )
     return;

   if ( incx != 1 ) {
     dmax = fabs( dx[0] );
     ii = 1;
     for ( i = incx ; i < n * incx ; i = i + incx ) {
       xmag = fabs( dx[i] );
       if ( xmag > dmax ) {
	 xindex = ii;
	 dmax = xmag;
       }
       ii++;
     }
     *pivot = dmax;  
     *index = xindex;
     
     return;
   }
   

   dmax = fabs( dx[0] );
   for ( i = 1 ; i < n ; i++ ) {
     xmag = fabs( dx[i] );
     if ( xmag > dmax ) {
       xindex = i;
       dmax = xmag;
     }
   }
   *pivot = dmax;
   *index = xindex;
   
   return;
};


inline void dswap(int n, double* dx, int incx, double* dy, int incy){
  /**
   * Interchange two vectors
   * Use unrolled loops for increments equal to one.
   */
  double  dtemp;
  int i, ix, iy, m, mp1;
  
  if (n <=0) return;
  if (incx == 1 && incy == 1){
    m = n %3;
    if(m !=0){
      for(i = 0; i<m; i++){
	dtemp = dx[i];
	dx[i] = dy[i];
	dy[i] = dtemp;
      }
      if (n <= 3) return;
    }
    mp1=m; // adjust for c indexing
    for(i=mp1; i<n; i+=3){
      dtemp = dx[i];
      dx[i] = dy[i];
      dy[i] = dtemp;
      
      dtemp = dx[i+1];
      dx[i+1] = dy[i+1];
      dy[i+1] = dtemp;
      
      dtemp = dx[i+2];
      dx[i+2] = dy[i+2];
      dy[i+2] = dtemp;
    }
  }else{
    ix = 0;
    iy = 0;
    if(incx < 0){
      ix = (-n+1)*incx+1;
    }
    if(incy <0){
      iy = (-n+1)*incy+1;
    }
    for(i=0;i<n;i++){
      dtemp = dx[ix];
      dx[ix]=dy[iy];
      dy[iy]=dtemp;
      ix = ix+incx;
      iy = iy+incy;
    }
  }
  return;
}

inline void dscal(int n, double da, double *dx, int incx){
  
  int m, i;
  
  if ( n <= 0 )
    return;
  
  /* Code for increments not equal to 1.  */
  
  if ( incx != 1 ) {
    for ( i = 0 ; i < n * incx ; i = i + incx ) 
      dx[i] = da * dx[i];
    return;
  }
  
  /* Code for increments equal to 1.  */

  /* Clean-up loop so remaining vector length is a multiple of 5.  */
  
  m = n % 5;
  if ( m != 0 ) {
    for ( i = 0 ; i < m ; i++ )
      dx[i] = da * dx[i];
    if ( n < 5 )
      return;
  }
  for ( i = m  ; i <n ; i = i + 5 ) {
    dx[i] = da * dx[i];
    dx[i+1] = da * dx[i+1];
    dx[i+2] = da * dx[i+2];
    dx[i+3] = da * dx[i+3];
    dx[i+4] = da * dx[i+4];
  }
  return;
}


inline void dger(int m, int n, double alpha, double *x, int incx, double * y, int incy, double * a, int lda){
  
  int info = 0;
  int i, j, ix, jy, kx;
  if( m < 0) info = 1;
  else if(n <0) info =2;
  else if(incx == 0) info = 5;
  else if(incy == 0) info = 7;
  else if(lda < max(1,m)) info = 9;
  if (info != 0){
    xerbla("dger", info);
    return;
  }
  if((m==0)||(n==0) ||(alpha == 0.0)){
    return;
  }

  /*     Start the operations. In this version the elements of A are
   *     accessed sequentially with one pass through A.
   */
  if(incy > 0){
    jy = 0 ; // adjust to c indexing  orig: = 1;
  }else{
    jy = -(n-1)*incy; // adjust to c indexing org: 1-(n-1)*incy
  }
  if(incx == 1){
    for(j=0; j<n; j++){
      if(y[jy] != 0){
	double temp = alpha* y[jy];
	for (i = 0; i<m; i++){
	  a[i*lda+j] += x[i]*temp;
	}
      }
      jy = jy+incy;
    }
  }else{
    if(incx > 0){
      kx = 0; // adjust to c indexing
    }else{
      kx = -(m-1)*incx; // adjust to c indexing
    }
    for(j=0; j<n; j++){
      if(y[jy] != 0){
	double  temp = alpha * y[jy];
	ix = kx;
	for (i=0; i<m; i++){
	  a[i*lda+j] += x[ix]*temp;
	  ix = ix + incx;
	}
      }
      jy = jy + incy;
    }
  }
}


inline void dllnu4(int m, int n,  HTA<double, 2,0> a, int lda, HTA<double, 2, 0> b, int ldb){
  double* pa= a.raw(); 
  double* pb =b.raw();
  cblas_dtrsm(CblasRowMajor, CblasLeft, CblasLower, CblasNoTrans, CblasUnit, m, n, 1.0, pa, lda, pb, ldb);
  return;
}


inline void dab4(int m, int n, int k, HTA<double,2, 0> a, int lda, HTA<double, 2, 0> b, int ldb, HTA<double, 2, 0> c, int ldc){
  double *pa = a.raw();
  double *pb = b.raw();
  double *pc = c.raw();
  cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans, m, n, k, -1.0, pa, lda, pb, ldb, 1.0, pc, ldc);
  return;
}

#endif // __HELPER_H__
