
#include <iostream>
#include "Shape.h"
#include <math.h>
#include "helper.h"
#include "utility.h"

void hta_dgetf2(HTA<double,2,1> a, int lda, HTA<int, 1,0> ipiv, int& info){
  info = 0;
  Shape<2> tile_shape = a[Tuple<2>(0,0)].shape();
  Shape<2> shape= a.shape();
  int tm_sz = tile_shape[0].size();          // tile size in m dimension
  int tn_sz = tile_shape[1].size(); // tile size in n dimension

  int tm = shape[0].size(); // number of tiles in m dimension
  int tn = shape[1].size(); // number of tiles in n dimension
  assert(tn == 1);           // Assume tn is 1 otherwise the code would be much more ugly.
  int m = tm*tm_sz; 
  int n = tn*tn_sz;          
  
  if(m<0){
    info = -1;
  }else if (n <0){
    info = -2;
  }
  if(info < 0){
    cout<<"error in function hta_dgetf2 with info "<<info<<endl;;
    return;
  }
  if(m ==0 || n == 0){
    return;
  }

  int min_mn = min(m,n);
  for(int j=0; j< min_mn; j++){
  /*
   * Find pivot and test for singularity.
   */
    timer_start(10);
    double pivot = 0.0;
    /* deal with the first block 
     * Modify the idamax to return also the pivot value;
     */
    int bj =  j/tm_sz;
    int delta = (bj+1)*tm_sz -j-1;
    int beta = (j/tn_sz+1)*tn_sz -j-1;
    int jp = 0;
    idamax(delta+1, &a(j, j), lda, &jp, &pivot);
    jp += j;


    for(int BJ = bj+1; BJ <tm; BJ++){
      /*
       * deal with the rest blocks
       */
      double tmp_pivot;
      int tmp_jp;
      idamax(tm_sz, &a(BJ*tm_sz,j), lda, &tmp_jp, &tmp_pivot);

      if(tmp_pivot > pivot){
	jp = tmp_jp+BJ*tm_sz;
	pivot = tmp_pivot;
      }
    }

    ipiv(Tuple<1>(j)) = jp;

    timer_stop(10);
    timer_start(11);
    if(pivot!=0){
      /* Apply the interchange to columns 1:N.*/

      if(jp != j){
	dswap(n, &a(j,0),1, &a(jp,0), 1);
      }

      if(j <m-1){
	/*  Compute elements J+1:M of J-th column. */
	/* Deal with the first block */
	
	dscal(delta, 1.0/a(j,j), &a(j+1,j),lda);
	for(int BJ = bj+1; BJ<tm; BJ++){
	  dscal(tm_sz, 1.0/a(j,j), &a(BJ*tm_sz,j), lda);
	}
      }
    }else if( info == 0){
      info = j;
    }
    timer_stop(11);
    timer_start(12);
    /*Update trailing submatrix.*/
    /* Deal with the first block */
    if(j < min(m,n)-1){
      
      cblas_dger(CblasRowMajor, delta, beta, -1.0, &a(j+1,j), lda, &a(j,j+1), 1, &a(j+1,j+1), lda);
      for(int BJ=bj+1;BJ<tm; BJ++){
	cblas_dger(CblasRowMajor, delta, tn_sz,-1.0, &a(BJ*tm_sz,j), lda, &a(j, j+1), 1, &a(BJ*tm_sz,j+1),lda);
      }
    }
    timer_stop(12);
  }
}


void hta_dlaswp(HTA<double,2,1> a, int lda, int bj, HTA<int,1,0> hipiv, int incx ){
  int * ipiv = hipiv.raw();
  int i, j, k, ix0, i1, i2, inc;
  Shape<2> s1 = a[Tuple<2>(0,0)].shape();
  int tn_sz = s1[1].size();
  int tm_sz = s1[0].size();
  int tn = a.shape()[1].size();
  int tm = a.shape()[0].size();

  if(incx > 0){
    ix0 = 0;
    inc = 1;
  }else if (incx < 0){
    printf("error\n");
    return;
  }else{
    return;
  }
  int ix  = 0;
  double * p, * q;
  for ( i= 0; i< tm_sz; i++){
    int ip = ipiv[ix];
    int ip_t = ip/tm_sz;
    int ip_r = ip%tm_sz;
     if ( (ip_t != bj) || (ip_r != i)){
      for(j = 0; j < tn; j++){
  	HTA<double, 2, 0> a_j = a[Tuple<2>(0, j)];
	HTA<double, 2, 0> b_j = a[Tuple<2>(ip_t, j)];
	p = & a_j(0, 0);
	q = & b_j(0, 0);
	for (k = 0; k<tn_sz; k++){
	  double temp = p[ix*lda+k];
	  p[ix*lda+k]=q[ip_r*lda+k];
	  q[ip_r*lda+k]=temp;
	}
      }
    }
    ix += inc;
  }
}



    
/*int main(int argc, char** argv){
  if (argc <2){
    cout<< "Usage: "<<argv[0]<<" size, tile_num\n";
    return -1;
  }
  int n = atoi(argv[1]);
  int first_tiling = atoi(argv[2]);
  int tile_sz = n/first_tiling;
  // Create the input matrix to be factorized.
  double* dData = new double[n*tile_sz];
  for (int i = 0; i < n*tile_sz; i++){
      dData[i] = i-3;
  }
  
  Tuple<2> tiling []= {Tuple<2>(first_tiling, 1), Tuple<2>(tile_sz, tile_sz)};
  Tuple<1> tiling1 [] = {Tuple<1>(first_tiling), Tuple<1>(tile_sz)};
  Allocator<double> data =Allocator<double>::makeBlock(n*tile_sz, dData); 
  HTA<double, 2, 1> htaa = Factory<double, 2, 1>::makeHTA(tiling, data, TILE);
  int info;
  
  Allocator<double> data1 =Allocator<double>::makeBlock(n);
  HTA<double,1,1> ipiv = Factory<double, 1, 1>::makeHTA(tiling1, data1, TILE);

  hta_dgetf2(htaa, ipiv, info);
  cout<<htaa<<endl;
  delete [] dData;
  return 0;

  } */     
