/*********************************************************************/
/*                     I.B.M. CONFIDENTIAL                           */
/*      (c) COPYRIGHT IBM CORP. 2005 ALL RIGHTS RESERVED.            */
/*                                                                   */
/*                LICENSED MATERIALS-PROPERTY OF IBM.                */
/*                                                                   */
/*********************************************************************/

/*
 * Version: $Id: ft_x.cpp,v 1.16 2007/01/14 07:32:40 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

/**
 * Description:
 * Solve a PDE using 3D FFT
 */

#include "npb-C.h"
#include "FT.h"
#include <iostream>

#define T_TOTAL 0
#define T_FT 1
#define T_DPERMUTE 2
#define T_TWIDDLE 3
#define T_CSUM 4
#define T_REDUCE 5
#define T_INITIAL 6
#define T_ASSIGN 7
#define T_INDEX 8

using namespace std;

void fft3(int dir, HTA_2H u1,int d);
void fft3_square(int dir, HTA_2H u1);
void fft3_rect(int dir, HTA_2H u1);

void compute_initial_conditions (HTA_2H u0, int d1, int d2, int d3);
void compute_indexmap (HTA_2H_DOUBLE twiddle, int d1, int d2, int d3, T t);
void checksum (HTA_2H u0, int kt, int d1, int d2, int d3, HTA_1H sum);

void compute_initial_conditions_inner (HTA_1H u0, int d1, int d2, int d3, T pos);
void compute_indexmap_inner (HTA_1H_DOUBLE twiddle, int d1, int d2, int d3, T pos, T t);
void checksum_inner (HTA_1H u0, int kt, int d1, int d2, int d3, LEAF sum, T pos);

static double ipow46(double a, int exponent) ;

static void verify (int d1, int d2, int d3, int nt,
		    bool *verified, char *Class);

void print_timers ();

complex<double> sums[NITER_DEFAULT];
int n1;
int n2;
int n3; 
int run (int argc, char** argv) {

  double mflops;
  char Class;
  bool verified = false;

  sleep(10);

  int niter = NITER_DEFAULT;
  int np1 = 1;
  int np2 = max (Traits::Default::nPlaces(), 1);

  int d = NX / NZ;
  
  n1 = 0;
  n2 = 0;
  n3 = d * np2 - 1; 
  assert (NX >= NZ);

  /* 2-level tiling */
  Tuple<3>::Seq tiling = (T(NZ/np2, NY, NX/(d*np2)), T(1, 1, d*np2), T(np2, 1, 1));
   
  /* part of fft */
  HTA_2H xnt = HTA_2H::alloc(tiling, NULL, ROW);   
  //  HTA_2H xnt2 = HTA_2H::alloc(tiling, NULL, ROW);   
  HTA_2H y = HTA_2H::alloc(tiling, NULL, ROW);  /* a temporary */  

  /* constant - is precomputed */
  HTA_2H_DOUBLE twiddle = HTA_2H_DOUBLE::alloc(tiling, NULL, ROW); 

  /* checksum */
  Sequence<Tuple<3> >tiling2 (Tuple<3> (1, 1, 1), Tuple<3> (d*np2, 1, 1));  
  HTA_1H sum = HTA_1H::alloc(tiling2, NULL, ROW);
  sum = complex<double>(0.0, 0.0);

  timer_start(T_TOTAL);
  
  timer_start (T_INITIAL);
  /* initialize xnt */
  //compute_initial_conditions(xnt, NZ, NY, NX);
  xnt.hmap (Initial(),1);
  timer_stop (T_INITIAL);
  
  timer_start (T_INDEX);
  if (d > 1) {
    compute_indexmap (twiddle()((R(0, n1), R(0, n2), R(0, n3/2))), NZ, NY, NX, T(0, 0, 0));
    compute_indexmap (twiddle()((R(0, n1), R(0, n2), R(n3/2+1, n3))), NZ, NY, NX, T(0, 0, 1));    
  } else {
    twiddle.hmap (IndexMap(Tuple<3>(0,0,0)),1);
  }

  timer_stop (T_INDEX);
  
  fft3(1, xnt, d);

  timer_start (T_ASSIGN);
  y = xnt;
  timer_stop (T_ASSIGN);
  for (int kt = 0; kt < niter; kt++) {

    timer_start (T_TWIDDLE);
    y *= twiddle;
    //y = y * twiddle;
    timer_stop (T_TWIDDLE);

    timer_start (T_ASSIGN);
    xnt = y;
    timer_stop (T_ASSIGN);

    fft3(-1, xnt, d);
    
    sum = complex<double>(0.0, 0.0);

    timer_start (T_CSUM);
    //    checksum(xnt, kt, NZ, NY ,NX, sum);        
    xnt.hmap (Checksum(), sum, 1);
    timer_stop (T_CSUM);

    timer_start (T_REDUCE);
    sums[kt] = sum.reduce(plus<complex<double> >());
    timer_stop (T_REDUCE);

    cout << "iter " << kt << " checksum: " << sums[kt] << endl;
  }

  timer_stop(T_TOTAL);
  verify(NX, NY, NZ, niter, &verified, &Class);

  IO::cout << endl << "Benchmark completed. " << endl;
  Traits::Default::barrier();

  double max_time = timer_read (T_TOTAL);
#if (HTA_TRAIT == HTA_TRAIT_MPI)
   double thread_time = timer_read(T_TOTAL);
   MPI_Allreduce (&thread_time,
	       &max_time,
	       1,
	       MPI_DOUBLE,
	       MPI_MAX,
	       MPI_COMM_WORLD);	       
#endif

   if (Traits::Default::myPlace() == 0) {
   
     print_timers();
     double total_time = max_time;
     
     if( total_time != 0.0) {
       mflops = 1.0e-6*(double)(NTOTAL) *
	 (14.8157+7.19641*log((double)(NTOTAL))
	  +  (5.23518+7.21113*log((double)(NTOTAL)))*niter)
	 /total_time;
     } else {
       mflops = 0.0;
     }
     
     c_print_results("FT_X", Class, NX, NY, NZ, niter, Traits::Default::nPlaces(),
		     total_time, mflops, "          floating point", verified, 
		     NPBVERSION, COMPILETIME,
		     CS1, CS2, CS3, CS4, CS5, CS6, CS7);
   }

}  

/** 
 * fft routine
 * @arg1 - direction of FT (forward = 1, backward = -1) 
 * @arg2 - input/output HTA
 * @arg3 - ratio of the first and third dimension
*/

void fft3(int dir, HTA_2H u1, int d)  
{
  if (d == 1)
    fft3_square(dir, u1);
   else
    fft3_rect(dir, u1);
}
void fft3_square(int dir, HTA_2H u1)  {

  if (dir ==  1) { 

    timer_start (T_FT);
    //u1.ft(2, 1);
    u1.hmap (FT(2,1), 1);
    //u1.ft(1, 1);
    u1.hmap (FT(1,1), 1);
    timer_stop (T_FT);

    timer_start (T_DPERMUTE);
    u1 = u1.dpermute(Tuple<2>(0, 2)); 
    timer_stop (T_DPERMUTE);

    timer_start (T_FT);
    //u1.ft(2, 1);
    u1.hmap (FT(2,1), 1);
    timer_stop (T_FT);

  } else {

    timer_start (T_FT);
    //u1.ft(2, -1);
    u1.hmap (FT(2, -1), 1);
    timer_stop (T_FT);

    timer_start (T_DPERMUTE);
    u1 = u1.dpermute(Tuple<2>(0, 2));
    timer_stop (T_DPERMUTE);
    
    timer_start (T_FT);
    //u1.ft(2, -1);
    u1.hmap (FT(2, -1), 1);
    //u1.ft(1, -1);
    u1.hmap (FT(1, -1), 1);
    timer_stop (T_FT);
  }  
}

/** 
 * fft routine
 * @arg1 - direction of FT (forward = 1, backward = -1) 
 * @arg2 - input/output HTA
 */
void fft3_rect(int dir, HTA_2H u1)  {
  
  if (dir ==  1) { 
    timer_start (T_FT);
    //u1.ft(2, 1);
    //u1.ft(1, 1);
    u1.hmap (FT(2, 1), 1);
    u1.hmap (FT(1, 1), 1);
    timer_stop (T_FT);
    
    timer_start (T_DPERMUTE);
    u1()((R(0, n1), R(0, n2), R(0, n3/2))) = ((HTA_2H) u1()((R(0, n1), R(0, n2), R(0, n3/2)))).dpermute(Tuple<2>(0, 2));
    u1()((R(0, n1), R(0, n2), R(n3/2+1, n3))) = ((HTA_2H) u1()((R(0, n1), R(0, n2), R(n3/2+1, n3)))).dpermute(Tuple<2>(0, 2));
    timer_stop (T_DPERMUTE);
    
    timer_start (T_FT);	
    //((HTA_2H) u1()((R(0, n1), R(0, n2), R(0, n3/2)))).ft (2, 1);
    ((HTA_2H) u1()((R(0, n1), R(0, n2), R(0, n3/2)))).hmap (FT (2, 1), 1);
    //((HTA_2H) u1()((R(0, n1), R(0, n2), R(n3/2+1, n3)))).ft(2, 1); 
    ((HTA_2H) u1()((R(0, n1), R(0, n2), R(n3/2+1, n3)))).hmap (FT(2, 1), 1); 
    timer_stop (T_FT);	
  } else {
    timer_start (T_FT);	
    //((HTA_2H) u1()((R(0, n1), R(0, n2), R(0, n3/2)))).ft (2, -1);
    //((HTA_2H) u1()((R(0, n1), R(0, n2), R(n3/2+1, n3)))).ft(2, -1); 
    ((HTA_2H) u1()((R(0, n1), R(0, n2), R(0, n3/2)))).hmap (FT(2, -1), 1);
    ((HTA_2H) u1()((R(0, n1), R(0, n2), R(n3/2+1, n3)))).hmap (FT(2, -1), 1); 
    timer_stop (T_FT);
    
    timer_start (T_DPERMUTE);  
    u1()((R(0, n1), R(0, n2), R(0, n3/2))) = ((HTA_2H) u1()((R(0, n1), R(0, n2), R(0, n3/2)))).dpermute(Tuple<2>(0, 2));
    u1()((R(0, n1), R(0, n2), R(n3/2+1, n3))) = ((HTA_2H) u1()((R(0, n1), R(0, n2), R(n3/2+1, n3)))).dpermute(Tuple<2>(0, 2));
    timer_stop (T_DPERMUTE);
    
    timer_start (T_FT);	
    //u1.ft(2, -1);
    //u1.ft(1, -1);
    u1.hmap (FT(2, -1), 1);
    u1.hmap (FT(1, -1), 1);
    timer_stop (T_FT);	    
  }  
}


/* GB: WIll be removed soon, after the bug in Index for CLASS A is fixed */
void compute_indexmap (HTA_2H_DOUBLE twiddle, int d1, int d2, int d3, T t)
{
  Shape<3>::iterator i = twiddle.shape().begin();
  Shape<3>::iterator end = twiddle.shape().end();
  for (; i != end; ++i) 
    if(twiddle(*i).isLocal()) 
       compute_indexmap_inner (twiddle(*i), d1, d2, d3, *i, t);
}

void compute_indexmap_inner (HTA_1H_DOUBLE twiddle, int d1, int d2, int d3, T pos, T t)
{
  int s1 = twiddle.memMap().leafSize()[0];
  int s2 = twiddle.memMap().leafSize()[1];
  int s3 = twiddle.memMap().leafSize()[2];
  
  int startZ = pos[2] * s3;
  int startY = pos[1] * s2;
  int startX = pos[0] * s1 + t[2] * s3;
  
  double pi = 3.141592653589793238;
  double ap = -4.0 * 1.0e-6 * pi * pi;
  
  for (int i = 0; i < s1; i++) {
    for (int j = 0; j < s2; j++) {
      for (int k = 0; k < s3; k++){
	int ii = (i + startX + d3/2) % d3 - d3/2;    
	int jj = (j + startY + d2/2) % d2 - d2/2;
	int kk = (k + startZ + d1/2) % d1 - d1/2;
	twiddle [T(i, j, k)] = exp(ap*double(kk*kk + jj*jj + ii*ii));
      }         
    }
  }
}


#if (HTA_TRAIT == HTA_TRAIT_UPC)
// main is defined in the upc runtime ... (cludge code in main.o)
int upc_main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  int ret = run(argc, argv);
  Traits::Default::finalize();
  return ret;
}
#else
int main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  int ret = run(argc, argv);
  Timers::printAll();
  Traits::Default::finalize();
  return ret;
}

/*--------------------------------------------------------------------
c-------------------------------------------------------------------*/

static void verify (int d1, int d2, int d3, int nt,
		    bool *verified, char *Class) {

/*--------------------------------------------------------------------
c-------------------------------------------------------------------*/

    int ierr, size, i;
    double err, epsilon;

/*--------------------------------------------------------------------
c   Sample size reference checksums
c-------------------------------------------------------------------*/
    
/*--------------------------------------------------------------------
c   Class S size reference checksums
c-------------------------------------------------------------------*/
    double vdata_real_s[6] = { 	5.546087004964e+02,
				5.546385409189e+02,
				 5.546148406171e+02,
				 5.545423607415e+02,
				 5.544255039624e+02,
				 5.542683411902e+02 };
    double vdata_imag_s[6] = {
				 4.845363331978e+02,
				 4.865304269511e+02,
				 4.883910722336e+02,
				 4.901273169046e+02,
				 4.917475857993e+02,
				 4.932597244941e+02 };
/*--------------------------------------------------------------------
c   Class W size reference checksums
c-------------------------------------------------------------------*/
    double vdata_real_w[6] = { 
				 5.673612178944e+02,
				 5.631436885271e+02,
				 5.594024089970e+02,
				 5.560698047020e+02,
				 5.530898991250e+02,
				 5.504159734538e+02 };
    double vdata_imag_w[6] = {
				 5.293246849175e+02,
				 5.282149986629e+02,
				 5.270996558037e+02, 
				 5.260027904925e+02, 
				 5.249400845633e+02,
				 5.239212247086e+02 };
/*--------------------------------------------------------------------
c   Class A size reference checksums
c-------------------------------------------------------------------*/
    double vdata_real_a[6] = { 
				 5.046735008193e+02,
				 5.059412319734e+02,
				 5.069376896287e+02,
				 5.077892868474e+02,
				 5.085233095391e+02,
				 5.091487099959e+02 };
    double vdata_imag_a[6] = { 
				 5.114047905510e+02,
				 5.098809666433e+02,
				 5.098144042213e+02,
				 5.101336130759e+02,
				 5.104914655194e+02,
				 5.107917842803e+02 };
/*--------------------------------------------------------------------
c   Class B size reference checksums
c-------------------------------------------------------------------*/
    double vdata_real_b[20] = { 
				  5.177643571579e+02,
				  5.154521291263e+02,
				  5.146409228649e+02,
				  5.142378756213e+02,
				  5.139626667737e+02,
				  5.137423460082e+02,
				  5.135547056878e+02,
				  5.133910925466e+02,
				  5.132470705390e+02,
				  5.131197729984e+02,
				  5.130070319283e+02,
				  5.129070537032e+02,
				  5.128182883502e+02,
				  5.127393733383e+02,
				  5.126691062020e+02,
				  5.126064276004e+02,
				  5.125504076570e+02,
				  5.125002331720e+02,
				  5.124551951846e+02,
				  5.124146770029e+02 };
    double vdata_imag_b[20] = { 
				  5.077803458597e+02,
				  5.088249431599e+02,                  
				  5.096208912659e+02,
				  5.101023387619e+02,                  
				  5.103976610617e+02,                  
				  5.105948019802e+02,                  
				  5.107404165783e+02,                  
				  5.108576573661e+02,                  
				  5.109577278523e+02,
				  5.110460304483e+02,                  
				  5.111252433800e+02,                  
				  5.111968077718e+02,                  
				  5.112616233064e+02,                  
				  5.113203605551e+02,                  
				  5.113735928093e+02,                  
				  5.114218460548e+02,
				  5.114656139760e+02,
				  5.115053595966e+02,
				  5.115415130407e+02,
				  5.115744692211e+02 };
/*--------------------------------------------------------------------
c   Class C size reference checksums
c-------------------------------------------------------------------*/
    double vdata_real_c[20] = { 
				  5.195078707457e+02,
				  5.155422171134e+02,
				  5.144678022222e+02,
				  5.140150594328e+02,
				  5.137550426810e+02,
				  5.135811056728e+02,
				  5.134569343165e+02,
				  5.133651975661e+02,
				  5.132955192805e+02,
				  5.132410471738e+02,
				  5.131971141679e+02,
				  5.131605205716e+02,
				  5.131290734194e+02,
				  5.131012720314e+02,
				  5.130760908195e+02,
				  5.130528295923e+02,
				  5.130310107773e+02,
				  5.130103090133e+02,
				  5.129905029333e+02,
				  5.129714421109e+02 };
    double vdata_imag_c[20] = {
				  5.149019699238e+02,
				  5.127578201997e+02,
				  5.122251847514e+02,
				  5.121090289018e+02,
				  5.121143685824e+02,
				  5.121496764568e+02,
				  5.121870921893e+02,
				  5.122193250322e+02,
				  5.122454735794e+02,
				  5.122663649603e+02,
				  5.122830879827e+02,
				  5.122965869718e+02,
				  5.123075927445e+02,
				  5.123166486553e+02,
				  5.123241541685e+02,
				  5.123304037599e+02,
				  5.123356167976e+02,
				  5.123399592211e+02,
				  5.123435588985e+02,
				  5.123465164008e+02 };

    epsilon = 1.0e-12;
    *verified = true;
    *Class = 'U';

    if (d1 == 64 &&
	d2 == 64 &&
	d3 == 64 &&
	nt == 6) {
	*Class = 'S';
	for (i = 0; i < nt; i++) {
            err = (real(sums[i]) - vdata_real_s[i]) / vdata_real_s[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
            err = (imag(sums[i]) - vdata_imag_s[i]) / vdata_imag_s[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
	}
    } else if (d1 == 128 &&
	       d2 == 128 &&
	       d3 == 32 &&
	       nt == 6) {
	*Class = 'W';
	for (i = 0; i < nt; i++) {
            err = (real(sums[i]) - vdata_real_w[i]) / vdata_real_w[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
            err = (imag(sums[i]) - vdata_imag_w[i]) / vdata_imag_w[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
	}
    } else if (d1 == 256 &&
	       d2 == 256 &&
	       d3 == 128 &&
	       nt == 6) {
	*Class = 'A';
	for (i = 0; i < nt; i++) {
            err = (real(sums[i]) - vdata_real_a[i]) / vdata_real_a[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
            err = (imag(sums[i]) - vdata_imag_a[i]) / vdata_imag_a[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
	}
    } else if (d1 == 512 &&
	       d2 == 256 &&
	       d3 == 256 &&
	       nt == 20) {
	*Class = 'B';
	for (i = 0; i < nt; i++) {
            err = (real(sums[i]) - vdata_real_b[i]) / vdata_real_b[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
            err = (imag(sums[i]) - vdata_imag_b[i]) / vdata_imag_b[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
	}
    } else if (d1 == 512 &&
	       d2 == 512 &&
	       d3 == 512 &&
	       nt == 20) {
	*Class = 'C';
	for (i = 0; i < nt; i++) {
            err = (real(sums[i]) - vdata_real_c[i]) / vdata_real_c[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
            err = (imag(sums[i]) - vdata_imag_c[i]) / vdata_imag_c[i];
            if (fabs(err) > epsilon) {
	      *verified = false;
	      break;
	    }
	}
    }
    
    if (*Class != 'U') {
	printf("Result verification successful\n");
    } else {
	printf("Result verification failed\n");
    }
    printf("class = %1c\n", *Class);
}


/* print and debug methods */
void print_timers ()

{
  char* name [] = { "total", "ft", "dpermute", "twiddle", "csum", "reduce", "initial", "assign", "index"};
  
  for (int i = T_TOTAL; i <= T_INDEX; i++)
    IO::cout << name[i] << " : " << timer_read (i) << endl;
}


#endif /* HTA_TRAIT */
