/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: error.h,v 1.1 2006/08/03 06:18:01 bikshand Exp $
 * Authors: Ganesh Bikshandi
 */

#ifndef  __SQ_SUM_H__
#define  __SQ_SUM_H__

#include <cassert>
#include "MapReduceHTA.h"

using namespace std;


struct dVector
{
  double val[5];
};

/**
 * map-reduce operator to compute the square sum and find maxiumum
 */
template<typename T, int DIM>
struct Error : public  OperatorHTA::MapReduceHTA<T, dVector, DIM>  
{
  Error(int level):
    OperatorHTA::MapReduceHTA<T, T, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {
    for  (int i =0; i < 5; i++)
      sqsum.val[i] = 0.0;
  }

  void reduce (const double & d) 
  { 
    for (int i = 0; i < 5; i++)
      sqsum.val[i] += d.val[i];
  }
  
  void map_tuple (const Tuple<DIM>& pos, const HTA<T, DIM> & u) {
    
    int iglob = pos[0];
    int jglob = pos[1];
    int kglob = pos[2];
    double u0[5];
    
    exact(iglob, jglob, kglob, &u0);
    for (int i = 0; i < 5; i++)
      {
	double tmp = u0[i] - u[T(i,0,0)];
	sqsum.val[i] +=  tmp * tmp;
      }
  }

  virtual const dVector& result () const 
  {
    return sqsum;
  }
  
  Operator::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }
  
  dVector sqsum;
}; // struct Error

#endif
