#include "lu.h"
#include "applu.h"
#include "ShapeH.h"
#include "MapReduceHTA.h"

typedef Triplet R;
typedef Tuple<3> T;


typedef HTA<double, 3> HTA_2H;

typedef Tuple<3> T;
int isiz1;
int isiz2;
int isiz3;

int np1;
int np2;
int np3;
/*--------------------------------------------------------------------
  c   xi-direction flux differences
  --------------------------------------------------------------------*/

 template<typename TYPE, int DIM>
    struct Loop1 :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> >  
    {
      Loop1 () {}
      
      Loop1 (int level) :
	OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {}
      
      void init (MapContext<TYPE, DIM>& context)
      {
	this->context_ = context;
      }
      
      void map_hta_none (HTA<TYPE, DIM,  
			  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
			  HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd) 
      {  
	flux[T(0, 0, 0)] = rsd[T(1, 0, 0)];
	TYPE u21 = rsd[T(1, 0, 0)] / rsd[T(0, 0, 0)];
	TYPE q = 0.50 * (  rsd[T(1, 0, 0)] * rsd[T(1, 0, 0)]
			   + rsd[T(2, 0, 0)] * rsd[T(2, 0, 0)]
			   + rsd[T(3, 0, 0)] * rsd[T(3, 0, 0)])
      / rsd[T(0, 0, 0)];
	flux[T(1, 0, 0)] = rsd[T(1, 0, 0)] * u21 + C2 * 
	  ( rsd[T(4, 0, 0)] - q );
	flux[T(2, 0, 0)] = rsd[T(2, 0, 0)] * u21;
	flux[T(3, 0, 0)] = rsd[T(3, 0, 0)] * u21;
	flux[T(4, 0, 0)] = ( C1 * rsd[T(4, 0, 0)] - C2 * q ) * u21; 
	assert (q != 0);
	assert (u21 != 0);
      }
      
      virtual const TYPE& result () const {} 
      
      bool doReduce () { return false; }
      
      OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_NONE; }
      
    };
  
  template<typename TYPE, int DIM>
    struct Loop3 :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> >  
    {
      Loop3 () {}
      
      Loop3 (int level) :
	OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {}
      
      void init (MapContext<TYPE, DIM>& context)
      {
	this->context_ = context;
      }
      
      void map_hta_none(  HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
			  HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
			  HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1) 
      {
	TYPE tmp =  1.0 / rsd0[T(0, 0, 0)];
	TYPE u21i = tmp * rsd0[T(1, 0, 0)];
	TYPE u31i = tmp * rsd0[T(2, 0, 0)];
	TYPE u41i = tmp * rsd0[T(3, 0, 0)];
	TYPE u51i = tmp * rsd0[T(4, 0, 0)];
	
	tmp = 1.0 / rsd1[T(0, 0, 0)];
	TYPE u21im1 = tmp * rsd1[T(1, 0, 0)];
	TYPE u31im1 = tmp * rsd1[T(2, 0, 0)];
	TYPE u41im1 = tmp * rsd1[T(3, 0, 0)];
	TYPE u51im1 = tmp * rsd1[T(4, 0, 0)];
    
	flux[T(1, 0, 0)] = (4.0/3.0) * tx3 *  ( u21i - u21im1 );
	flux[T(2, 0, 0)] = tx3 * ( u31i - u31im1 );
	flux[T(3, 0, 0)] = tx3 * ( u41i - u41im1 );
	flux[T(4, 0, 0)] = 0.50 * ( 1.0 - C1*C5 )
	  * tx3 * ( ( u21i * u21i + u31i * u31i + u41i * u41i )
		- ( u21im1*u21im1 + u31im1*u31im1 + u41im1*u41im1 ) )
	  + (1.0/6.0) * tx3 * ( u21i*u21i - u21im1*u21im1 )
	  + C1 * C5 * tx3 * ( u51i - u51im1 ); 	
      }
      
      virtual const TYPE& result () const {} 
      
      bool doReduce () { return false; }
      
      OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_NONE; }
      
    };

void erhs1 (HTA_2H rsd, HTA_2H frct, HTA_2H flux)
{  

  double dsspm = dssp;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = rsd.shape().size()[0];
  int ty = rsd.shape().size()[1]; 
  int tz = rsd.shape().size()[2];

  int L1 = 1;
  int L2 = isiz1+2;
  int sx, sy, sz;

  ShapeH<3> shape(T(tx, ty, tz));
  
  /* loop 1 */
  shape =  Shape<3> ((R(L1, L2), R(jst, jend),  R(kst, kend)));
  sx = (L2 - L1 + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape1 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  Loop1<double, 3> loop1(1);
  HTA_2H t0 = flux()(shape1)();
  HTA_2H t1 = rsd()(shape1)(); 
  
  t0.mapReduceHTA (loop1, t1);
  
  /*loop 2 */
  // I do not use map because the computation is simpler
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, kend)));  
  sx = (iend - ist+1)*tx;
  sy = (jend - jst+1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape2 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  frct()(shape2)() =  frct()(shape2)() -  tx2 * (flux()(shape2 + T(1, 0, 0))() - flux()(shape2 -  T(1, 0, 0))());  

  //return;

  /* loop 3 */
  //cout << "here 1 " << endl;

  shape = Shape<3>((R(ist, L2), R(jst, jend), R(kst, kend)));  
  sx = (L2 - ist+1)*tx;
  sy = (jend - jst+1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape3 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  Loop3<double, 3> loop3(1);
  HTA_2H t2 = flux()(shape3)();
  HTA_2H t3 = rsd()(shape3)();
  HTA_2H t4 = rsd()(shape3 - T(1, 0, 0))();
  t2.mapReduceHTA (loop3, t3, t4);

  /* loop 4 */
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, kend)));
  sx = (iend-ist+1)*tx;
  sy = (jend-jst+1)*ty;
  sz = (kend-kst+1)*tz;
  ShapeH<3> shape4 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  
  //cout  << "dx " << dx1 << " " << tx1 << endl;

  frct()(shape4)((R(0, 0), R(0,0), R(0,0))) = 
    frct()(shape4)((R(0, 0), R(0,0), R(0,0)))
    + tx1  * dx1 * (rsd()(shape4 -  T(1, 0, 0))((R(0, 0), R(0,0), R(0,0)))
		    - 2.0 * rsd()(shape4)((R(0, 0), R(0,0), R(0,0)))
		    +       	    rsd()(shape4 + T(1, 0, 0))((R(0, 0), R(0,0), R(0,0))));
    
  frct()(shape4)((R(1,1), R(0,0), R(0,0))) = 
    frct()(shape4)((R(1,1), R(0,0), R(0,0)))
    + tx3 * C3 * C4 * ( flux()(shape4 + T(1, 0, 0))((R(1,1), R(0,0), R(0,0))) - flux()(shape4)((R(1,1), R(0,0), R(0,0))))
    + dx2 * tx1 * (            rsd()(shape4 - T(1, 0, 0))((R(1,1), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(1,1), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(1, 0, 0))((R(1,1), R(0,0), R(0,0)))) ;
  
  frct()(shape4)((R(2, 2), R(0,0), R(0,0))) = 
    frct()(shape4)((R(2,2), R(0,0), R(0,0)))
    + tx3 * C3 * C4 * ( flux()(shape4 + T(1, 0, 0))((R(2, 2), R(0,0), R(0,0))) - flux()(shape4)((R(2,2), R(0,0), R(0,0))) )
    + dx3 * tx1 * (            rsd()(shape4 -  T(1, 0, 0))((R(2,2), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(2,2), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(1, 0, 0))((R(2,2), R(0,0), R(0,0))));
  
  frct()(shape4)((R(3,3), R(0,0), R(0,0))) = 
    frct()(shape4)((R(3,3), R(0,0), R(0,0)))
    + tx3 * C3 * C4 * ( flux()(shape4 + T(1, 0, 0))((R(3,3), R(0,0), R(0,0))) - flux()(shape4)((R(3,3), R(0,0), R(0,0))) )
    + dx4 * tx1 * (            rsd()(shape4 - T(1, 0, 0))((R(3,3), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(3,3), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(1, 0, 0))((R(3,3), R(0,0), R(0,0))));

  frct()(shape4)((R(4, 4), R(0,0), R(0,0))) = 
    frct()(shape4)((R(4, 4), R(0,0), R(0,0)))
    + tx3 * C3 * C4 * ( flux()(shape4 + T(1, 0, 0))((R(4, 4), R(0,0), R(0,0))) - flux()(shape4)((R(4, 4), R(0,0), R(0,0))) )
    + dx5 * tx1 * (            rsd()(shape4 - T(1, 0, 0))((R(4, 4), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(4, 4), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(1, 0, 0))((R(4, 4), R(0,0), R(0,0))));
  

  
  /*--------------------------------------------------------------------
    c   Fourth-order dissipation
    --------------------------------------------------------------------*/  
  
  //cout << "dssp " << dssp << endl;

  frct((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3,3), R(jst, jend), R(kst, kend)))() = 
    frct((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3,3), R(jst, jend), R(kst, kend)))()
    - dssp * ( 5.0 * rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3, 3), R(jst, jend), R(kst, kend)))()
	       - 4.0 * rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend)))()
	       + rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(5, 5), R(jst, jend), R(kst, kend)))());
  
  frct((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend)))() = 
    frct((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend)))()
    - dssp * ( - 4.0 * rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3, 3), R(jst, jend), R(kst, kend)))()
		+ 6.0 * rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend)))()
		- 4.0 * rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(5, 5), R(jst, jend), R(kst, kend)))()
		+           rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(6, 6), R(jst, jend), R(kst, kend)))());
  
  /* loop 5 */
  int ist1 = 2;
  int iend1 = isiz1+1;
  shape = Shape<3>((R(ist1, iend1), R(jst, jend), R(kst, kend)));
  sx = (iend1 - ist1 + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape5 = shape.intersect (Shape<3>((R(3, sx-4), R(1,sy-2), R(1,sz-2))));
  frct()(shape5)() = frct()(shape5)()
    - dsspm * (            rsd()(shape5 - T(2, 0, 0))()
			   - 4.0 * rsd()(shape5 - T(1,0,0))()
			   + 6.0 * rsd()(shape5)()
			   - 4.0 * rsd()(shape5 + T(1, 0, 0))()
			   +           rsd()(shape5 + T(2, 0, 0))());
  
  
  frct((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend)))() = 
    frct((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend)))()
    - dsspm * (             rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-3, isiz1-3), R(jst, jend), R(kst, kend)))()
			    - 4.0 * rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-2, isiz1-2), R(jst, jend), R(kst, kend)))()
			    + 6.0 * rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend)))()
			    - 4.0 * rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend)))());
  
  frct((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend)))() = 
    frct((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend)))()
    - dsspm * (             rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-2, isiz1-2), R(jst, jend), R(kst, kend)))()
			    - 4.0 * rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend)))()
			    + 5.0 * rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend)))()); 
  
}



template<typename TYPE, int DIM>
struct Loop6 :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> >  
{
  Loop6 () {}
  
  Loop6 (int level) :
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {}
  
  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_none ( HTA<TYPE, DIM, Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd) 
  {  
    flux[T(0, 0, 0)] = rsd[T(2, 0, 0)];
    TYPE u31 = rsd[T(2, 0, 0)] / rsd[T(0, 0, 0)];
    TYPE q = 0.50 * (  rsd[T(1, 0, 0)] * rsd[T(1, 0, 0)]
		       + rsd[T(2, 0, 0)] * rsd[T(2, 0, 0)]
		       + rsd[T(3, 0, 0)] * rsd[T(3, 0, 0)] )
      / rsd[T(0, 0, 0)];
    flux[T(1, 0, 0)] = rsd[T(1, 0, 0)] * u31;
    flux[T(2, 0, 0)] = rsd[T(2, 0, 0)] * u31 + C2 * 
      ( rsd[T(4, 0, 0)] - q );
    flux[T(3, 0, 0)] = rsd[T(3, 0, 0)] * u31;
    flux[T(4, 0, 0)] = ( C1 * rsd[T(4, 0, 0)] - C2 * q ) * u31;
    assert (q != 0);
    assert (u31 != 0);
  }
  
  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
  
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_NONE; }
  
};

template<typename TYPE, int DIM>
struct Loop8 :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> >  
{
  Loop8 () {}
  
  Loop8 (int level) :
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {}
  
  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_none (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1) 
  {
    TYPE tmp =  1.0 / rsd0[T(0, 0, 0)];
    TYPE u21j = tmp * rsd0[T(1, 0, 0)];
    TYPE u31j = tmp * rsd0[T(2, 0, 0)];
    TYPE u41j = tmp * rsd0[T(3, 0, 0)];
    TYPE u51j = tmp * rsd0[T(4, 0, 0)];
    
    tmp = 1.0 / rsd1[T(0, 0, 0)];
    TYPE u21jm1 = tmp * rsd1[T(1, 0, 0)];
    TYPE u31jm1 = tmp * rsd1[T(2, 0, 0)];
    TYPE u41jm1 = tmp * rsd1[T(3, 0, 0)];
    TYPE u51jm1 = tmp * rsd1[T(4, 0, 0)];
    
    flux[T(1, 0, 0)] = ty3 * ( u21j - u21jm1 );
    flux[T(2, 0, 0)] = (4.0/3.0) * ty3 *
      ( u31j - u31jm1 );
    flux[T(3, 0, 0)] = ty3 * ( u41j - u41jm1 );
    flux[T(4, 0, 0)] = 0.50 * ( 1.0 - C1*C5 )
      * ty3 * ( ( u21j * u21j + u31j * u31j + u41j * u41j )
		- ( u21jm1*u21jm1 + u31jm1*u31jm1 + u41jm1*u41jm1 ) )
      + (1.0/6.0) * ty3 * ( u31j*u31j - u31jm1*u31jm1 )
      + C1 * C5 * ty3 * ( u51j - u51jm1 ); 	
  }
  
  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
  
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_NONE; }
  
};

void erhs2 (HTA_2H rsd, HTA_2H frct, HTA_2H flux)
{  


  
  double dsspm = dssp;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = rsd.shape().size()[0];
  int ty = rsd.shape().size()[1]; 
  int tz = rsd.shape().size()[2];

  int L1 = 1;
  int L2 = isiz2+2;
  int sx, sy, sz;

  ShapeH<3> shape(T(tx, ty, tz));
  
  /* loop 1 */
  shape =  Shape<3> ((R(ist, iend), R(L1, L2),  R(kst, kend)));
  sx = (iend - ist + 1)*tx;
  sy = (L2 - L1 + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape1 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  Loop6<double, 3> loop6(1);
  HTA_2H t0 = flux()(shape1)();
  HTA_2H t1 = rsd()(shape1)(); 
  
  t0.mapReduceHTA (loop6, t1);
  
  /*loop 2 */
  // I do not use mapReduce because the computation is simpler
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, kend)));  
  sx = (iend - ist+1)*tx;
  sy = (jend - jst+1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape2 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  frct()(shape2)() =  frct()(shape2)() -  ty2 * (flux()(shape2 + T(0, 1, 0))() - flux()(shape2 -  T(0, 1, 0))());  

  //return;

  /* loop 3 */
  shape = Shape<3>((R(ist, iend), R(jst, L2), R(kst, kend)));  
  sx = (iend - ist+1)*tx;
  sy = (L2 - ist +1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape3 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  Loop8<double, 3> loop8(1);
  HTA_2H t2 = flux()(shape3)();
  HTA_2H t3 = rsd()(shape3)();
  HTA_2H t4 = rsd()(shape3 - T(0, 1, 0))();
  t2.mapReduceHTA (loop8, t3, t4);

  /* loop 4 */
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, kend)));
  sx = (iend-ist+1)*tx;
  sy = (jend-jst+1)*ty;
  sz = (kend-kst+1)*tz;
  ShapeH<3> shape4 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  
  cout  << "dx " << dx1 << " " << tx1 << endl;

  frct()(shape4)((R(0, 0), R(0,0), R(0,0))) = 
    frct()(shape4)((R(0, 0), R(0,0), R(0,0)))
    + ty1  * dy1 * (rsd()(shape4 -  T(0, 1, 0))((R(0, 0), R(0,0), R(0,0)))
		    - 2.0 * rsd()(shape4)((R(0, 0), R(0,0), R(0,0)))
		    +       	    rsd()(shape4 + T(0, 1, 0))((R(0, 0), R(0,0), R(0,0))));
    
  frct()(shape4)((R(1,1), R(0,0), R(0,0))) = 
    frct()(shape4)((R(1,1), R(0,0), R(0,0)))
    + ty3 * C3 * C4 * ( flux()(shape4 + T(0, 1, 0))((R(1,1), R(0,0), R(0,0))) - flux()(shape4)((R(1,1), R(0,0), R(0,0))))
    + dy2 * ty1 * (            rsd()(shape4 - T(0, 1, 0))((R(1,1), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(1,1), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 1, 0))((R(1,1), R(0,0), R(0,0)))) ;
  
  frct()(shape4)((R(2, 2), R(0,0), R(0,0))) = 
    frct()(shape4)((R(2,2), R(0,0), R(0,0)))
    + ty3 * C3 * C4 * ( flux()(shape4 + T(0, 1, 0))((R(2, 2), R(0,0), R(0,0))) - flux()(shape4)((R(2,2), R(0,0), R(0,0))) )
    + dy3 * ty1 * (            rsd()(shape4 -  T(0, 1, 0))((R(2,2), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(2,2), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 1, 0))((R(2,2), R(0,0), R(0,0))));
  
  frct()(shape4)((R(3,3), R(0,0), R(0,0))) = 
    frct()(shape4)((R(3,3), R(0,0), R(0,0)))
    + ty3 * C3 * C4 * ( flux()(shape4 + T(0, 1, 0))((R(3,3), R(0,0), R(0,0))) - flux()(shape4)((R(3,3), R(0,0), R(0,0))) )
    + dy4 * ty1 * (            rsd()(shape4 - T(0, 1, 0))((R(3,3), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(3,3), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 1, 0))((R(3,3), R(0,0), R(0,0))));
  frct()(shape4)((R(4, 4), R(0,0), R(0,0))) = 
    frct()(shape4)((R(4, 4), R(0,0), R(0,0)))
    + ty3 * C3 * C4 * ( flux()(shape4 + T(0, 1, 0))((R(4, 4), R(0,0), R(0,0))) - flux()(shape4)((R(4, 4), R(0,0), R(0,0))) )
    + dy5 * ty1 * (            rsd()(shape4 - T(0, 1, 0))((R(4, 4), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(4, 4), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 1, 0))((R(4, 4), R(0,0), R(0,0))));
    
  /*--------------------------------------------------------------------
    c   Fourth-order dissipation
    --------------------------------------------------------------------*/  
  
  cout << "dssp " << dssp << endl;
  
  frct((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(3, 3), R(kst, kend)))() = 
    frct((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(3, 3), R(kst, kend)))()
    - dssp * ( 5.0 * rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(3, 3), R(kst, kend)))()
	       - 4.0 * rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend)))()
	       + rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(5, 5), R(kst, kend)))());
  
  frct((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend)))() = 
    frct((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend)))()
    - dssp * ( - 4.0 * rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(3, 3), R(kst, kend)))()
	       + 6.0 * rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend)))()
	       - 4.0 * rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(5, 5), R(kst, kend)))()
	       +           rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(6, 6), R(kst, kend)))());
  
  /* loop 5 */
  int jst1 = 2;
  int jend1 = isiz2+1;
  shape = Shape<3>((R(ist, iend), R(jst1, jend1), R(kst, kend)));
  sx = (iend - ist + 1)*tx;
  sy = (jend1 - jst1 + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape5 = shape.intersect (Shape<3>((R(1, sx-2), R(3,sy-4), R(1,sz-2))));
  frct()(shape5)() = frct()(shape5)()
    - dsspm * (            rsd()(shape5 - T(0, 2, 0))()
			   - 4.0 * rsd()(shape5 - T(0, 1, 0))()
			   + 6.0 * rsd()(shape5)()
			   - 4.0 * rsd()(shape5 + T(0, 1, 0))()
			   +           rsd()(shape5 + T(0, 2, 0))());
  
  frct((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1, isiz2-1), R(kst, kend)))() = 
    frct((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1, isiz2-1), R(kst, kend)))()
    - dsspm * (             rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-3,isiz2-3), R(kst, kend)))()
			    - 4.0 * rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-2,isiz2-2), R(kst, kend)))()
			    + 6.0 * rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1,isiz2-1), R(kst, kend)))()
			    - 4.0 * rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend)))());
  
  frct((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend)))() = 
    frct((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend)))()
    - dsspm * (             rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-2,isiz2-2), R(kst, kend)))()
			    - 4.0 * rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1, isiz2-1), R(kst, kend)))()
			    + 5.0 * rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend)))()); 
  
}



template<typename TYPE, int DIM>
struct Loop12 :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> >  
{
  Loop12 () {}
  
  Loop12 (int level) :
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {}
  
  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_none (HTA<TYPE, DIM,  
		      Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd) 
  {  
    flux[T(0, 0, 0)] = rsd[T(3, 0, 0)];
    TYPE u41 = rsd[T(3, 0, 0)] / rsd[T(0, 0, 0)];
    TYPE q = 0.50 * (  rsd[T(1, 0, 0)] * rsd[T(1, 0, 0)]
		       + rsd[T(2, 0, 0)] * rsd[T(2, 0, 0)]
		       + rsd[T(3, 0, 0)] * rsd[T(3, 0, 0)] )
      / rsd[T(0, 0, 0)];
    flux[T(1, 0, 0)] = rsd[T(1, 0, 0)] * u41;
    flux[T(2, 0, 0)] = rsd[T(2, 0, 0)] * u41;
    flux[T(3, 0, 0)] = rsd[T(3, 0, 0)] * u41 + C2 * ( rsd[T(4, 0, 0)] - q);
    flux[T(4, 0, 0)] = ( C1 * rsd[T(4, 0, 0)] - C2 * q ) * u41;
    assert (q != 0);
    assert (u41 != 0);
  }
  
  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
  
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_NONE; }
  
};

template<typename TYPE, int DIM>
struct Loop14 :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> >  
{
  Loop14 () {}
  
  Loop14 (int level) :
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > (level) {}
  
  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_none (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1) 
  {
    TYPE tmp =  1.0 / rsd0[T(0, 0, 0)];
    TYPE u21j = tmp * rsd0[T(1, 0, 0)];
    TYPE u31j = tmp * rsd0[T(2, 0, 0)];
    TYPE u41j = tmp * rsd0[T(3, 0, 0)];
    TYPE u51j = tmp * rsd0[T(4, 0, 0)];
    
    tmp = 1.0 / rsd1[T(0, 0, 0)];
    TYPE u21jm1 = tmp * rsd1[T(1, 0, 0)];
    TYPE u31jm1 = tmp * rsd1[T(2, 0, 0)];
    TYPE u41jm1 = tmp * rsd1[T(3, 0, 0)];
    TYPE u51jm1 = tmp * rsd1[T(4, 0, 0)];
    
    flux[T(1, 0, 0)] = tz3 * ( u21j - u21jm1 );
    flux[T(2, 0, 0)] =  tz3 * ( u31j - u31jm1 );     
    flux[T(3, 0, 0)] =  (4.0/3.0) * tz3 *
      ( u41j - u41jm1 );
    flux[T(4, 0, 0)] = 0.50 * ( 1.0 - C1*C5 )
      * tz3 * ( ( u21j * u21j + u31j * u31j + u41j * u41j )
		- ( u21jm1*u21jm1 + u31jm1*u31jm1 + u41jm1*u41jm1 ) )
      + (1.0/6.0) * tz3 * ( u41j*u41j - u41jm1*u41jm1 )
      + C1 * C5 * tz3 * ( u51j - u51jm1 ); 	
  }
  
  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
  
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_NONE; }
  
};

void erhs3 (HTA_2H rsd, HTA_2H frct, HTA_2H flux)
{  

  double dsspm = dssp;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = rsd.shape().size()[0];
  int ty = rsd.shape().size()[1]; 
  int tz = rsd.shape().size()[2];

  int L1 = 1;
  int L2 = isiz3+2;
  int sx, sy, sz;

  ShapeH<3> shape(T(tx, ty, tz));
  
  /* loop 1 */
  shape =  Shape<3> ((R(ist, iend), R(jst, jend),  R(L1, L2)));
  sx = (iend - ist + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (L2 - L1 + 1)*tz;
  ShapeH<3> shape1 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  Loop12<double, 3> loop12(1);
  HTA_2H t0 = flux()(shape1)();
  HTA_2H t1 = rsd()(shape1)(); 
  
  t0.mapReduceHTA (loop12, t1);
  
  /*loop 2 */
  // I do not use mapReduce because the computation is simpler
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, kend)));  
  sx = (iend - ist+1)*tx;
  sy = (jend - jst+1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape2 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  frct()(shape2)() =  frct()(shape2)() -  tz2 * (flux()(shape2 + T(0,  0, 1))() - flux()(shape2 -  T(0, 0, 1))());  

  //return;

  /* loop 3 */
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, L2)));  
  sx = (iend - ist+1)*tx;
  sy = (jend - jst +1)*ty;
  sz = (L2 - kst+1)*tz;
  ShapeH<3> shape3 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  Loop14<double, 3> loop14(1);
  HTA_2H t2 = flux()(shape3)();
  HTA_2H t3 = rsd()(shape3)();
  HTA_2H t4 = rsd()(shape3 - T(0, 0, 1))();
  t2.mapReduceHTA (loop14, t3, t4);

  /* loop 4 */
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, kend)));
  sx = (iend-ist+1)*tx;
  sy = (jend-jst+1)*ty;
  sz = (kend-kst+1)*tz;
  ShapeH<3> shape4 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  
  cout  << "dx " << dx1 << " " << tx1 << endl;

  frct()(shape4)((R(0, 0), R(0,0), R(0,0))) = 
    frct()(shape4)((R(0, 0), R(0,0), R(0,0)))
    + tz1  * dz1 * (rsd()(shape4 -  T(0, 0, 1))((R(0, 0), R(0,0), R(0,0)))
		    - 2.0 * rsd()(shape4)((R(0, 0), R(0,0), R(0,0)))
		    +       	    rsd()(shape4 + T(0, 0, 1))((R(0, 0), R(0,0), R(0,0))));
    
  frct()(shape4)((R(1,1), R(0,0), R(0,0))) = 
    frct()(shape4)((R(1,1), R(0,0), R(0,0)))
    + tz3 * C3 * C4 * ( flux()(shape4 + T(0, 0, 1))((R(1,1), R(0,0), R(0,0))) - flux()(shape4)((R(1,1), R(0,0), R(0,0))))
    + dz2 * tz1 * (            rsd()(shape4 - T(0, 0, 1))((R(1,1), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(1,1), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 0, 1))((R(1,1), R(0,0), R(0,0)))) ;
  
  frct()(shape4)((R(2, 2), R(0,0), R(0,0))) = 
    frct()(shape4)((R(2,2), R(0,0), R(0,0)))
    + tz3 * C3 * C4 * ( flux()(shape4 + T(0, 0, 1))((R(2, 2), R(0,0), R(0,0))) - flux()(shape4)((R(2,2), R(0,0), R(0,0))) )
    + dz3 * tz1 * (            rsd()(shape4 -  T(0, 0, 1))((R(2,2), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(2,2), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 0, 1))((R(2,2), R(0,0), R(0,0))));
  
  frct()(shape4)((R(3,3), R(0,0), R(0,0))) = 
    frct()(shape4)((R(3,3), R(0,0), R(0,0)))
    + tz3 * C3 * C4 * ( flux()(shape4 + T(0, 0, 1))((R(3,3), R(0,0), R(0,0))) - flux()(shape4)((R(3,3), R(0,0), R(0,0))) )
    + dz4 * tz1 * (            rsd()(shape4 - T(0, 0, 1))((R(3,3), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(3,3), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 0, 1))((R(3,3), R(0,0), R(0,0))));
  frct()(shape4)((R(4, 4), R(0,0), R(0,0))) = 
    frct()(shape4)((R(4, 4), R(0,0), R(0,0)))
    + tz3 * C3 * C4 * ( flux()(shape4 + T(0, 0, 1))((R(4, 4), R(0,0), R(0,0))) - flux()(shape4)((R(4, 4), R(0,0), R(0,0))) )
    + dz5 * tz1 * (            rsd()(shape4 - T(0, 0, 1))((R(4, 4), R(0,0), R(0,0)))
			       - 2.0 * rsd()(shape4)((R(4, 4), R(0,0), R(0,0)))
			       +           rsd()(shape4 + T(0, 0, 1))((R(4, 4), R(0,0), R(0,0))));
    
  /*--------------------------------------------------------------------
    c   Fourth-order dissipation
    --------------------------------------------------------------------*/  
  
  cout << "dssp " << dssp << endl;
  
  frct((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend), R(jst, jend), R(3, 3)))() = 
    frct((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(3, 3)))()
    - dssp * ( 5.0 * rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(3, 3)))()
	       - 4.0 * rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(4, 4)))()
	       + rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(5, 5)))());
  
  frct((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(4, 4)))() = 
    frct((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend), R(4, 4)))()
    - dssp * ( - 4.0 * rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend), R(3, 3)))()
	       + 6.0 * rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend), R(4, 4)))()
	       - 4.0 * rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend), R(5, 5)))()
	       +           rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend), R(jst, jend), R(6, 6)))());
  
  /* loop 5 */
  int kst1 = 2;
  int kend1 = isiz2+1;
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst1, kend1)));
  sx = (iend - ist + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend1 - kst1 + 1)*tz;
  ShapeH<3> shape5 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(3,sz-4))));
  frct()(shape5)() = frct()(shape5)()
    - dsspm * (            rsd()(shape5 - T(0, 0, 2))()
			   - 4.0 * rsd()(shape5 - T(0, 0, 1))()
			   + 6.0 * rsd()(shape5)()
			   - 4.0 * rsd()(shape5 + T(0, 0, 1))()
			   +           rsd()(shape5 + T(0, 0, 2))());
  
  frct((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3-1, isiz3-1)))() = 
    frct((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3-1, isiz3-1)))()
    - dsspm * (             rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3-3,isiz3-3)))()
			    - 4.0 * rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3-2,isiz3-2)))()
			    + 6.0 * rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend),R(isiz3-1,isiz3-1)))()
			    - 4.0 * rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend),R(isiz3, isiz3 )))());
  
  frct((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3, isiz3)))() = 
    frct((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3, isiz3)))()
    - dsspm * (             rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3-2,isiz3-2)))()
			    - 4.0 * rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3-1, isiz3-1)))()
			    + 5.0 * rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))((R(ist, iend),  R(jst, jend), R(isiz3, isiz3)))()); 
  
}



static void exact( int i, int j, int k, double u000ijk[5] ) {

/*--------------------------------------------------------------------
c
c   compute the exact solution at (i,j,k)
c
--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c  local variables
--------------------------------------------------------------------*/
  int m;
  double xi, eta, zeta;

  xi  = ((double)i) / (double) (ISIZ1 - 1);
  eta  = ((double)j) / (double) (ISIZ2 - 1);
  zeta = ((double)k) / (double) (ISIZ3 - 1);

  for (m = 0; m < 5; m++) {
    u000ijk[m] =  ce[m][0]
      + ce[m][1] * xi
      + ce[m][2] * eta
      + ce[m][3] * zeta
      + ce[m][4] * xi * xi
      + ce[m][5] * eta * eta
      + ce[m][6] * zeta * zeta
      + ce[m][7] * xi * xi * xi
      + ce[m][8] * eta * eta * eta
      + ce[m][9] * zeta * zeta * zeta
      + ce[m][10] * xi * xi * xi * xi
      + ce[m][11] * eta * eta * eta * eta
      + ce[m][12] * zeta * zeta * zeta * zeta;
  }
}


static void exact (int i, int j, int k, HTA<double, 3> h) {

/*--------------------------------------------------------------------
c
c   compute the exact solution at (i,j,k)
c
--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c  local variables
--------------------------------------------------------------------*/
  int m;
  double xi, eta, zeta;

  xi  = ((double)i) / (double) (ISIZ1 - 1);
  eta  = ((double)j) / (double) (ISIZ2 - 1);
  zeta = ((double)k) / (double) (ISIZ3 - 1);

  for (m = 0; m < 5; m++) {
    h[T(m, 0, 0)] =  ce[m][0]
      + ce[m][1] * xi
      + ce[m][2] * eta
      + ce[m][3] * zeta
      + ce[m][4] * xi * xi
      + ce[m][5] * eta * eta
      + ce[m][6] * zeta * zeta
      + ce[m][7] * xi * xi * xi
      + ce[m][8] * eta * eta * eta
      + ce[m][9] * zeta * zeta * zeta
      + ce[m][10] * xi * xi * xi * xi
      + ce[m][11] * eta * eta * eta * eta
      + ce[m][12] * zeta * zeta * zeta * zeta;
  }
}
