void jacld (HTA_2H u, HTA_2H a, HTA_2H b, HTA_2H c, HTA_2H d)
{
/*--------------------------------------------------------------------
c   compute the lower triangular part of the jacobian matrix
--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c  local variables
--------------------------------------------------------------------*/
  int i, j;
  double  r43;
  double  c1345;
  double  c34;

  r43 = ( 4.0 / 3.0 );
  c1345 = C1 * C3 * C4 * C5;
  c34 = C3 * C4;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = u.shape().size()[0];
  int ty = u.shape().size()[1]; 
  int tz = u.shape().size()[2];
  int sx, sy, sz;

  ShapeH<3> shape1(T(tx, ty, tz));
  shape1 =  Shape<3> ((R(ist, iend), R(jst, jend),  R(kst, kend)));
  sx = (iend - ist + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape = shape1.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  /*--------------------------------------------------------------------
    c   form the block daigonal
    --------------------------------------------------------------------*/

  HTA_2H tmp1 = 1.0 / u()(shape)(T(0, 0, 0));
  HTA_2H tmp2 = tmp1 * tmp1;
  HTA_2H tmp3 = tmp1 * tmp2;

  d()(shape)(T(0, 0, 0)) =  1.0
    + dt * 2.0 * (   tx1 * dx1
		     + ty1 * dy1
		     + tz1 * dz1 );

  d()(shape)(T(0, 1, 0)) =  0.0;
  d()(shape)(T(0, 2, 0)) =  0.0;
  d()(shape)(T(0, 3, 0)) =  0.0;
  d()(shape)(T(0, 4, 0)) =  0.0;

  d()(shape)(T(1, 0, 0)) =  (dt * 2.0)
    * (  (  tx1 * ( - r43 * c34) 
	    + ty1 * ( -     c34 )
	    + tz1 * ( -     c34 ) ) 
	 * u()(shape)(T(1, 0, 0)) ) * tmp2;
  
  d()(shape)(T(1, 1, 0)) =  1.0
    + dt * 2.0 
    * (  tx1 * r43 * c34 * tmp1
	 + ty1 *       c34 * tmp1
	 + tz1 *       c34 * tmp1 )
    + dt * 2.0 * (   tx1 * dx2
		     + ty1 * dy2
		     + tz1 * dz2  );
  d()(shape)(T(1, 2, 0)) = 0.0;
  d()(shape)(T(1, 3, 0)) = 0.0;
  d()(shape)(T(1, 4, 0)) = 0.0;

  d()(shape)(T(2, 0, 0)) = dt * 2.0
    * (  (tx1 * ( -       c34 )
	  + ty1 * ( - r43 * c34 )
	  + tz1 * ( -       c34) ) 
	 * u()(shape)(T(2, 0, 0)) * tmp2);
  
  d()(shape)(T(2, 1, 0)) = 0.0;
  d()(shape)(T(2, 2, 0)) = 1.0
    + dt * 2.0
    * (  tx1 *       c34 * tmp1
	 + ty1 * r43 * c34 * tmp1
	 + tz1 *       c34 * tmp1 )
    + dt * 2.0 * (  tx1 * dx3
		    + ty1 * dy3
		    + tz1 * dz3 );
  
  d()(shape)(T(2, 3, 0)) = 0.0;
  d()(shape)(T(2, 4, 0)) = 0.0;
  
  d()(shape)(T(3, 0, 0)) = dt * 2.0
    * (  (tx1 * ( -       c34 ) 
	  + ty1 * ( -       c34 ) 
	  + tz1 * ( - r43 * c34 ) ) 
	 * u()(shape)(T(3, 0, 0)) * tmp2 );
  
  d()(shape)(T(3, 1, 0)) = 0.0;
  d()(shape)(T(3, 2, 0)) = 0.0;
  d()(shape)(T(3, 3, 0)) = 1.0
    + dt * 2.0
    * (  tx1 *       c34 * tmp1
	 + ty1 *       c34 * tmp1
	 + tz1 * r43 * c34 * tmp1 )
    + dt * 2.0 * (  tx1 * dx4
		    + ty1 * dy4
		    + tz1 * dz4 );
  d()(shape)(T(3, 4, 0)) = 0.0;

  d()(shape)(T(4, 0, 0)) = dt * 2.0
    * ( tx1 * ( - ( r43*c34 - c1345 ) * u()(shape)(T(1, 0, 0)).pow2()  * tmp3
		- ( c34 - c1345 ) * u()(shape)(T(2, 0, 0)).pow2()  * tmp3
		- ( c34 - c1345 ) * u()(shape)(T(3, 0, 0)).pow2()  * tmp3
		- ( c1345 ) * u()(shape)(T(4, 0, 0)) ) * tmp2 
	+ ty1 * ( - ( c34 - c1345 ) * u()(shape)(T(1, 0, 0)).pow2()  * tmp3 
		  - ( r43*c34 - c1345 ) * u()(shape)(T(2, 0, 0)).pow2()  * tmp3
		  - ( c34 - c1345 ) * u()(shape)(T(3, 0, 0)).pow2()  * tmp3 
		  - ( c1345 ) * u()(shape)(T(4, 0, 0)) ) * tmp2
	+ tz1 * ( - ( c34 - c1345 ) * u()(shape)(T(1, 0, 0)).pow2()  * tmp3
		  - ( c34 - c1345 ) * u()(shape)(T(2, 0, 0)).pow2()  * tmp3
		  - ( r43*c34 - c1345 ) * u()(shape)(T(3, 0, 0)).pow2()  * tmp3
		  - ( c1345 ) *  u()(shape)(T(4, 0, 0)) ) )  * tmp2;

  d()(shape)(T(4, 1, 0)) = dt * 2.0
    * ( (tx1 * ( r43*c34 - c1345 ) 
	 + ty1 * (     c34 - c1345 ) 
	 + tz1 * (     c34 - c1345 )) 
	* u()(shape)(T(1, 0, 0))) * tmp2;
  
  d()(shape)(T(4, 2, 0)) = dt * 2.0
    * ( (tx1 * ( c34 - c1345 ) 
	 + ty1 * ( r43*c34 -c1345 ) 
	 + tz1 * ( c34 - c1345 )) 
	* u()(shape)(T(2, 0, 0))) * tmp2;
  
  d()(shape)(T(4, 3, 0)) = dt * 2.0
    * ( (tx1 * ( c34 - c1345 ) 
	 + ty1 * ( c34 - c1345 ) 
	 + tz1 * ( r43*c34 - c1345 )) 
	* u()(shape)(T(3, 0, 0))) * tmp2;
  
  d()(shape)(T(4, 4, 0)) = 1.0
    + dt * 2.0 * ( tx1 * c1345 * tmp1
		   + ty1 * c1345 * tmp1
		   + tz1 * c1345 * tmp1 )
    + dt * 2.0 * (  tx1 * dx5
		    +  ty1 * dy5
		    +  tz1 * dz5 );

  /*--------------------------------------------------------------------
    c   form the first block sub-diagonal
    --------------------------------------------------------------------*/
  tmp1 = 1.0 / u()(shape - T(0, 0, 1))(T(0, 0, 0));
  tmp2 = tmp1 * tmp1;
  tmp3 = tmp1 * tmp2;

  a()(shape)(T(0, 0, 0)) = - dt * tz1 * dz1;
  a()(shape)(T(0, 1, 0)) =   0.0;
  a()(shape)(T(0, 2, 0)) =   0.0;
  a()(shape)(T(0, 3, 0)) = - dt * tz2;
  a()(shape)(T(0, 4, 0)) =   0.0;

  a()(shape)(T(1, 0, 0)) = - dt * tz2
    * ( - ( u()(shape - T(0, 0, 1))(T(1, 0, 0)) * u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2 )
    - dt * tz1 * ( - c34 * tmp2 * u()(shape - T(0, 0, 1))(T(1, 0, 0)) );
  a()(shape)(T(1, 1, 0)) = - dt * tz2 * ( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 )
    - dt * tz1 * c34 * tmp1
    - dt * tz1 * dz2 ;
  a()(shape)(T(1, 2, 0)) = 0.0;
  a()(shape)(T(1, 3, 0)) = - dt * tz2 * ( u()(shape - T(0, 0, 1))(T(1, 0, 0)) * tmp1 );
  a()(shape)(T(1, 4, 0)) = 0.0;

  a()(shape)(T(2, 0, 0)) = - dt * tz2
    * ( - ( u()(shape - T(0, 0, 1))(T(2, 0, 0))*u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2 )
    - dt * tz1 * ( - c34 * tmp2 * u()(shape - T(0, 0, 1))(T(2, 0, 0)) );
  a()(shape)(T(2, 1, 0)) = 0.0;
  a()(shape)(T(2, 2, 0)) = - dt * tz2 * ( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 )
    - dt * tz1 * ( c34 * tmp1 )
    - dt * tz1 * dz3;
  a()(shape)(T(2, 3, 0)) = - dt * tz2 * ( u()(shape - T(0, 0, 1))(T(2, 0, 0)) * tmp1 );
  a()(shape)(T(2, 4, 0)) = 0.0;

  a()(shape)(T(3, 0, 0)) = - dt * tz2
    * ( - ( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 ) *( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 )
	+ 0.50 * C2
	* ( ( u()(shape - T(0, 0, 1))(T(1, 0, 0)) * u()(shape - T(0, 0, 1))(T(1, 0, 0))
	      + u()(shape - T(0, 0, 1))(T(2, 0, 0)) * u()(shape - T(0, 0, 1))(T(2, 0, 0))
	      + u()(shape - T(0, 0, 1))(T(3, 0, 0)) * u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2 ) )
    - dt * tz1 * ( - r43 * c34 * tmp2 * u()(shape - T(0, 0, 1))(T(3, 0, 0)) );
  a()(shape)(T(3, 1, 0)) = - dt * tz2
    * ( - C2 * ( u()(shape - T(0, 0, 1))(T(1, 0, 0)) * tmp1 ) );
  a()(shape)(T(3, 2, 0)) = - dt * tz2
    * ( - C2 * ( u()(shape - T(0, 0, 1))(T(2, 0, 0)) * tmp1 ) );
  a()(shape)(T(3, 3, 0)) = - dt * tz2 * ( 2.0 - C2 )
    * ( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 )
    - dt * tz1 * ( r43 * c34 * tmp1 )
    - dt * tz1 * dz4;
  a()(shape)(T(3, 4, 0)) = - dt * tz2 * C2;

  a()(shape)(T(4, 0, 0)) = - dt * tz2
    * ( ( C2 * (  u()(shape - T(0, 0, 1))(T(1, 0, 0)) * u()(shape - T(0, 0, 1))(T(1, 0, 0))
		  + u()(shape - T(0, 0, 1))(T(2, 0, 0)) * u()(shape - T(0, 0, 1))(T(2, 0, 0))
		  + u()(shape - T(0, 0, 1))(T(3, 0, 0)) * u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2
	  - C1 * ( u()(shape - T(0, 0, 1))(T(4, 0, 0)) * tmp1 ) )
	* ( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 ) )* tmp2 * u()(shape)(T(3, 0, 0))
    - dt * tz1
    * ( - ( c34 - c1345 ) * tmp3 * (u()(shape - T(0, 0, 1))(T(1, 0, 0))*u()(shape - T(0, 0, 1))(T(1, 0, 0)))
	- ( c34 - c1345 ) * tmp3 * (u()(shape - T(0, 0, 1))(T(2, 0, 0))*u()(shape - T(0, 0, 1))(T(2, 0, 0)))
	- ( r43*c34 - c1345 )* tmp3 * (u()(shape - T(0, 0, 1))(T(3, 0, 0))*u()(shape - T(0, 0, 1))(T(3, 0, 0)))
	- c1345 * tmp2 * u()(shape - T(0, 0, 1))(T(4, 0, 0)) );
  a()(shape)(T(4, 1, 0)) = - dt * tz2
    * ( - C2 * ( u()(shape - T(0, 0, 1))(T(1, 0, 0))*u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2 )
    - dt * tz1 * ( c34 - c1345 ) * tmp2 * u()(shape - T(0, 0, 1))(T(1, 0, 0));
  a()(shape)(T(4, 2, 0)) = - dt * tz2
    * ( - C2 * ( u()(shape - T(0, 0, 1))(T(2, 0, 0))*u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2 )
    - dt * tz1 * ( c34 - c1345 ) * tmp2 * u()(shape - T(0, 0, 1))(T(2, 0, 0));
  a()(shape)(T(4, 3, 0)) = - dt * tz2
    * ( C1 * ( u()(shape - T(0, 0, 1))(T(4, 0, 0)) * tmp1 )
	- 0.50 * C2
	* ( (  u()(shape - T(0, 0, 1))(T(1, 0, 0))*u()(shape - T(0, 0, 1))(T(1, 0, 0))
	       + u()(shape - T(0, 0, 1))(T(2, 0, 0))*u()(shape - T(0, 0, 1))(T(2, 0, 0))
	       + 3.0*u()(shape - T(0, 0, 1))(T(3, 0, 0))*u()(shape - T(0, 0, 1))(T(3, 0, 0)) ) * tmp2 ) )
    - dt * tz1 * ( r43*c34 - c1345 ) * tmp2 * u()(shape - T(0, 0, 1))(T(3, 0, 0));
  a()(shape)(T(4, 4, 0)) = - dt * tz2
    * ( C1 * ( u()(shape - T(0, 0, 1))(T(3, 0, 0)) * tmp1 ) )
    - dt * tz1 * c1345 * tmp1
    - dt * tz1 * dz5;

  /*--------------------------------------------------------------------
    c   form the second block sub-diagonal
    --------------------------------------------------------------------*/
  tmp1 = 1.0 / u()(shape - T(0, 1, 0))(T(0, 0, 0));
  tmp2 = tmp1 * tmp1;
  tmp3 = tmp1 * tmp2;

  b()(shape)(T(0, 0, 0)) = - dt * ty1 * dy1;
  b()(shape)(T(0, 1, 0)) =   0.0;
  b()(shape)(T(0, 2, 0)) = - dt * ty2;
  b()(shape)(T(0, 3, 0)) =   0.0;
  b()(shape)(T(0, 4, 0)) =   0.0;

  b()(shape)(T(1, 0, 0)) = - dt * ty2
    * ( - ( u()(shape - T(0, 1, 0))(T(1, 0, 0))*u()(shape - T(0, 1, 0))(T(2, 0, 0)) ) * tmp2 )
    - dt * ty1 * ( - c34 * tmp2 * u()(shape - T(0, 1, 0))(T(1, 0, 0)) );
  b()(shape)(T(1, 1, 0)) = - dt * ty2 * ( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 )
    - dt * ty1 * ( c34 * tmp1 )
    - dt * ty1 * dy2;
  b()(shape)(T(1, 2, 0)) = - dt * ty2 * ( u()(shape - T(0, 1, 0))(T(1, 0, 0)) * tmp1 );
  b()(shape)(T(1, 3, 0)) = 0.0;
  b()(shape)(T(1, 4, 0)) = 0.0;

  b()(shape)(T(2, 0, 0)) = - dt * ty2
    * ( - ( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 ) *( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 )
	+ 0.50 * C2 * ( (  u()(shape - T(0, 1, 0))(T(1, 0, 0)) * u()(shape - T(0, 1, 0))(T(1, 0, 0))
			   + u()(shape - T(0, 1, 0))(T(2, 0, 0)) * u()(shape - T(0, 1, 0))(T(2, 0, 0))
			   + u()(shape - T(0, 1, 0))(T(3, 0, 0)) * u()(shape - T(0, 1, 0))(T(3, 0, 0)) )
			* tmp2 ) )
    - dt * ty1 * ( - r43 * c34 * tmp2 * u()(shape - T(0, 1, 0))(T(2, 0, 0)) );
  b()(shape)(T(2, 1, 0)) = - dt * ty2
    * ( - C2 * ( u()(shape - T(0, 1, 0))(T(1, 0, 0)) * tmp1 ) );
  b()(shape)(T(2, 2, 0)) = - dt * ty2 * ( ( 2.0 - C2 )
				       * ( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 ) )
    - dt * ty1 * ( r43 * c34 * tmp1 )
    - dt * ty1 * dy3;
  b()(shape)(T(2, 3, 0)) = - dt * ty2
    * ( - C2 * ( u()(shape - T(0, 1, 0))(T(3, 0, 0)) * tmp1 ) );
  b()(shape)(T(2, 4, 0)) = - dt * ty2 * C2;

  b()(shape)(T(3, 0, 0)) = - dt * ty2
    * ( - ( u()(shape - T(0, 1, 0))(T(2, 0, 0))*u()(shape - T(0, 1, 0))(T(3, 0, 0)) ) * tmp2 )
    - dt * ty1 * ( - c34 * tmp2 * u()(shape - T(0, 1, 0))(T(3, 0, 0)) );
  b()(shape)(T(3, 1, 0)) = 0.0;
  b()(shape)(T(3, 2, 0)) = - dt * ty2 * ( u()(shape - T(0, 1, 0))(T(3, 0, 0)) * tmp1 );
  b()(shape)(T(3, 3, 0)) = - dt * ty2 * ( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 )
    - dt * ty1 * ( c34 * tmp1 )
    - dt * ty1 * dy4;
  b()(shape)(T(3, 4, 0)) = 0.0;

  b()(shape)(T(4, 0, 0)) = - dt * ty2
    * ( ( C2 * (  u()(shape - T(0, 1, 0))(T(1, 0, 0)) * u()(shape - T(0, 1, 0))(T(1, 0, 0))
		  + u()(shape - T(0, 1, 0))(T(2, 0, 0)) * u()(shape - T(0, 1, 0))(T(2, 0, 0))
		  + u()(shape - T(0, 1, 0))(T(3, 0, 0)) * u()(shape - T(0, 1, 0))(T(3, 0, 0)) ) * tmp2
	  - C1 * ( u()(shape - T(0, 1, 0))(T(4, 0, 0)) * tmp1 ) )
	* ( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 ) )
    - dt * ty1
    * ( - (     c34 - c1345 )*tmp3*u()(shape - T(0, 1, 0))(T(1, 0, 0)).pow2()
	- ( r43*c34 - c1345 )*tmp3*u()(shape - T(0, 1, 0))(T(2, 0, 0)).pow2()
	- (     c34 - c1345 )*tmp3*u()(shape - T(0, 1, 0))(T(3, 0, 0)).pow2()
	- c1345*tmp2*u()(shape - T(0, 1, 0))(T(4, 0, 0)) );
  b()(shape)(T(4, 1, 0)) = - dt * ty2
    * ( - C2 * ( u()(shape - T(0, 1, 0))(T(1, 0, 0))*u()(shape - T(0, 1, 0))(T(2, 0, 0)) ) * tmp2 )
    - dt * ty1
    * ( c34 - c1345 ) * tmp2 * u()(shape - T(0, 1, 0))(T(1, 0, 0));
  b()(shape)(T(4, 2, 0)) = - dt * ty2
    * ( C1 * ( u()(shape - T(0, 1, 0))(T(4, 0, 0)) * tmp1 )
	- 0.50 * C2 
	* ( (  u()(shape - T(0, 1, 0))(T(1, 0, 0))*u()(shape - T(0, 1, 0))(T(1, 0, 0))
	       + 3.0 * u()(shape - T(0, 1, 0))(T(2, 0, 0))*u()(shape - T(0, 1, 0))(T(2, 0, 0))
	       + u()(shape - T(0, 1, 0))(T(3, 0, 0))*u()(shape - T(0, 1, 0))(T(3, 0, 0)) ) * tmp2 ) )
    - dt * ty1
    * ( r43*c34 - c1345 ) * tmp2 * u()(shape - T(0, 1, 0))(T(2, 0, 0));
  b()(shape)(T(4, 3, 0)) = - dt * ty2
    * ( - C2 * ( u()(shape - T(0, 1, 0))(T(2, 0, 0))*u()(shape - T(0, 1, 0))(T(3, 0, 0)) ) * tmp2 )
    - dt * ty1 * ( c34 - c1345 ) * tmp2 * u()(shape - T(0, 1, 0))(T(3, 0, 0));
  b()(shape)(T(4, 4, 0)) = - dt * ty2
    * ( C1 * ( u()(shape - T(0, 1, 0))(T(2, 0, 0)) * tmp1 ) )
    - dt * ty1 * c1345 * tmp1
    - dt * ty1 * dy5;

  /*--------------------------------------------------------------------
    c   form the third block sub-diagonal
    --------------------------------------------------------------------*/
  tmp1 = 1.0 / u()(shape - T(1, 0, 0))(T(0, 0, 0));
  tmp2 = tmp1 * tmp1;
  tmp3 = tmp1 * tmp2;

  c()(shape)(T(0, 0, 0)) = - dt * tx1 * dx1;
  c()(shape)(T(0, 1, 0)) = - dt * tx2;
  c()(shape)(T(0, 2, 0)) =   0.0;
  c()(shape)(T(0, 3, 0)) =   0.0;
  c()(shape)(T(0, 4, 0)) =   0.0;

  c()(shape)(T(1, 0, 0)) = - dt * tx2
    * ( - ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 ) *( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 )
	+ C2 * 0.50 * (  u()(shape - T(1, 0, 0))(T(1, 0, 0)) * u()(shape - T(1, 0, 0))(T(1, 0, 0))
			 + u()(shape - T(1, 0, 0))(T(2, 0, 0)) * u()(shape - T(1, 0, 0))(T(2, 0, 0))
			 + u()(shape - T(1, 0, 0))(T(3, 0, 0)) * u()(shape - T(1, 0, 0))(T(3, 0, 0)) ) * tmp2 )
    - dt * tx1 * ( - r43 * c34 * tmp2 * u()(shape - T(1, 0, 0))(T(1, 0, 0)) );
  c()(shape)(T(1, 1, 0)) = - dt * tx2
    * ( ( 2.0 - C2 ) * ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 ) )
    - dt * tx1 * ( r43 * c34 * tmp1 )
    - dt * tx1 * dx2;
  c()(shape)(T(1, 2, 0)) = - dt * tx2
    * ( - C2 * ( u()(shape - T(1, 0, 0))(T(2, 0, 0)) * tmp1 ) );
  c()(shape)(T(1, 3, 0)) = - dt * tx2
    * ( - C2 * ( u()(shape - T(1, 0, 0))(T(3, 0, 0)) * tmp1 ) );
  c()(shape)(T(1, 4, 0)) = - dt * tx2 * C2;

  c()(shape)(T(2, 0, 0)) = - dt * tx2
    * ( - ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * u()(shape - T(1, 0, 0))(T(2, 0, 0)) ) * tmp2 )
    - dt * tx1 * ( - c34 * tmp2 * u()(shape - T(1, 0, 0))(T(2, 0, 0)) );
  c()(shape)(T(2, 1, 0)) = - dt * tx2 * ( u()(shape - T(1, 0, 0))(T(2, 0, 0)) * tmp1 );
  c()(shape)(T(2, 2, 0)) = - dt * tx2 * ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 )
    - dt * tx1 * ( c34 * tmp1 )
    - dt * tx1 * dx3;
  c()(shape)(T(2, 3, 0)) = 0.0;
  c()(shape)(T(2, 4, 0)) = 0.0;

  c()(shape)(T(3, 0, 0)) = - dt * tx2
    * ( - ( u()(shape - T(1, 0, 0))(T(1, 0, 0))*u()(shape - T(1, 0, 0))(T(3, 0, 0)) ) * tmp2 )
    - dt * tx1 * ( - c34 * tmp2 * u()(shape - T(1, 0, 0))(T(3, 0, 0)) );
  c()(shape)(T(3, 1, 0)) = - dt * tx2 * ( u()(shape - T(1, 0, 0))(T(3, 0, 0)) * tmp1 );
  c()(shape)(T(3, 2, 0)) = 0.0;
  c()(shape)(T(3, 3, 0)) = - dt * tx2 * ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 )
    - dt * tx1 * ( c34 * tmp1 )
    - dt * tx1 * dx4;
  c()(shape)(T(3, 4, 0)) = 0.0;

  c()(shape)(T(4, 0, 0)) = - dt * tx2
    * ( ( C2 * (  u()(shape - T(1, 0, 0))(T(1, 0, 0)) * u()(shape - T(1, 0, 0))(T(1, 0, 0))
		  + u()(shape - T(1, 0, 0))(T(2, 0, 0)) * u()(shape - T(1, 0, 0))(T(2, 0, 0))
		  + u()(shape - T(1, 0, 0))(T(3, 0, 0)) * u()(shape - T(1, 0, 0))(T(3, 0, 0)) ) * tmp2
	  - C1 * ( u()(shape - T(1, 0, 0))(T(4, 0, 0)) * tmp1 ) )
	* ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 ) )
    - dt * tx1
    * ( - ( r43*c34 - c1345 ) * tmp3 * u()(shape - T(1, 0, 0))(T(1, 0, 0)).pow2()
	- (     c34 - c1345 ) * tmp3 * u()(shape - T(1, 0, 0))(T(2, 0, 0)).pow2()
	- (     c34 - c1345 ) * tmp3 * u()(shape - T(1, 0, 0))(T(3, 0, 0)).pow2()
	- c1345 * tmp2 * u()(shape - T(1, 0, 0))(T(4, 0, 0)) );
 
  c()(shape)(T(4, 1, 0)) = - dt * tx2
    * ( C1 * ( u()(shape - T(1, 0, 0))(T(4, 0, 0)) * tmp1 )
	- 0.50 * C2
	* ( (  3.0*u()(shape - T(1, 0, 0))(T(1, 0, 0))*u()(shape - T(1, 0, 0))(T(1, 0, 0))
	       + u()(shape - T(1, 0, 0))(T(2, 0, 0))*u()(shape - T(1, 0, 0))(T(2, 0, 0))
	       + u()(shape - T(1, 0, 0))(T(3, 0, 0))*u()(shape - T(1, 0, 0))(T(3, 0, 0)) ) * tmp2 ) )
    - dt * tx1
    * ( r43*c34 - c1345 ) * tmp2 * u()(shape - T(1, 0, 0))(T(1, 0, 0));

  c()(shape)(T(4, 2, 0)) = - dt * tx2
    * ( - C2 * ( u()(shape - T(1, 0, 0))(T(2, 0, 0))*u()(shape - T(1, 0, 0))(T(1, 0, 0)) ) * tmp2 )
    - dt * tx1
    * (  c34 - c1345 ) * tmp2 * u()(shape - T(1, 0, 0))(T(2, 0, 0));
  
  c()(shape)(T(4, 3, 0)) = - dt * tx2
    * ( - C2 * ( u()(shape - T(1, 0, 0))(T(3, 0, 0))*u()(shape - T(1, 0, 0))(T(1, 0, 0)) ) * tmp2 )
    - dt * tx1
    * (  c34 - c1345 ) * tmp2 * u()(shape - T(1, 0, 0))(T(3, 0, 0));
  
  c()(shape)(T(4, 4, 0)) = - dt * tx2
    * ( C1 * ( u()(shape - T(1, 0, 0))(T(1, 0, 0)) * tmp1 ) )
    - dt * tx1 * c1345 * tmp1
    - dt * tx1 * dx5;
}
