int reminder_x;
int reminder_y;
int reminder_z;

#include "flux.h"
#include "exact.h"
#include "globals.h"
#include "blts.h"
#include "buts.h"
//#include "exact_hta.h"

#define T_total 0
#define T_rhs   1
#define T_blts  2
#define T_buts  3
#define T_exch1 4
#define T_exch2 5
#define T_misc1 6
#define T_misc2 7
#define T_misc3 8
#define T_misc4 9
#define T_misc5 10
#define T_wait 11
#define T_comm1  12
#define T_comm2  13
#define T_comm3  14
#define T_comm4  15
#define T_comm5  16
#define T_comm6  17
#define T_zero  18

HTA_2H uInner;
HTA_2H rsdInner;
HTA_2H fluxInner;
HTA_2H frctInner;

/* print and debug methods */

void print_timers ()

{
  char* name [] = { "total", "rhs", "blts", "buts", "exch1", "exch2", "misc1", "misc2", "misc3", "misc4",
		    "misc5", "wait", "comm1", "comm2", "comm3", "comm4", "comm5", "comm6", "zero"};
  
  for (int i = T_total; i <= T_zero; i++)
    IO::cout << name[i] << " : " << timer_read (i) << endl;
}



int nPlanes;


static void verify(double xcr[5], double xce[5], double xci,
		   char *Class, bool *verified);

template<typename TYPE, int DIM>
struct SetBV_E :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  SetBV_E() {}

  SetBV_E (int level):
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {}

  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > h) 
  {        
    Tuple<DIM> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
  
    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) 
      {
	tpos[1] = spos[1];
	for (int j = 0; j < size[1]; j++, tpos[1]++) {
	  tpos[2] = spos[2];    
	  for (int k = 0; k < size[2]; k++, tpos[2]++) { 
	    
	    exact (ISIZ1-1, tpos[1], tpos[2], 
		   i, j, k,		   
		   h);
	  }    
	}
      }
  }
  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
    
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }

};

template<typename TYPE, int DIM>
struct SetBV_W :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  SetBV_W() {}

  SetBV_W (int level):
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {}

  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > h) 
  {    
    Tuple<DIM> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
    
    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) 
      {
	tpos[1] = spos[1];
	for (int j = 0; j < size[1]; j++, tpos[1]++) {
	  tpos[2] = spos[2];    	
	  for (int k = 0; k < size[2]; k++, tpos[2]++) { 
	    exact (0, tpos[1], tpos[2],
		   i, j, k,
		   h);
	  }
	}
        
      }
  }

  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
    
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }

};


template<typename TYPE, int DIM>
struct SetBV_N :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  SetBV_N() {}

  SetBV_N (int level):
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {}

  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > h) 
  {    
    Tuple<DIM> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
    
    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) 
      {
	tpos[1] = spos[1];
	for (int j = 0; j < size[1]; j++, tpos[1]++) {
	  tpos[2] = spos[2];    
	  for (int k = 0; k < size[2]; k++, tpos[2]++) { 
	    exact (tpos[0], 0, tpos[2], 
		   i, j, k,
		   h);
	  }
	}
      }
  }

  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
    
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }

};

template<typename TYPE, int DIM>
struct SetBV_S :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  SetBV_S() {}

  SetBV_S (int level):
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {}

  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > h) 
  {    
    Tuple<DIM> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
    
    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) 
      {
	tpos[1] = spos[1];
	for (int j = 0; j < size[1]; j++, tpos[1]++) {
	  tpos[2] = spos[2];    	
	  for (int k = 0; k < size[2]; k++, tpos[2]++) { 
	    exact (tpos[0], ISIZ2-1, tpos[2], 
		   i, j, k,		   
		   h);
	  }
	}
      }
  }

  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
    
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }

};

template<typename TYPE, int DIM>
struct SetBV_F :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  SetBV_F() {}

  SetBV_F (int level):
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {}

  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > h) 
  {    

    Tuple<DIM> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);

    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) 
      {
	tpos[1] = spos[1];
	for (int j = 0; j < size[1]; j++, tpos[1]++) {
	  tpos[2] = spos[2];
	  for (int k = 0; k < size[2]; k++, tpos[2]++) { 
	    exact (tpos[0], tpos[1],0, 
		   i, j, k,
		   h);
	  }
	}
      }
  }

  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
    
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }

};

template<typename TYPE, int DIM>
struct SetBV_B :  public OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  SetBV_B () {}

  SetBV_B (int level):
    OperatorHTA::MapReduceHTA<TYPE, TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {}

  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > h) 
  {    
    Tuple<DIM> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
    
    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) 
      {
	tpos[1] = spos[1];
	for (int j = 0; j < size[1]; j++, tpos[1]++) {
	  tpos[2] = spos[2];	
	  for (int k = 0; k < size[2]; k++, tpos[2]++) { 	
	    exact (tpos[0], tpos[1], ISIZ3-1,
		   i, j, k,
		   h);
	  }
	}
      }
  }

  virtual const TYPE& result () const {} 
  
  bool doReduce () { return false; }
    
  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }

};

struct SetIV 
{
  void operator() (HTA_2H h) 
  {
    double  pxi, peta, pzeta;
    double  ue_1jk[5],ue_nx0jk[5],ue_i1k[5],
      ue_iny0k[5],ue_ij1[5],ue_ijnz[5];
      
    int iglob, jglob, kglob;

    Tuple<4> size = h.shape().size();
    Tuple<4> spos = h.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
    //cout << tmp << spos << endl;

    Tuple<4> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) {
      tpos[1] = spos[1];
      for (int j = 0; j < size[1]; j++, tpos[1]++) {
	tpos[2] = spos[2];
	for (int k = 0; k < size[2]; k++, tpos[2]++) { 

	  iglob = tpos[0];
	  jglob = tpos[1];
	  kglob = tpos[2];
	  
	  //cout << "GBDEBUG " << tpos << endl;

	  double zeta = (double) iglob / (double) (ISIZ1 - 1);
	  double eta =  (double) jglob / (double) (ISIZ2 - 1);
	  double xi =   (double) kglob / (double) (ISIZ3 - 1);   
	  
	  exact (iglob, jglob, 0, ue_1jk);
	  exact (iglob,jglob, ISIZ3-1, ue_nx0jk);
	  exact (iglob,0,kglob, ue_i1k);
	  exact (iglob,ISIZ2-1,kglob, ue_iny0k);
	  exact (0,jglob, kglob, ue_ij1);
	  exact (ISIZ1-1,jglob, kglob, ue_ijnz);
	  
	  for (int m = 0; m < 5; m++) {
	    pxi =   ( 1.0 - xi ) * ue_1jk[m]
	      + xi   * ue_nx0jk[m];
	    peta =  ( 1.0 - eta ) * ue_i1k[m]
	      + eta   * ue_iny0k[m];
	    pzeta = ( 1.0 - zeta ) * ue_ij1[m]
	      + zeta   * ue_ijnz[m];
	    
	    //cout << " here " << Tuple<3>(m, 0, 0) << " " <<  h.shape() << endl;
	    
	    double tmp = pxi + peta + pzeta
	      - pxi * peta - peta * pzeta - pzeta * pxi
	      + pxi * peta * pzeta;
	    
	    h[T(i, j, k, m)] = tmp;
	    // cout << tmp << endl;
	  }
	}
      }
    }
  }
  
};

struct RSD
{
  void operator () (HTA_2H rsd)
  {    
    Tuple<4> size = rsd.shape().size();
    Tuple<4> spos = rsd.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);

    Tuple<4> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) {
      tpos[1] = spos[1];
      for (int j = 0; j < size[1]; j++, tpos[1]++) {	
	tpos[2] = spos[2];
	for (int k = 0; k < size[2]; k++, tpos[2]++) { 
	  
	  int iglob = tpos[0];
	  int jglob = tpos[1];
	  int kglob = tpos[2];
	  
	  double xi = ( (double)(iglob) ) / (double) (ISIZ1 - 1 );
	  double eta = ( (double)(jglob) ) / (double) (ISIZ2 - 1 );
	  double zeta = ( (double)(kglob) ) / (double) (ISIZ3- 1 );
	  
	  for (int m = 0; m < 5; m++) {

	    rsd[T(i, j, k, m)] =  ce[m][0]
	      + ce[m][1] * xi
	      + ce[m][2] * eta
	      + ce[m][3] * zeta
	      + ce[m][4] * xi * xi
	      + ce[m][5] * eta * eta
	      + ce[m][6] * zeta * zeta
	      + ce[m][7] * xi * xi * xi
	      + ce[m][8] * eta * eta * eta
	      + ce[m][9] * zeta * zeta * zeta
	      + ce[m][10] * xi * xi * xi * xi
	      + ce[m][11] * eta * eta * eta * eta
	      + ce[m][12] * zeta * zeta * zeta * zeta;
	  }	  	
	}
      }
    }
  }
};
  
struct dVector
{
  double val[5];
};

int tcount = 0;

/**
 * map-reduce operator to compute the square sum and find maxiumum
 */
template<typename TYPE, int DIM>
struct Error : public  OperatorHTA::MapReduceHTA<TYPE, dVector, DIM, Traits::HTA_DEFAULT_TRAIT<double, 4> >  
{
  Error () {}
  
  void init (MapContext<TYPE, DIM>& context)
  {
    this->context_ = context;
  }
  
  Error(int level):
    OperatorHTA::MapReduceHTA<TYPE, dVector, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> > (level) {
    for  (int i = 0; i < 5; i++)
      sqsum.val[i] = 0.0;
  }
  
  void reduce (const dVector & d) 
  { 
    for (int i = 0; i < 5; i++)
      sqsum.val[i] += d.val[i];
  }
  
  void map_hta_tuple (const Tuple<DIM>& pos, HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 4> >  u) {
    
    double u0[5];    
    
    tcount++;

    Tuple<DIM> size = u.shape().size();
    Tuple<4> spos = u.memMap().leafPos();
    Tuple<4> tmp = spos / Tuple<4>(isiz1+4, isiz2+4, isiz3+4, 1);
    spos = spos - (tmp * 4)  - Tuple<4>(2,2,2,0);
  
    //    cout << "here " << spos  << " " <<  tmp << endl;

    Tuple<DIM> tpos = spos;
    for (int i = 0; i < size[0]; i++, tpos[0]++) {
      tpos[1] = spos[1];
      for (int j = 0; j < size[1]; j++, tpos[1]++) {	
	tpos[2] = spos[2];
	for (int k = 0; k < size[2]; k++, tpos[2]++) { 	
	  int iglob = tpos[0];
	  int jglob = tpos[1];
	  int kglob = tpos[2];  
	  exact(iglob, jglob, kglob, u0);
	  for (int m = 0; m < 5; m++) {
	    double tmp = u0[m] - u[T(i, j, k, m)];
	    sqsum.val[m] +=  tmp * tmp;
	  }
	}
      }
    }
  }

  virtual const dVector& result () const 
  {
    return sqsum;
  }

  bool doReduce () { return true; }

  OperatorHTA::MapReduceKey keyType () const { return OperatorHTA::MR_TUPLE; }
  
  dVector sqsum;
}; // struct Error



/*--------------------------------------------------------------------
--------------------------------------------------------------------*/

static void setcoeff(void) 
{
  
  /*--------------------------------------------------------------------
    c   set up coefficients
    --------------------------------------------------------------------*/
  
  dxi = 1.0 / (double) ( ISIZ1 - 1 );
  deta = 1.0 / (double) ( ISIZ2 - 1 );
  dzeta = 1.0 / (double) ( ISIZ3 - 1 );
  
  tx1 = 1.0 / (double) ( dxi * dxi );
  tx2 = 1.0 / (double) ( 2.0 * dxi );
  tx3 = 1.0 / (double) dxi;

  ty1 = 1.0 / (double) ( deta * deta );
  ty2 = 1.0 / (double) ( 2.0 * deta );
  ty3 = 1.0 / (double) deta;

  tz1 = 1.0 / (double) ( dzeta * dzeta );
  tz2 = 1.0 / (double) ( 2.0 * dzeta );
  tz3 = 1.0 / (double) dzeta;

  ii1 = 1;
  ii2 = ISIZ1 - 2;
  ji1 = 1;
  ji2 = ISIZ2 - 3;
  ki1 = 2;
  ki2 = ISIZ3 - 2;

/*--------------------------------------------------------------------
c   diffusion coefficients
--------------------------------------------------------------------*/
  dx1 = 0.75;
  dx2 = dx1;
  dx3 = dx1;
  dx4 = dx1;
  dx5 = dx1;

  dy1 = 0.75;
  dy2 = dy1;
  dy3 = dy1;
  dy4 = dy1;
  dy5 = dy1;

  dz1 = 1.00;
  dz2 = dz1;
  dz3 = dz1;
  dz4 = dz1;
  dz5 = dz1;

/*--------------------------------------------------------------------
c   fourth difference dissipation
--------------------------------------------------------------------*/
  dssp = ( max (dx1, max(dy1, dz1) ) ) / 4.0;

/*--------------------------------------------------------------------
c   coefficients of the exact solution to the first pde
--------------------------------------------------------------------*/
  ce[0][0] = 2.0;
  ce[0][1] = 0.0;
  ce[0][2] = 0.0;
  ce[0][3] = 4.0;
  ce[0][4] = 5.0;
  ce[0][5] = 3.0;
  ce[0][6] = 5.0e-01;
  ce[0][7] = 2.0e-02;
  ce[0][8] = 1.0e-02;
  ce[0][9] = 3.0e-02;
  ce[0][10] = 5.0e-01;
  ce[0][11] = 4.0e-01;
  ce[0][12] = 3.0e-01;

/*--------------------------------------------------------------------
c   coefficients of the exact solution to the second pde
--------------------------------------------------------------------*/
  ce[1][0] = 1.0;
  ce[1][1] = 0.0;
  ce[1][2] = 0.0;
  ce[1][3] = 0.0;
  ce[1][4] = 1.0;
  ce[1][5] = 2.0;
  ce[1][6] = 3.0;
  ce[1][7] = 1.0e-02;
  ce[1][8] = 3.0e-02;
  ce[1][9] = 2.0e-02;
  ce[1][10] = 4.0e-01;
  ce[1][11] = 3.0e-01;
  ce[1][12] = 5.0e-01;

/*--------------------------------------------------------------------
c   coefficients of the exact solution to the third pde
--------------------------------------------------------------------*/
  ce[2][0] = 2.0;
  ce[2][1] = 2.0;
  ce[2][2] = 0.0;
  ce[2][3] = 0.0;
  ce[2][4] = 0.0;
  ce[2][5] = 2.0;
  ce[2][6] = 3.0;
  ce[2][7] = 4.0e-02;
  ce[2][8] = 3.0e-02;
  ce[2][9] = 5.0e-02;
  ce[2][10] = 3.0e-01;
  ce[2][11] = 5.0e-01;
  ce[2][12] = 4.0e-01;

/*--------------------------------------------------------------------
c   coefficients of the exact solution to the fourth pde
--------------------------------------------------------------------*/
  ce[3][0] = 2.0;
  ce[3][1] = 2.0;
  ce[3][2] = 0.0;
  ce[3][3] = 0.0;
  ce[3][4] = 0.0;
  ce[3][5] = 2.0;
  ce[3][6] = 3.0;
  ce[3][7] = 3.0e-02;
  ce[3][8] = 5.0e-02;
  ce[3][9] = 4.0e-02;
  ce[3][10] = 2.0e-01;
  ce[3][11] = 1.0e-01;
  ce[3][12] = 3.0e-01;

/*--------------------------------------------------------------------
c   coefficients of the exact solution to the fifth pde
--------------------------------------------------------------------*/
  ce[4][0] = 5.0;
  ce[4][1] = 4.0;
  ce[4][2] = 3.0;
  ce[4][3] = 2.0;
  ce[4][4] = 1.0e-01;
  ce[4][5] = 4.0e-01;
  ce[4][6] = 3.0e-01;
  ce[4][7] = 5.0e-02;
  ce[4][8] = 4.0e-02;
  ce[4][9] = 3.0e-02;
  ce[4][10] = 1.0e-01;
  ce[4][11] = 3.0e-01;
  ce[4][12] = 2.0e-01;
}

void setiv (HTA_1H u)
{
   //set initial values
  // SetIV<double, 4> setiv (1);
  HTA_1H t = u()((R(2, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  //  t.mapReduceHTA (setiv);
  t.hmap (SetIV(), 0);
}

void setbv (HTA_1H u)
{

   //set boundary values;
  SetBV_E<double, 4> setbv_e (1);
  SetBV_W<double, 4> setbv_w (1);
  SetBV_N<double, 4> setbv_n (1);
  SetBV_S<double, 4> setbv_s (1);
  SetBV_F<double, 4> setbv_f (1);
  SetBV_B<double, 4> setbv_b (1);

  /* a really good usage of HTAs */
  /* marred by the poor syntax and several mapReduce classes (see above) */

  //north face
  ((HTA_1H) u((R(0, 0), R(0, np2-1), R(0, np3-1), R(0,0)))((R(2, 2), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)))).mapReduceHTA (setbv_w);

  //south face
  ((HTA_1H) u((R(np1-1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+1+reminder_x, isiz1+1+reminder_x), R(2, isiz2+1), R(2, isiz3+1),  R(0, 4)))).mapReduceHTA (setbv_e);

  //east face
  ((HTA_1H) u((R(0, np1-1), R(0, 0), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(2, 2), R(2, isiz3+1),  R(0, 4)))).mapReduceHTA (setbv_n);

  //west face
  ((HTA_1H) u((R(0, np1-1), R(np2-1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+1+reminder_y, isiz2+1+reminder_y), R(2, isiz3+1),  R(0, 4)))).mapReduceHTA (setbv_s);

  //front face
  ((HTA_1H) u((R(0, np1-1), R(0, np2-1), R(0, 0), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(2, 2),  R(0, 4)))).mapReduceHTA (setbv_f);

  //back face
  ((HTA_1H) u((R(0, np1-1), R(0, np2-1), R(np3-1, np3-1), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3+1+reminder_z, isiz3+1+reminder_z),  R(0, 4)))).mapReduceHTA (setbv_b);

}

void initRsdViews (HTA_1H rsd)
{
  
  rsdShadowSouth = rsd((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+2, isiz1+3), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  rsdBorderNorth =   rsd((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(2, 3), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  
  rsdShadowNorth = rsd((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(0, 1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  rsdBorderSouth =   rsd((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
 

  rsdShadowWest = rsd((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+2, isiz2+3), R(2, isiz3+1), R(0, 4)));
  rsdBorderEast =   rsd((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(2, 3), R(2, isiz3+1), R(0, 4)));
  
  rsdShadowEast = rsd((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(0, 1), R(2, isiz3+1), R(0, 4)));
  rsdBorderWest =   rsd((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2, isiz2+1), R(2, isiz3+1), R(0, 4)));  

  rsdShadowFront = rsd((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3+2, isiz3+3), R(0, 4)));
  rsdBorderBack =   rsd((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1),  R(2, isiz2+1), R(2, 3), R(0, 4)));
  
  rsdShadowBack = rsd((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(0, 1), R(0, 4)));
  rsdBorderFront =   rsd((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3, isiz3+1), R(0, 4)));
  
  //rsdShadowSouth2 = rsd((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(1, 1), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4)));
  //rsdBorderNorth2 = rsd((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+1, isiz1+1), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4)));

//   rsdShadowEast2 = rsd((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(1, 1), R(2 ,isiz3+1), R(0, 4)));
//   rsdBorderWest2 = rsd((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+1, isiz2+1), R(2 ,isiz3+1), R(0, 4)));

//   rsdShadowBack2 = rsd((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2 ,isiz2+1), R(1,1), R(0, 4)));
//   rsdBorderFront2 =  rsd((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3+1, isiz3+1), R(0, 4)));
}


void initUViews (HTA_1H rsd)
{
  
  uShadowSouth = rsd((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+2, isiz1+3), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  uBorderNorth =   rsd((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(2, 3), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  
  uShadowNorth = rsd((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(0, 1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  uBorderSouth =   rsd((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
 

  uShadowWest = rsd((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+2, isiz2+3), R(2, isiz3+1), R(0, 4)));
  uBorderEast =   rsd((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(2, 3), R(2, isiz3+1), R(0, 4)));
  
  uShadowEast = rsd((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(0, 1), R(2, isiz3+1), R(0, 4)));
  uBorderWest =   rsd((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2, isiz2+1), R(2, isiz3+1), R(0, 4)));  

  uShadowFront = rsd((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3+2, isiz3+3), R(0, 4)));
  uBorderBack =   rsd((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1),  R(2, isiz2+1), R(2, 3), R(0, 4)));
  
  uShadowBack = rsd((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(0, 1), R(0, 4)));
  uBorderFront =   rsd((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3, isiz3+1), R(0, 4))); 
}

void erhs (HTA_1H rsd, 
	   HTA_1H frct, 
	   HTA_1H flux)
{  
  //  RSD<double, 4> rsdMap(1);
  HTA_1H tmp0 = rsd()((R(2, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  //tmp0.mapReduceHTA (rsdMap);
  tmp0.hmap (RSD(), 0);

  rsdShadowSouth = rsdBorderNorth;
  rsdShadowNorth = rsdBorderSouth;

  frct = 0.0;

  erhs1 (rsdInner, frctInner, fluxInner, rsd, frct, flux); //HTA version
    
  rsdShadowEast = rsdBorderWest;
  rsdShadowWest = rsdBorderEast;
 
  erhs2 (rsdInner, frctInner, fluxInner, rsd, frct, flux); //HTA version

  rsdShadowBack = rsdBorderFront;
  rsdShadowFront = rsdBorderBack;
  erhs3 (rsdInner, frctInner, fluxInner, rsd, frct, flux); //HTA version

}


void rhs (HTA_1H u, HTA_1H rsd, HTA_1H flux, HTA_1H frct)
{

  rsd =  frct;
  -rsd;
 
  uShadowSouth = uBorderNorth;
  uShadowNorth = uBorderSouth;

  erhs1 (uInner, rsdInner, fluxInner, u, rsd, flux); 

  uShadowEast = uBorderWest;
  uShadowWest = uBorderEast;
    

  erhs2 (uInner, rsdInner, fluxInner, u, rsd, flux); //HTA version
  
  uShadowBack = uBorderFront;
  uShadowFront = uBorderBack;

  erhs3 (uInner, rsdInner, fluxInner, u, rsd, flux); //HTA version  
}

void fill (bool* ind, int m)
{
  
  for (int i = 0; i < np1; i++)
    for (int j = 0; j < np2; j++)
      for (int k = 0; k < np3; k++)
	{
	  //	  assert (i*np2*np3 + j * np3 + k < 8);
	  if (i + j + k == m)
	    ind[i*np2*np3 + j * np3 + k] = true;
	  else
	    ind[i*np2*np3 + j * np3 + k] = false;
	}  
}

void print (bool* ind)
{      
  for (int i = 0; i < np1; i++)
    for (int j = 0; j < np2; j++)
      for (int k = 0; k < np3; k++)
	{
	  cout <<  ind[i*np2*np3 + j * np3 + k] << endl;
	}
}

// void exchange_1 (HTA_2H h, HTA_2H v)
// {
  
//   timer_start (T_comm1);
//   //  cout <<  "Here 1" << endl;
//   v((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(1, 1), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4))) =
//     h((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+1, isiz1+1), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4)));
//   timer_stop (T_comm1);
//   //cout << "Here 3 " << endl;

//   timer_start (T_wait); 
//   //MPI_Barrier (MPI_COMM_WORLD);
//   timer_stop (T_wait); 
  
//   timer_start (T_comm2);
//   v((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(1, 1), R(2 ,isiz3+1), R(0, 4))) =
//     h((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+1, isiz2+1), R(2 ,isiz3+1), R(0, 4)));
//   timer_stop (T_comm2);

//   timer_start (T_wait); 
//   //MPI_Barrier (MPI_COMM_WORLD);
//   timer_stop (T_wait); 
  
//   timer_start (T_comm3);
//   v((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2 ,isiz2+1), R(1,1), R(0, 4))) =
//     h((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(isiz3+1, isiz3+1), R(0, 4)));
//   timer_stop (T_comm3);
  
//   timer_start (T_wait); 
//   //MPI_Barrier (MPI_COMM_WORLD);
//   timer_stop (T_wait); 
// }

void exchange_1 (HTA_2H hyperplane1[6], HTA_2H hyperplane2[6])
{
  hyperplane1[0] = hyperplane2[1];

  hyperplane1[2] = hyperplane2[3];

  hyperplane1[4] = hyperplane2[5];
}

// void exchange_2 (HTA_1H h, HTA_1H v)
// {
//   timer_start (T_comm4);
//   v((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+2, isiz1+2), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4))) =
//     h((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(2, 2), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4)));
//   timer_stop (T_comm4);
  
//   timer_start (T_wait); 
//   //MPI_Barrier (MPI_COMM_WORLD);
//   timer_stop (T_wait); 
  
//   timer_start (T_comm5);
//   v((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+2, isiz2+2), R(2 ,isiz3+1), R(0, 4))) =
//     h((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(2, 2), R(2 ,isiz3+1), R(0, 4)));
//   timer_stop (T_comm5);
  
//   timer_start (T_wait); 
//   //MPI_Barrier (MPI_COMM_WORLD);
//   timer_stop (T_wait); 

//   timer_start (T_comm6);
//   v((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2 ,isiz2+1), R(isiz3+2, isiz3+2), R(0, 4))) =
//     h((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(2, 2), R(0, 4)));
//   timer_stop (T_comm6);

//   timer_start (T_wait); 
//   //MPI_Barrier (MPI_COMM_WORLD);
//   timer_stop (T_wait); 
// }

void zero (HTA_1H h)
{  
  h((R(0, 0), R(0, np2-1), R(0, np3-1), R(0,0)))((R(2,2), R(2, isiz2+1), R(2, isiz3+1), R(0, 4))) = 0.0;
  h((R(np1-1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+1,isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4))) = 0.0;
  
  h((R(0, np1-1), R(0, 0), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(2,2), R(2, isiz3+1), R(0, 4))) = 0.0;
  h((R(0, np1-1), R(np2-1, np2-1), R(0, np3-1), R(0,0)))((R(2,isiz1+1), R(isiz2+1, isiz2+1), R(2, isiz3+1), R(0, 4))) = 0.0;
 
  h((R(0, np1-1), R(0, np2-1), R(0, 0), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(2, 2), R(0, 4))) = 0.0;
  h((R(0, np1-1), R(0, np2-1), R(np3-1, np3-1), R(0,0)))((R(2,isiz1+1), R(2, isiz2+1), R(isiz3+1, isiz3+1), R(0, 4))) = 0.0;
}

dVector error (HTA_1H u)
{
  int L1 = 2;
  int L2 = isiz1 + 1;
  int L3 = 2;
  int L4 = isiz2 + 1;
  int L5 = 2;
  int L6 = isiz3 + 1;

  //  cout << "Hi " << isiz1 << " " << isiz2 << " " << isiz3 << endl;

  Error<double, 4> err(1);
  ShapeH<4> shape2(T(np1, np2, np3, 1));
  shape2 =  Shape<4> ((R(L1, L2), R(L3, L4),  R(L5, L6), R(0, 4)));
  int sx = (L2 - L1 + 1)*np1 + reminder_x;
  int sy = (L4 - L3 + 1)*np2 + reminder_y;
  int sz = (L6 - L5 + 1)*np3 + reminder_z;
  ShapeH<4> shapeInner = shape2.intersect (Shape<4>((R(0, sx-1), R(0, sy-1), R(0,sz-1), R(0,4))));
  HTA_1H uInner = u()(shapeInner);
  uInner.mapReduceHTA (err);

  //zero (u);
  //HTA_1H uInner = u()((R(2, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4)));
  //uInner.mapReduceHTA (err);
  //cout <<"hello " << tcount << " " << isiz1 << uInner.memMap().leafSize() << endl;
  return err.result();
}

void ssor (HTA_1H u, 
	   HTA_1H rsd, 
	   HTA_1H flux, 
	   HTA_1H frct)
{
  r43 = ( 4.0 / 3.0 );
  c1345 = C1 * C3 * C4 * C5;
  c34 = C3 * C4;
 
  nPlanes = np1 + np2 + np3 - 2;
  const int size = np1 * np2 * np3;
  bool* ind  = new bool [size];
  bool* indN = new bool [size];
    
  double tmp = 1.0 / (omega * (2.0 - omega));
  
  timer_start (T_rhs);
  rhs (u, rsd, flux, frct);
  timer_stop (T_rhs);

  //pre-compute the shadow regions of all the hyperplanes;
  HTA_1H hyperplane_blts[nPlanes][6];
  for (int m = 0; m < nPlanes; m++) 
    {
      fill (ind, m);
      
      hyperplane_blts[m][0] = rsd(ind)((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(1, 1), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4))) ;
      hyperplane_blts[m][1] = rsd(ind)((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+1, isiz1+1), R(2, isiz2+1), 
										    R(2 ,isiz3+1), R(0, 4)));
      
      hyperplane_blts[m][2] = rsd(ind)((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(1, 1), R(2 ,isiz3+1), R(0, 4)));
      hyperplane_blts[m][3] = rsd(ind)((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+1, isiz2+1), 
										  R(2 ,isiz3+1), R(0, 4)));

      hyperplane_blts[m][4] =   rsd(ind)((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2 ,isiz2+1), R(1,1), R(0, 4)));
      hyperplane_blts[m][5] =   rsd(ind)((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), 
										      R(isiz3+1, isiz3+1), R(0, 4)));   
    }
  
  //pre-compute the shadow regions of all the hyperplanes;
  HTA_1H hyperplane_buts[nPlanes][6];
  for (int m = nPlanes-1; m >=0 ; m--) 
    {
      fill (ind, m);
      
      hyperplane_buts[m][0] = rsd(ind)((R(0, np1-2), R(0, np2-1), R(0, np3-1), R(0,0)))((R(isiz1+2, isiz1+2), R(2, isiz2+1), 
										  R(2 ,isiz3+1), R(0, 4)));
      hyperplane_buts[m][1] = rsd(ind)((R(1, np1-1), R(0, np2-1), R(0, np3-1), R(0,0)))((R(2, 2), R(2, isiz2+1), R(2 ,isiz3+1), R(0, 4)));
      
      hyperplane_buts[m][2] = rsd(ind)((R(0, np1-1), R(0, np2-2), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(isiz2+2, isiz2+2), 
											 R(2 ,isiz3+1), R(0, 4)));
      hyperplane_buts[m][3] = rsd(ind)((R(0, np1-1), R(1, np2-1), R(0, np3-1), R(0,0)))((R(2, isiz1+1), R(2, 2), R(2 ,isiz3+1), R(0, 4)));

      hyperplane_buts[m][4] = rsd(ind)((R(0, np1-1), R(0, np2-1), R(0, np3-2), R(0,0)))((R(2, isiz1+1), R(2 ,isiz2+1), R(isiz3+2, isiz3+2), R(0, 4))); 
      hyperplane_buts[m][5] = rsd(ind)((R(0, np1-1), R(0, np2-1), R(1, np3-1), R(0,0)))((R(2, isiz1+1), R(2, isiz2+1), R(2, 2), R(0, 4)));
    }


  timer_start (T_total);
  //itmax = 1;
  //int rank = 0;
  for (int i = 0; i < itmax; i++) 
    { 
      timer_start (T_misc1);
      rsd *= dt;
      timer_stop (T_misc1);

      for (int m = 0; m < nPlanes; m++)
	{	  
	  //	  cout << "iteration  " << m << " " << rank << endl;      
	  fill (ind, m);
	  timer_start (T_blts);
	  //blts (rsd, u, ind);
	  rsdInner.hmap (JacldBlts(),
	  	 uInner,
	  	 0,
	  	 ind);
	  timer_stop (T_blts);

          timer_start (T_wait); 
          //MPI_Barrier (MPI_COMM_WORLD);
          timer_stop (T_wait); 

	  timer_start (T_misc2);
	  fill (indN, m+1);		  
	  timer_stop (T_misc2);

          timer_start (T_wait); 
          //MPI_Barrier (MPI_COMM_WORLD);
          timer_stop (T_wait); 

	  timer_start (T_exch1);
	  if (m < (nPlanes - 1)) 
	    {
	      Traits::Default::async();
	      exchange_1 (hyperplane_blts[m+1], hyperplane_blts[m]);
	      //exchange_1 (rsd(ind), rsd(indN));	
	      //exchange_1 (rsd, rsd);	
	      Traits::Default::sync();	      
	    }
	  timer_stop (T_exch1);	

          timer_start (T_wait); 
          //MPI_Barrier (MPI_COMM_WORLD);
          timer_stop (T_wait); 
	}
      
      for (int m = nPlanes - 1; m >= 0; m--)
	{	  
	  // cout << "iteration  " << m << endl;      
	  
	  timer_start (T_misc3);
	  fill (ind, m);
	  timer_stop (T_misc3);
	  
	  //jacu (u, a, b, c, d, ind);
	  
	  //buts (rsd, d, a, b, c, ind);
	  timer_start (T_buts);
	  //buts (rsd, u, ind);
	  rsdInner.hmap (JacuButs(),
			 uInner,
			 0,
			 ind);
	  timer_stop (T_buts);
	 
          timer_start (T_wait); 
          //MPI_Barrier (MPI_COMM_WORLD);
          timer_stop (T_wait); 

	  timer_start (T_misc4);
	  fill (indN, m-1);	
	  timer_stop (T_misc4);
	  
          timer_start (T_wait); 
          //MPI_Barrier (MPI_COMM_WORLD);
          timer_stop (T_wait); 

	  timer_start (T_exch2);
	  if (m > 0) 
	    {
              Traits::Default::async();
	      exchange_1(hyperplane_buts[m-1], hyperplane_buts[m]);
	      //exchange_2 (rsd(ind), rsd(indN));	
	      //exchange_2 (rsd, rsd);	
              Traits::Default::sync();
	    }
	  timer_stop (T_exch2);
	
          timer_start (T_wait); 
          //MPI_Barrier (MPI_COMM_WORLD);
          timer_stop (T_wait); 
	}

      timer_start (T_wait); 
      //MPI_Barrier (MPI_COMM_WORLD);
      timer_stop (T_wait); 

      timer_start (T_misc5);
      u += rsd * tmp ;
      timer_stop (T_misc5);
      
      timer_start (T_rhs);
      rhs (u, rsd, flux, frct);      
      timer_stop (T_rhs);

      timer_start (T_zero);
      zero (rsd);
      timer_stop (T_zero);           
    }

  timer_stop (T_total);
  
  delete [] ind;
  delete [] indN;
  //cout << u()((R(2, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4))) << endl;
    
}

void initShape ()
{
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = np1;
  int ty = np2;
  int tz = np3;
  int sx, sy, sz;

  ShapeH<4> shape1(T(tx, ty, tz, 1));
  shape1 =  Shape<4> ((R(ist, iend), R(jst, jend),  R(kst, kend), R(0, 4)));
  sx = (iend - ist + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend - kst + 1)*tz;
  iShape = shape1.intersect (Shape<4>((R(1, sx-2), R(1,sy-2), R(1,sz-2), R(0,4))));
  
  iShapeB  = iShape + T(0, 0, 1,0);
  iShapeE  = iShape + T(0, 1, 0,0);
  iShapeS  = iShape + T(1, 0, 0,0);
 
  iShapeF  = iShape - T(0, 0, 1,0);
  iShapeW  = iShape - T(0, 1, 0,0);
  iShapeN  = iShape - T(1, 0, 0,0);

}

int run(char** argv)
{
  //sleep (20);

  ipr = IPR_DEFAULT;
  inorm = INORM_DEFAULT;
  itmax = ITMAX_DEFAULT;
  dt = DT_DEFAULT;
  omega = OMEGA_DEFAULT;
  
  int fnum = max (Traits::Default::nPlaces(), 32);
  int ndim = log((double) fnum)/log(2.0) + 0.00001;
  np1   = pow(2.0, (double) (ndim/2));
  if (ndim % 2 == 1) np1 = np1 + np1;
  np2   = fnum / np1;
  np3 = np1;

  if (itmax == 300) { /* CLASS = W */
    if (fnum == 3) {
      np1 = 3;
      np2 = 3;
      np3 = 3;
    } else if (fnum == 9) {
      np1 = 3;
      np2 = 3;
      np3 = 3;
    }else if (fnum == 33){
      np1 = 11;
      np2 = 3;
      np3 = 11;
    } else if (fnum == 99) {
      np1 = 33;
      np2 = 3;
      np3 = 33;
    } else if (fnum == 121) {
      np1 = 11;
      np2 = 11;
      np3 = 11;
    }    
  }  


  //  int factor = atoi (argv[argc-1]);  
 
  int factor = 1;
  np1 *= factor; 
  np2 *= factor; 
  np3 *= factor; 

  isiz1 = ceil ((double) ISIZ1  / (double) np1);
  isiz2 = ceil ((double) ISIZ2  / (double) np2);
  isiz3 = ceil ((double) ISIZ3  / (double) np3);

  GX = (isiz1+4) * np1;
  GY = (isiz2+4) * np2;
  GZ = (isiz3+4) * np3;

  cout << "ITMAX " << itmax << " " << ndim << " " << np1 << " " << np2 << " " << np3 << factor << endl;
  nx0 = ISIZ1;
  ny0 = ISIZ2;
  nz0 = ISIZ3;

  setcoeff();

  HTA_1H u    = HTA_1H::alloc (1, (Tuple<4>((isiz1+4), isiz2+4, isiz3+4, 5), Tuple<4>(np1, np2, np3, 1)), NULL, TILE);
  //  cout << "Here 1 " << endl;
  HTA_1H rsd  = HTA_1H::alloc (1, (Tuple<4>((isiz1+4), isiz2+4, isiz3+4, 5), Tuple<4>(np1, np2, np3, 1)), NULL, TILE);
  //  cout << "Here 2 " << endl;
  HTA_1H flux = HTA_1H::alloc (1, (Tuple<4>((isiz1+4), isiz2+4, isiz3+4, 5), Tuple<4>(np1, np2, np3, 1)), NULL, TILE);
  // cout << "Here 3" << endl;
  HTA_1H frct = HTA_1H::alloc (1, (Tuple<4>((isiz1+4), isiz2+4, isiz3+4, 5), Tuple<4>(np1, np2, np3, 1)), NULL, TILE);
  //cout << "Here 4 " << endl;  
  

  initRsdViews(rsd);

  initUViews(u);

  initShape();
  
  reminder_x = -(isiz1 * np1 - ISIZ1);
  reminder_y = -(isiz2 * np2 - ISIZ2);
  reminder_z = -(isiz3 * np3 - ISIZ3);

  cout << "Here " << reminder_x << " " << reminder_y << " " << reminder_z << endl;
  
  setiv (u);
 
  setbv (u);

  //  cout << u()((R(2, isiz1+1), R(2, isiz2+1), R(2, isiz3+1), R(0, 4))) << endl;

  int L1 = 1;
  int L2 = isiz1 + 2;
  int L3 = 1;
  int L4 = isiz2 + 2;
  int L5 = 1;
  int L6 = isiz3 + 2;

  ShapeH<4> shape2(T(np1, np2, np3, 1));
  shape2 =  Shape<4> ((R(L1-1, L2+1), R(L3-1, L4+1),  R(L5-1, L6+1), R(0, 4)));
  int sx = (L2 - L1 + 3)*np1 + reminder_x;
  int sy = (L4 - L3 + 3)*np2 + reminder_y;
  int sz = (L6 - L5 + 3)*np3 + reminder_z;
  ShapeH<4> shapeInner = shape2.intersect (Shape<4>((R(1, sx-2), R(1,sy-2), R(1,sz-2), R(0,4))));
  rsdInner = rsd()(shapeInner);
  fluxInner = flux()(shapeInner);
  frctInner = frct()(shapeInner);
  uInner = u()(shapeInner);

  // cout << shapeInner << endl;
  erhs (rsd, frct, flux);
 
  //  return;
  
  //ssor(a, b, c, d, u, rsd, flux, frct);
  ssor(u, rsd, flux, frct);

  double max_time = timer_read (T_total);
#if (HTA_TRAIT == HTA_TRAIT_MPI)
  double thread_time = timer_read(T_total);
  MPI_Allreduce (&thread_time,
		 &max_time,
		 1,
		 MPI_DOUBLE,
		 MPI_MAX,
		 MPI_COMM_WORLD);
  //Just for verification...
  cout << "Thread time : " << thread_time << endl;
#endif

  maxtime = max_time;

  cout << "total TIME: " << maxtime << endl;

  dVector err = error (u);

  double errnm[5];
  double rsdnm[5]; //TODO: Not yet computed
  double frc; //TODO: Not yet computed

  for (int i = 0; i < 5; i++) {
    errnm[i] = sqrt (err.val[i]/((ISIZ1-2)*(ISIZ2-2)*(ISIZ3-2)));
    //   cout << "error " << errnm[i]  << endl;
  }
  
  if (Traits::Default::myPlace() == 0) {
    
    bool verified = false;
    char Class;
    double mflops;
    verify (rsdnm, errnm, frc, &Class, &verified);
    mflops = (double)itmax*(1984.77*(double)nx0
			    *(double)ny0
			    *(double)nz0
			    -10923.3*pow2((double)( nx0+ny0+nz0 )/3.0)
			    +27770.9* (double)( nx0+ny0+nz0 )/3.0
			    -144010.0)
      / (maxtime*1000000.0);
    
    print_timers();
    
    c_print_results("LU", Class, nx0,
		    ny0, nz0, itmax, fnum,
		    maxtime, mflops, "          floating point", verified, 
		    NPBVERSION, COMPILETIME, CS1, CS2, CS3, CS4, CS5, CS6, 
		    "(none)");
    
  }  

  return 0;
  
}


/*--------------------------------------------------------------------
--------------------------------------------------------------------*/

static void verify(double xcr[5], double xce[5], double xci,
		   char *Class, bool *verified) {
  
  /*--------------------------------------------------------------------
    c  verification routine                         
    --------------------------------------------------------------------*/
  
  double xcrref[5],xceref[5],xciref, 
    xcrdif[5],xcedif[5],xcidif,
    epsilon, dtref;
  int m;

/*--------------------------------------------------------------------
c   tolerance level
--------------------------------------------------------------------*/
  epsilon = 1.0e-08;

  *Class = 'U';
  *verified = TRUE;

  for (m = 0; m < 5; m++) {
    xcrref[m] = 1.0;
    xceref[m] = 1.0;
  }
  xciref = 1.0;

  if ( nx0 == 12 && ny0 == 12 && nz0 == 12 && itmax == 50)  {
    *Class = 'S';
    dtref = 5.0e-1;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of residual, for the (12X12X12) grid,
c   after 50 time steps, with  DT = 5.0d-01
--------------------------------------------------------------------*/
    xcrref[0] = 1.6196343210976702e-02;
    xcrref[1] = 2.1976745164821318e-03;
    xcrref[2] = 1.5179927653399185e-03;
    xcrref[3] = 1.5029584435994323e-03;
    xcrref[4] = 3.4264073155896461e-02;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of solution error, for the (12X12X12) grid,
c   after 50 time steps, with  DT = 5.0d-01
--------------------------------------------------------------------*/
    xceref[0] = 6.4223319957960924e-04;
    xceref[1] = 8.4144342047347926e-05;
    xceref[2] = 5.8588269616485186e-05;
    xceref[3] = 5.8474222595157350e-05;
    xceref[4] = 1.3103347914111294e-03;

/*--------------------------------------------------------------------
c   Reference value of surface integral, for the (12X12X12) grid,
c   after 50 time steps, with DT = 5.0d-01
--------------------------------------------------------------------*/
    xciref = 7.8418928865937083;

  } else if ( nx0 == 33 && ny0 == 33 && nz0 == 33 && itmax == 300) {

    *Class = 'W';   /* SPEC95fp size */
    dtref = 1.5e-3;
/*--------------------------------------------------------------------
c   Reference values of RMS-norms of residual, for the (33x33x33) grid,
c   after 300 time steps, with  DT = 1.5d-3
--------------------------------------------------------------------*/
    xcrref[0] =   0.1236511638192e+02;
    xcrref[1] =   0.1317228477799e+01;
    xcrref[2] =   0.2550120713095e+01;
    xcrref[3] =   0.2326187750252e+01;
    xcrref[4] =   0.2826799444189e+02;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of solution error, for the (33X33X33) grid,
--------------------------------------------------------------------*/
    xceref[0] =   0.4867877144216;
    xceref[1] =   0.5064652880982e-01;
    xceref[2] =   0.9281818101960e-01;
    xceref[3] =   0.8570126542733e-01;
    xceref[4] =   0.1084277417792e+01;

/*--------------------------------------------------------------------
c   Reference value of surface integral, for the (33X33X33) grid,
c   after 300 time steps, with  DT = 1.5d-3
--------------------------------------------------------------------*/
    xciref    =   0.1161399311023e+02;

  } else if ( nx0 == 64 && ny0 == 64 && nz0 == 64 && itmax == 250) {

    *Class = 'A';
    dtref = 2.0e+0;
/*--------------------------------------------------------------------
c   Reference values of RMS-norms of residual, for the (64X64X64) grid,
c   after 250 time steps, with  DT = 2.0d+0.0
--------------------------------------------------------------------*/
    xcrref[0] = 7.7902107606689367e+02;
    xcrref[1] = 6.3402765259692870e+01;
    xcrref[2] = 1.9499249727292479e+02;
    xcrref[3] = 1.7845301160418537e+02;
    xcrref[4] = 1.8384760349464247e+03;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of solution error, for the (64X64X64) grid,
c   after 250 time steps, with  DT = 2.0d+0.0
--------------------------------------------------------------------*/
    xceref[0] = 2.9964085685471943e+01;
    xceref[1] = 2.8194576365003349;
    xceref[2] = 7.3473412698774742;
    xceref[3] = 6.7139225687777051;
    xceref[4] = 7.0715315688392578e+01;

/*--------------------------------------------------------------------
c   Reference value of surface integral, for the (64X64X64) grid,
c   after 250 time steps, with DT = 2.0d+0.0
--------------------------------------------------------------------*/
    xciref = 2.6030925604886277e+01;

    } else if ( nx0 == 102 && ny0 == 102 && nz0 == 102 && itmax == 250) {

      *Class = 'B';
      dtref = 2.0e+0;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of residual, for the (102X102X102) grid,
c   after 250 time steps, with  DT = 2.0d+0.0
--------------------------------------------------------------------*/
      xcrref[0] = 3.5532672969982736e+03;
      xcrref[1] = 2.6214750795310692e+02;
      xcrref[2] = 8.8333721850952190e+02;
      xcrref[3] = 7.7812774739425265e+02;
      xcrref[4] = 7.3087969592545314e+03;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of solution error, for the (102X102X102) 
c   grid, after 250 time steps, with  DT = 2.0d+0.0
--------------------------------------------------------------------*/
      xceref[0] = 1.1401176380212709e+02;
      xceref[1] = 8.1098963655421574;
      xceref[2] = 2.8480597317698308e+01;
      xceref[3] = 2.5905394567832939e+01;
      xceref[4] = 2.6054907504857413e+02;

/*--------------------------------------------------------------------
c   Reference value of surface integral, for the (102X102X102) grid,
c   after 250 time steps, with DT = 2.0d+0.0
--------------------------------------------------------------------*/
      xciref = 4.7887162703308227e+01;

      } else if ( nx0 == 162 && ny0 == 162 && nz0 == 162 && itmax == 250) {

	*Class = 'C';
	dtref = 2.0e+0;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of residual, for the (162X162X162) grid,
c   after 250 time steps, with  DT = 2.0d+0.0
--------------------------------------------------------------------*/
	xcrref[0] = 1.03766980323537846e+04;
	xcrref[1] = 8.92212458801008552e+02;
	xcrref[2] = 2.56238814582660871e+03;
	xcrref[3] = 2.19194343857831427e+03;
	xcrref[4] = 1.78078057261061185e+04;

/*--------------------------------------------------------------------
c   Reference values of RMS-norms of solution error, for the (162X162X162) 
c   grid, after 250 time steps, with  DT = 2.0d+0.0
--------------------------------------------------------------------*/
	xceref[0] = 2.15986399716949279e+02;
	xceref[1] = 1.55789559239863600e+01;
	xceref[2] = 5.41318863077207766e+01;
	xceref[3] = 4.82262643154045421e+01;
	xceref[4] = 4.55902910043250358e+02;

/*--------------------------------------------------------------------
c   Reference value of surface integral, for the (162X162X162) grid,
c   after 250 time steps, with DT = 2.0d+0.0
--------------------------------------------------------------------*/
	xciref = 6.66404553572181300e+01;
      } else {
	*verified = FALSE;
      }

/*--------------------------------------------------------------------
c    verification test for residuals if gridsize is either 12X12X12 or 
c    64X64X64 or 102X102X102 or 162X162X162
--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c    Compute the difference of solution values and the known reference values.
--------------------------------------------------------------------*/
  for (m = 0; m < 5; m++) {
           
    // xcrdif[m] = fabs((xcr[m]-xcrref[m])/xcrref[m]);
    xcedif[m] = fabs((xce[m]-xceref[m])/xceref[m]);
           
  }
  //  xcidif = fabs((xci - xciref)/xciref);

/*--------------------------------------------------------------------
c    Output the comparison of computed results to known cases.
--------------------------------------------------------------------*/

  if (*Class != 'U') {
    printf("\n Verification being performed for Class %1c\n", *Class);
    printf(" Accuracy setting for epsilon = %20.13e\n", epsilon);
    if (fabs(dt-dtref) > epsilon) {  
      *verified = FALSE;
      *Class = 'U';
      printf(" DT does not match the reference value of %15.8e\n", dtref);
    }
  } else {
    printf(" Unknown Class\n");
  }

  if (*Class != 'U') {
    printf(" Comparison of RMS-norms of residual\n");
  } else {
    printf(" RMS-norms of residual\n");
  }

  /*  for (m = 0; m < 5; m++) {
    if (*Class  ==  'U') {
      printf("          %2d  %20.13e\n", m, xcr[m]);
    } else if (xcrdif[m] > epsilon) {
      *verified = FALSE;
      printf(" FAILURE: %2d  %20.13e%20.13e%20.13e\n",
	     m,xcr[m],xcrref[m],xcrdif[m]);
    } else {
      printf("          %2d  %20.13e%20.13e%20.13e\n",
	     m,xcr[m],xcrref[m],xcrdif[m]);
    }
    }*/

  if (*Class != 'U') {
    printf(" Comparison of RMS-norms of solution error\n");
  } else {
    printf(" RMS-norms of solution error\n");
  }
        
  for (m = 0; m < 5; m++) {
    if (*Class  ==  'U') {
      printf("          %2d  %20.13e\n", m, xce[m]);
    } else if (xcedif[m] > epsilon) {
      *verified = FALSE;
      printf(" FAILURE: %2d  %20.13e%20.13e%20.13e\n",
	     m,xce[m],xceref[m],xcedif[m]);
    } else {
      printf("          %2d  %20.13e%20.13e%20.13e\n",
	     m,xce[m],xceref[m],xcedif[m]);
    }
  }
        
  if (*Class != 'U') {
    printf(" Comparison of surface integral\n");
  } else {
    printf(" Surface integral\n");
  }

  /*
  if (*Class  ==  'U') {
    printf("              %20.13e\n", xci);
  } else if (xcidif > epsilon) {
    *verified = FALSE;
    printf(" FAILURE:     %20.13e%20.13e%20.13e\n", 
	   xci, xciref, xcidif);
  } else {
    printf("              %20.13e%20.13e%20.13e\n",
	   xci, xciref, xcidif);
  }
  */

  if (*Class  ==  'U') {
    printf(" No reference values provided\n");
    printf(" No verification performed\n");
  } else if (*verified) {
    printf(" Verification Successful\n");
  } else {
    printf(" Verification failed\n");
  }
}


#if (HTA_TRAIT == HTA_TRAIT_UPC)
// main is defined in the upc runtime ... (cludge code in main.o)
int upc_main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  int ret = run(argc, argv);
  Traits::Default::finalize();
  return ret;
}
#else
int main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  /* some ebugging some
  int rank = Traits::Default::myPlace();
  char ranks[50];
  sprintf (ranks, "/home/bikshand/htalib/benchmarks/NAS_PRIVATE/bin/rank%d/", rank);
  cout << ranks << endl;
  chdir (ranks);
  */
  int ret = run(argv);
  Traits::Default::finalize();
  return ret;
}
#endif /* HTA */
