
template<typename TYPE, int DIM>
struct Invert   
{
    
  void operator() ( HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > V, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > V0, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > V1, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > V2, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > U, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > U0, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > U1, 		     
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > U2		     
		    ) 
  {    
    double tmat[5][5];    
    double tmp1, tmp2, tmp, tmp3;
    
    double a[5][5];
    double b[5][5];
    double c[5][5];
    double d[5][5];
    double sum[5] ;

    Tuple<DIM> size = V.shape().size();
    for (int i = 0; i < size[0]; i++)
      for (int j = 0; j < size[1]; j++)
	for (int k = 0; k < size[2]; k++)
	  {
	    HTA<TYPE, DIM>  v = V(T(i,j,k));
	    HTA<TYPE, DIM>  v0 = V0(T(i,j,k));
	    HTA<TYPE, DIM>  v1 = V1(T(i,j,k));
	    HTA<TYPE, DIM>  v2 = V2(T(i,j,k));
	   
	    HTA<TYPE, DIM>  u = U(T(i,j,k));
	    HTA<TYPE, DIM>  u0 = U0(T(i,j,k));
	    HTA<TYPE, DIM>  u1 = U1(T(i,j,k));
	    HTA<TYPE, DIM>  u2 = U2(T(i,j,k));	   	    	
	    
	    tmp1 = 1.0 / u[0];
	    tmp2 = tmp1 * tmp1;
	    tmp3 = tmp1 * tmp2;
	    
	    d[0][0] =  1.0
	      + dt * 2.0 * (tx1 * dx1
			       + ty1 * dy1
			       + tz1 * dz1 );
	    d[0][1] =  0.0;
	    d[0][2] =  0.0;
	    d[0][3] =  0.0;
	    d[0][4] =  0.0;
	    
	    d[1][0] =  dt * 2.0
	      * (  tx1 * ( - r43 * c34 * tmp2 * u[1] )
		   + ty1 * ( -       c34 * tmp2 * u[1] )
		   + tz1 * ( -       c34 * tmp2 * u[1] ) );
	    
	    //    assert (d[1][0] != 0);
	    
	    d[1][1] =  1.0
	      + dt * 2.0 
	      * (  tx1 * r43 * c34 * tmp1
		   + ty1 *       c34 * tmp1
		   + tz1 *       c34 * tmp1 )
	      + dt * 2.0 * (   tx1 * dx2
			       + ty1 * dy2
			       + tz1 * dz2  );
	    d[1][2] = 0.0;
	    d[1][3] = 0.0;
	    d[1][4] = 0.0;
    
	    d[2][0] = dt * 2.0
	      * (  tx1 * ( -       c34 * tmp2 * u[2] )
		   + ty1 * ( - r43 * c34 * tmp2 * u[2] )
		   + tz1 * ( -       c34 * tmp2 * u[2] ) );
	    d[2][1] = 0.0;
	    d[2][2] = 1.0
	      + dt * 2.0
	      * (  tx1 *       c34 * tmp1
		   + ty1 * r43 * c34 * tmp1
		   + tz1 *       c34 * tmp1 )
	      + dt * 2.0 * (  tx1 * dx3
			      + ty1 * dy3
			      + tz1 * dz3 );
	    d[2][3] = 0.0;
	    d[2][4] = 0.0;
      
	    d[3][0] = dt * 2.0
	      * (  tx1 * ( -       c34 * tmp2 * u[3] )
		   + ty1 * ( -       c34 * tmp2 * u[3] )
		   + tz1 * ( - r43 * c34 * tmp2 * u[3] ) );
	    d[3][1] = 0.0;
	    d[3][2] = 0.0;
	    d[3][3] = 1.0
	      + dt * 2.0
	      * (  tx1 *       c34 * tmp1
		   + ty1 *       c34 * tmp1
		   + tz1 * r43 * c34 * tmp1 )
	      + dt * 2.0 * (  tx1 * dx4
			      + ty1 * dy4
			      + tz1 * dz4 );
	    d[3][4] = 0.0;
      
	    d[4][0] = dt * 2.0
	      * ( tx1 * ( - ( r43*c34 - c1345 ) * tmp3 * ( pow2(u[1]) )
			  - ( c34 - c1345 ) * tmp3 * ( pow2(u[2]) )
			  - ( c34 - c1345 ) * tmp3 * ( pow2(u[3]) )
			  - ( c1345 ) * tmp2 * u[4] )
		  + ty1 * ( - ( c34 - c1345 ) * tmp3 * ( pow2(u[1]) )
			    - ( r43*c34 - c1345 ) * tmp3 * ( pow2(u[2]) )
			    - ( c34 - c1345 ) * tmp3 * ( pow2(u[3]) )
			    - ( c1345 ) * tmp2 * u[4] )
		  + tz1 * ( - ( c34 - c1345 ) * tmp3 * ( pow2(u[1]) )
			    - ( c34 - c1345 ) * tmp3 * ( pow2(u[2]) )
			    - ( r43*c34 - c1345 ) * tmp3 * ( pow2(u[3]) )
			    - ( c1345 ) * tmp2 * u[4] ) );
	    d[4][1] = dt * 2.0
	      * ( tx1 * ( r43*c34 - c1345 ) * tmp2 * u[1]
		  + ty1 * (     c34 - c1345 ) * tmp2 * u[1]
		  + tz1 * (     c34 - c1345 ) * tmp2 * u[1] );
	    d[4][2] = dt * 2.0
	      * ( tx1 * ( c34 - c1345 ) * tmp2 * u[2]
		  + ty1 * ( r43*c34 -c1345 ) * tmp2 * u[2]
		  + tz1 * ( c34 - c1345 ) * tmp2 * u[2] );
	    d[4][3] = dt * 2.0
	      * ( tx1 * ( c34 - c1345 ) * tmp2 * u[3]
		  + ty1 * ( c34 - c1345 ) * tmp2 * u[3]
		  + tz1 * ( r43*c34 - c1345 ) * tmp2 * u[3] );
	    d[4][4] = 1.0
	      + dt * 2.0 * ( tx1 * c1345 * tmp1
			     + ty1 * c1345 * tmp1
			     + tz1 * c1345 * tmp1 )
	      + dt * 2.0 * (  tx1 * dx5
			      +  ty1 * dy5
			      +  tz1 * dz5 );

      
	    tmp1 = 1.0 / (double) u0[0];
	    tmp2 = tmp1 * tmp1;
	    tmp3 = tmp1 * tmp2;
    
	    a[0][0] = - dt * tz1 * dz1;
	    a[0][1] =   0.0;
	    a[0][2] =   0.0;
	    a[0][3] = - dt * tz2;
	    a[0][4] =   0.0;
    
	    a[1][0] = - dt * tz2
	      * ( - ( u0[1]*u0[3] ) * tmp2 )
	      - dt * tz1 * ( - c34 * tmp2 * u0[1] );
	    a[1][1] = - dt * tz2 * ( u0[3] * tmp1 )
	      - dt * tz1 * c34 * tmp1
	      - dt * tz1 * dz2 ;
	    a[1][2] = 0.0;
	    a[1][3] = - dt * tz2 * ( u0[1] * tmp1 );
	    a[1][4] = 0.0;
    
	    a[2][0] = - dt * tz2
	      * ( - ( u0[2]*u0[3] ) * tmp2 )
	      - dt * tz1 * ( - c34 * tmp2 * u0[2] );
	    a[2][1] = 0.0;
	    a[2][2] = - dt * tz2 * ( u0[3] * tmp1 )
	      - dt * tz1 * ( c34 * tmp1 )
	      - dt * tz1 * dz3;
	    a[2][3] = - dt * tz2 * ( u0[2] * tmp1 );
	    a[2][4] = 0.0;
    
	    a[3][0] = - dt * tz2
	      * ( - ( u0[3] * tmp1 ) *( u0[3] * tmp1 )
		  + 0.50 * C2
		  * ( ( u0[1] * u0[1]
			+ u0[2] * u0[2]
			+ u0[3] * u0[3] ) * tmp2 ) )
	      - dt * tz1 * ( - r43 * c34 * tmp2 * u0[3] );
	    a[3][1] = - dt * tz2
	      * ( - C2 * ( u0[1] * tmp1 ) );
	    a[3][2] = - dt * tz2
	      * ( - C2 * ( u0[2] * tmp1 ) );
	    a[3][3] = - dt * tz2 * ( 2.0 - C2 )
	      * ( u0[3] * tmp1 )
	      - dt * tz1 * ( r43 * c34 * tmp1 )
	      - dt * tz1 * dz4;
	    a[3][4] = - dt * tz2 * C2;
    
	    a[4][0] = - dt * tz2
	      * ( ( C2 * (  u0[1] * u0[1]
			    + u0[2] * u0[2]
			    + u0[3] * u0[3] ) * tmp2
		    - C1 * ( u0[4] * tmp1 ) )
		  * ( u0[3] * tmp1 ) )
	      - dt * tz1
	      * ( - ( c34 - c1345 ) * tmp3 * (u0[1]*u0[1])
		  - ( c34 - c1345 ) * tmp3 * (u0[2]*u0[2])
		  - ( r43*c34 - c1345 )* tmp3 * (u0[3]*u0[3])
		  - c1345 * tmp2 * u0[4] );
	    a[4][1] = - dt * tz2
	      * ( - C2 * ( u0[1]*u0[3] ) * tmp2 )
	      - dt * tz1 * ( c34 - c1345 ) * tmp2 * u0[1];
	    a[4][2] = - dt * tz2
	      * ( - C2 * ( u0[2]*u0[3] ) * tmp2 )
	      - dt * tz1 * ( c34 - c1345 ) * tmp2 * u0[2];
	    a[4][3] = - dt * tz2
	      * ( C1 * ( u0[4] * tmp1 )
		  - 0.50 * C2
		  * ( (  u0[1]*u0[1]
			 + u0[2]*u0[2]
			 + 3.0*u0[3]*u0[3] ) * tmp2 ) )
	      - dt * tz1 * ( r43*c34 - c1345 ) * tmp2 * u0[3];
	    a[4][4] = - dt * tz2
	      * ( C1 * ( u0[3] * tmp1 ) )
	      - dt * tz1 * c1345 * tmp1
	      - dt * tz1 * dz5;
 
	    
	    tmp1 = 1.0 / u1[0];
	    tmp2 = tmp1 * tmp1;
	    tmp3 = tmp1 * tmp2;

	    b[0][0] = - dt * ty1 * dy1;
	    b[0][1] =   0.0;
	    b[0][2] = - dt * ty2;
	    b[0][3] =   0.0;
	    b[0][4] =   0.0;

	    b[1][0] = - dt * ty2
	      * ( - ( u1[1]*u1[2] ) * tmp2 )
	      - dt * ty1 * ( - c34 * tmp2 * u1[1] );
	    b[1][1] = - dt * ty2 * ( u1[2] * tmp1)
	      - dt * ty1 * ( c34 * tmp1 )
	      - dt * ty1 * dy2;
	    b[1][2] = - dt * ty2 * ( u1[1] * tmp1 );
	    b[1][3] = 0.0;
	    b[1][4] = 0.0;

	    b[2][0] = - dt * ty2
	      * ( - ( u1[2] * tmp1 ) *( u1[2] * tmp1 )
		  + 0.50 * C2 * ( (  u1[1] * u1[1]
				     + u1[2] * u1[2]
				     + u1[3] * u1[3] )
				  * tmp2 ) )
	      - dt * ty1 * ( - r43 * c34 * tmp2 * u1[2] );
	    b[2][1] = - dt * ty2
	      * ( - C2 * ( u1[1] * tmp1 ) );
	    b[2][2] = - dt * ty2 * ( ( 2.0 - C2 )
					   * ( u1[2] * tmp1 ) )
	      - dt * ty1 * ( r43 * c34 * tmp1 )
	      - dt * ty1 * dy3;
	    b[2][3] = - dt * ty2
	      * ( - C2 * ( u1[3] * tmp1 ) );
	    b[2][4] = - dt * ty2 * C2;

	    b[3][0] = - dt * ty2
	      * ( - ( u1[2]*u1[3] ) * tmp2 )
	      - dt * ty1 * ( - c34 * tmp2 * u1[3] );
	    b[3][1] = 0.0;
	    b[3][2] = - dt * ty2 * ( u1[3] * tmp1 );
	    b[3][3] = - dt * ty2 * ( u1[2] * tmp1 )
	      - dt * ty1 * ( c34 * tmp1 )
	      - dt * ty1 * dy4;
	    b[3][4] = 0.0;

	    b[4][0] = - dt * ty2
	      * ( ( C2 * (  u1[1] * u1[1]
			    + u1[2] * u1[2]
			    + u1[3] * u1[3] ) * tmp2
		    - C1 * ( u1[4] * tmp1 ) )
		  * ( u1[2] * tmp1 ) )
	      - dt * ty1
	      * ( - (     c34 - c1345 )*tmp3*(pow2(u1[1]))
		  - ( r43*c34 - c1345 )*tmp3*(pow2(u1[2]))
		  - (     c34 - c1345 )*tmp3*(pow2(u1[3]))
		  - c1345*tmp2*u1[4] );
	    b[4][1] = - dt * ty2
	      * ( - C2 * ( u1[1]*u1[2] ) * tmp2 )
	      - dt * ty1
	      * ( c34 - c1345 ) * tmp2 * u1[1];
	    b[4][2] = - dt * ty2
	      * ( C1 * ( u1[4] * tmp1 )
		  - 0.50 * C2 
		  * ( (  u1[1]*u1[1]
			 + 3.0 * u1[2]*u1[2]
			 + u1[3]*u1[3] ) * tmp2 ) )
	      - dt * ty1
	      * ( r43*c34 - c1345 ) * tmp2 * u1[2];
	    b[4][3] = - dt * ty2
	      * ( - C2 * ( u1[2]*u1[3] ) * tmp2 )
	      - dt * ty1 * ( c34 - c1345 ) * tmp2 * u1[3];
	    b[4][4] = - dt * ty2
	      * ( C1 * ( u1[2] * tmp1 ) )
	      - dt * ty1 * c1345 * tmp1
	      - dt * ty1 * dy5;
          
	    tmp1 = 1.0 / u2[0];
	    tmp2 = tmp1 * tmp1;
	    tmp3 = tmp1 * tmp2;
    
	    c[0][0] = - dt * tx1 * dx1;
	    c[0][1] = - dt * tx2;
	    c[0][2] =   0.0;
	    c[0][3] =   0.0;
	    c[0][4] =   0.0;
    
	    c[1][0] = - dt * tx2
	      * ( - ( u2[1] * tmp1 ) *( u2[1] * tmp1 )
		  + C2 * 0.50 * (  u2[1] * u2[1]
				   + u2[2] * u2[2]
				   + u2[3] * u2[3] ) * tmp2 )
	      - dt * tx1 * ( - r43 * c34 * tmp2 * u2[1] );
	    c[1][1] = - dt * tx2
	      * ( ( 2.0 - C2 ) * ( u2[1] * tmp1 ) )
	      - dt * tx1 * ( r43 * c34 * tmp1 )
	      - dt * tx1 * dx2;
	    c[1][2] = - dt * tx2
	      * ( - C2 * ( u2[2] * tmp1 ) );
	    c[1][3] = - dt * tx2
	      * ( - C2 * ( u2[3] * tmp1 ) );
	    c[1][4] = - dt * tx2 * C2;
    
	    c[2][0] = - dt * tx2
	      * ( - ( u2[1] * u2[2] ) * tmp2 )
	      - dt * tx1 * ( - c34 * tmp2 * u2[2] );
	    c[2][1] = - dt * tx2 * ( u2[2] * tmp1);
	    c[2][2] = - dt * tx2 * ( u2[1] * tmp1)
	      - dt * tx1 * ( c34 * tmp1 )
	      - dt * tx1 * dx3;
	    c[2][3] = 0.0;
	    c[2][4] = 0.0;
    
	    c[3][0] = - dt * tx2
	      * ( - ( u2[1]*u2[3] ) * tmp2 )
	      - dt * tx1 * ( - c34 * tmp2 * u2[3] );
	    c[3][1] = - dt * tx2 * ( u2[3] * tmp1);
	    c[3][2] = 0.0;
	    c[3][3] = - dt * tx2 * ( u2[1] * tmp1)
	      - dt * tx1 * ( c34 * tmp1 )
	      - dt * tx1 * dx4;
	    c[3][4] = 0.0;
    
	    c[4][0] = - dt * tx2
	      * ( ( C2 * (  u2[1] * u2[1]
			    + u2[2] * u2[2]
			    + u2[3] * u2[3] ) * tmp2
		    - C1 * ( u2[4] * tmp1 ) )
		  * ( u2[1] * tmp1 ) )
	      - dt * tx1
	      * ( - ( r43*c34 - c1345 ) * tmp3 * ( pow2(u2[1]) )
		  - (     c34 - c1345 ) * tmp3 * ( pow2(u2[2]) )
		  - (     c34 - c1345 ) * tmp3 * ( pow2(u2[3]) )
		  - c1345 * tmp2 * u2[4] );
	    c[4][1] = - dt * tx2
	      * ( C1 * ( u2[4] * tmp1 )
		  - 0.50 * C2
		  * ( (  3.0*u2[1]*u2[1]
			 + u2[2]*u2[2]
			 + u2[3]*u2[3] ) * tmp2 ) )
	      - dt * tx1
	      * ( r43*c34 - c1345 ) * tmp2 * u2[1];
	    c[4][2] = - dt * tx2
	      * ( - C2 * ( u2[2]*u2[1] ) * tmp2 )
	      - dt * tx1
	      * (  c34 - c1345 ) * tmp2 * u2[2];
	    c[4][3] = - dt * tx2
	      * ( - C2 * ( u2[3]*u2[1] ) * tmp2 )
	      - dt * tx1
	      * (  c34 - c1345 ) * tmp2 * u2[3];
	    c[4][4] = - dt * tx2
	      * ( C1 * ( u2[1] * tmp1 ) )
	      - dt * tx1 * c1345 * tmp1
	      - dt * tx1 * dx5;  

	    
	    for (int i = 0; i < 5; i++)
	      {
		sum[i] = 0 ;
		for (int j = 0 ; j < 5; j++)
		  sum[i] +=  (a[i][j] * v0[j] + 
			      b[i][j] * v1[j] + 
			      c[i][j] * v2[j]);
	      }
	    
	    for (int i = 0; i < 5; i++)
	      v[i] = v[i] - omega * sum[i];
	    
	    //cout << "here " << v << endl;
	    
	    for (int m = 0; m < 5; m++) 
	      {
		tmat[m][0] = d[m][0];
		tmat[m][1] = d[m][1];
		tmat[m][2] = d[m][2];
		tmat[m][3] = d[m][3];
		tmat[m][4] = d[m][4];
		/* assert (tmat[m][0]!=0);
		   assert (tmat[m][1]!=0);
		   assert (tmat[m][2]!=0);
		   assert (tmat[m][3]!=0);
		   assert (tmat[m][4]!=0);*/
	      }
	    
	    tmp1 = 1.0 / tmat[0][0];
	    tmp = tmp1 * tmat[1][0];
	    tmat[1][1] =  tmat[1][1]
	      - tmp * tmat[0][1];
	    tmat[1][2] =  tmat[1][2]
	      - tmp * tmat[0][2];
	    tmat[1][3] =  tmat[1][3]
	      - tmp * tmat[0][3];
	    tmat[1][4] =  tmat[1][4]
	      - tmp * tmat[0][4];
	    
	    v[1] = v[1]
	      - v[0] * tmp;
	    
	    tmp = tmp1 * tmat[2][0];
	    tmat[2][1] =  tmat[2][1]
	      - tmp * tmat[0][1];
	    tmat[2][2] =  tmat[2][2]
	      - tmp * tmat[0][2];
	    tmat[2][3] =  tmat[2][3]
	      - tmp * tmat[0][3];
	    tmat[2][4] =  tmat[2][4]
	      - tmp * tmat[0][4];
	    v[2] = v[2]
	      - v[0] * tmp;
	    
	    
	    tmp = tmp1 * tmat[3][0];
	    tmat[3][1] =  tmat[3][1]
	      - tmp * tmat[0][1];
	    tmat[3][2] =  tmat[3][2]
	      - tmp * tmat[0][2];
	    tmat[3][3] =  tmat[3][3]
	      - tmp * tmat[0][3];
	    tmat[3][4] =  tmat[3][4]
	      - tmp * tmat[0][4];
	    v[3] = v[3]
	      - v[0] * tmp;
	    
	    tmp = tmp1 * tmat[4][0];
	    tmat[4][1] =  tmat[4][1]
	      - tmp * tmat[0][1];
	    tmat[4][2] =  tmat[4][2]
	      - tmp * tmat[0][2];
	    tmat[4][3] =  tmat[4][3]
	      - tmp * tmat[0][3];
	    tmat[4][4] =  tmat[4][4]
	      - tmp * tmat[0][4];
	    v[4] = v[4]
      - v[0] * tmp;
	    
	    
	    tmp1 = 1.0 / tmat[ 1][1];
	    tmp = tmp1 * tmat[ 2][1];
	    tmat[2][2] =  tmat[2][2]
	      - tmp * tmat[1][2];
	    tmat[2][3] =  tmat[2][3]
	      - tmp * tmat[1][3];
	    tmat[2][4] =  tmat[2][4]
	      - tmp * tmat[1][4];
	    v[2] = v[2]
	      - v[1] * tmp;
	    
	    tmp = tmp1 * tmat[3][1];
	    tmat[3][2] =  tmat[3][2]
	      - tmp * tmat[1][2];
	    tmat[3][3] =  tmat[3][3]
	      - tmp * tmat[1][3];
	    tmat[3][4] =  tmat[3][4]
	      - tmp * tmat[1][4];
	    v[3] = v[3]
	      - v[1] * tmp;
	    
	    tmp = tmp1 * tmat[4][1];
	    tmat[4][2] =  tmat[4][2]
	      - tmp * tmat[1][2];
	    tmat[4][3] =  tmat[4][3]
	      - tmp * tmat[1][3];
	    tmat[4][4] =  tmat[4][4]
	      - tmp * tmat[1][4];
	    v[4] = v[4]
	      - v[1] * tmp;
	    
	    
	    tmp1 = 1.0 / tmat[2][2];
	    tmp = tmp1 * tmat[3][2];
	    tmat[3][3] =  tmat[3][3]
	      - tmp * tmat[2][3];
	    tmat[3][4] =  tmat[3][4]
	      - tmp * tmat[2][4];
	    v[3] = v[3]
	      - v[2] * tmp;
	    
	    tmp = tmp1 * tmat[4][2];
	    tmat[4][3] =  tmat[4][3]
	      - tmp * tmat[2][3];
	    tmat[4][4] =  tmat[4][4]
	      - tmp * tmat[2][4];
	    v[4] = v[4]
	      - v[2] * tmp;
	    
	    tmp1 = 1.0 / tmat[3][3];
	    tmp = tmp1 * tmat[4][3];
	    tmat[4][4] =  tmat[4][4]
	      - tmp * tmat[3][4];
	    v[4] = v[4]
	      - v[3] * tmp;
	    
	    /*--------------------------------------------------------------------
	      c   back substitution
	      --------------------------------------------------------------------*/
	    
	    
	    v[4] = v[4]
	      / tmat[4][4];
	    
	    v[3] = v[3]
	      - tmat[3][4] * v[4];
	    v[3] = v[3]
	      / tmat[3][3];
	    
	    v[2] = v[2]
	      - tmat[2][3] * v[3]
	      - tmat[2][4] * v[4];
	    v[2] = v[2]
      / tmat[2][2];
	    
	    v[1] = v[1]
	      - tmat[1][2] * v[2]
	      - tmat[1][3] * v[3]
	      - tmat[1][4] * v[4];
	    v[1] = v[1]
	      / tmat[1][1];
	    
	    v[0] = v[0]
	      - tmat[0][1] * v[1]
	      - tmat[0][2] * v[2]
	      - tmat[0][3] * v[3]
	      - tmat[0][4] * v[4];
	    v[0] = v[0]
	      / tmat[0][0];
	    
	    //    cout << "there " << v << endl;
	  }  
  }
};

void blts (HTA_2H v, HTA_2H u, bool* mask)
{    
  r43 = ( 4.0 / 3.0 );
  c1345 = C1 * C3 * C4 * C5;
  c34 = C3 * C4;
  

  //inversion cannnot be vectorized 
  //hence map is the only option.
  v()(iShape).hmap(Invert<double, 3>(),		  	   
		   v()(iShape - T(0, 0, 1)),
		   v()(iShape - T(0, 1, 0)),
		   v()(iShape - T(1, 0, 0)),
		   u()(iShape),
		   u()(iShape - T(0, 0, 1)),
		   u()(iShape - T(0, 1, 0)),
		   u()(iShape - T(1, 0, 0)),		   
		   1,
		   mask);
}
