#include "lu.h"
#include "applu.h"
#include "ShapeH.h"
#include "MapReduceHTA.h"

typedef Triplet R;
typedef Tuple<3> T;

ShapeH<3> iShape;

typedef HTA<double, 3> HTA_2H;

typedef Tuple<3> T;
int isiz1;
int isiz2;
int isiz3;

int np1;
int np2;
int np3;
/*--------------------------------------------------------------------
  c   xi-direction flux differences
  --------------------------------------------------------------------*/

template<typename TYPE, int DIM>
struct Loop1    {
  void operator() (HTA<TYPE, DIM,  
		   Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		   HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd) 
  {  
    flux[0] = rsd[1];
    TYPE u21 = rsd[1] / rsd[0];
    TYPE q = 0.50 * (  rsd[1] * rsd[1]
		       + rsd[2] * rsd[2]
		       + rsd[3] * rsd[3])
      / rsd[0];
    flux[1] = rsd[1] * u21 + C2 * 
      ( rsd[4] - q );
    flux[2] = rsd[2] * u21;
    flux[3] = rsd[3] * u21;
    flux[4] = ( C1 * rsd[4] - C2 * q ) * u21; 
    assert (q != 0);
    assert (u21 != 0);
  }      
};

template<typename TYPE, int DIM>
struct Loop2 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux2
		    )
  {
    for (int i = 0; i < 5; i++)
      frct0[i] =  frct0[i] -  tx2 * (flux2[i] - flux1[i]);  
  }
};

template<typename TYPE, int DIM>
struct Loop3    {
  void operator()(  HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1) 
  {
    TYPE tmp =  1.0 / rsd0[0];
    TYPE u21i = tmp * rsd0[1];
    TYPE u31i = tmp * rsd0[2];
    TYPE u41i = tmp * rsd0[3];
    TYPE u51i = tmp * rsd0[4];
    
    tmp = 1.0 / rsd1[0];
    TYPE u21im1 = tmp * rsd1[1];
    TYPE u31im1 = tmp * rsd1[2];
    TYPE u41im1 = tmp * rsd1[3];
    TYPE u51im1 = tmp * rsd1[4];
    
    flux[1] = (4.0/3.0) * tx3 *  ( u21i - u21im1 );
    flux[2] = tx3 * ( u31i - u31im1 );
    flux[3] = tx3 * ( u41i - u41im1 );
    flux[4] = 0.50 * ( 1.0 - C1*C5 )
      * tx3 * ( ( u21i * u21i + u31i * u31i + u41i * u41i )
		- ( u21im1*u21im1 + u31im1*u31im1 + u41im1*u41im1 ) )
      + (1.0/6.0) * tx3 * ( u21i*u21i - u21im1*u21im1 )
      + C1 * C5 * tx3 * ( u51i - u51im1 ); 	
  }     
};

template<typename TYPE, int DIM>
struct Loop4 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd2,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux1) 
  { 
    frct[0] =  frct[0]
      + tx1  * dx1 * (rsd1[0]
		      - 2.0 * rsd0[0]
		      +       	    rsd2[0]);    
    frct[1] = frct[1]
      + tx3 * C3 * C4 * ( flux1[1] - flux0[1])
      + dx2 * tx1 * (            rsd1[1]
				 - 2.0 * rsd0[1]
				 +           rsd2[1]) ;    
    frct[2] =  frct[2]
      + tx3 * C3 * C4 * ( flux1[2] - flux0[2] )
      + dx3 * tx1 * (            rsd1[2]
				 - 2.0 * rsd0[2]
				 +           rsd2[2]);    
    frct[3] =  frct[3]
      + tx3 * C3 * C4 * ( flux1[3] - flux0[3] )
      + dx4 * tx1 * (            rsd1[3]
				 - 2.0 * rsd0[3]
				 +           rsd2[3]);    
    frct[4] = 
      frct[4]
      + tx3 * C3 * C4 * ( flux1[4] - flux0[4] )
      + dx5 * tx1 * (            rsd1[4]
				 - 2.0 * rsd0[4]
				 +           rsd2[4]);
  }
};


template<typename TYPE, int DIM>
struct Loop5 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd2,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd3,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd4) 
  {
    for (int i = 0; i < 5; i++)
      frct[i] = frct[i] - dssp * (rsd2[i]
			   - 4.0 * rsd1[i]
			   + 6.0 * rsd0[i]
			   - 4.0 * rsd3[i]
			   + rsd4[i]);  
  }
};


template<typename TYPE, int DIM>
struct Special1 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd3,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd4) 
  {
    for (int i = 0; i < 5; i++)
      frct[i] = frct[i] - 
	dssp * (
		+ 5.0 * rsd0[i]
		- 4.0 * rsd3[i]
		+ rsd4[i]);  
  }
};

template<typename TYPE, int DIM>
struct Special2 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd3,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd4) 
  {
    for (int i = 0; i < 5; i++)
      frct[i] = frct[i] - 
	dssp * (
		- 4.0 * rsd1[i]
		+ 6.0 * rsd0[i]
		- 4.0 * rsd3[i]
		+ rsd4[i]);  
  }
};


template<typename TYPE, int DIM>
struct Special3 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd2,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd3)
  {
    for (int i = 0; i < 5; i++)
      frct[i] = frct[i] - 
	dssp * (rsd2[i]
		- 4.0 * rsd1[i]
		+ 6.0 * rsd0[i]
		- 4.0 * rsd3[i]);		
  }
};

template<typename TYPE, int DIM>
struct Special4 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd2)
  {
    for (int i = 0; i < 5; i++)
      frct[i] = frct[i] - 
	dssp * (rsd2[i]
		- 4.0 * rsd1[i]
		+ 5.0 * rsd0[i]);
  }
};


void erhs1 (HTA_2H rsd, HTA_2H frct, HTA_2H flux)
{  

  double dsspm = dssp;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = rsd.shape().size()[0];
  int ty = rsd.shape().size()[1]; 
  int tz = rsd.shape().size()[2];

  int L1 = 1;
  int L2 = isiz1+2;
  int sx, sy, sz;

  ShapeH<3> shape(T(tx, ty, tz));
  
  /* loop 1 */
  shape =  Shape<3> ((R(L1, L2), R(jst, jend),  R(kst, kend)));
  sx = (L2 - L1 + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape1 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  Loop1<double, 3> loop1;
  flux()(shape1).hmap (loop1,
		       rsd()(shape1),
		       0);

  frct()(iShape).hmap(Loop2<double, 3>(), 
		      flux()(iShape - T(1, 0, 0)),
		      flux()(iShape + T(1, 0, 0)),
		      0);

  shape = Shape<3>((R(ist, L2), R(jst, jend), R(kst, kend)));  
  sx = (L2 - ist+1)*tx;
  sy = (jend - jst+1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape3 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  Loop3<double, 3> loop3;
  flux()(shape3).hmap(loop3,
		      rsd()(shape3),
		      rsd()(shape3 - T(1, 0, 0)),
		      0);

 
  frct()(iShape).
    hmap (Loop4<double, 3>(),
	  rsd()(iShape),
	  rsd()(iShape - T(1,0,0)),
	  rsd()(iShape + T(1,0,0)),
	  flux()(iShape),
	  flux()(iShape + T(1,0,0)),
	  0);
  
  /*--------------------------------------------------------------------
    c   Fourth-order dissipation
    --------------------------------------------------------------------*/  
  
  //cout << "dssp " << dssp << endl;
 
 
  frct((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3, 3), R(jst, jend), R(kst, kend)))
    .hmap (Special1<double, 3>(),
	   rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3, 3), R(jst, jend), R(kst, kend))),
	   rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend))),
	   rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(5, 5), R(jst, jend), R(kst, kend))),
	   0);

  //special case (II)
  frct((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend))).
    hmap (Special2<double, 3>(),
	  rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(4, 4), R(jst, jend), R(kst, kend))),
	  rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(3, 3), R(jst, jend), R(kst, kend))),
	  rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(5, 5), R(jst, jend), R(kst, kend))),
	  rsd((R(0, 0), R(0, ty-1), R(0, tz-1)))((R(6, 6), R(jst, jend), R(kst, kend))),
	  0);
  
  /* loop 5 */
  int ist1 = 2;
  int iend1 = isiz1+1;
  shape = Shape<3>((R(ist1, iend1), R(jst, jend), R(kst, kend)));
  sx = (iend1 - ist1 + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape5 = shape.intersect (Shape<3>((R(3, sx-4), R(1,sy-2), R(1,sz-2))));
 
  frct()(shape5).hmap (Loop5<double, 3>(),
		       rsd()(shape5),
		       rsd()(shape5-T(1,0,0)),
		       rsd()(shape5-T(2,0,0)),
		       rsd()(shape5+T(1,0,0)),
		       rsd()(shape5+T(2,0,0)),
		       0);
  
  frct((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend))).
    hmap (Special3<double, 3>(),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend))),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-2, isiz1-2), R(jst, jend), R(kst, kend))),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-3, isiz1-3), R(jst, jend), R(kst, kend))),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend))),
	  0);
  
  frct((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend))).
    hmap (Special4<double, 3>(),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1, isiz1), R(jst, jend), R(kst, kend))),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-1, isiz1-1), R(jst, jend), R(kst, kend))),
	  rsd((R(tx-1, tx-1), R(0, ty-1), R(0, tz-1)))((R(isiz1-2, isiz1-2), R(jst, jend), R(kst, kend))),
	  0);
}



template<typename TYPE, int DIM>
struct Loop6{
           
  void operator() ( HTA<TYPE, DIM, Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd) 
  {  
    flux[0] = rsd[2];
    TYPE u31 = rsd[2] / rsd[0];
    TYPE q = 0.50 * (  rsd[1] * rsd[1]
		       + rsd[2] * rsd[2]
		       + rsd[3] * rsd[3] )
      / rsd[0];
    flux[1] = rsd[1] * u31;
    flux[2] = rsd[2] * u31 + C2 * 
      ( rsd[4] - q );
    flux[3] = rsd[3] * u31;
    flux[4] = ( C1 * rsd[4] - C2 * q ) * u31;
    assert (q != 0);
    assert (u31 != 0);
  }
  
    
      
};

template<typename TYPE, int DIM>
struct Loop8{
    
    
    
  void operator() (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1) 
  {
    TYPE tmp =  1.0 / rsd0[0];
    TYPE u21j = tmp * rsd0[1];
    TYPE u31j = tmp * rsd0[2];
    TYPE u41j = tmp * rsd0[3];
    TYPE u51j = tmp * rsd0[4];
    
    tmp = 1.0 / rsd1[0];
    TYPE u21jm1 = tmp * rsd1[1];
    TYPE u31jm1 = tmp * rsd1[2];
    TYPE u41jm1 = tmp * rsd1[3];
    TYPE u51jm1 = tmp * rsd1[4];
    
    flux[1] = ty3 * ( u21j - u21jm1 );
    flux[2] = (4.0/3.0) * ty3 *
      ( u31j - u31jm1 );
    flux[3] = ty3 * ( u41j - u41jm1 );
    flux[4] = 0.50 * ( 1.0 - C1*C5 )
      * ty3 * ( ( u21j * u21j + u31j * u31j + u41j * u41j )
		- ( u21jm1*u21jm1 + u31jm1*u31jm1 + u41jm1*u41jm1 ) )
      + (1.0/6.0) * ty3 * ( u31j*u31j - u31jm1*u31jm1 )
      + C1 * C5 * ty3 * ( u51j - u51jm1 ); 	
  }
  
    
      
};


template<typename TYPE, int DIM>
struct Loop10 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd2,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux1) 
  { 
    frct[0] =  frct[0]
      + ty1  * dy1 * (rsd1[0]
		      - 2.0 * rsd0[0]
		      +       	    rsd2[0]);    
    frct[1] = frct[1]
      + ty3 * C3 * C4 * ( flux1[1] - flux0[1])
      + dy2 * ty1 * (            rsd1[1]
				 - 2.0 * rsd0[1]
				 +           rsd2[1]) ;    
    frct[2] =  frct[2]
      + ty3 * C3 * C4 * ( flux1[2] - flux0[2] )
      + dy3 * ty1 * (            rsd1[2]
				 - 2.0 * rsd0[2]
				 +           rsd2[2]);    
    frct[3] =  frct[3]
      + ty3 * C3 * C4 * ( flux1[3] - flux0[3] )
      + dy4 * ty1 * (            rsd1[3]
				 - 2.0 * rsd0[3]
				 +           rsd2[3]);    
    frct[4] = 
      frct[4]
      + ty3 * C3 * C4 * ( flux1[4] - flux0[4] )
      + dy5 * ty1 * (            rsd1[4]
				 - 2.0 * rsd0[4]
				 +           rsd2[4]);
  }
};

void erhs2 (HTA_2H rsd, HTA_2H frct, HTA_2H flux)
{    
  double dsspm = dssp;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = rsd.shape().size()[0];
  int ty = rsd.shape().size()[1]; 
  int tz = rsd.shape().size()[2];

  int L1 = 1;
  int L2 = isiz2+2;
  int sx, sy, sz;

  ShapeH<3> shape(T(tx, ty, tz));
  
  /* loop 1 */
  shape =  Shape<3> ((R(ist, iend), R(L1, L2),  R(kst, kend)));
  sx = (iend - ist + 1)*tx;
  sy = (L2 - L1 + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape1 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  Loop6<double, 3> loop6;
  flux()(shape1).hmap(loop6, rsd()(shape1), 0);
  
  double tmp = tx2;
  tx2 = ty2;
  frct()(iShape).hmap (Loop2<double, 3>(), 
		       flux()(iShape - T(0, 1, 0)),
		       flux()(iShape + T(0, 1, 0)),
		       0);

  tx2 = tmp;

  /* loop 3 */
  shape = Shape<3>((R(ist, iend), R(jst, L2), R(kst, kend)));  
  sx = (iend - ist+1)*tx;
 sy = (L2 - ist +1)*ty;
  sz = (kend - kst+1)*tz;
  ShapeH<3> shape3 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  Loop8<double, 3> loop8;
   flux()(shape3).hmap (loop8,
		      rsd()(shape3),
		      rsd()(shape3 - T(0, 1, 0)),
		      0);
			    
   frct()(iShape).hmap (Loop10<double, 3>(),
	  rsd()(iShape),
	  rsd()(iShape - T(0,1,0)),
	  rsd()(iShape + T(0,1,0)),
	  flux()(iShape),
	  flux()(iShape + T(0,1,0)),
	  0);
  
  /*--------------------------------------------------------------------
    c   Fourth-order dissipation
    --------------------------------------------------------------------*/    

  //special case (I)
  frct((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(3, 3), R(kst, kend))).
    hmap (Special1<double, 3>(),
	  rsd((R(0, tx-1),  R(0, 0), R(0, tz-1)))((R(ist, iend),  R(3, 3), R(kst, kend))),
	  rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend))),
	  rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(5, 5), R(kst, kend))),
	  0);
  
  //special case (II)
  frct((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend))).
    hmap (Special2<double, 3>(),
	   rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(4, 4), R(kst, kend))),
	   rsd((R(0, tx-1),  R(0, 0), R(0, tz-1)))((R(ist, iend),  R(3, 3), R(kst, kend))),
	   rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(5, 5), R(kst, kend))),
	   rsd((R(0, tx-1), R(0, 0), R(0, tz-1)))((R(ist, iend), R(6, 6), R(kst, kend))),
	   0);
  
  int jst1 = 2;
  int jend1 = isiz2+1;
  shape = Shape<3>((R(ist, iend), R(jst1, jend1), R(kst, kend)));
  sx = (iend - ist + 1)*tx;
  sy = (jend1 - jst1 + 1)*ty;
  sz = (kend - kst + 1)*tz;
  ShapeH<3> shape5 = shape.intersect (Shape<3>((R(1, sx-2), R(3,sy-4), R(1,sz-2))));
  frct()(shape5).hmap (Loop5<double, 3>(),
		       rsd()(shape5),
		       rsd()(shape5-T(0,1,0)),
		       rsd()(shape5-T(0,2,0)),
		       rsd()(shape5+T(0,1,0)),
		       rsd()(shape5+T(0,2,0)),
		       0);
  
  frct((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1, isiz2-1), R(kst, kend))).
    hmap (Special3<double, 3>(),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1, isiz2-1), R(kst, kend))),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-2, isiz2-2), R(kst, kend))),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-3, isiz2-3),  R(kst, kend))),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend))),
	  0);

  frct((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend))).
    hmap (Special4<double, 3>(),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2, isiz2), R(kst, kend))),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-1, isiz2-1), R(kst, kend))),
	  rsd((R(0, tx-1), R(ty-1, ty-1), R(0, tz-1)))((R(ist, iend), R(isiz2-2, isiz2-2), R(kst, kend))),
	  0);
}



template<typename TYPE, int DIM>
struct Loop12{
           
  void operator() (HTA<TYPE, DIM,  
		      Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd) 
  {  
    flux[0] = rsd[3];
    TYPE u41 = rsd[3] / rsd[0];
    TYPE q = 0.50 * (  rsd[1] * rsd[1]
		       + rsd[2] * rsd[2]
		       + rsd[3] * rsd[3] )
      / rsd[0];
    flux[1] = rsd[1] * u41;
    flux[2] = rsd[2] * u41;
    flux[3] = rsd[3] * u41 + C2 * ( rsd[4] - q);
    flux[4] = ( C1 * rsd[4] - C2 * q ) * u41;
    assert (q != 0);
    assert (u41 != 0);
  }
  
    
      
};

template<typename TYPE, int DIM>
struct Loop14{
   
  void operator() (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux, 
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		      HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1) 
  {
    TYPE tmp =  1.0 / rsd0[0];
    TYPE u21j = tmp * rsd0[1];
    TYPE u31j = tmp * rsd0[2];
    TYPE u41j = tmp * rsd0[3];
    TYPE u51j = tmp * rsd0[4];
    
    tmp = 1.0 / rsd1[0];
    TYPE u21jm1 = tmp * rsd1[1];
    TYPE u31jm1 = tmp * rsd1[2];
    TYPE u41jm1 = tmp * rsd1[3];
    TYPE u51jm1 = tmp * rsd1[4];
    
    flux[1] = tz3 * ( u21j - u21jm1 );
    flux[2] =  tz3 * ( u31j - u31jm1 );     
    flux[3] =  (4.0/3.0) * tz3 *
      ( u41j - u41jm1 );
    flux[4] = 0.50 * ( 1.0 - C1*C5 )
      * tz3 * ( ( u21j * u21j + u31j * u31j + u41j * u41j )
		- ( u21jm1*u21jm1 + u31jm1*u31jm1 + u41jm1*u41jm1 ) )
      + (1.0/6.0) * tz3 * ( u41j*u41j - u41jm1*u41jm1 )
      + C1 * C5 * tz3 * ( u51j - u51jm1 ); 	
  }      
};

template<typename TYPE, int DIM>
struct Loop16 {
  void operator () (HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > frct, 
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd1,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > rsd2,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux0,
		    HTA<TYPE, DIM,  Traits::HTA_DEFAULT_TRAIT<double, 3> > flux1) 
  { 
    frct[0] =  frct[0]
      + tz1  * dz1 * (rsd1[0]
		      - 2.0 * rsd0[0]
		      +       	    rsd2[0]);    
    frct[1] = frct[1]
      + tz3 * C3 * C4 * ( flux1[1] - flux0[1])
      + dz2 * tz1 * (            rsd1[1]
				 - 2.0 * rsd0[1]
				 +           rsd2[1]) ;    
    frct[2] =  frct[2]
      + tz3 * C3 * C4 * ( flux1[2] - flux0[2] )
      + dz3 * tz1 * (            rsd1[2]
				 - 2.0 * rsd0[2]
				 +           rsd2[2]);    
    frct[3] =  frct[3]
      + tz3 * C3 * C4 * ( flux1[3] - flux0[3] )
      + dz4 * tz1 * (            rsd1[3]
				 - 2.0 * rsd0[3]
				 +           rsd2[3]);    
    frct[4] = 
      frct[4]
      + tz3 * C3 * C4 * ( flux1[4] - flux0[4] )
      + dz5 * tz1 * (            rsd1[4]
				 - 2.0 * rsd0[4]
				 +           rsd2[4]);
  }
};

void erhs3 (HTA_2H rsd, HTA_2H frct, HTA_2H flux)
{  

  double dsspm = dssp;
  
  int jst = 2;
  int jend = isiz2+1;
  int ist = 2;
  int iend = isiz1+1;
  int kst = 2;
  int kend = isiz3+1;
  int tx = rsd.shape().size()[0];
  int ty = rsd.shape().size()[1]; 
  int tz = rsd.shape().size()[2];

  int L1 = 1;
  int L2 = isiz3+2;
  int sx, sy, sz;

  ShapeH<3> shape(T(tx, ty, tz));
  
  /* loop 1 */
  shape =  Shape<3> ((R(ist, iend), R(jst, jend),  R(L1, L2)));
  sx = (iend - ist + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (L2 - L1 + 1)*tz;
  ShapeH<3> shape1 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));

  Loop12<double, 3> loop12;
  flux()(shape1).hmap (loop12,
		       rsd()(shape1),
		       0);

  /*loop 2 */
  double tmp = tx2;
  tx2 = tz2;

  frct()(iShape).hmap (Loop2<double, 3>(),
		       flux()(iShape - T(0, 0, 1)),
		       flux()(iShape + T(0, 0, 1)),
		       0);
  tx2 = tmp;

  /* loop 3 */
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst, L2)));  
  sx = (iend - ist+1)*tx;
  sy = (jend - jst +1)*ty;
  sz = (L2 - kst+1)*tz;
  ShapeH<3> shape3 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(1,sz-2))));
  Loop14<double, 3> loop14;  
  flux()(shape3).hmap(loop14,
		      rsd()(shape3),
		      rsd()(shape3 - T(0, 0, 1)),
		      0);

  /* loop 4 */
 
  frct()(iShape).hmap (Loop16<double, 3>(),
	  rsd()(iShape),
	  rsd()(iShape - T(0,0, 1)),
	  rsd()(iShape + T(0, 0,1)),
	  flux()(iShape),
	  flux()(iShape + T(0,0,1)),
	  0);
 

  /*--------------------------------------------------------------------
    c   Fourth-order dissipation
    --------------------------------------------------------------------*/  
 
  
  //special case (I)
  frct((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend), R(jst, jend), R(3, 3))).
    hmap (Special1<double, 3>(),
	  rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(3, 3))),
	  rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(4, 4))),
	  rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(5, 5))),
	  0);
  //special case (II)
  frct((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend), R(jst, jend), R(4, 4))).
    hmap (Special2<double, 3>(),
	    rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(4, 4))),
	    rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(3, 3))),
	    rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(5, 5))),
	    rsd((R(0, tx-1), R(0, ty-1), R(0, 0)))((R(ist, iend),  R(jst, jend),R(6, 6))),
	    0);
    
  /* loop 5 */
  int kst1 = 2;
  int kend1 = isiz2+1;
  shape = Shape<3>((R(ist, iend), R(jst, jend), R(kst1, kend1)));
  sx = (iend - ist + 1)*tx;
  sy = (jend - jst + 1)*ty;
  sz = (kend1 - kst1 + 1)*tz;
  ShapeH<3> shape5 = shape.intersect (Shape<3>((R(1, sx-2), R(1,sy-2), R(3,sz-4))));
  frct()(shape5).hmap (Loop5<double, 3>(),
		       rsd()(shape5),
		       rsd()(shape5-T(0,0,1)),
		       rsd()(shape5-T(0,0,2)),
		       rsd()(shape5+T(0,0,1)),
		       rsd()(shape5+T(0,0,2)),
		       0);
     
  frct((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3-1, isiz3-1))).
    hmap (Special3<double, 3>(),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3-1, isiz3-1))),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3-2, isiz3-2))),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3-3, isiz3-3))),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3, isiz3))),
	  0);
  
  frct((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3, isiz3))).
    hmap (Special4<double, 3>(),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3, isiz3))),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3-1, isiz3-1))),
	  rsd((R(0, tx-1), R(0, ty-1), R(tz-1, tz-1)))(( R(ist, iend), R(jst, jend), R(isiz3-2, isiz3-2))),
	  0);    
}



static void exact( int i, int j, int k, double u000ijk[5] ) {

/*--------------------------------------------------------------------
c
c   compute the exact solution at (i,j,k)
c
--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c  local variables
--------------------------------------------------------------------*/
  int m;
  double xi, eta, zeta;

  xi  = ((double)i) / (double) (ISIZ1 - 1);
  eta  = ((double)j) / (double) (ISIZ2 - 1);
  zeta = ((double)k) / (double) (ISIZ3 - 1);

  for (m = 0; m < 5; m++) {
    u000ijk[m] =  ce[m][0]
      + ce[m][1] * xi
      + ce[m][2] * eta
      + ce[m][3] * zeta
      + ce[m][4] * xi * xi
      + ce[m][5] * eta * eta
      + ce[m][6] * zeta * zeta
      + ce[m][7] * xi * xi * xi
      + ce[m][8] * eta * eta * eta
      + ce[m][9] * zeta * zeta * zeta
      + ce[m][10] * xi * xi * xi * xi
      + ce[m][11] * eta * eta * eta * eta
      + ce[m][12] * zeta * zeta * zeta * zeta;
  }
}


static void exact (int i, int j, int k, HTA<double, 3> h) {

/*--------------------------------------------------------------------
c
c   compute the exact solution at (i,j,k)
c
--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c  local variables
--------------------------------------------------------------------*/
  int m;
  double xi, eta, zeta;

  xi  = ((double)i) / (double) (ISIZ1 - 1);
  eta  = ((double)j) / (double) (ISIZ2 - 1);
  zeta = ((double)k) / (double) (ISIZ3 - 1);

  for (m = 0; m < 5; m++) {
    h[m] =  ce[m][0]
      + ce[m][1] * xi
      + ce[m][2] * eta
      + ce[m][3] * zeta
      + ce[m][4] * xi * xi
      + ce[m][5] * eta * eta
      + ce[m][6] * zeta * zeta
      + ce[m][7] * xi * xi * xi
      + ce[m][8] * eta * eta * eta
      + ce[m][9] * zeta * zeta * zeta
      + ce[m][10] * xi * xi * xi * xi
      + ce[m][11] * eta * eta * eta * eta
      + ce[m][12] * zeta * zeta * zeta * zeta;
  }
}
