typedef class HTA<double, 3> Grid;
typedef class HTA<double, 3> Leaf;

struct Interp
{ 
  void operator () (Grid dst, Grid src) 
  {
    int mm3 = src.shape().size()[0];
    int mm2 = src.shape().size()[1];
    int mm1 = src.shape().size()[2];
    
    int i3, i2, i1;

    const Tuple<3> step_src  = src.memMap().linearStep() * src.shape().step();
    int step_z_0 = step_src[0];
    int step_z_1 = step_src[1];
    int step_z_2 = step_src[2];
    
    const Tuple<3> step_dst = dst.memMap().linearStep() * dst.shape().step();
    int step_u_0 = step_dst[0];
    int step_u_1 = step_dst[1];

    const double* z = src.raw();
    double* u = dst.raw();

    double* z1 = new double [2 * mm1];
    double* z2 = new double [2 * mm1];
    double* z3 = new double [2 * mm1];
    
    for (i3 = 0; i3 < mm3-1; i3++) {
      for (i2 = 0; i2 < mm2-1; i2++) {
	for (i1 = 0; i1 < mm1; i1++) {
	  z1[i1] = z[i3 * step_z_0 + (i2+1) * step_z_1 + i1] + z[i3 * step_z_0 + i2 * step_z_1 + i1];
	  z2[i1] = z[(i3+1) * step_z_0 + i2 * step_z_1 + i1] + z[i3 * step_z_0 + i2 * step_z_1 + i1];
	  z3[i1] = z[(i3+1) * step_z_0 + (i2+1) * step_z_1 + i1] + z[(i3+1) * step_z_0 + i2 * step_z_1 + i1] + z1[i1];
	}
	for (i1 = 0; i1 < mm1-1; i1++) {
	  u[2*i3 * step_u_0 + 2*i2 * step_u_1 + 2*i1] = u[2*i3 * step_u_0 + 2*i2 * step_u_1 + 2*i1]
	    +z[i3 * step_z_0 + i2 * step_z_1 + i1];
	  u[2*i3 * step_u_0 + 2*i2 * step_u_1 + 2*i1+1] = u[2*i3 * step_u_0 + 2*i2 * step_u_1 + 2*i1+1]
	    +0.5*(z[i3 * step_z_0 + i2 * step_z_1 + (i1+1)]+z[i3 * step_z_0 + i2 * step_z_1 + i1]);
	}
	for (i1 = 0; i1 < mm1-1; i1++) {
	  u[2*i3 * step_u_0 + (2*i2+1) * step_u_1 + 2*i1] = u[2*i3 * step_u_0 + (2*i2+1) * step_u_1 + 2*i1]
	    +0.5 * z1[i1];
	  u[2*i3 * step_u_0 + (2*i2+1) * step_u_1 + 2*i1+1] = u[2*i3 * step_u_0 + (2*i2+1) * step_u_1 + 2*i1+1]
			+0.25*( z1[i1] + z1[(i1+1)] );
	}
	
	for (i1 = 0; i1 < mm1-1; i1++) {
	  u[(2*i3+1) * step_u_0 + 2*i2 * step_u_1 + 2*i1] = u[(2*i3+1) * step_u_0 + 2*i2 * step_u_1 + 2*i1]
	    +0.5 * z2[i1];
	  u[(2*i3+1) * step_u_0 + 2*i2 * step_u_1 + 2*i1+1] = u[(2*i3+1) * step_u_0 + 2*i2 * step_u_1 + 2*i1+1]
	    +0.25*( z2[i1] + z2[(i1+1)] );
	}
	
	for (i1 = 0; i1 < mm1-1; i1++) {
	  u[(2*i3+1) * step_u_0 + (2*i2+1) * step_u_1 + 2*i1] = u[(2*i3+1) * step_u_0 + (2*i2+1) * step_u_1 + 2*i1]
	    +0.25* z3[i1];
	  u[(2*i3+1) * step_u_0 + (2*i2+1) * step_u_1 + 2*i1+1] = u[(2*i3+1) * step_u_0 + (2*i2+1) * step_u_1 + 2*i1+1]
	    +0.125*( z3[i1] + z3[(i1+1)] );
	}
      }
    }   
    
    delete [] z1;
    delete [] z2;
    delete [] z3;
  }
};

struct InterpSpecial
{ 
  void operator () (Grid dst, Grid src) 
  {
    int mm3 = src.shape().size()[0];
    int mm2 = src.shape().size()[1];
    int mm1 = src.shape().size()[2];
    
    int n3 = dst.shape().size()[0];
    int n2 = dst.shape().size()[1];
    int n1 = dst.shape().size()[2];
    
    int i3, i2, i1;

    const Tuple<3> step_src  = src.memMap().linearStep() * src.shape().step();
    int step_z_0 = step_src[0];
    int step_z_1 = step_src[1];
    int step_z_2 = step_src[2];
    
    const Tuple<3> step_dst = dst.memMap().linearStep() * dst.shape().step();
    int step_u_0 = step_dst[0];
    int step_u_1 = step_dst[1];

    const double* z = src.raw();
    double* u = dst.raw();
    
    double* z1 = new double [2 * mm1];
    double* z2 = new double [2 * mm1];
    double* z3 = new double [2 * mm1];

    int d1, t1, d2, t2, d3, t3;
    if(n1==3){
      d1 = 1;
      t1 = 0;
    }else{
      d1 = 0;
      t1 = -1;
    }
    
    if(n2==3){
      d2 = 1;
      t2 = 0;
    }else{
      d2 = 0;
      t2 = -1;
    }
    
    if(n3==3){
      d3 = 1;
      t3 = 0;
    }	else{
      d3 = 0;
      t3 = -1;
    }
    
    for  (i3 = d3;  i3 < mm3-1; i3++)
      {
	for  (i2 = d2; i2 < mm2-1; i2++) 
	  {
	    for  (i1=d1; i1 < mm1-1; i1++)
	      u[(2*i1-d1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-d3]=
		u[(2*i1-d1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-d3]
		+z[(i1) * step_z_0 + (i2) * step_z_1 + i3];
	    
	    for  (i1=0; i1 < mm1-1; i1++)
	      u[(2*i1-t1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-d3]=
		u[(2*i1-t1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-d3]
		+0.50*(z[(i1+1) * step_z_0 + (i2) * step_z_1 + i3]+
			z[(i1) * step_z_0 + (i2) * step_z_1 + i3]);
	  }
	for (i2=0; i2 < mm2-1; i2++)
	  {
	    for  (i1=d1; i1 < mm1-1; i1++)
	      u[(2*i1-d1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-d3]=
		u[(2*i1-d1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-d3]	   
		+0.50*(z[(i1) * step_z_0 + (i2+1) * step_z_1 + i3]+
			z[(i1) * step_z_0 + (i2) * step_z_1 + i3]);
	    
	    for  (i1=0; i1 < mm1-1; i1++)
	      u[(2*i1-t1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-d3]
		=u[(2*i1-t1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-d3]       
		+0.250*(z[(i1+1) * step_z_0 + (i2+1) * step_z_1 + i3]+
			 z[(i1+1) * step_z_0 + (i2) * step_z_1 + i3]+
			 z[(i1) * step_z_0 + (  i2+1) * step_z_1 + i3]+
			 z[(i1) * step_z_0 + (  i2) * step_z_1 + i3]);
	  }
      }	 
    
    for  (i3=0; i3 < mm3-1; i3++)
      {
	for  (i2=d2; i2 < mm2-1; i2++)
	  {
	    for  (i1=d1; i1 < mm1-1; i1++)		      
	      u[(2*i1-d1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-t3]=
		u[(2*i1-d1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-t3]		
		+0.50*(z[(i1) * step_z_0 + (i2) * step_z_1 + i3+1]+
			z[(i1) * step_z_0 + (i2) * step_z_1 + i3]);
	    
	    for (i1=0; i1 < mm1-1; i1++)
	      u[(2*i1-t1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-t3]=
		u[(2*i1-t1) * step_u_0 + (2*i2-d2) * step_u_1 + 2*i3-t3]
		+0.250*(z[(i1+1) * step_z_0 + (i2) * step_z_1 + i3+1]+
			 z[(i1) * step_z_0 + (i2) * step_z_1 + i3+1] +
			 z[(i1+1) * step_z_0 + (i2) * step_z_1 + i3  ]+
			 z[(i1) * step_z_0 + (i2) * step_z_1 + i3  ]);
	  }
	
		
	for  (i2=0; i2 < mm2-1; i2++)
	  {
	    for (i1=d1; i1 < mm1-1; i1++)
	      u[(2*i1-d1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-t3]=
		u[(2*i1-d1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-t3]
		+0.250*(z[(i1) * step_z_0 + (i2+1) * step_z_1 + i3+1]+
			 z[(i1) * step_z_0 + (i2) * step_z_1 + i3+1]+
			 z[(i1) * step_z_0 + (i2+1) * step_z_1 + i3]+
			 z[(i1) * step_z_0 + (i2) * step_z_1 + i3]);
	    
	    for (i1=0; i1 < mm1-1; i1++)
		      
	      u[(2*i1-t1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-t3]=
		u[(2*i1-t1) * step_u_0 + (2*i2-t2) * step_u_1 + 2*i3-t3]
		+0.1250*(z[(i1+1) * step_z_0 + (i2+1) * step_z_1 + i3+1]+
			  z[(i1+1) * step_z_0 + (i2) * step_z_1 + i3+1]
			  +z[(i1  ) * step_z_0 + (i2+1) * step_z_1 + i3+1]+
			  z[(i1  ) * step_z_0 + (i2) * step_z_1 + i3+1]
			  +z[(i1+1) * step_z_0 + (i2+1) * step_z_1 + i3  ]+
			  z[(i1+1) * step_z_0 + (i2) * step_z_1 + i3  ]
			  +z[(i1  ) * step_z_0 + (i2+1) * step_z_1 + i3  ]+
			  z[(i1  ) * step_z_0 + (i2) * step_z_1 + i3  ]);
	  }
      }
  }
  
};

