struct Rprj3
{
  Rprj3 (int d0, int d1, int d2):
    d0_(d0), d1_(d1), d2_(d2) {}
  
  void operator() ( Grid dst, 		     
		    Grid arg)
  {
    double w0 = 0.5;
    double w1 = 0.25;
    double w2 = 0.125;
    double w3 = 0.0625;    
    
    const Tuple<3> step_arg = arg.memMap().linearStep() * arg.shape().step();
    int step_arg_0 = step_arg[0];
    int step_arg_1 = step_arg[1];
    int step_arg_2 = step_arg[2];
    
    const Tuple<3> step_dst = dst.memMap().linearStep() * dst.shape().step();
    int step_dst_0 = step_dst[0];
    int step_dst_1 = step_dst[1];

    int size_0 = dst.shape().size()[0]; // most significant
    int size_1 = dst.shape().size()[1];
    int size_2 = dst.shape().size()[2];
    
    double u1[2 * size_2];
    double u2[2 * size_2];
    
    const double* a = arg.raw();
    double* d = dst.raw();

    int i_a, j_a, k_a;

    for (int i = 1; i < size_0 - 1; i++) {
      i_a = 2 * i - d0_;
      for (int j = 1; j < size_1 - 1; j++) {
	j_a = 2 * j - d1_;
        for (int k = 1; k < size_2; k++) {
	  k_a = 2 * k - d2_;
	  
	  // can the compiler hoist + k from the index computation here ?
	  u1[k_a-1] = 
	    a[i_a * step_arg_0    + (j_a-1) * step_arg_1 + k_a - 1] + 
	    a[i_a * step_arg_0   +  (j_a+1) * step_arg_1 + k_a - 1] + 
	    a[(i_a-1) * step_arg_0 + j_a * step_arg_1    + k_a - 1] + 
	    a[(i_a+1) * step_arg_0 + j_a * step_arg_1    + k_a - 1];
	  u2[k_a-1] = 
	    a[(i_a-1) * step_arg_0 + (j_a-1) * step_arg_1 + k_a - 1] +
	    a[(i_a-1) * step_arg_0 + (j_a+1) * step_arg_1 + k_a - 1] + 
	    a[(i_a+1) * step_arg_0 + (j_a-1) * step_arg_1 + k_a - 1] + 
	    a[(i_a+1) * step_arg_0 + (j_a+1) * step_arg_1 + k_a - 1] ;
	}
       
        for (int k = 1; k < size_2 - 1; k++) {    

	  k_a = 2 * k - d2_;

	  double y2 = a[(i_a-1) * step_arg_0 + (j_a-1) * step_arg_1 + k_a] +
	    a[(i_a + 1) * step_arg_0 + (j_a - 1) * step_arg_1 + k_a] + 
	    a[(i_a - 1) * step_arg_0 + (j_a + 1) * step_arg_1 + k_a] + 
	    a[(i_a + 1) * step_arg_0 + (j_a + 1) * step_arg_1 + k_a];
	    
	  double x2 = a[i_a * step_arg_0 + (j_a-1) * step_arg_1 + k_a] + 
	    a[i_a * step_arg_0 + (j_a+1) * step_arg_1 + k_a] + 
	    a[(i_a - 1) * step_arg_0 + j_a * step_arg_1 + k_a] + 
	    a[(i_a + 1) * step_arg_0 + j_a * step_arg_1 + k_a]; 
	    
	  d[i * step_dst_0  + j * step_dst_1 + k] = 
	    w0 * a[i_a * step_arg_0 + j_a * step_arg_1 + k_a]  +
	    w1 * (x2 + a[i_a * step_arg_0 + j_a * step_arg_1 + k_a-1] + a[i_a * step_arg_0 + j_a * step_arg_1 + k_a+1]) +
	    w2 * (y2 + u1[k_a-1] + u1[k_a+1]) +
	    w3 * (u2[k_a-1] + u2[k_a+1]);
	}
      }
    }

    //    delete [] u1;
    //delete [] u2;
  }


private:

  int d0_;
  int d1_;
  int d2_;
};
