/*=================================PROTOTYPES=========================================================================*/
int ny; 
void mg3P    (Grid* r , Grid& v , Grid* u , double a[4] , double c[4] , int k);
Grid conv (Grid s , double w[4]);
Grid conv2 (Grid s , double w[4] , int , Tuple<3>);
void interp  (Grid& z , Grid& u , int k);
void interp_real (Grid z , Grid u );
void interp_inner_special (Leaf z , Leaf u );
void rprj3   (Grid& r , Grid& s , int k);
void resid   (Grid& u , Grid& v , Grid&r , double a[4] , int k);
void psinv   (Grid& r , Grid& s , double c[4] , int k);
void zran3   (Grid& z , int gx , int gy , int gz);
void norm2u3 (Grid& r , double *rnm2 , double *rnmu , int nx = NX_DEFAULT , int ny = NY_DEFAULT , int nz = NZ_DEFAULT);
void rep_nrm (Grid &x , char* callee , int k );
void comm3    (Grid& u);
void comm3_ex (Grid& u , int s1 , int s2 , int s3);
void zero3   (Grid& z);
void print_timers(void);
void zran3_inner (Leaf z , Tuple<3> tileIndex , int gx , int gy , int gz);
void comm_ex_ew (Grid& u);
void comm_ex_ns (Grid& u);
void comm_ex_fb (Grid& u);
void setup (Grid* r , Grid* u , int nprocs);
/*=======================================INIT===================================================================*/
template<typename T , int N , int DIM>
struct XchgData 
{
  Tuple<DIM> maxKey[N];
  T maxVal[N];
  Tuple<DIM> minKey[N];
  T minVal[N];
};
template<typename T , int N , int DIM>
struct FilterN : public Operator::MapReduce<T , XchgData<T , N , DIM> , DIM>  
{
  FilterN() : minIdx_(0) , maxIdx_(0) { 
    for (int i = 0; i < N; i++) {
      data_.minVal[i] = 1.0;
      data_.maxVal[i] = 0.0;
    }
  }
  void reduce (const XchgData<T , N , DIM>& d) 
  { 
    for (int i = 0; i < N; i++) { 
      map_tuple(d.maxKey[i] , d.maxVal[i]);
      map_tuple(d.minKey[i] , d.minVal[i]);
    }
  }
  void map_tuple (const Tuple<DIM>& key , const T& val) 
  { 
    if (val > data_.maxVal[minIdx_]) {
            data_.maxVal[minIdx_] = val;
      data_.maxKey[minIdx_] = key;
      T min = val;
      for (int i = 0; i < N; i++) {
	if (data_.maxVal[i] < min) {
	  min = data_.maxVal[i];
	  minIdx_ = i;
	}
      }
    } 
    if (val > 0.0 && val < data_.minVal[maxIdx_]) {
            data_.minVal[maxIdx_] = val;
      data_.minKey[maxIdx_] = key;
      T max = val;
      for (int i = 0; i < N; i++) {
	if (data_.minVal[i] > max) {
	  max = data_.minVal[i];
	  maxIdx_ = i;
	}
      }
    }
  }   
  virtual const XchgData<T , N , DIM>& result () const 
  {
    return data_;
  }
  Operator::MapReduceKey keyType () const { return Operator::MR_TUPLE; }
  XchgData<T , N , DIM> data_;
    }; 
/*=======================================COMM===================================================================*/
template<typename T , int DIM>
struct SquareSum : public Operator::MapReduce<T , SquareSum<T , DIM> , DIM>  
{
  SquareSum() : sqsum(0) , max (0) {}
  void reduce (const SquareSum<T , DIM>& d) 
  { 
    sqsum += d.sqsum;
    Operator::setIfLarger<double>()(max , d.max);
  }
  void map_none (const T& val) {
    sqsum += std::pow(val , 2.0);
    Operator::setIfLarger<double>()(max , std::fabs(val));
  }
  virtual const SquareSum<T , DIM>& result () const 
  {
    return *this;
  }
  Operator::MapReduceKey keyType () const { return Operator::MR_NONE; }
  T sqsum;
  T max;
}; 
/*=============================================MAIN================================================================*/
int run(int ntiles = -1)
{ 
         int nPlaces = Traits::Default::nPlaces();
   if (ntiles <= 0)
     ntiles = nPlaces;
   IO::   IO::   assert (ntiles == 1 || ntiles % 8 == 0);
      lt = LT_DEFAULT;
   Grid r[lt];
   Grid u[lt];  
   setup (r , u , ntiles);
   int sx = r[lt-1].shape().size()[0];
   int sy = r[lt-1].shape().size()[1];
   int sz = r[lt-1].shape().size()[2];
   nx = NX_DEFAULT / sx;
   ny = NY_DEFAULT / sy;
   nz = NZ_DEFAULT / sz;
   int nit = NIT_DEFAULT;
   Sequence<Tuple<3> > tiling(Tuple<3> (nx+2 , ny+2 , nz+2) , Tuple<3> (sx , sy , sz));
   Grid v = HTA<double , 3>::alloc(1 , tiling , NULL , order);
   zran3 (v , NX_DEFAULT , NY_DEFAULT , NZ_DEFAULT);
   zero3 (u[lt-1]);
      debug_vec[0] = 0;
   if ( (NX_DEFAULT != NY_DEFAULT) || (NX_DEFAULT != NY_DEFAULT) ) Class = 'U';
   else if( NX_DEFAULT == 32 && nit == 4 )   Class = 'S';
   else if( NX_DEFAULT == 64 && nit == 40 )  Class = 'W';
   else if( NX_DEFAULT == 256 && nit == 20 ) Class = 'B';
   else if( NX_DEFAULT == 512 && nit == 20 ) Class = 'C';
   else if( NX_DEFAULT == 256 && nit == 4 )  Class = 'A';
   else 	Class = 'U';
   a[0] = -8.0/3.0;
   a[1] =  0.0;
   a[2] =  1.0/6.0;
   a[3] =  1.0/12.0;
   if (Class == 'A' || Class == 'S' || Class =='W') {
     c[0] =  -3.0/8.0;
     c[1] =  1.0/32.0;
     c[2] =  -1.0/64.0;
     c[3] =   0.0;
   } else {
     c[0] =  -3.0/17.0;
     c[1] =  1.0/33.0;
     c[2] =  -1.0/61.0;
     c[3] =   0.0;
   }
      resid (u[lt-1] , v , r[lt-1] , a , lt-1);
   norm2u3 (r[lt-1] , &rnm2 , &rnmu);
   for (int it = 0; it < nit; it++) {
     IO::     mg3P (r , v , u , a , c , lt-1);
     resid (u[lt-1] , v , r[lt-1] , a , lt-1);
   }
   norm2u3 (r[lt-1] , &rnm2 , &rnmu);
      IO::   Traits::Default::barrier();
   if (Traits::Default::myPlace() == 0) {
     bool verified = false;
     double verify_value = 0.0;
     if (Class != 'U') {
       if (Class == 'S')  verify_value = 0.530770700573e-04;
       else if (Class == 'W')  verify_value = 0.250391406439e-17;  
       else if (Class == 'A')   verify_value = 0.2433365309e-5;
       else if (Class == 'B')   verify_value = 0.180056440132e-5;
       else if (Class == 'C') verify_value = 0.570674826298e-06;
       
       if ( fabs( rnm2 - verify_value ) <= epsilon ) {
	 verified = true;
	 IO::	 IO::	 IO::       } else {
	 verified = false;
	 IO::	 IO::	 IO::       }
     } else {
       verified = false;
       IO::       IO::     }
     int nn = NX_DEFAULT * NY_DEFAULT * NZ_DEFAULT;
     double t =           if (t != 0)
       mflops = 58. * nit * nn * 1.0E-6 / t;
     else
       mflops = 0.0;
     c_print_results("MG" , Class , NX_DEFAULT , NY_DEFAULT , NZ_DEFAULT , nit , ntiles , t , mflops , "floating point" , verified , NPBVERSION , COMPILETIME , CS1 , CS2 , CS3 , CS4 , CS5 , CS6 , CS7);      
     IO::     print_timers ();
     IO::   }
   return 0;
}
/*=============================COMP=========================================================*/
void mg3P (Grid* r , Grid& v , Grid* u , double a[4] , double c[4] , int k)
{
  if (k > 0) {
    rprj3 (r[k] , r[k-1] , k);
    mg3P (r , v , u , a , c , k-1);      
    if (k != lt-1) zero3 (u[k]);
    interp (u[k-1] , u[k] , k);
    k == lt-1 ? resid (u[k] , v , r[k] , a , k) : resid (u[k] , r[k] , r[k] , a , k) ;
    psinv (r[k] , u[k] , c , k);
  } else  {
    zero3 (u[k]);
    psinv (r[k] , u[k] , c , k); 
  }  
}
void resid (Grid& u , Grid& v , Grid& r , double a[4] , int k)
{     
    r = v - u.lconv(a);  
    comm3 (r);
  if (debug_vec[0]) 
    rep_nrm (r , "resid" , k);
} 
void psinv (Grid& r , Grid& u , double c[4] , int k)
{  
    u += r.lconv(c);
    comm3 (u);
  if (debug_vec[0]) rep_nrm (u , "psinv" , k);
} 
void rprj3 (Grid& h , Grid& s , int k)
{
    Tuple<3> hSize = h.shape().size();
  Tuple<3> sSize = s.shape().size();
  int d1 = hSize[0] == sSize[0] ? 0 : 1;
  int d2 = hSize[1] == sSize[1] ? 0 : 1;
  int d3 = hSize[2] == sSize[2] ? 0 : 1;
  Grid r = h((Triplet (d1 , hSize[0]-1 , d1+1) ,       Triplet (d2 , hSize[1]-1 , d2+1) ,       Triplet (d3 , hSize[2]-1 , d3+1)));
  Tuple<3> rSize = r.shape().size();
  int nx = rSize[0];
  int ny = rSize[1];
  int nz = rSize[2];
  int sx = r (T(0 , 0 , 0)).shape().size()[0];
  int sy = r (T(0 , 0 , 0)).shape().size()[1];
  int sz = r (T(0 , 0 , 0)).shape().size()[2];
  int iSize = s (T(0 , 0 , 0)).shape().size()[0];
  int jSize = s (T(0 , 0 , 0)).shape().size()[1];
  int kSize = s (T(0 , 0 , 0)).shape().size()[2];
  Sequence<Triplet> t_args2(Triplet (2-d1 , 2 * (iSize-2) - d1 , 2) , Triplet (2-d2 , 2 * (jSize-2) - d2 , 2) , Triplet(2-d3 , 2 * (kSize-2) - d3 , 2));    
  Sequence<Triplet> t_args4(Triplet (1 , iSize-2) , Triplet (1 , jSize-2) , Triplet(1 , kSize-2));
  Shape<3> s2 (t_args2);
  Shape<3> s4 (t_args4);
  double a[4];
  a[0] = 0.5;
  a[1] = 0.25;
  a[2] = 0.125;
  a[3] = 0.0625;
  Grid tmp = r.lconv(a);
  s()(s4) = tmp()(s2); 
    comm3 (s);
  if (debug_vec[0]) rep_nrm (s , "rprj3" , k-1);
 } 
Grid conv (Grid s , double w[4])
{
    Tuple<3> size = s(T(0 , 0 , 0)).shape().size();  
  Sequence<Tuple<3> > tiling(size , s.shape().size());  
  Grid u1 = HTA<double , 3>::alloc(1 , tiling , NULL , order);
  Grid u2 = HTA<double , 3>::alloc(1 , tiling , NULL , order); 
  Grid r = HTA<double , 3>::alloc(1 , tiling , NULL , order);   
  Triplet I(1 , size[0]-2);
  Triplet J(1 , size[1]-2);
  Triplet K(0 , size[2]-1);
  Triplet RI(1 , size[0] - 2);
  Triplet RJ(1 , size[1] - 2);
  Triplet RK(0 , size[2] - 1);
  
  u1()((RI , RJ , RK)) =  
    (Grid) s()((I , J-1 , K)) + 
    (Grid) s()((I , J+1 , K)) +
    (Grid) s()((I-1 , J , K)) +
    (Grid) s()((I+1 , J , K));
  u2()((RI , RJ , RK)) =  
    (Grid) s()((I-1 , J-1 , K)) +
    (Grid) s()((I-1 , J+1 , K)) +
    (Grid) s()((I+1 , J-1 , K)) +
    (Grid) s()((I+1 , J+1 , K));
  K = Triplet(1 , size[2] - 2);
  RK = Triplet(1 , size[2] - 2);
  r()((RI , RJ , RK)) = 
    w[0] *  (Grid) s()((I , J , K)) + 
    w[1] * ((Grid) u1()((RI , RJ , RK))  +   (Grid) s()((I , J , K-1)) +  (Grid) s()((I , J , K+1)))  + 
    w[2] * ((Grid) u2()((RI , RJ , RK))  +  (Grid) u1()((RI , RJ , RK-1)) + (Grid) u1()((RI , RJ , RK+1))) + 
    w[3] * ((Grid) u2()((RI , RJ , RK-1)) +  (Grid) u2()((RI , RJ , RK+1))); 
  return r;
}
Grid conv2 (Grid s , double w[4] , int step , Tuple<3> d)
{
    Tuple<3> size = s(T(0 , 0 , 0)).shape().size();  
  Tuple<3> rsize = Tuple<3>::max((size - 2) / step , 1) + 2;  
  Sequence<Tuple<3> > tiling(rsize , s.shape().size());  
  Grid u1 = HTA<double , 3>::alloc(1 , tiling , NULL , order);
  Grid u2 = HTA<double , 3>::alloc(1 , tiling , NULL , order); 
  Grid u3 = HTA<double , 3>::alloc(1 , tiling , NULL , order);
  Grid u4 = HTA<double , 3>::alloc(1 , tiling , NULL , order); 
  u1 = 0.0;
  u2 = 0.0;
  u3 = 0.0;
  u4  = 0.0;
  Grid r = HTA<double , 3>::alloc(1 , tiling , NULL , order);   
  r = 0.0;
  Triplet I(2*2-d[0] , 2*(rsize[0]-1)-d[0] , step);
  Triplet J(2*2-d[1] , 2*(rsize[1]-1)-d[1] , step);
  Triplet K(2*2-d[2] , 2*rsize[2]-d[2] , step);
  Triplet RI(1 , rsize[0] - 2);
  Triplet RJ(1 , rsize[1] - 2);
  Triplet RK(1 , rsize[2] - 1);
      
  u1()((RI , RJ , RK)) =  
    (Grid) s()((I , J-1 , K-1)) + 
    (Grid) s()((I , J+1 , K-1)) +
    (Grid) s()((I-1 , J , K-1)) +
    (Grid) s()((I+1 , J , K-1));
  u2()((RI , RJ , RK)) =  
    (Grid) s()((I-1 , J-1 , K-1)) +
    (Grid) s()((I-1 , J+1 , K-1)) +
    (Grid) s()((I+1 , J-1 , K-1)) +
    (Grid) s()((I+1 , J+1 , K-1));
  K = Triplet(2*2-d[2] , 2*(rsize[2]-1)-d[2] , step);
  RK = Triplet(1 , rsize[2] - 2);
       
  u3()((RI , RJ , RK)) =  
    (Grid) s()((I , J-1 , K)) + 
    (Grid) s()((I , J+1 , K)) +
    (Grid) s()((I-1 , J , K)) +
    (Grid) s()((I+1 , J , K));
  u4()((RI , RJ , RK)) =  
    (Grid) s()((I-1 , J-1 , K)) +
    (Grid) s()((I-1 , J+1 , K)) +
    (Grid) s()((I+1 , J-1 , K)) +
    (Grid) s()((I+1 , J+1 , K));
   r()((RI , RJ , RK)) = 
    w[0] *  (Grid) s()((I , J , K)) + 
    w[1] * ((Grid) u3()((RI , RJ , RK))  +   (Grid) s()((I , J , K-1)) +  (Grid) s()((I , J , K+1)))  + 
    w[2] * ((Grid) u4()((RI , RJ , RK))  +  (Grid) u1()((RI , RJ , RK)) + (Grid) u1()((RI , RJ , RK+1))) + 
    w[3] * ((Grid) u2()((RI , RJ , RK+1)) +  (Grid) u2()((RI , RJ , RK))); 
  
  return r;
}
void interp (Grid& z , Grid& u , int k)
{
    int s1 = z.shape().size()[0] == u.shape().size()[0] ? 1 : 2;
  int s2 = z.shape().size()[1] == u.shape().size()[1] ? 1 : 2;
  int s3 = z.shape().size()[2] == u.shape().size()[2] ? 1 : 2;
  Sequence<Triplet> t_args(Triplet (s1-1 , u.shape().size()[0]-1 , s1) ,    Triplet (s2-1 , u.shape().size()[1]-1 , s2) ,    Triplet (s3-1 , u.shape().size()[2]-1 , s3));
  Grid tmp_u = u (t_args);
  Shape<3>::iterator iz = z.shape().begin();
  Shape<3>::iterator iu = tmp_u.shape().begin();   
  Shape<3>::iterator iu_end = tmp_u.shape().end();
  if (z.shape() == u.shape()) {
      interp_real(z , u);
  } else {
    for (; iu != iu_end; iu++ , iz++)
      if (z(*iz).isLocal() && u(*iu).isLocal()) 
	interp_inner_special(z(*iz) , u(*iu));
    comm3_ex (u , s1 , s2 , s3);
  }  
    if (debug_vec[0]) rep_nrm (z , "interp" , k-1);
  if (debug_vec[0]) rep_nrm (u , "interp" , k);
}
void interp_real (Grid z , Grid u )
{  
    int iSize = z (T(0 , 0 , 0)).shape().size()[0];
  int jSize = z (T(0 , 0 , 0)).shape().size()[1];
  int kSize = z (T(0 , 0 , 0)).shape().size()[2];
  int nx = z.shape().size()[0];
  int ny = z.shape().size()[1];
  Grid tmpz1 = HTA<double , 3>::alloc(1 , tiling , NULL , order);
  Grid tmpz2 = HTA<double , 3>::alloc(1 , tiling , NULL , order);
  Grid tmpz3 = HTA<double , 3>::alloc(1 , tiling , NULL , order);
  Triplet I(0 , iSize-2);
  Triplet J(0 , jSize-2);
  Triplet K(0 , kSize-1);
  tmpz1()((I , J , K)) = (Grid) z()((I , J + 1 , K)) + (Grid) z()((I , J , K));
  tmpz2()((I , J , K)) = (Grid) z()((I + 1 , J , K)) + (Grid) z()((I , J , K));
  tmpz3()((I , J , K)) = (Grid) z()((I + 1 , J + 1 , K)) + (Grid) z()((I + 1 , J , K)) + (Grid) tmpz1()((I , J , K));    
  Triplet X(0 , iSize-2);
  Triplet Y(0 , jSize-2);
  Triplet Z(0 , kSize-2);
  Triplet A(0 , 2 * (iSize-2) , 2);
  Triplet B(0 , 2 * (jSize-2) , 2);
  Triplet C(0 , 2 * (kSize-2) , 2);
  u()((A , B , C)) =     (Grid) u()((A , B , C)) +        (Grid)     z()((X , Y , Z));  
  u()((A + 1 , B , C)) =     (Grid) u()((A + 1 , B , C)) + 0.5 *  (Grid) tmpz2()((X , Y , Z));
  u()((A , B + 1 , C)) =     (Grid) u()((A , B + 1 , C)) + 0.5 *  (Grid) tmpz1()((X , Y , Z));
  u()((A + 1 , B + 1 , C)) =     (Grid) u()((A + 1 , B + 1 , C)) + 0.25 * (Grid) tmpz3()((X , Y , Z));    
  u()((A , B , C + 1)) = (Grid) u()((A , B , C + 1)) + 0.5 *  ((Grid) z()(((X , Y , Z + 1))) + (Grid) z()((X , Y , Z)));
  u()((A + 1 , B , C + 1)) = (Grid) u()((A + 1 , B , C + 1)) + 0.25 * ((Grid) tmpz2()((X , Y , Z)) + (Grid) tmpz2()((X , Y , Z + 1) ));
  u()((A , B + 1 , C + 1)) = (Grid) u()((A , B + 1 , C + 1)) + 0.25 * ((Grid) tmpz1()((X , Y , Z)) + (Grid) tmpz1()((X , Y , Z + 1)));
  u()((A + 1 , B + 1 , C + 1)) = (Grid) u()((A + 1 , B + 1 , C + 1)) + 0.125 *((Grid) tmpz3()((X , Y , Z)) + (Grid) tmpz3()((X , Y , Z + 1)));
  }
void interp_inner_special (Leaf z , Leaf u)
{  
  int n1 = u.shape().size()[0]; 
  int n2 = u.shape().size()[1]; 
  int n3 = u.shape().size()[2]; 
  int d1 = n1 == 3 ? 1 : 0;
  int t1 = n1 == 3 ? 0 : -1;
  int d2 = n2 == 3 ? 1 : 0;
  int t2 = n2 == 3 ? 0 : -1;
  int d3 = n3 == 3 ? 1 : 0;
  int t3 = n3 == 3 ? 0 : -1;
  int mm1 = z.shape().size()[0];
  int mm2 = z.shape().size()[1];
  int mm3 = z.shape().size()[2];
  for (int i3 = d3; i3 < mm3-1; i3++)
    {
      for (int i2 = d2; i2 < mm2-1; i2++)
	{
	  for (int i1 = d1; i1 < mm1-1; i1++)
	    u [T(2*i1-d1 , 2*i2-d2 , 2*i3-d3)] += z [T(i1 , i2 , i3)];
						 
	  for (int i1 = 0; i1 < mm1-1; i1++)
	    u [T(2*i1-t1 , 2*i2-d2 , 2*i3-d3)] += 0.5 * (z[T(i1+1 , i2 , i3)] + z[T(i1 , i2 , i3)]);
	}
      for (int i2 = 0; i2 < mm2-1; i2++)
	{
	  for (int i1 = d1; i1 < mm1-1; i1++)
	    u [T(2*i1-d1 , 2*i2-t2 , 2*i3-d3)] += 0.5 * (z[T(i1 , i2+1 , i3)] + z[T(i1 , i2 , i3)]); 
	  for (int i1 = 0; i1 < mm1-1; i1++)
	    u [T(2*i1-t1 , 2*i2-t2 , 2*i3-d3)] += 0.25 * ( z[T(i1+1 , i2+1 , i3)] + z[T(i1+1 , i2 , i3)]
						     + z[T(i1 , i2+1 , i3)] + z[T(i1 , i2 , i3)]);
	}	           
    }
  for (int i3 = 0; i3 < mm3-1; i3++)
    {
      for (int i2 = d2; i2 < mm2-1; i2++)
	{
	  for (int i1 = d1; i1 < mm1-1; i1++)
	    u [T(2*i1-d1 , 2*i2-d2 , 2*i3-t3)] += 0.5 * (z[T(i1 , i2 , i3+1)] + z[T(i1 , i2 , i3)]);
	  
	  for (int i1 = 0; i1 < mm1-1; i1++)
	    u [T(2*i1-t1 , 2*i2-d2 , 2*i3-t3)] += 0.25 * ( z[T(i1+1 , i2 , i3+1)] + z[T(i1 , i2 , i3+1)]
						     + z[T(i1+1 , i2 , i3)]   + z[T(i1 , i2 , i3)]);
	}
      for (int i2 = 0; i2 < mm2-1; i2++)
	{
	  for (int i1 = d1; i1 < mm1-1; i1++)
	    u [T(2*i1-d1 , 2*i2-t2 , 2*i3-t3)] += 0.25 * (z[T(i1 , i2+1 , i3+1)] + z[T(i1 , i2 , i3+1)]
						      + z[T(i1 , i2+1 , i3)] + z[T(i1 , i2 , i3)]);
	  
	  for (int i1 = 0; i1 < mm1-1; i1++)
	    u [T(2*i1-t1 , 2*i2-t2 , 2*i3-t3)] += 0.125 * (z[T(i1+1 , i2+1 , i3+1)] + z [T(i1+1 , i2 , i3+1)]
						      +z[T(i1 , i2+1 , i3+1)] + z [T(i1 , i2 , i3+1)]
						      +z[T(i1+1 , i2+1 , i3  )] + z [T(i1+1 , i2 , i3)]
						      +z[T(i1 , i2+1 , i3  )] + z [T(i1 , i2 , i3)] );
	}      
    }
}
void zero3 (Grid& z)
{
    z = 0.0;
  }
/*========================================================COMM========================================================*/
void comm3 (Grid& u)
{
    int NX = u.shape().size()[0] - 1;
  int NY = u.shape().size()[1] - 1;
  int NZ = u.shape().size()[2] - 1;
  int nx = u(T(0 , 0 , 0)).shape().size()[0] - 1;
  int ny = u(T(0 , 0 , 0)).shape().size()[1] - 1;
  int nz = u(T(0 , 0 , 0)).shape().size()[2] - 1;
  
  Traits::Default::async();
  if (NX > 0) 
    u((R(0 , NX-1) , R(0 , NY) , R(0 , NZ)))((R(nx , nx) , R(1 , ny-1) , R(1 , nz-1))) = u((R(1 , NX) , R(0 , NY) , R(0 , NZ)))((R(1 , 1) , R(1 , ny-1) , R(1 , nz-1)));
  u((R(NX , NX) , R(0 , NY) , R(0 , NZ)))((R(nx , nx) , R(1 , ny-1) , R(1 , nz-1))) = u((R(0 , 0) , R(0 , NY) , R(0 , NZ)))((R(1 , 1) , R(1 , ny-1) , R(1 , nz-1)));
  if (NX > 0) 
    u((R(1 , NX) , R(0 , NY) , R(0 , NZ)))((R(0 , 0) , R(1 , ny-1) , R(1 , nz-1))) = u((R(0 , NX-1) , R(0 , NY) , R(0 , NZ)))((R(nx-1 , nx-1) , R(1 , ny-1) , R(1 , nz-1)));
  u((R(0 , 0) , R(0 , NY) , R(0 , NZ)))((R(0 , 0) , R(1 , ny-1) , R(1 , nz-1))) = u((R(NX , NX) , R(0 , NY) , R(0 , NZ)))((R(nx-1 , nx-1) , R(1 , ny-1) , R(1 , nz-1)));
  Traits::Default::sync();
  Traits::Default::async();
    if (NY > 0) 
    u((R(0 , NX) , R(0 , NY-1) , R(0 , NZ)))((R(0 , nx) , R(ny , ny) , R(1 , nz-1))) = u((R(0 , NX) , R(1 , NY) , R(0 , NZ)))((R(0 , nx) , R(1 , 1) , R(1 , nz-1)));
  u((R(0 , NX) , R(NY , NY) , R(0 , NZ)))((R(0 , nx) , R(ny , ny) , R(1 , nz-1))) = u((R(0 , NX) , R(0 , 0) , R(0 , NZ)))((R(0 , nx) , R(1 , 1) , R(1 , nz-1)));
  if (NY > 0) 
    u((R(0 , NX) , R(1 , NY) , R(0 , NZ)))((R(0 , nx) , R(0 , 0) , R(1 , nz-1))) = u((R(0 , NX) , R(0 , NY-1) , R(0 , NZ)))((R(0 , nx) , R(ny-1 , ny-1) , R(1 , nz-1)));
  u((R(0 , NX) , R(0 , 0) , R(0 , NZ)))((R(0 , nx) , R(0 , 0) , R(1 , nz-1))) = u((R(0 , NX) , R(NY , NY) , R(0 , NZ)))((R(0 , nx) , R(ny-1 , ny-1) , R(1 , nz-1)));
  Traits::Default::sync();
  Traits::Default::async();
  if (NZ > 0) 
    u((R(0 , NX) , R(0 , NY) , R(0 , NZ-1)))((R(0 , nx) , R(0 , ny) , R(nz , nz))) = u((R(0 , NX) , R(0 , NY) , R(1 , NZ)))((R(0 , nx) , R(0 , ny) , R(1 , 1)));
  u((R(0 , NX) , R(0 , NY) , R(NZ , NZ)))((R(0 , nx) , R(0 , ny) , R(nz , nz))) = u((R(0 , NX) , R(0 , NY) , R(0 , 0)))((R(0 , nx) , R(0 , ny) , R(1 , 1)));
  if (NZ > 0) 
    u((R(0 , NX) , R(0 , NY) , R(1 , NZ)))((R(0 , nx) , R(0 , ny) , R(0 , 0))) = u((R(0 , NX) , R(0 , NY) , R(0 , NZ-1)))((R(0 , nx) , R(0 , ny) , R(nz-1 , nz-1)));
  u((R(0 , NX) , R(0 , NY) , R(0 , 0)))((R(0 , nx) , R(0 , ny) , R(0 , 0))) = u((R(0 , NX) , R(0 , NY) , R(NZ , NZ)))((R(0 , nx) , R(0 , ny) , R(nz-1 , nz-1)));   
  Traits::Default::sync();
  }
 void comm3_ex (Grid& u , int s1 , int s2 , int s3)
 {
   if (s1 == 2) comm_ex_ns (u);
   if (s2 == 2) comm_ex_ew (u);
   if (s3 == 2) comm_ex_fb (u);
 }
 void comm_ex_ns (Grid& u)
 {
   int NX = u.shape().size()[0] - 1;
   int NY = u.shape().size()[1] - 1;
   int NZ = u.shape().size()[2] - 1;
   int nx = u (T(0 , 0 , 0)).shape().size()[0] - 1; 
   int ny = u (T(0 , 0 , 0)).shape().size()[1] - 1; 
   int nz = u (T(0 , 0 , 0)).shape().size()[2] - 1; 
   u((SR(0 , NX , 2) , R(0 , NY) , R(0 , NZ)))((R(2 , 2) , R(0 , ny) , R(0 , nz))) = u((SR(1 , NX , 2) , R(0 , NY) , R(0 , NZ)))((R(1 , 1) , R(0 , ny) , R(0 , nz)));
   u((SR(2 , NX-1 , 2) , R(0 , NY) , R(0 , NZ)))((R(0 , 1) , R(0 , ny) , R(0 , nz))) = u((SR(1 , NX-1 , 2) , R(0 , NY) , R(0 , NZ)))((R(1 , 2) , R(0 , ny) , R(0 , nz)));
   u((R(0 , 0) , R(0 , NY) , R(0 , NZ)))((R(0 , 1) , R(0 , ny) , R(0 , nz))) = u((R(NX , NX) , R(0 , NY) , R(0 , NZ)))((R(1 , 2) , R(0 , ny) , R(0 , nz)));
}
 void comm_ex_ew (Grid& u)
 {
   int NX = u.shape().size()[0] - 1;
   int NY = u.shape().size()[1] - 1;
   int NZ = u.shape().size()[2] - 1;
   int nx = u (T(0 , 0 , 0)).shape().size()[0] - 1; 
   int ny = u (T(0 , 0 , 0)).shape().size()[1] - 1; 
   int nz = u (T(0 , 0 , 0)).shape().size()[2] - 1; 
   u((R(0 , NX) , SR(0 , NY , 2) , R(0 , NZ)))((R(0 , nx) , R(2 , 2) , R(0 , nz))) = u((R(0 , NX) , SR(0 , NY , 2) , R(0 , NZ)))((R(0 , nx) , R(1 , 1) , R(0 , nz)));
   u((R(0 , NX) , SR(2 , NY-1 , 2) , R(0 , NZ)))((R(0 , nx) , R(0 , 1) , R(0 , nz))) = u((R(0 , NX) , SR(1 , NY-1 , 2) , R(0 , NZ)))((R(0 , nx) , R(1 , 2) , R(0 , nz)));
   u((R(0 , NX) , R(0 , 0) , R(0 , NZ)))((R(0 , nx) , R(0 , 1) , R(0 , nz))) = u((R(0 , NX) , R(NY , NY) , R(0 , NZ)))((R(0 , nx) , R(1 , 2) , R(0 , nz)));
 }
void comm_ex_fb (Grid& u)
 {
   int NX = u.shape().size()[0] - 1;
   int NY = u.shape().size()[1] - 1;
   int NZ = u.shape().size()[2] - 1;
   int nx = u (T(0 , 0 , 0)).shape().size()[0] - 1; 
   int ny = u (T(0 , 0 , 0)).shape().size()[1] - 1; 
   int nz = u (T(0 , 0 , 0)).shape().size()[2] - 1; 
   
   u((R(0 , NX) , R(0 , NY) , SR(0 , NZ , 2)))((R(0 , nx) , R(0 , ny) , R(2 , 2))) = u((R(0 , NX) , R(0 , NY) , SR(1 , NZ , 2)))((R(0 , nx) , R(0 , ny) , R(1 , 1)));
   u((R(0 , NX) , R(0 , NY) , SR(1 , NZ-1 , 2)))((R(0 , nx) , R(0 , ny) , R(0 , 1))) = u((R(0 , NX) , R(0 , NY) , SR(1 , NZ-1 , 2)))((R(0 , nx) , R(0 , ny) , R(1 , 2)));
   u((R(0 , NX) , R(0 , NY) , R(0 , 0)))((R(0 , nx) , R(0 , ny) , R(0 , 1))) = u((R(0 , NX) , R(0 , NY) , R(NZ , NZ)))((R(0 , nx) , R(0 , ny) , R(1 , 2)));
}
 void norm2u3 (Grid& r , double *rnm2 , double *rnmu , int nx , int ny , int nz)
 {
      double n =(double) nx * ny * nz;
   int tile_x = r (T(0 , 0 , 0)).shape().size()[0]; 
   int tile_y = r (T(0 , 0 , 0)).shape().size()[1]; 
   int tile_z = r (T(0 , 0 , 0)).shape().size()[2]; 
   Sequence<Triplet> shape_args1 (Triplet(1 , tile_x-2) , Triplet(1 , tile_y-2) , Triplet(1 , tile_z-2) );
   Shape<3> s1 [] = { r.shape() , Shape<3>(shape_args1) };
   SquareSum <double , 3> mr = SquareSum <double , 3>();
   r.tileAt(2 , s1).mapReduce(mr);
   *rnm2 = sqrt (mr.sqsum / (double) n);
   *rnmu = mr.max;
    }
 void rep_nrm (Grid &x , char* callee , int k )
 {
         int div = (int) pow(2.0 , lt-k-1);
   int nx = NX_DEFAULT / div;
   int ny = NY_DEFAULT / div;
   int nz = NZ_DEFAULT / div;
       }
/*==================================================INIT==============================================================*/
 void zran3 (Grid & z , int gx , int gy , int gz)
 {
   zero3 (z);
   Shape<3>::iterator iz = z.shape().begin();   
   Shape<3>::iterator iz_end = z.shape().end();   
   for (; iz != iz_end; iz++) {
     if (z(*iz).isLocal()) 
       zran3_inner(z(*iz) , *iz , gx , gy , gz);
   }
      FilterN <double , 10 , 3> mr = FilterN <double , 10 , 3>();
   z.mapReduce(mr);
   zero3 (z);
      for (int i = 0; i < 10; ++ i) {
     z [mr.result().minKey[i]] = -1;
     z [mr.result().maxKey[i]] = 1;
   }
 }
void zran3_inner (Leaf z , Tuple<3> tileIndex , int gx , int gy , int gz)
{
  double a = pow(5.0 , 13);
  int nx = z.shape().size()[0] - 2;
  int ny = z.shape().size()[1] - 2;
  int nz = z.shape().size()[2] - 2;
  Tuple<3> localIndexStart = tileIndex * Tuple<3> (nx , ny , nz);
  int pi = localIndexStart[0]  + localIndexStart[1] * gx + localIndexStart[2] * gx * gy;
  double ai = RandUtils::power (a , pi); 
  double a1 = RandUtils::power (a , gx);
  double a2 = RandUtils::power (a , gx*gy);
  double x0 = 314159265.0;
  double rdummy = RandUtils::randlc (&x0 , ai);
  for (int k = 1; k < z.shape().size()[2]-1; k++) {
    double x1 = x0;
    for (int j = 1; j < z.shape().size()[1]-1; j++) {
	  double x2 = x1;
	  for (int i = 1; i < z.shape().size()[0]-1; i++) 
	    z [T(i , j , k)] = RandUtils::randlc (&x2 , a);
	  rdummy = RandUtils::randlc (&x1 , a1);
    }
    rdummy = RandUtils::randlc (&x0 , a2);
  }
}	      
/*=====================================================ALLOC==============================================================*/
void setup (Grid* r , Grid* u , int nprocs)
{    
  int pi[3];
  int ng[3];
  int di[3];
  int si[3];
  double log_p  = log(double(nprocs)+0.0001)/log(2.0);
  int dx     = int(log_p/3);
  pi[0] = (int) pow(2.0 , dx);
  int dy     = int((log_p-dx)/2);
  pi[1]  = (int) pow (2.0 , dy);
  pi[2]  = nprocs/(pi[0]*pi[1]);
  ng[0] = NX_DEFAULT;
  ng[1] = NY_DEFAULT;
  ng[2] = NZ_DEFAULT;
  for (int i = 0; i < 3; i++)
    di[i] =   pi[i]; 
  for (int k = lt-1; k >=0; k--)
    {
      for (int i = 0; i < 3; i++)
	{  
	  di[i] = di[i] - (ng[i] % di[i] );
	  int totalSize = ng[i] + di[i] * 2;
	  si[i] = totalSize / di[i];
	  ng[i] /= 2; 
	}
                  
      r[k] = HTA<double , 3>::alloc (1 , (Tuple<3>(si[0] , si[1] , si[2]) , Tuple<3>(di[0] , di[1] , di[2])) , NULL , order);
      u[k] = HTA<double , 3>::alloc (1 , (Tuple<3>(si[0] , si[1] , si[2]) , Tuple<3>(di[0] , di[1] , di[2])) , NULL , order);
      zero3 (r[k]);
      zero3 (u[k]);
    }
}
