/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: ConstDistribution.h,v 1.11.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Christoph von Praun
 */


#ifndef __CONST_DISTRIBUTION_H__
#define __CONST_DISTRIBUTION_H__

using namespace std;

#include <cassert>
#include "Distribution.h"
#include "Tuple.h"

template <int DIM>
class ConstDistribution : public Distribution<DIM>
{
public:

  /** 
   * @return the node corresponding to an index 
   */
  virtual int operator[] (int idx) const 
  { 
    return place_; 
  }
   
  virtual int operator[] (const Tuple<DIM>& t) const 
  { 
    return place_; 
  }  


  virtual int place () const 
  { 
    return place_; 
  }

  /** 
   * @return the 'course' within the node, i.e. the block / tile number 
   * relative to the node.
   */
  virtual int course (int idx) const
  {
    return idx;
  }

  virtual bool operator== (const Distribution<DIM>& rhs) const
  {
    bool ret = false;
    const ConstDistribution<DIM>* tmp = dynamic_cast<const ConstDistribution*>(&rhs);
    if (tmp != NULL) 
      ret = (tmp->place_ == this->place_);
    return ret;
  }

  /**
   * factory method
   */
  static const Distribution<DIM>& get (int i)
  {
    assert (i >= 0 && i < MAX_);
    
    if (!initDone_) 
      init_();
    return pool_[i];
  }

  virtual int nPlaces () const 
  {
    return 1;
  }

  static const int MAX_ = 256;

  virtual Distribution<DIM>* clone() const 
  {
    return new ConstDistribution (*this);
  }

public:
  ConstDistribution () {}

  ConstDistribution (int node) : place_ (node) 
  {
    assert (node >= 0);
    assert (node < MAX_); // current limitation
  }

  virtual ~ConstDistribution () {}

  /** 
   * initializes the pool of distributions 
   */
  static void init_ ()
  {
    assert (!initDone_);
    for (int i = 0; i < MAX_; ++i) {
      pool_[i].place_ = i;
      pool_[i].disableRefCounting_();
    }
    initDone_ = true;
  }

  /** 
   * pool of constant distributions 
   */
  static ConstDistribution<DIM> pool_[];

  static bool initDone_;

  int place_;

  template <int D> 
  friend ostream& operator << (ostream& ost, const ConstDistribution<D>& d);

  void print_(ostream& ost, int indent = 0) const 
  {
    ost << "ConstDistribution(node=" << place_ << ")" << endl;
  }

};

template<int DIM> 
ostream& operator << (ostream& ost, const ConstDistribution<DIM>& d) 
{  
  d.print_(ost);
  return ost;
}

template<int DIM>
bool ConstDistribution<DIM>::initDone_ = false;

template<int DIM>
ConstDistribution<DIM> ConstDistribution<DIM>::pool_[ConstDistribution<DIM>::MAX_];

#endif /* __CONST_DISTRIBUTION_H__ */
