/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: CustomDistribution.h,v 1.6.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Christoph von Praun
 */


#ifndef __CUSTOM_DISTRIBUTION_H__
#define __CUSTOM_DISTRIBUTION_H__

using namespace std;

#include <cassert>
#include "AbstractArray.h"
#include "Distribution.h"
#include "Tuple.h"


template <typename T, int DIM>
class CustomDistribution : public Distribution<DIM>
{
public:

  CustomDistribution (const AbstractArray<T, DIM>& arr) : arr_ (arr) 
  {
    assert (arr_.level() > 0);
  }

  virtual ~CustomDistribution () {}

  /** 
   * @return the node correspnoding to an index 
   */
  virtual int operator[] (int idx) const { 
    return arr_.accessTile(idx).home();
  }
  
  virtual int operator[] (const Tuple<DIM>& t) const 
  { 
    return arr_.accessTile(t).home();
  }  

  /** 
   * @return the 'course' within the node, i.e. the block / tile number 
   * relative to the node.
   */
  virtual int course (int idx) const
  {
    int home = arr_.accessTile(idx).home();
    int course = 0;
    for (int i = 0; i < idx; ++i) {
      if (arr_.accessTile(i).home() == home)
	course ++;
    }
    return course;
  }

  virtual bool operator== (const Distribution<DIM>& rhs) const
  {
    bool ret = false;
    const CustomDistribution<T, DIM>* tmp = dynamic_cast<const CustomDistribution<T, DIM>*>(&rhs);
    if (tmp != NULL) 
      ret = (&(tmp->arr_) == &(this->arr_));
    return ret;
  }

  virtual int nPlaces () const 
  {
    assert (false);
    return 0;
  }

  virtual Distribution<DIM> * clone() const{
    return new CustomDistribution(this->arr_);
  }

private:

  const AbstractArray<T, DIM>& arr_;

  template <typename T2, int DIM2> 
  friend ostream& operator << (ostream& ost, const CustomDistribution<T2, DIM2>& d);

  void print_(ostream& ost, int indent = 0) const 
  {
    ost << "CustomDistribution()" << endl;
  }

};

template<typename T, int DIM> 
ostream& operator << (ostream& ost, const CustomDistribution<T, DIM>& d) 
{  
  d.print_(ost);
  return ost;
}

#endif /* __CUSTOM_DISTRIBUTION_H__ */
