/**
 * IBM SOFTWARE DISCLIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Distribution.h,v 1.45 2007/02/14 16:09:58 bikshand Exp $
 * Authors: Christoph von Praun
 */


#ifndef __DISTRIBUTION_H__
#define __DISTRIBUTION_H__

#include "util/RefCounted.h"
#include "Tuple.h"

using namespace std;

/**
 * Tile to processor mapping function
 */
template <int DIM>
class Distribution : public RefCounted 
{

public:
  
  virtual ~Distribution () {}

  /**
   * @param idx   tile idx
   * @return      the id of the processor where tile idx maps to
   */
  virtual int operator[] (int idx) const = 0;

  virtual int operator[] (const Tuple<DIM>& t) const = 0;
  
  /**
   * @return  id of the place if this is a constant distribution, 
   *          otherwise -1;
   */
  virtual int place () const { return -1; }

  /** 
   * @return the 'course' within the node, i.e. the block / tile number 
   * relative to the node.
   */
  virtual int course (int idx) const = 0;

  virtual bool operator== (const Distribution<DIM>& rhs) const = 0;

  bool operator!= (const Distribution<DIM>& rhs) const
  {
    bool isEqual = this->operator==(rhs);
    return !isEqual;
  }

  virtual Distribution<DIM>* clone() const 
  {
    assert (false);
  }

  virtual int nPlaces() const = 0;

private:

  template <int D> 
  friend ostream& operator << (ostream& ost, const Distribution<D>& d);

  virtual void print_(ostream& ost, int indent = 0) const = 0;

};

template<int DIM> 
ostream& operator << (ostream& ost, const Distribution<DIM>& d) {  
  d.print_(ost, 0);
  return ost;
}

#endif /* __DISTRIBUTION_H__ */
