/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: HTACache.h,v 1.8 2006/07/08 03:37:08 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __HTA_CACHE_H__
#define  __HTA_CACHE_H__

#include "AbstractArray.h"
#include "util/Cache.h"
#include "util/Finalizable.h"

using namespace std;

/* internal class used in support of the HTACache */
template <typename T, int DIM>
class HTAHashAndCompare_ : public HashAndCompare <AbstractArray<T, DIM> > 
{
public:

  static int hash (const T& v) 
  { 
    return v.hash();
  }
  
  static bool equal (const T& arg1, const T& arg2) 
  {
    return 
      T::memConformable(arg1, arg2) &&
      T::shapeConformable(arg1, arg2);
    return true;
  }
};

template <typename T, int DIM>
class HTACache 
  : public Cache <AbstractArray<T, DIM> , HTAHashAndCompare_<AbstractArray<T, DIM>, DIM> >, public Finalizable
{

public:

  static AbstractArray<T, DIM>* take (const AbstractArray<T, DIM>& key) 
  { 
    HTACache<T, DIM>& inst = getUniqueInstance_();
    AbstractArray<T, DIM>* ret = inst.take_(key);
    HTA_DEBUG(3, "HTACache::take arr=" << ret << " cache-fill=" << inst.fill());
    return ret;
  }
  
  static bool put (AbstractArray<T, DIM>* value) 
  {
    assert (value != NULL);
    HTACache<T, DIM>& inst = getUniqueInstance_();
    bool ret = value->isCachable() ? inst.put_(value) : false;
    HTA_DEBUG(3, "HTACache::put arr=" << value << " cache-fill=" << inst.fill());
    return ret;
  }

private:
  
  HTACache(int size) : 
    Cache<AbstractArray<T, DIM> , HTAHashAndCompare_<AbstractArray<T, DIM>, DIM> >(size), 
    Finalizable()
  {}
  
  virtual ~HTACache()
  {}

  virtual void finalize () 
  {
    HTA_DEBUG(1, "HTACache::finalize");
    this->clean_();
  }
  
  static  HTACache<T, DIM>& getUniqueInstance_() {
    if (uniqueInstance_ == NULL) 
      uniqueInstance_ = new  HTACache<T, DIM>(5);
    return *uniqueInstance_;
  }

  static HTACache<T, DIM>* uniqueInstance_;
  
};

template <typename T, int DIM>
HTACache<T, DIM>* HTACache<T, DIM>::uniqueInstance_ = NULL;

#endif  /* __HTA_CACHE_H__ */
