/** Copyright (c) 2007 Basilio B. Fraguela Universidade da Coruna */

#ifndef _HELPERINDXSHAPE_
#define _HELPERINDXSHAPE_

#include <cassert>
#include <iostream>

#include "Triplet.h"
#include "Tuple.h"

template<class T, int DIM, typename TRAIT>
class HTA;

template<class T, int DIM, typename TRAIT, int FILL>
class HelperIndxShape;

template<class T, int DIM, typename TRAIT>
class HelperIndxShape<T, DIM, TRAIT, 1> {

  public:

  const HTA<T, DIM, TRAIT>& hta_; 
  typename Triplet::Seq seq_;
  
  typedef HelperIndxShape<T, DIM, TRAIT, 1> RType;

  HelperIndxShape(const HTA<T, DIM, TRAIT> &hta, const Triplet &t)
    : hta_(hta), seq_(t)
    {
      //std::cout << "HelperIndxShape built with " << t << std::endl;
    }

  typename HelperIndxShape<T, DIM, TRAIT, 2>::RType operator[] (const Triplet &t)
  {
    return HelperIndxShape<T, DIM, TRAIT, 2>(*this, t);
  }

};

template<class T, int DIM, typename TRAIT, int FILL>
class HelperIndxShape {
  HelperIndxShape<T, DIM, TRAIT, 1> &orix_;
public:

  typedef HelperIndxShape<T, DIM, TRAIT, FILL> RType;
  
  HelperIndxShape(HelperIndxShape<T, DIM, TRAIT, 1> &orix, const Triplet &t) 
  : orix_(orix)
  {
    //std::cout << "HelperIndxShape extended with " << t << std::endl;
    orix_.seq_[FILL - 1] = t;
  }

  HelperIndxShape(const HTA<T, DIM, TRAIT>& hta, const Tuple<DIM>& tuple, const Triplet &t)
   : orix_(hta, Triplet(tuple[0]))
  {
    for(int i = 1; i < (FILL - 1); ++i) 
		orix_.seq[i] = Triplet(tuple[i]);
	orix_.seq_[FILL - 1] = t;
  }

  typename HelperIndxShape<T, DIM, TRAIT, FILL+1>::RType operator[] (const Triplet &t)
  {
    return HelperIndxShape<T, DIM, TRAIT, FILL+1>(orix_, t);
  }
  
};

template<class T, int DIM, typename TRAIT>
class HelperIndxShape<T, DIM, TRAIT, DIM> {
  const HTA<T, DIM, TRAIT> &hta_;
  Shape<DIM> s_;
  
  static typename Triplet::Seq BuildTripletSeq(const Tuple<DIM>& tuple) {
    typename Triplet::Seq ret;
	for(int i = 0; i < (DIM - 1); ++i) 
		ret[i] = Triplet(tuple[i]);
	return ret;
  }

public:
  typedef typename TRAIT::Type RType;
	
  HelperIndxShape(HelperIndxShape<T, DIM, TRAIT, 1> &orix, const Triplet &t)
    : hta_(orix.hta_), s_((orix.seq_, t))
	//v_((orix.seq_[DIM - 1] = t, (orix.hta_)[Shape<DIM>(orix.seq_)]))
    {
      //std::cout << "HelperIndxShape " << s_ << std::endl;
    }
    
  HelperIndxShape(const HTA<T, DIM, TRAIT>& hta, const Tuple<DIM>& tuple, const Triplet &t)
   : hta_(hta), s_((BuildTripletSeq(tuple), t))
    // v_(hta[Shape<DIM>((BuildTripletSeq(tuple), t))])
  {
    //std::cout << "HelperIndxShape " << s_ << std::endl;
  }

  operator typename TRAIT::Type () const { return hta_[s_]; }

  typename TRAIT::Type operator= (const typename TRAIT::Type& val) { return hta_[s_] = val; }
};

template<class T, typename TRAIT>
class HelperIndxShape<T, 1, TRAIT, 1> {
  const HTA<T, 1, TRAIT> &hta_;
  Shape<1> s_;

public:

  typedef typename TRAIT::Type RType;
	
  HelperIndxShape(const HTA<T, 1, TRAIT> &hta, const Triplet &t)
    : hta_(hta), s_(typename Triplet::Seq(t))
    { }
	
  operator typename TRAIT::Type () const { return hta_[s_]; }

  typename TRAIT::Type operator= (const typename TRAIT::Type& val) { return hta_[s_] = val; }
};

#endif
