/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: IbmESSL.h,v 1.3 2006/08/20 21:13:46 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __IBM_ESSL_H__
#define  __IBM_ESSL_H__

#define _COMPLEX_

#include <iostream>

#include "essl.h"
#include "MathKernel.h"
#include "debug/Tracing.h"
#include "util/complex.h"

using namespace std;

/** 
 * Wrapper for IBM ESSL 
 * - this code does not work yet -
 */
template<typename T>
class IbmESSL : public MathKernel<T> {
  
public:
 
  IbmESSL (int size, const int* stride) :
    MathKernel<T>(), 
    size_(size), 
    stride_(stride[1]), 
    naux1_((size > 2048) ? (int) (20000.0 + (2.28 * (double) size)) : 20000), // magic formula in ESSL
    aux1_(new double[naux1_]),
    naux2_((size > 2048) ? (int) (20000.0 + (2.28 * (double) size)) : 20000), // magic formula in ESSL
    aux2_(new double[naux2_]),
    auxScale_(0.0), 
    auxIsign_(0)
  { 
    assert (0); // this code does not work yet, it has not been tested.
  }

  virtual ~IbmESSL () 
  {
    delete[] aux1_;
    delete[] aux2_;
  }

  virtual void dftForward (T* arr) 
  {
    // check alignment on doubleword boundary
    assert ((((unsigned) arr) & 0x7u) == 0);

    HTA_DEBUG (1, "IbmESSL ft-forward: stride=" << stride_ << " size=" << size_);
    double scale = 1.0;
    int isign = 1;
    // initAux_(isign, scale); // CVP: for some reason this does not work - the compute call 
    // terminates with an error
    dcft (1, NULL, stride_, 0, NULL, stride_, 0, size_, 1, isign, scale, aux1_, naux1_, aux2_, naux2_); 
    // (int, void *, int, int, void *, int, int, int, int, int, double, double *, int, double *, int);
    // compute
    dcft (0, arr, stride_, 0, arr, stride_, 0, size_, 1, isign, scale, aux1_, naux1_, aux2_, naux2_);
  }
  
  virtual void dftBackward (T* arr) 
  {
    double scale = 1.0 / (double) size_;
    int isign = -1;
    //initAux_(isign, scale); 
    dcft (1, NULL, stride_, 0, NULL, stride_, 0, size_, 1, isign, scale, aux1_, naux1_, aux2_, naux2_);
    // compute
    dcft (0, arr, stride_, 0, arr, stride_, 0, size_, 1, isign, scale, aux1_, naux1_, aux2_, naux2_);
  }
  
private:

  void initAux_ (int isign, double scale) 
  {
    // init aux1
    //if (scale != auxScale_ || isign != auxIsign_) {
      dcft (1, NULL, stride_, 0, NULL, stride_, 0, size_, 1, 1, 1.0, aux1_, naux1_, aux2_, naux2_);
      //auxScale_ = scale;
      //auxIsign_ = isign;
      //}
  }

  const int size_;
  const int stride_;
  const int naux1_;
  double* aux1_; 
  const int naux2_;
  double* aux2_;
  double auxScale_;
  int auxIsign_;
};

#endif /* __IBM_ESSL_H__ */
