/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: IntelMKL.h,v 1.8 2006/08/08 16:43:32 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __INTEL_MKL_H__
#define  __INTEL_MKL_H__

#include <iostream>

#include "MathKernel.h"
#include "mkl_dfti.h"
#include "mkl_spblas.h"
#include "debug/Tracing.h"

using namespace std;

/** 
 * Wrapper for Intel MKL
 */
template<typename T>
class IntelMKL : public MathKernel<T> {
  
public:
 
  IntelMKL (int size, const int* stride) :
    MathKernel<T>()
  { 
    long status;
    status = DftiCreateDescriptor (&handle_, DFTI_DOUBLE, DFTI_COMPLEX, 1, size);
    checkStatus_(status);
    status = DftiSetValue (handle_, DFTI_INPUT_STRIDES, stride);
    checkStatus_(status);
    status = DftiSetValue (handle_, DFTI_OUTPUT_STRIDES, stride);
    checkStatus_(status);
    status = DftiCommitDescriptor (handle_);
    checkStatus_(status);
  }
  
  virtual ~IntelMKL () 
  {
    DftiFreeDescriptor (&handle_);
  }

  void dftForward (T* arr) 
  {
    long status = DftiComputeForward (handle_, arr);
    checkStatus_(status);
  }
  
  void dftBackward (T* arr) 
  {
    long status = DftiComputeBackward (handle_, arr);
    checkStatus_(status);
  }
  
  /** 
   * Computes matrix - matrix product of a sparse matrix stored in the
   * compressed sparse column format 
   */
  static void dcscmm (char *transa, 
		      int *m, int *n, int *k, 
		      T *alpha, 
		      char *matdescra, T *val, int *indx,  int *pntrb, int *pntre, 
		      T *b, int *ldb, 
		      T *beta, 
		      T *c, int *ldc) 
  {
    mkl_dcscmm(transa, m, n, k, alpha, matdescra, val, indx, pntrb, pntre, b, ldb, beta, c, ldc);
  }

  static void dcsrmm (char *transa, 
		      int *m, int *n, int *k, 
		      T *alpha, 
		      char *matdescra, T *val, int *indx,  int *pntrb, int *pntre, 
		      T *b, int *ldb, 
		      T *beta, 
		      T *c, int *ldc) 
  {
    mkl_dcsrmm(transa, m, n, k, alpha, matdescra, val, indx, pntrb, pntre, b, ldb, beta, c, ldc);
  }
    
private:
  
  inline void checkStatus_ (long status) 
  {
    if (status) 
      HTA_WARN (0, "IntelMKL::checkStatus_ error:" << DftiErrorMessage(status));
  }
  
  DFTI_DESCRIPTOR_HANDLE handle_;

};

#endif /* __INTEL_MKL_H__ */
