/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: LeafMemMapping.h,v 1.10.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */


#ifndef __LEAF_MEM_MAPPING_H__
#define __LEAF_MEM_MAPPING_H__

#include <cassert>

#include "MemMapping.h"
#include "Tuple.h"


using namespace std;


template <int DIM>
class LeafMemMapping : public MemMapping <DIM>
{

public:

  /*
   * @param off       offset in the physical representation
   * @param nlevels   number of levels of the HTA where this leaf belongs to
   * @param tiling    full tiling information
   * @param leaf_pos  position where this tile starts in the array 
   */
  LeafMemMapping (int off, int nlevels,
		  const typename Tuple<DIM>::Seq& tiling,
		  const Tuple<DIM> position, 
		  const Tuple<DIM>* linear_step = NULL) :
    MemMapping<DIM>(ROW, 0, off, nlevels, tiling, linear_step), 
    leafPos_(position)
  { 
    HTA_DEBUG (2, "LeafMemMapping::<init> position=" << position << " this= " << *this);
  }
  
  virtual ~LeafMemMapping () 
  {
    //cout << "here " << this << endl;
  }
  
  virtual const MemMapping<DIM>* select (const MemMapping<DIM>** children, 
					 const Tuple<DIM>& tiling) const
  {
    assert (false);
    return NULL;
  }
  
  virtual const MemMapping<DIM>* clone () const
  {
    return  new LeafMemMapping<DIM> (this->offset(), 0, this->tiling(), this->leafPos_ / this->leafSize());
  }

  virtual const MemMapping<DIM>* clone_helper (int offset) const
  {
    return  new LeafMemMapping<DIM> (offset, 0, this->tiling(), this->leafPos_ / this->leafSize());
  }
  
  /**
   * @return  the distribution of the idx's tile in the 
   *          enumation of the tiles according to the ordering
   */
  virtual const LeafMemMapping<DIM>& operator[] (int idx) const 
  {
    assert (false);
    return *this;
  }
  
  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling, 
					 const Distribution<DIM>& dist) const 
  {
    return create (tiling);
  }

  virtual const MemMapping<DIM>* create (typename Tuple<DIM>::Seq& tiling) const 
  {
    // assert (this->leafPos() == Tuple<DIM>::zero); -- why this assertion CVP1 ??
    return new LeafMemMapping<DIM> (this->offset(), 
				    this->nlevels_, 
				    tiling, 
				    Tuple<DIM>::zero);
  }

  /**
   * @return the position that this array has within the 
   *         flattened overall array.
   */
  virtual const Tuple<DIM>& leafPos() const 
  {
    return leafPos_;
  }

private:
  
  virtual void print_(ostream& ost, int indent = 0) const 
  {
    ost << "LeafMemMapping<" << DIM << ">(" 
	<< "this=" << this
	<< " offset=" << this->offset()
	<< "card=" << this->linearCard()
	<< " linearstep=" << this->linearStep() << ")";
  }

  const Tuple<DIM> leafPos_;
};


#endif /* __LEAF_MEM_MAPPING_H__ */
