#ifndef  __MAPS_H__
#define  __MAPS_H__

#include "Array.h"
#include "tbb/blocked_range.h"
#include "tbb/parallel_for.h"

template<typename T, int DIM, typename TRAIT, typename OP>
class Hmap1Body {
  OP& my_op;
  size_t my_level;
  bool* my_mask;
  Array<T,DIM,TRAIT>* array;
 public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //    printf("hmap1body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j]; // what is "this"?  
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ( (!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level);
    }
  }
  Hmap1Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level, bool* mask) : array(a), my_op(op), my_level(level), my_mask(mask) 
  {}
};

template<typename T, int DIM, typename TRAIT>
template<typename OP>
 void Array<T, DIM, TRAIT>::hmap (OP& op,
				  size_t level,
				  bool* mask)
  
{
  if(this->level() == level)
    {
      //      cout << "Here " << endl;
      assert (this->isLocal());	
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)));	
    }  else  {
      update_on_read(mask);
      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap1Body<T, DIM, TRAIT, OP>(this, op, level, mask));
	// for (int j=0 ; j < this->numTiles_; j++) 
// 	  {
// 	    int i = this->localTiles_[j];   
// 	    typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	    if ( (!mask || mask[i]) && tile->isLocal()) 
// 	      tile->hmap (op, level);
// 	  }
	set_on_write(mask);
    }
}

template<typename T, int DIM, typename TRAIT, typename T1, int DIM1, typename TRAIT1, typename OP>
class Hmap2Body {
  Array<T, DIM, TRAIT>* array;
  OP& my_op;
  size_t my_level;
  Array<T1, DIM1, TRAIT1>* my_arg1;
  bool* my_mask;
public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //printf("hmap2body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j];   
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ( (!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level, my_arg1->tileAt(i));
    }
  }
  Hmap2Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level,
	    Array<T1, DIM1, TRAIT1>* arg1, bool* mask) : array(a), my_op(op), my_level(level), my_arg1(arg1), my_mask(mask) 
  {}
};

template<typename T, int DIM, typename TRAIT>
template<typename OP, class T1, int DIM1, typename TRAIT1>
 void Array<T, DIM, TRAIT>::hmap (OP& op,
				  size_t level,
				  Array<T1, DIM1, TRAIT1>* arg1,
				  bool* mask)
  
{
  if(this->level() == level)
    {
      //      cout << "Here " << endl;
      assert (this->isLocal());	
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)), 
	  HTA<T1, DIM1, TRAIT1> (static_cast<typename TRAIT1::TypeImpl&>(*arg1)));
	
    }  else  {
      arg1->update_on_read(mask);
      update_on_read(mask);
      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap2Body<T, DIM, TRAIT, T1, DIM1, TRAIT1, OP>(this, op, level, arg1, mask));
      // for (int j=0 ; j < this->numTiles_; j++) 
// 	{
	  
// 	  int i = this->localTiles_[j];   
// 	  typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	  if ( (!mask || mask[i]) && tile->isLocal()) 
// 	    tile->hmap (op, level, arg1->tileAt(i));
// 	}
      arg1->set_on_write(mask);
      set_on_write(mask);
    }
}

template<typename T, int DIM, typename TRAIT, typename T1, int DIM1, typename TRAIT1, typename T2, int DIM2, typename TRAIT2, typename OP>
class Hmap3Body {
  Array<T, DIM, TRAIT>* array;
  OP& my_op;
  size_t my_level;
  Array<T1, DIM1, TRAIT1>* my_arg1;
  Array<T2, DIM2, TRAIT2>* my_arg2;
  bool* my_mask;
public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //    printf("hmap3body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j];   
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ( (!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level, my_arg1->tileAt(i), my_arg2->tileAt(i));
    }
  }
  Hmap3Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level,
	    Array<T1, DIM1, TRAIT1>* arg1, Array<T2, DIM2, TRAIT2>* arg2, bool* mask) : array(a), my_op(op), my_level(level), my_arg1(arg1), my_arg2(arg2), my_mask(mask)
  {}
};

template<typename T, int DIM, typename TRAIT>
template<typename OP, class T1, int DIM1, typename TRAIT1, class T2, int DIM2, typename TRAIT2>
 void Array<T, DIM, TRAIT>::hmap (OP& op,
				  size_t level,
				  Array<T1, DIM1, TRAIT1>* arg1,
				  Array<T2, DIM2, TRAIT2>* arg2,
				  bool* mask)
	 
{
  //  cout << "Inside hmap " << endl;
  if(this->level() == level)
    {      
      // cout << "Here " << endl;
      assert (this->isLocal());	
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)), 
	  HTA<T1, DIM1, TRAIT1> (static_cast<typename TRAIT1::TypeImpl&>(*arg1)), 
	  HTA<T2, DIM2, TRAIT2> (static_cast<typename TRAIT2::TypeImpl&>(*arg2)));

    }  else  {
      arg1->update_on_read(mask);
      arg2->update_on_read(mask);
      update_on_read(mask);
      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap3Body<T, DIM, TRAIT, T1, DIM1, TRAIT1, T2, DIM2, TRAIT2, OP>(this, op, level, arg1, arg2, mask));
      // for (int j=0 ; j < this->numTiles_; j++) 
// 	{
// 	  int i = this->localTiles_[j];   
// 	  typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	  if ( (!mask || mask[i]) && tile->isLocal()) 
// 	    tile->hmap (op, level, arg1->tileAt(i), arg2->tileAt(i));
// 	}
      arg1->set_on_write(mask);
      arg2->set_on_write(mask);
      set_on_write(mask);
      
    }
}

template<typename T, int DIM, typename TRAIT, typename T1, int DIM1, typename TRAIT1, typename T2, int DIM2, typename TRAIT2, typename T3, int DIM3, typename TRAIT3, typename OP>
class Hmap4Body {
  Array<T, DIM, TRAIT>* array;
  OP& my_op;
  size_t my_level;
  Array<T1, DIM1, TRAIT1>* my_arg1;
  Array<T2, DIM2, TRAIT2>* my_arg2;
  Array<T3, DIM3, TRAIT3>* my_arg3;
  bool* my_mask;
public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //    printf("hmap4body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j];  
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ((!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level, my_arg1->tileAt(i), my_arg2->tileAt(i), my_arg3->tileAt(i));
    }
  }
  Hmap4Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level,
	    Array<T1, DIM1, TRAIT1>* arg1, Array<T2, DIM2, TRAIT2>* arg2,
	    Array<T3, DIM3, TRAIT3>* arg3, bool* mask) :
    array(a), my_op(op), my_arg1(arg1), my_arg2(arg2), my_arg3(arg3), my_mask(mask)
    {}
};

template<typename T, int DIM, typename TRAIT>
template<typename OP, class T1, int DIM1, typename TRAIT1, class T2, int DIM2, typename TRAIT2, class T3, int DIM3, typename TRAIT3>
 void Array<T, DIM, TRAIT>::hmap (OP& op,
					   size_t level,
					   Array<T1, DIM1, TRAIT1>* arg1,
					   Array<T2, DIM2, TRAIT2>* arg2,
					   Array<T3, DIM3, TRAIT3>* arg3,
					   bool* mask) 
{
  if(this->level() == level)
    {
      assert (this->isLocal());	
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)), 
	  HTA<T1, DIM1, TRAIT1> (static_cast<typename TRAIT1::TypeImpl&>(*arg1)), 
	  HTA<T2, DIM2, TRAIT2> (static_cast<typename TRAIT2::TypeImpl&>(*arg2)),
	  HTA<T3, DIM3, TRAIT3> (static_cast<typename TRAIT3::TypeImpl&>(*arg3)));      
    }  else  {
      arg1->update_on_read(mask);
      arg2->update_on_read(mask);
      arg3->update_on_read(mask);
      update_on_read(mask);

      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap4Body<T, DIM, TRAIT, T1, DIM1, TRAIT1, T2, DIM2, TRAIT2, T3, DIM3, TRAIT3, OP>(this, op, level, arg1, arg2, arg3, mask));
      // for (int j=0 ; j < this->numTiles_; j++) 
// 	{
// 	 int i = this->localTiles_[j];  
// 	typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	if ((!mask || mask[i]) && tile->isLocal()) 
// 	  tile->hmap (op, level, arg1->tileAt(i), arg2->tileAt(i), arg3->tileAt(i));
//       }
      arg1->set_on_write(mask);
      arg2->set_on_write(mask);
      arg3->set_on_write(mask);
      set_on_write(mask);
    }
}

template<typename T, int DIM, typename TRAIT, typename OP>
class Hmap5Body {
  Array<T, DIM, TRAIT>* array;
  OP& my_op;
  size_t my_level;
  Array<T, DIM, TRAIT>* my_arg1;
  Array<T, DIM, TRAIT>* my_arg2;
  Array<T, DIM, TRAIT>* my_arg3;
  Array<T, DIM, TRAIT>* my_arg4;
  bool* my_mask;
public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //    printf("hmap5body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j];  
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ( (!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level, my_arg1->tileAt(i), my_arg2->tileAt(i), my_arg3->tileAt(i), my_arg4->tileAt(i));
    }
  }
  Hmap5Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level, 
	    Array<T, DIM, TRAIT>* arg1, Array<T, DIM, TRAIT>* arg2, 
	    Array<T, DIM, TRAIT>* arg3, Array<T, DIM, TRAIT>* arg4, 
	    bool* mask) : array(a),
    my_op(op), my_level(level), my_arg1(arg1), my_arg2(arg2),
    my_arg3(arg3), my_arg4(arg4), my_mask(mask)
  {}
};

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1 ,
					   Array<T, DIM, TRAIT>* arg2 ,
					   Array<T, DIM, TRAIT>* arg3 ,
					   Array<T, DIM, TRAIT>* arg4,
					   bool* mask) 
{
  if(this->level() == level)
    {
      assert (this->isLocal());	
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)), 
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg1)), 
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg2)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg3)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg4)));

    }  else  {
      arg1->update_on_read(mask);
      arg2->update_on_read(mask);
      arg3->update_on_read(mask);
      arg4->update_on_read(mask);
      update_on_read(mask);
      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap5Body<T, DIM, TRAIT, OP>(this, op, level, arg1, arg2, arg3, arg4, mask));
      // for (int j=0 ; j < this->numTiles_; j++) 
// 	{
// 	  int i = this->localTiles_[j];  
// 	  typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	  if ( (!mask || mask[i]) && tile->isLocal()) 
// 	  tile->hmap (op, level, arg1->tileAt(i), arg2->tileAt(i), arg3->tileAt(i), arg4->tileAt(i));
// 	}
      arg1->set_on_write(mask);
      arg2->set_on_write(mask);
      arg3->set_on_write(mask);
      arg4->set_on_write(mask);
      set_on_write(mask);
    }
}

template<typename T, int DIM, typename TRAIT, typename OP>
class Hmap6Body {
  Array<T, DIM, TRAIT>* array;
  OP& my_op;
  size_t my_level;
  Array<T, DIM, TRAIT>* my_arg1;
  Array<T, DIM, TRAIT>* my_arg2;
  Array<T, DIM, TRAIT>* my_arg3;
  Array<T, DIM, TRAIT>* my_arg4;
  Array<T, DIM, TRAIT>* my_arg5;
  bool* my_mask;
public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //    printf("hmap6body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j];   
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ( (!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level, my_arg1->tileAt(i), my_arg2->tileAt(i), my_arg3->tileAt(i), my_arg4->tileAt(i), my_arg5->tileAt(i));
    }
  }
  Hmap6Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level,
	    Array<T, DIM, TRAIT>* arg1, Array<T, DIM, TRAIT>* arg2,
	    Array<T, DIM, TRAIT>* arg3, Array<T, DIM, TRAIT>* arg4,
	    Array<T, DIM, TRAIT>* arg5, bool* mask) :
    array(a), my_level(level), my_arg1(arg1), my_arg2(arg2),
    my_arg3(arg3), my_arg4(arg4), my_arg5(arg5), my_mask(mask)
  {}
};

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1 ,
					   Array<T, DIM, TRAIT>* arg2 ,
					   Array<T, DIM, TRAIT>* arg3 ,
					   Array<T, DIM, TRAIT>* arg4 ,
					   Array<T, DIM, TRAIT>* arg5,
					   bool* mask) 
{
  if(this->level() == level)
    {
      assert (this->isLocal());	
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)), 
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg1)), 
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg2)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg3)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg4)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg5)));	  
    }  else  {
      arg1->update_on_read(mask);
      arg2->update_on_read(mask);
      arg3->update_on_read(mask);
      arg4->update_on_read(mask);
      arg5->update_on_read(mask);
      update_on_read(mask);
      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap6Body<T, DIM, TRAIT, OP>(this, op, level, arg1, arg2, arg3, arg4, arg5, mask));
      // for (int j=0 ; j < this->numTiles_; j++) 
// 	{
// 	  int i = this->localTiles_[j];   
// 	typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	if ( (!mask || mask[i]) && tile->isLocal()) 
// 	  tile->hmap (op, level, arg1->tileAt(i), arg2->tileAt(i), arg3->tileAt(i), arg4->tileAt(i), arg5->tileAt(i));
//       }
      arg1->set_on_write(mask);
      arg2->set_on_write(mask);
      arg3->set_on_write(mask);
      arg4->set_on_write(mask);
      arg5->set_on_write(mask);
      set_on_write(mask);
    }
}
  
template<typename T, int DIM, typename TRAIT, typename OP>
class Hmap7Body {
  Array<T, DIM, TRAIT>* array;
  OP& my_op;
  size_t my_level;
  Array<T, DIM, TRAIT>* my_arg1;
  Array<T, DIM, TRAIT>* my_arg2;
  Array<T, DIM, TRAIT>* my_arg3;
  Array<T, DIM, TRAIT>* my_arg4;
  Array<T, DIM, TRAIT>* my_arg5;
  Array<T, DIM, TRAIT>* my_arg6;
  Array<T, DIM, TRAIT>* my_arg7;
  bool* my_mask;
public:
  void operator() ( const tbb::blocked_range<int>& range ) const {
    int j_end = range.end();
    //    printf("hmap7body\n");
    for (int j = range.begin(); j < j_end; j++) {
      int i = array->localTiles_[j];   
      typename TRAIT::TypeTileImpl* tile = array->tileAt (i);
      if ( (!my_mask || my_mask[i]) && tile->isLocal()) 
	tile->hmap (my_op, my_level, my_arg1->tileAt(i), 
		    my_arg2->tileAt(i), my_arg3->tileAt(i), 
		    my_arg4->tileAt(i), my_arg5->tileAt(i),
		    my_arg6->tileAt(i), my_arg7->tileAt(i));
    }
  }
  Hmap7Body(Array<T, DIM, TRAIT>* a, OP& op, size_t level,
	    Array<T, DIM, TRAIT>* arg1, Array<T, DIM, TRAIT>* arg2,
	    Array<T, DIM, TRAIT>* arg3, Array<T, DIM, TRAIT>* arg4,
	    Array<T, DIM, TRAIT>* arg5, Array<T, DIM, TRAIT>* arg6,
	    Array<T, DIM, TRAIT>* arg7, bool* mask) :
    array(a), my_op(op), my_level(level), my_arg1(arg1), my_arg2(arg2),
    my_arg3(arg3), my_arg4(arg4), my_arg5(arg5), my_arg6(arg6),
    my_arg7(arg7), my_mask(mask) 
  {}
	    
};

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1 ,
					   Array<T, DIM, TRAIT>* arg2 ,
					   Array<T, DIM, TRAIT>* arg3 ,
					   Array<T, DIM, TRAIT>* arg4 ,
					   Array<T, DIM, TRAIT>* arg5,
					   Array<T, DIM, TRAIT>* arg6,
					   Array<T, DIM, TRAIT>* arg7,
					   bool* mask) 

{
  if(this->level() == level)
    {
      assert (this->isLocal());	
      //cout << "here " << level << this->level() << endl;
      op (HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*this)), 
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg1)), 
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg2)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg3)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg4)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg5)),
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg6)),	  
	  HTA<T, DIM, TRAIT> (static_cast<typename TRAIT::TypeImpl&>(*arg7)));	  
    }  else  {
      assert (false);
      tbb::parallel_for(tbb::blocked_range<int>(0, this->numTiles_), Hmap7Body<T, DIM, TRAIT, OP>(this, op, level, arg1, arg2, arg3, arg4, arg5, arg6, arg7, mask));
      // for (int j=0 ; j < this->numTiles_; j++) 
// 	{
	  
// 	  int i = this->localTiles_[j];   
// 	  typename TRAIT::TypeTileImpl* tile = this->tileAt (i);
// 	  if ( (!mask || mask[i]) && tile->isLocal()) 
// 	  tile->hmap (op, level, arg1->tileAt(i), 
// 		      arg2->tileAt(i), arg3->tileAt(i), arg4->tileAt(i), arg5->tileAt(i),
// 		      arg6->tileAt(i), arg7->tileAt(i));
//       }
    }
}
  
template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (int n,
					   const Shape<DIM>* const* s1,
					   const Shape<DIM>* const* s2,	     
					   OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1,
					   bool* mask)
    
{
  assert (this->level() > 0);
       
  // (1) allocate / compute the children
  // responsibility to de-allocate is taken by the newly created 
  // MemMapping and HTA respectively
  switch (DIM) {
  case 3:  
  case 4:
    hmap3D_(n, s1, s2, op, level, arg1, mask);
    break;
  default:
    break;
  }
}

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (int n,
					   const Shape<DIM>* const* s1,
					   const Shape<DIM>* const* s2,
					   const Shape<DIM>* const* s3,	     
					   OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1,
					   Array<T, DIM, TRAIT>* arg2,
					   bool* mask)    
{
  assert (this->level() > 0);
       
  // (1) allocate / compute the children
  // responsibility to de-allocate is taken by the newly created 
  // MemMapping and HTA respectively
  switch (DIM) {
  case 3:  
  case 4:
    hmap3D_(n, s1, s2, s3, op, level, arg1, arg2, mask);
    break;
  default:
    break;
  }
}

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (int n,
					   const Shape<DIM>* const* s1,
					   const Shape<DIM>* const* s2,
					   const Shape<DIM>* const* s3,	     
					   const Shape<DIM>* const* s4,	     
					   OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1,
					   Array<T, DIM, TRAIT>* arg2,
					   Array<T, DIM, TRAIT>* arg3,
					   bool* mask)    
{
  assert (this->level() > 0);
       
  // (1) allocate / compute the children
  // responsibility to de-allocate is taken by the newly created 
  // MemMapping and HTA respectively
  switch (DIM) {
  case 3:  
  case 4:
    hmap3D_(n, s1, s2, s3, s4, op, level, arg1, arg2, arg3, mask);
    break;
  default:
    break;
  }
}

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (int n,
					   const Shape<DIM>* const* s1,
					   const Shape<DIM>* const* s2,
					   const Shape<DIM>* const* s3,	     
					   const Shape<DIM>* const* s4,	     
					   const Shape<DIM>* const* s5,	     
					   OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1,
					   Array<T, DIM, TRAIT>* arg2,
					   Array<T, DIM, TRAIT>* arg3,
					   Array<T, DIM, TRAIT>* arg4,
					   bool* mask)
    
{
  assert (this->level() > 0);
       
  // (1) allocate / compute the children
  // responsibility to de-allocate is taken by the newly created 
  // MemMapping and HTA respectively
  switch (DIM) {
  case 3:  
  case 4:
    hmap3D_(n, s1, s2, s3, s4, s5, op, level, arg1, arg2, arg3, arg4, mask);
    break;
  default:
    break;
  }
}



template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (int n,
					   const Shape<DIM>* const* s1,
					   const Shape<DIM>* const* s2,
					   const Shape<DIM>* const* s3,	     
					   const Shape<DIM>* const* s4,	     
					   const Shape<DIM>* const* s5,	     
					   const Shape<DIM>* const* s6,	     
					   OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1,
					   Array<T, DIM, TRAIT>* arg2,
					   Array<T, DIM, TRAIT>* arg3,
					   Array<T, DIM, TRAIT>* arg4,
					   Array<T, DIM, TRAIT>* arg5,
					   bool* mask)
    
{
  assert (this->level() > 0);
       
  // (1) allocate / compute the children
  // responsibility to de-allocate is taken by the newly created 
  // MemMapping and HTA respectively
  switch (DIM) {
  case 3:  
  case 4:
    hmap3D_(n, s1, s2, s3, s4, s5, s6, op, level, arg1, arg2, arg3, arg4, arg5, mask);
    break;
  default:
    break;
  }
}
 
template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap (int n,
					   const Shape<DIM>* const* s1,
					   const Shape<DIM>* const* s2,
					   const Shape<DIM>* const* s3,	     
					   const Shape<DIM>* const* s4,	     
					   const Shape<DIM>* const* s5,	     
					   const Shape<DIM>* const* s6,
					   const Shape<DIM>* const* s7,
					   const Shape<DIM>* const* s8,	     
					   OP& op,
					   size_t level,
					   Array<T, DIM, TRAIT>* arg1,
					   Array<T, DIM, TRAIT>* arg2,
					   Array<T, DIM, TRAIT>* arg3,
					   Array<T, DIM, TRAIT>* arg4,
					   Array<T, DIM, TRAIT>* arg5,
					   Array<T, DIM, TRAIT>* arg6,
					   Array<T, DIM, TRAIT>* arg7,
					   bool* mask)    
{
  assert (this->level() > 0);
       
  // (1) allocate / compute the children
  // responsibility to de-allocate is taken by the newly created 
  // MemMapping and HTA respectively
  switch (DIM) {
  case 3:  
  case 4:
    hmap3D_(n, s1, s2, s3, s4, s5, s6, s7, s8, op, level, arg1, arg2, arg3, arg4, arg5, arg6, arg7, mask);
    break;
  default:
    break;
  }
}

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap3D_ (int n,
					      const Shape<DIM>* const* s,
					      const Shape<DIM>* const* s0,
					      OP& op,
					      size_t level,
					      Array<T, DIM, TRAIT>* arg0,
					      bool* mask,
					      int shape_off)
{ 
  if (n == 0) {
    this->hmap (op,
		level,
		arg0);
  } else {
    const Shape<DIM>& shape = s[0]->shapeAt (shape_off);
    const Shape<DIM>& arg0_s = s0[0]->shapeAt (shape_off);
      
    const Tuple<DIM>& low = shape.low();
    const Tuple<DIM>& size = shape.size();
    const Tuple<DIM>& step_shape =  shape.linearStep();    
    const Tuple<DIM>& step_dst_shape =  this->shape_.linearStep();    
    const Tuple<DIM>& arg0_low = arg0_s.low();
      
    int low_0 = low[0];
    int size_0 = size[0];
    int step_dst_0 = step_dst_shape[0];
    int step_shape_0 = step_shape[0];
    int arg0_off_0 = arg0_low[0] - low_0;
      
    int low_1 = low[1];
    int size_1 = size[1];
    int step_dst_1 = step_dst_shape[1];
    int step_shape_1 = step_shape[1];    
    int arg0_off_1 = arg0_low[1] - low_1;

    int low_2 = low[2];
    int size_2 = size[2];
    int step_dst_2 = step_dst_shape[2];
    int step_shape_2 = step_shape[2];
    int arg0_off_2 = arg0_low[2] - low_2;

    if (level != this->level())
      {      
	for (int k = low_0; k < size_0 + low_0; ++k)  
	  for (int j = low_1; j < size_1 + low_1; ++j)
	    for (int i = low_2; i < size_2 + low_2; ++i) 
	      {	    
		int offset = i * step_dst_2 + j * step_dst_1 + k * step_dst_0;				
		typename TRAIT::TypeTileImpl* tile = tileAt(offset);
		if ((!mask || mask[offset]) && tile->isLocal()) 
		  {
		    int offset_arg0 = (i + arg0_off_2) * step_dst_2 + (j + arg0_off_1) * step_dst_1 + (k + arg0_off_0) * step_dst_0;

		    int shape_off = i * step_shape_2 + j * step_shape_1 + k * step_shape_0;

		    tile->hmap3D_ (n-1,
				   s+1,
				   s0+1,
				   op,
				   level,
				   arg0->tileAt(offset_arg0),
				   NULL,
				   shape_off);	    
		  }
	      }
      }else {	
	this->tileAt(shape)->hmap (op,
				   level,
				   arg0->tileAt(arg0_s));
      } 
  }

}

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap3D_ (int n,
					      const Shape<DIM>* const* s,
					      const Shape<DIM>* const* s0,
					      const Shape<DIM>* const* s1,
					      OP& op,
					      size_t level,
					      Array<T, DIM, TRAIT>* arg0,
					      Array<T, DIM, TRAIT>* arg1,
					      bool* mask,
					      int shape_off )
{ 
  if (n == 0) {
    this->hmap (op,
		level,
		arg0,
		arg1);
  } else {
    const Shape<DIM>& shape = s[0]->shapeAt (shape_off);
    const Shape<DIM>& arg0_s = s0[0]->shapeAt (shape_off);
    const Shape<DIM>& arg1_s = s1[0]->shapeAt (shape_off);
      
    const Tuple<DIM>& low = shape.low();
    const Tuple<DIM>& size = shape.size();
    const Tuple<DIM>& step_shape =  shape.linearStep();    
    const Tuple<DIM>& step_dst_shape =  this->shape_.linearStep();    
    const Tuple<DIM>& arg0_low = arg0_s.low();
    const Tuple<DIM>& arg1_low = arg1_s.low();
      
    int low_0 = low[0];
    int size_0 = size[0];
    int step_dst_0 = step_dst_shape[0];
    int step_shape_0 = step_shape[0];
    int arg0_off_0 = arg0_low[0] - low_0;
    int arg1_off_0 = arg1_low[0] - low_0;
      
    int low_1 = low[1];
    int size_1 = size[1];
    int step_dst_1 = step_dst_shape[1];
    int step_shape_1 = step_shape[1];    
    int arg0_off_1 = arg0_low[1] - low_1;
    int arg1_off_1 = arg1_low[1] - low_1;

    int low_2 = low[2];
    int size_2 = size[2];
    int step_dst_2 = step_dst_shape[2];
    int step_shape_2 = step_shape[2];
    int arg0_off_2 = arg0_low[2] - low_2;
    int arg1_off_2 = arg1_low[2] - low_2;

    if (level != this->level())
      {      
	for (int k = low_0; k < size_0 + low_0; ++k)  
	  for (int j = low_1; j < size_1 + low_1; ++j)
	    for (int i = low_2; i < size_2 + low_2; ++i) 
	      {	    
		int offset = i * step_dst_2 + j * step_dst_1 + k * step_dst_0;				
		typename TRAIT::TypeTileImpl* tile = tileAt(offset);
		if ((!mask || mask[offset]) && tile->isLocal()) 
		  {
		    int offset_arg0 = (i + arg0_off_2) * step_dst_2 + (j + arg0_off_1) * step_dst_1 + (k + arg0_off_0) * step_dst_0;
		    int offset_arg1 = (i + arg1_off_2) * step_dst_2 + (j + arg1_off_1) * step_dst_1 + (k + arg1_off_0) * step_dst_0;

		    int shape_off = i * step_shape_2 + j * step_shape_1 + k * step_shape_0;

		    tile->hmap3D_ (n-1,
				   s+1,
				   s0+1,
				   s1+1,
				   op,
				   level,
				   arg0->tileAt(offset_arg0),
				   arg1->tileAt(offset_arg1),
				   NULL,
				   shape_off);	    
		  }
	      }
      }else {	
	this->tileAt(shape)->hmap (op,
				   level,
				   arg0->tileAt(arg0_s),
				   arg1->tileAt(arg1_s));
      } 
  }

}


template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap3D_ (int n,
					      const Shape<DIM>* const* s,
					      const Shape<DIM>* const* s0,
					      const Shape<DIM>* const* s1,
					      const Shape<DIM>* const* s2,		
					      OP& op,
					      size_t level,
					      Array<T, DIM, TRAIT>* arg0,
					      Array<T, DIM, TRAIT>* arg1,
					      Array<T, DIM, TRAIT>* arg2,
					      bool* mask,
					      int shape_off )
{ 
  if (n == 0) {
    this->hmap (op,
		level,
		arg0,
		arg1,
		arg2);
  } else {
    const Shape<DIM>& shape = s[0]->shapeAt (shape_off);
    const Shape<DIM>& arg0_s = s0[0]->shapeAt (shape_off);
    const Shape<DIM>& arg1_s = s1[0]->shapeAt (shape_off);
    const Shape<DIM>& arg2_s = s2[0]->shapeAt (shape_off);
      
    const Tuple<DIM>& low = shape.low();
    const Tuple<DIM>& size = shape.size();
    const Tuple<DIM>& step_shape =  shape.linearStep();    
    const Tuple<DIM>& step_dst_shape =  this->shape_.linearStep();    
    const Tuple<DIM>& arg0_low = arg0_s.low();
    const Tuple<DIM>& arg1_low = arg1_s.low();
    const Tuple<DIM>& arg2_low = arg2_s.low();
      
    int low_0 = low[0];
    int size_0 = size[0];
    int step_dst_0 = step_dst_shape[0];
    int step_shape_0 = step_shape[0];
    int arg0_off_0 = arg0_low[0] - low_0;
    int arg1_off_0 = arg1_low[0] - low_0;
    int arg2_off_0 = arg2_low[0] - low_0;
      
    int low_1 = low[1];
    int size_1 = size[1];
    int step_dst_1 = step_dst_shape[1];
    int step_shape_1 = step_shape[1];    
    int arg0_off_1 = arg0_low[1] - low_1;
    int arg1_off_1 = arg1_low[1] - low_1;
    int arg2_off_1 = arg2_low[1] - low_1;

    int low_2 = low[2];
    int size_2 = size[2];
    int step_dst_2 = step_dst_shape[2];
    int step_shape_2 = step_shape[2];
    int arg0_off_2 = arg0_low[2] - low_2;
    int arg1_off_2 = arg1_low[2] - low_2;
    int arg2_off_2 = arg2_low[2] - low_2;

    if (level != this->level())
      {      
	for (int k = low_0; k < size_0 + low_0; ++k)  
	  for (int j = low_1; j < size_1 + low_1; ++j)
	    for (int i = low_2; i < size_2 + low_2; ++i) 
	      {	    
		int offset = i * step_dst_2 + j * step_dst_1 + k * step_dst_0;				
		typename TRAIT::TypeTileImpl* tile = tileAt(offset);
		if ((!mask || mask[offset]) && tile->isLocal()) 
		  {
		    int offset_arg0 = (i + arg0_off_2) * step_dst_2 + (j + arg0_off_1) * step_dst_1 + (k + arg0_off_0) * step_dst_0;
		    int offset_arg1 = (i + arg1_off_2) * step_dst_2 + (j + arg1_off_1) * step_dst_1 + (k + arg1_off_0) * step_dst_0;
		    int offset_arg2 = (i + arg2_off_2) * step_dst_2 + (j + arg2_off_1) * step_dst_1 + (k + arg2_off_0) * step_dst_0;

		    int shape_off = i * step_shape_2 + j * step_shape_1 + k * step_shape_0;

		    tile->hmap3D_ (n-1,
				   s+1,
				   s0+1,
				   s1+1,
				   s2+1,
				   op,
				   level,
				   arg0->tileAt(offset_arg0),
				   arg1->tileAt(offset_arg1),
				   arg2->tileAt(offset_arg2),
				   NULL,
				   shape_off);	    
		  }
	      }
      }else {	
	this->tileAt(shape)->hmap (op,
				   level,
				   arg0->tileAt(arg0_s),
				   arg1->tileAt(arg1_s),
				   arg2->tileAt(arg2_s));
      } 
  }

}

template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap3D_ (int n,
					      const Shape<DIM>* const* s,
					      const Shape<DIM>* const* s0,
					      const Shape<DIM>* const* s1,
					      const Shape<DIM>* const* s2,		
					      const Shape<DIM>* const* s3,		
					      OP& op,
					      size_t level,
					      Array<T, DIM, TRAIT>* arg0,
					      Array<T, DIM, TRAIT>* arg1,
					      Array<T, DIM, TRAIT>* arg2,
					      Array<T, DIM, TRAIT>* arg3,
					      bool* mask,
					      int shape_off )
{ 
  if (n == 0) {
    this->hmap (op,
		level,
		arg0,
		arg1,
		arg2,
		arg3);
  } else {
    const Shape<DIM>& shape = s[0]->shapeAt (shape_off);
    const Shape<DIM>& arg0_s = s0[0]->shapeAt (shape_off);
    const Shape<DIM>& arg1_s = s1[0]->shapeAt (shape_off);
    const Shape<DIM>& arg2_s = s2[0]->shapeAt (shape_off);
    const Shape<DIM>& arg3_s = s3[0]->shapeAt (shape_off);
      
    const Tuple<DIM>& low = shape.low();
    const Tuple<DIM>& size = shape.size();
    const Tuple<DIM>& step_shape =  shape.linearStep();    
    const Tuple<DIM>& step_dst_shape =  this->shape_.linearStep();    
    const Tuple<DIM>& arg0_low = arg0_s.low();
    const Tuple<DIM>& arg1_low = arg1_s.low();
    const Tuple<DIM>& arg2_low = arg2_s.low();
    const Tuple<DIM>& arg3_low = arg3_s.low();
      
    int low_0 = low[0];
    int size_0 = size[0];
    int step_dst_0 = step_dst_shape[0];
    int step_shape_0 = step_shape[0];
    int arg0_off_0 = arg0_low[0] - low_0;
    int arg1_off_0 = arg1_low[0] - low_0;
    int arg2_off_0 = arg2_low[0] - low_0;
    int arg3_off_0 = arg3_low[0] - low_0;
      
    int low_1 = low[1];
    int size_1 = size[1];
    int step_dst_1 = step_dst_shape[1];
    int step_shape_1 = step_shape[1];    
    int arg0_off_1 = arg0_low[1] - low_1;
    int arg1_off_1 = arg1_low[1] - low_1;
    int arg2_off_1 = arg2_low[1] - low_1;
    int arg3_off_1 = arg3_low[1] - low_1;

    int low_2 = low[2];
    int size_2 = size[2];
    int step_dst_2 = step_dst_shape[2];
    int step_shape_2 = step_shape[2];
    int arg0_off_2 = arg0_low[2] - low_2;
    int arg1_off_2 = arg1_low[2] - low_2;
    int arg2_off_2 = arg2_low[2] - low_2;
    int arg3_off_2 = arg3_low[2] - low_2;

    if (level != this->level())
      {      
	for (int k = low_0; k < size_0 + low_0; ++k)  
	  for (int j = low_1; j < size_1 + low_1; ++j)
	    for (int i = low_2; i < size_2 + low_2; ++i) 
	      {	    
		int offset = i * step_dst_2 + j * step_dst_1 + k * step_dst_0;				
		typename TRAIT::TypeTileImpl* tile = tileAt(offset);
		if ((!mask || mask[offset]) && tile->isLocal()) 
		  {
		    int offset_arg0 = (i + arg0_off_2) * step_dst_2 + (j + arg0_off_1) * step_dst_1 + (k + arg0_off_0) * step_dst_0;
		    int offset_arg1 = (i + arg1_off_2) * step_dst_2 + (j + arg1_off_1) * step_dst_1 + (k + arg1_off_0) * step_dst_0;
		    int offset_arg2 = (i + arg2_off_2) * step_dst_2 + (j + arg2_off_1) * step_dst_1 + (k + arg2_off_0) * step_dst_0;
		    int offset_arg3 = (i + arg3_off_2) * step_dst_2 + (j + arg3_off_1) * step_dst_1 + (k + arg3_off_0) * step_dst_0;

		    int shape_off = i * step_shape_2 + j * step_shape_1 + k * step_shape_0;

		    tile->hmap3D_ (n-1,
				   s+1,
				   s0+1,
				   s1+1,
				   s2+1,
				   s3+1,
				   op,
				   level,
				   arg0->tileAt(offset_arg0),
				   arg1->tileAt(offset_arg1),
				   arg2->tileAt(offset_arg2),
				   arg3->tileAt(offset_arg3),
				   NULL,
				   shape_off);	    
		  }
	      }
      }else {	
	this->tileAt(shape)->hmap (op,
				   level,
				   arg0->tileAt(arg0_s),
				   arg1->tileAt(arg1_s),
				   arg2->tileAt(arg2_s),
				   arg3->tileAt(arg3_s));
      } 
  }

}


template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap3D_ (int n,
					      const Shape<DIM>* const* s,
					      const Shape<DIM>* const* s0,
					      const Shape<DIM>* const* s1,
					      const Shape<DIM>* const* s2,		
					      const Shape<DIM>* const* s3,		
					      const Shape<DIM>* const* s4,		
					      OP& op,
					      size_t level,
					      Array<T, DIM, TRAIT>* arg0,
					      Array<T, DIM, TRAIT>* arg1,
					      Array<T, DIM, TRAIT>* arg2,
					      Array<T, DIM, TRAIT>* arg3,
					      Array<T, DIM, TRAIT>* arg4,
					      bool* mask,
					      int shape_off )
{ 
  if (n == 0) {
    this->hmap (op,
		level,
		arg0,
		arg1,
		arg2,
		arg3,
		arg4);
  } else {
    const Shape<DIM>& shape = s[0]->shapeAt (shape_off);
    const Shape<DIM>& arg0_s = s0[0]->shapeAt (shape_off);
    const Shape<DIM>& arg1_s = s1[0]->shapeAt (shape_off);
    const Shape<DIM>& arg2_s = s2[0]->shapeAt (shape_off);
    const Shape<DIM>& arg3_s = s3[0]->shapeAt (shape_off);
    const Shape<DIM>& arg4_s = s4[0]->shapeAt (shape_off);
      
    const Tuple<DIM>& low = shape.low();
    const Tuple<DIM>& size = shape.size();
    const Tuple<DIM>& step_shape =  shape.linearStep();    
    const Tuple<DIM>& step_dst_shape =  this->shape_.linearStep();    
    const Tuple<DIM>& arg0_low = arg0_s.low();
    const Tuple<DIM>& arg1_low = arg1_s.low();
    const Tuple<DIM>& arg2_low = arg2_s.low();
    const Tuple<DIM>& arg3_low = arg3_s.low();
    const Tuple<DIM>& arg4_low = arg4_s.low();
      
    int low_0 = low[0];
    int size_0 = size[0];
    int step_dst_0 = step_dst_shape[0];
    int step_shape_0 = step_shape[0];
    int arg0_off_0 = arg0_low[0] - low_0;
    int arg1_off_0 = arg1_low[0] - low_0;
    int arg2_off_0 = arg2_low[0] - low_0;
    int arg3_off_0 = arg3_low[0] - low_0;
    int arg4_off_0 = arg4_low[0] - low_0;
      
    int low_1 = low[1];
    int size_1 = size[1];
    int step_dst_1 = step_dst_shape[1];
    int step_shape_1 = step_shape[1];    
    int arg0_off_1 = arg0_low[1] - low_1;
    int arg1_off_1 = arg1_low[1] - low_1;
    int arg2_off_1 = arg2_low[1] - low_1;
    int arg3_off_1 = arg3_low[1] - low_1;
    int arg4_off_1 = arg4_low[1] - low_1;

    int low_2 = low[2];
    int size_2 = size[2];
    int step_dst_2 = step_dst_shape[2];
    int step_shape_2 = step_shape[2];
    int arg0_off_2 = arg0_low[2] - low_2;
    int arg1_off_2 = arg1_low[2] - low_2;
    int arg2_off_2 = arg2_low[2] - low_2;
    int arg3_off_2 = arg3_low[2] - low_2;
    int arg4_off_2 = arg4_low[2] - low_2;

    if (level != this->level())
      {      
	for (int k = low_0; k < size_0 + low_0; ++k)  
	  for (int j = low_1; j < size_1 + low_1; ++j)
	    for (int i = low_2; i < size_2 + low_2; ++i) 
	      {	    
		int offset = i * step_dst_2 + j * step_dst_1 + k * step_dst_0;				
		typename TRAIT::TypeTileImpl* tile = tileAt(offset);
		if ((!mask || mask[offset]) && tile->isLocal()) 
		  {
		    int offset_arg0 = (i + arg0_off_2) * step_dst_2 + (j + arg0_off_1) * step_dst_1 + (k + arg0_off_0) * step_dst_0;
		    int offset_arg1 = (i + arg1_off_2) * step_dst_2 + (j + arg1_off_1) * step_dst_1 + (k + arg1_off_0) * step_dst_0;
		    int offset_arg2 = (i + arg2_off_2) * step_dst_2 + (j + arg2_off_1) * step_dst_1 + (k + arg2_off_0) * step_dst_0;
		    int offset_arg3 = (i + arg3_off_2) * step_dst_2 + (j + arg3_off_1) * step_dst_1 + (k + arg3_off_0) * step_dst_0;
		    int offset_arg4 = (i + arg4_off_2) * step_dst_2 + (j + arg4_off_1) * step_dst_1 + (k + arg4_off_0) * step_dst_0;

		    int shape_off = i * step_shape_2 + j * step_shape_1 + k * step_shape_0;

		    tile->hmap3D_ (n-1,
				   s+1,
				   s0+1,
				   s1+1,
				   s2+1,
				   s3+1,
				   s4+1,
				   op,
				   level,
				   arg0->tileAt(offset_arg0),
				   arg1->tileAt(offset_arg1),
				   arg2->tileAt(offset_arg2),
				   arg3->tileAt(offset_arg3),
				   arg4->tileAt(offset_arg4),
				   NULL,
				   shape_off);	    
		  }
	      }
      }else {	
	this->tileAt(shape)->hmap (op,
				   level,
				   arg0->tileAt(arg0_s),
				   arg1->tileAt(arg1_s),
				   arg2->tileAt(arg2_s),
				   arg3->tileAt(arg3_s),
				   arg4->tileAt(arg4_s));
      } 
  }

}


template<typename T, int DIM, typename TRAIT>
template<typename OP>

 void Array<T, DIM, TRAIT>::hmap3D_ (int n,
					      const Shape<DIM>* const* s,
					      const Shape<DIM>* const* s0,
					      const Shape<DIM>* const* s1,
					      const Shape<DIM>* const* s2,		
					      const Shape<DIM>* const* s3,		
					      const Shape<DIM>* const* s4,		
					      const Shape<DIM>* const* s5,		
					      const Shape<DIM>* const* s6,		
					      OP& op,
					      size_t level,
					      Array<T, DIM, TRAIT>* arg0,
					      Array<T, DIM, TRAIT>* arg1,
					      Array<T, DIM, TRAIT>* arg2,
					      Array<T, DIM, TRAIT>* arg3,
					      Array<T, DIM, TRAIT>* arg4,
					      Array<T, DIM, TRAIT>* arg5,
					      Array<T, DIM, TRAIT>* arg6,
					      bool* mask,
					      int shape_off )
{ 
  if (n == 0) {
    this->hmap (op,
		level,
		arg0,
		arg1,
		arg2,
		arg3,
		arg4,
		arg5,
		arg6);
  } else {
    const Shape<DIM>& shape = s[0]->shapeAt (shape_off);
    const Shape<DIM>& arg0_s = s0[0]->shapeAt (shape_off);
    const Shape<DIM>& arg1_s = s1[0]->shapeAt (shape_off);
    const Shape<DIM>& arg2_s = s2[0]->shapeAt (shape_off);
    const Shape<DIM>& arg3_s = s3[0]->shapeAt (shape_off);
    const Shape<DIM>& arg4_s = s4[0]->shapeAt (shape_off);
    const Shape<DIM>& arg5_s = s5[0]->shapeAt (shape_off);
    const Shape<DIM>& arg6_s = s6[0]->shapeAt (shape_off);
      
    const Tuple<DIM>& low = shape.low();
    const Tuple<DIM>& size = shape.size();
    const Tuple<DIM>& step_shape =  shape.linearStep();    
    const Tuple<DIM>& step_dst_shape =  this->shape_.linearStep();    
    const Tuple<DIM>& arg0_low = arg0_s.low();
    const Tuple<DIM>& arg1_low = arg1_s.low();
    const Tuple<DIM>& arg2_low = arg2_s.low();
    const Tuple<DIM>& arg3_low = arg3_s.low();
    const Tuple<DIM>& arg4_low = arg4_s.low();
    const Tuple<DIM>& arg5_low = arg5_s.low();
    const Tuple<DIM>& arg6_low = arg6_s.low();
      
    int low_0 = low[0];
    int size_0 = size[0];
    int step_dst_0 = step_dst_shape[0];
    int step_shape_0 = step_shape[0];
    int arg0_off_0 = arg0_low[0] - low_0;
    int arg1_off_0 = arg1_low[0] - low_0;
    int arg2_off_0 = arg2_low[0] - low_0;
    int arg3_off_0 = arg3_low[0] - low_0;
    int arg4_off_0 = arg4_low[0] - low_0;
    int arg5_off_0 = arg5_low[0] - low_0;
    int arg6_off_0 = arg6_low[0] - low_0;
      
    int low_1 = low[1];
    int size_1 = size[1];
    int step_dst_1 = step_dst_shape[1];
    int step_shape_1 = step_shape[1];    
    int arg0_off_1 = arg0_low[1] - low_1;
    int arg1_off_1 = arg1_low[1] - low_1;
    int arg2_off_1 = arg2_low[1] - low_1;
    int arg3_off_1 = arg3_low[1] - low_1;
    int arg4_off_1 = arg4_low[1] - low_1;
    int arg5_off_1 = arg5_low[1] - low_1;
    int arg6_off_1 = arg6_low[1] - low_1;

    int low_2 = low[2];
    int size_2 = size[2];
    int step_dst_2 = step_dst_shape[2];
    int step_shape_2 = step_shape[2];
    int arg0_off_2 = arg0_low[2] - low_2;
    int arg1_off_2 = arg1_low[2] - low_2;
    int arg2_off_2 = arg2_low[2] - low_2;
    int arg3_off_2 = arg3_low[2] - low_2;
    int arg4_off_2 = arg4_low[2] - low_2;
    int arg5_off_2 = arg5_low[2] - low_2;
    int arg6_off_2 = arg6_low[2] - low_2;

    if (level != this->level())
      {      
	for (int j=0 ; j < this->numTiles_; j++) 
	  {
	    int i = this->localTiles_[j];   	
	    typename TRAIT::TypeTileImpl* tile = this->tileAt (i);

	    if ((!mask || mask[i]) && tile->isLocal()) 
	      {
		tile->hmap3D_ (n-1,
			       s+1,
			       s0+1,
			       s1+1,
			       s2+1,
			       s3+1,
			       s4+1,
			       s5+1,
			       s6+1,
			       op,
			       level,
			       arg0->tileAt(i),
			       arg1->tileAt(i),
			       arg2->tileAt(i),
			       arg3->tileAt(i),
			       arg4->tileAt(i),
			       arg5->tileAt(i),
			       arg6->tileAt(i),
			       NULL,
			       i);	
	      }
	  }
      } else {	
	this->tileAt(shape)->hmap (op,
				   level,
				   arg0->tileAt(arg0_s),
				   arg1->tileAt(arg1_s),
				   arg2->tileAt(arg2_s),
				   arg3->tileAt(arg3_s),
				   arg4->tileAt(arg4_s),
				   arg5->tileAt(arg5_s),
				   arg6->tileAt(arg6_s));
      } 
  }
  
}

#endif /* __MAPS_H__*/
