/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: NasFT.h,v 1.9 2007/01/09 23:22:32 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __NAS_FT_H__
#define  __NAS_FT_H__

//#include <math.h>
//#include <complex.h>
#include "util/complex"
//#include <cmath>
// #include <stdlib.h>

#define FFTBLOCK	1
#define FFTBLOCKPAD	1
#define	PI	3.141592653589793238

using namespace std;


#include <iostream>

#include "MathKernel.h"
#include "debug/Tracing.h"

using namespace std;

double& real (double& x) { assert(false);}
double& imag (double& x) { assert(false);}

/** 
 * Wrapper for Intel MKL
 */
template<typename T>
class NasFT : public MathKernel<T> {
  
private:
  void init (int n) 
  {
    /**
     * GBDEBUG: Should deallocate this 
     */
    u = new T[n];
    
    int m,nu,ku,i,j,ln;
    double t, ti;
    
    nu = size_;
    m = log2(size_);
    real (u[0]) = (double)m;
    imag (u[0]) = 0.0;
    ku = 1;
    ln = 1;
    
    for (j = 1; j <= m; j++) {
      t = PI / ln;
      
      for (i = 0; i <= ln - 1; i++) {
	ti = i * t;
	real (u[i+ku]) = cos(ti);
	imag (u[i+ku]) = sin(ti);
      }
      
      ku = ku + ln;
      ln = 2 * ln;
    }   
  }
public:
  
  NasFT (int size, const int* stride) :
    size_ (size),
    stride_ (stride),
    MathKernel<T>()
  {     
    if (!u) init (size_);
  }

    virtual ~NasFT () 
  {

  }

  void dftForward (T* x) 
  {
    assert (typeid(T) == typeid(T));
    T* in =  x;
    if (stride_[1] > 1)
      {
	in = new T [size_];
	for (int i = 0; i < size_; i++)
	  in[i] =  x[i * stride_[1]];
      }

    T y[size_];
    cfftz (1, log2(size_), size_,   in,  y);
    
    int m = log2(size_);
   
    /* copy is not required always */
    copy_ (x, y, in, m);

    if (in != x)
      delete [] in;

  }
  
  void dftBackward (T* x) 
  {
    assert (typeid(T) == typeid(T));
    T* in =   x;
    if (stride_[1] > 1)
      {
	in =  new T [size_];
	for (int i = 0; i < size_; i++)
	  in[i] =  x[i * stride_[1]];
      }
    T y[size_];
    cfftz (-1, log2(size_), size_,  in,  y);    
     
    int m = log2(size_);

    /* copy is not required always */
    copy_ (x, y, in, m);

    if (in != x)
      delete [] in;

  }
  
  void copy_ (T* x, T* y, T* in, int m)
  {
    if (stride_[1] > 1) {
      if (m % 2 == 1) {
	for (int j = 0; j < size_; j++) {
	  for (int i = 0; i < FFTBLOCK; i++) {
	    x[j * stride_[1]] = y[j];
	  }
	}
      } else {
	for (int j = 0; j < size_; j++) {
	  for (int i = 0; i < FFTBLOCK; i++) {
	    x[j * stride_[1]] = in[j];
	  }
	}
      }
      //     cout << "here " << stride_[1] << endl;
    } else if (m % 2 == 1) {
      for (int j = 0; j < size_; j++) {
	for (int i = 0; i < FFTBLOCK; i++) {
	  real (x[j]) = real (y[j]);
	  imag (x[j]) = imag (y[j]);
	}
      }
      //cout << "here 2" << endl;
    }
  }

  /** 
   * Computes matrix - matrix product of a sparse matrix stored in the
   * compressed sparse column format 
   */
  static void dcscmm (char *transa, 
		      int *m, int *n, int *k, 
		      T *alpha, 
		      char *matdescra, T *val, int *indx,  int *pntrb, int *pntre, 
		      T *b, int *ldb, 
		      T *beta, 
		      T *c, int *ldc) 
  {
    assert (false);
  }

  static void dcsrmm (char *transa, 
		      int *m, int *n, int *k, 
		      T *alpha, 
		      char *matdescra, T *val, int *indx,  int *pntrb, int *pntre, 
		      T *b, int *ldb, 
		      T *beta, 
		      T *c, int *ldc) 
  {
    assert (false);
  }
  
private:

  int size_;
  const int *stride_;
public:
  static T* u ;

  inline void checkStatus_ (long status) 
  {
  }
  
/*--------------------------------------------------------------------
c-------------------------------------------------------------------*/

  void cfftz (int is, int m, int n, 
		   T* x,
		   T* y) 
{
/*--------------------------------------------------------------------
c-------------------------------------------------------------------*/

/*--------------------------------------------------------------------
c   Computes NY N-point complex-to-complex FFTs of X using an algorithm due
c   to Swarztrauber.  X is both the input and the output array, while Y is a 
c   scratch array.  It is assumed that N = 2^M.  Before calling CFFTZ to 
c   perform FFTs, the array U must be initialized by calling CFFTZ with IS 
c   set to 0 and M set to MX, where MX is the maximum value of M for any 
c   subsequent call.
c-------------------------------------------------------------------*/

    int i,j,l,mx;

/*--------------------------------------------------------------------
c   Check if input parameters are invalid.
c-------------------------------------------------------------------*/
    mx = (int)(real (u[0]));
    if ((is != 1 && is != -1) || m < 1 || m > mx) {
	printf("CFFTZ: Either U has not been initialized, or else\n"
	       "one of the input parameters is invalid%5d%5d%5d\n",
	       is, m, mx);
	exit(1);
    }

/*--------------------------------------------------------------------
c   Perform one variant of the Stockham FFT.
c-------------------------------------------------------------------*/
    for (l = 1; l <= m; l+=2) {
        fftz2 (is, l, m, n, FFTBLOCK, FFTBLOCKPAD, u, x, y);
        if (l == m) break;
	fftz2 (is, l + 1, m, n, FFTBLOCK, FFTBLOCKPAD, u, y, x);
    }

/*--------------------------------------------------------------------
c   Copy Y to X.
c-------------------------------------------------------------------*/
  //MOVED TO copy_ 
}


/*--------------------------------------------------------------------
c-------------------------------------------------------------------*/

  void fftz2 (int is, int l, int m, int n, int ny, int ny1,
	      T* u, 
	      T* x,
	      T* y) 
  {
    /*--------------------------------------------------------------------
      c-------------------------------------------------------------------*/

    /*--------------------------------------------------------------------
      c   Performs the L-th iteration of the second variant of the Stockham FFT.
      c-------------------------------------------------------------------*/
    
    int k,n1,li,lj,lk,ku,i,j,i11,i12,i21,i22;
    T u1,x11,x21;
    
    /*--------------------------------------------------------------------
      c   Set initial parameters.
      c-------------------------------------------------------------------*/
    
    n1 = n / 2;
    if (l-1 == 0) {
	lk = 1;
    } else {
	lk = 2 << ((l - 1)-1);
    }
    if (m-l == 0) {
	li = 1;
    } else {
	li = 2 << ((m - l)-1);
    }
    lj = 2 * lk;
    ku = li;

    for (i = 0; i < li; i++) {
      
        i11 = i * lk;
        i12 = i11 + n1;
        i21 = i * lj;
        i22 = i21 + lk;
        if (is >= 1) {
          real(u1) = real(u[ku+i]);
          imag(u1) = imag(u[ku+i]);
        } else {
          real (u1) = real(u[ku+i]);
          imag (u1) = -imag(u[ku+i]);
        }

/*--------------------------------------------------------------------
c   This loop is vectorizable.
c-------------------------------------------------------------------*/
        for (k = 0; k < lk; k++) {
	    for (j = 0; j < ny; j++) {
	      assert (j < 1 );
	      assert (i11+k < n);
	      double x11real, x11imag;
	      double x21real, x21imag;
	      x11real = real (x[i11+k]);
	      x11imag = imag (x[i11+k]);
	      x21real = real (x[i12+k]);
	      x21imag = imag (x[i12+k]);
	      real (y[i21+k]) = x11real + x21real;
	      imag (y[i21+k]) = x11imag + x21imag;
	      real (y[i22+k]) = real (u1) * (x11real - x21real)
		- imag (u1) * (x11imag - x21imag);
	      imag (y[i22+k]) = real (u1) * (x11imag - x21imag)
		+ imag (u1) * (x11real - x21real);
	    }
	}
    }
}

};

template<>
complex<double>* NasFT<complex<double> >::u = NULL;

template<>
double* NasFT<double>::u = NULL;


#endif /* __INTEL_MKL_H__ */
