/**
 * IBM SOFTWARE DISCLIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Overlap.h,v 1.1.4.1 2007/08/13 16:47:15 brodman2 Exp $
 * Authors: Jia Guo
 */

#ifndef __OVERLAP_H__
#define __OVERLAP_H__

#include "Tuple.h"
#include "util/RefCounted.h" 
using namespace std;

enum OMode {
  ZERO,
  PERIODIC,
  DEFINED
};


template <int DIM>
class Overlap: public RefCounted
{


public: 
  Overlap (const Tuple<DIM>& pos, const Tuple<DIM>& neg, 
	   const OMode mode, bool needUpdate= false) :
    RefCounted(),
    pos_dir_(pos), 
    neg_dir_(neg), 
    mode_(mode),
    needUpdate_(needUpdate)
  {}

  Overlap(Overlap<DIM> & ol):
    pos_dir_(ol.pos_dir_), 
    neg_dir_(ol.neg_dir_), 
    mode_(ol.mode_),
    needUpdate_(ol.needUpdate_)
  {}

  ~Overlap (){
  }

  bool needUpdate() const 
  {
    return needUpdate_;
  }

  inline void setUpdate(bool update)
  {
    needUpdate_ = update;
  }
  inline OMode mode() const
  { 
    return mode_;
  }

  inline Tuple<DIM> pos() const{
    return pos_dir_;
  }
  
  inline Tuple<DIM> neg() const{
    return neg_dir_;
  }


 

private:



  /*The overlapped size in the positive direction*/
  const Tuple<DIM> pos_dir_;
  /*The overlapped size in the negative direction */
  const Tuple<DIM> neg_dir_;
  /* the mode for overlapping 
   * 0: no wrap around
   * 1: wrap around 
   * 2: no wrap but with ghost region
   */
  const OMode mode_;
  bool needUpdate_;
};

template <int DIM>
ostream& operator << (ostream& ost, const Overlap<DIM>& o) 
{
  return ost << o.pos()  << ":" << o.neg()<< endl ;
}

#endif /* end __OVERLAP_H__ */
