/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: ReduceExpr.h,v 1.3.2.2 2007/08/14 17:58:44 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __REDUCE_EXPR_H__
#define  __REDUCE_EXPR_H__

#include <cassert>
#include <functional>
#include <iostream>

using namespace std;

/**  
 * Deferred Expression Evaluation Mechanism for 
 * Reduction Operation (reduce)
 */

template<typename OP, typename TRAIT>
class ReduceExpr {
  
public:
  
  ReduceExpr (const OP& op,
	      const typename TRAIT::Type& arg1,
	      const int& dim,
	      const unsigned& all_to_all_mask = 0x0) :
    arg1_ (arg1), 
    dim_  (dim),
    all_to_all_mask_ (all_to_all_mask),
    op_ (op)
  { }
  
  ~ReduceExpr () {}

  void eval (typename TRAIT::Type& dst) const
  {
    dst = typename TRAIT::TypeScalar (); //BBF used to be 0
    eval_ (dst);
  }
  
  // JCB - Why was this commented out?
  operator typename TRAIT::Type () const 
  {  
     typename TRAIT::Type dst = arg1_.clone();     
     dst = typename TRAIT::TypeScalar  (); //BBF used to be 0
     eval_ (dst);
     return  dst;
  } 
private:

    void eval_ (typename TRAIT::Type& dst) const
  {    
    arg1_.wrapped_->reduce_(op_,			    
			    dim_,
			    *dst.wrapped_
			    );
  }

  const OP& op_;
  const typename TRAIT::Type arg1_;
  const int dim_;
  const unsigned all_to_all_mask_;
};


#endif
