/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: TransExpr.h,v 1.2.4.2 2007/08/14 17:58:44 brodman2 Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __TRANS_EXPR_H__
#define  __TRANS_EXPR_H__

#include <cassert>
#include <functional>
#include <iostream>

using namespace std;

/**  
 * Deferred Expression Evaluation Mechanism for 
 * Reduction Operation (reduce)
 */

#define TRANSPOSE 0
#define DTRANSPOSE  1

template<typename TRAIT, int TYPE>
class TransExpr {
  
public:
  
  TransExpr (const typename TRAIT::Type& arg1,
             const Tuple<2>& trans_dims,
	     const unsigned& permute_flag) : 
    arg1_ (arg1),
    trans_dims_ (trans_dims),
    permute_flag_ (permute_flag) {}
  
  ~TransExpr () {}
  
  void eval (typename TRAIT::Type& dst) const
  {
    eval_ (*dst.wrapped_);
  }
  

  void eval (typename TRAIT::TypeImpl& dst) const
  {
    eval_ (dst);
  }
  
  operator typename TRAIT::Type () const 
  {  
     typename TRAIT::Type dst = arg1_.clone();
     eval_ (*dst.wrapped_);
     return  dst;
  } 
private:

  void eval_ (typename TRAIT::TypeImpl& dst) const
  { 
    if (TYPE == TRANSPOSE) 
      {   
	
        Tuple<TRAIT::dim> permVector;
        for (int i = 0; i < TRAIT::dim; ++i)
	  permVector[i] = i;   
	
        permVector = permVector.swap(trans_dims_[0],trans_dims_[1]); 
        arg1_.wrapped_->actual_().copy(&permVector, &dst, permute_flag_);
      } else if (TYPE == DTRANSPOSE) {
	arg1_.wrapped_->actual_().dpermute(trans_dims_, &dst);
      }
  }
  
  const typename TRAIT::Type arg1_;
  const Tuple<2> trans_dims_;
  const unsigned permute_flag_;
};

#endif
