/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Counters.h,v 1.6 2006/05/19 14:36:37 vonpraun Exp $
 * Authors: Christoph von Praun
 */

#ifndef __COUNTERS_H__
#define __COUNTERS_H__

#include <cassert>
#include <iostream>

using namespace std;

namespace Counters {

#if (HTA_DEBUG_COUNT == 1)
  static long long counters_[10];
  static bool enabled_ = true;
#endif  /* HTA_DEBUG_COUNT */
 
  inline void enable ()
  {
#if (HTA_DEBUG_COUNT == 1)
    enabled_ = true;
#endif /* HTA_DEBUG_COUNT */
  }
  
  inline void disable ()
  {
#if (HTA_DEBUG_COUNT == 1)
    enabled_ = false;
#endif /* HTA_DEBUG_COUNT */
  }

  inline void inc (int id, int i=1) 
  {

#if (HTA_DEBUG_COUNT == 1)
    assert (id >= 0 && id < 10);
    if (enabled_)
      counters_[id] += i;
#endif  /* HTA_DEBUG_COUNT */
  }

  inline long long get (int id) 
  {
#if (HTA_DEBUG_COUNT == 1)
    assert (id >= 0 && id < 10);
    return counters_[id];
#else
    return 0;
#endif  /* HTA_DEBUG_COUNT */
  }

  inline void printAll () 
  {
#if (HTA_DEBUG_COUNT == 1)
    cout << endl;
    cout << "htalib counters ----------------------------------------------------------------------- " << endl;
    for (int i = 0; i < 10; ++ i) 
      cout << "COUNTER[" << i << "] " << counters_[i] << endl;
    cout << "------------------------------------------------------------------------------------- " << endl;
#endif  /* HTA_DEBUG_COUNT */   
  }
}

#endif /* __COUNTERS_H__ */
