/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Timers.h,v 1.2 2006/05/19 18:32:10 vonpraun Exp $
 * Authors: Christoph von Praun
 */

#ifndef __TIMERS_H__
#define __TIMERS_H__

#include <cassert>
#include <iostream>

#include "util/Timer.h"

using namespace std;

namespace Timers {

#if (HTA_DEBUG_TIME == 1)
  static Timer timers_[10]; 
  static bool enabled_ = true;
#endif  /* HTA_DEBUG_TIME */
 
  inline void enable ()
  {
#if (HTA_DEBUG_TIME == 1)
    enabled_ = true;
#endif /* HTA_DEBUG_TIME */
  }
  
  inline void disable ()
  {
#if (HTA_DEBUG_TIME == 1)
    enabled_ = false;
#endif /* HTA_DEBUG_TIME */
  }

  inline void start (int i) 
  {
#if (HTA_DEBUG_TIME == 1)
    assert (i >= 0 && i < 10);
    if (enabled_)
      timers_[i].start();
#endif  /* HTA_DEBUG_TIME */
  }

  inline void stop (int i) 
  {
#if (HTA_DEBUG_TIME == 1)
    assert (i >= 0 && i < 10);
    if (enabled_)
      timers_[i].stop();
#endif  /* HTA_DEBUG_TIME */
  }

  inline void printAll () 
  {
#if (HTA_DEBUG_TIME == 1)
    cout << endl;
    cout << "htalib timers ----------------------------------------------------------------------- " << endl;
    for (int i = 0; i < 10; ++ i) 
      cout << "TIMER[" << i << "] " << timers_[i] << endl;
    cout << "------------------------------------------------------------------------------------- " << endl;
#endif  /* HTA_DEBUG_TIME */   
  }
}

#endif /* __TIMERS_H__ */
