/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Tracing.h,v 1.3 2006/05/16 14:20:02 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */


#ifndef __TRACING_H__
#define __TRACING_H__

#include <iostream>
#include "Traits.h"

#ifndef HTA_DLEVEL
#define HTA_DLEVEL -1
#endif

#ifndef HTA_WLEVEL
#define HTA_WLEVEL -1
#endif

namespace Tracing {

  static int dlevel_ = HTA_DLEVEL;
  static int wlevel_ = HTA_WLEVEL;

  static void dlevel (int l) 
  {
    dlevel_ = l;
  }

  static void wlevel (int l) 
  {
    wlevel_ = l;
  }
} 

/* debugging */
#if (HTA_DLEVEL >= 0)
#define HTA_DEBUG(L,X)    { if (L <= Tracing::dlevel_) { std::cout << "DEBG[" << L << ", " << Traits::Default::myPlace() << "]  " << X << std::endl << std::flush; } }
#else
#define HTA_DEBUG(L,X)    { ; }
#endif /* HTA_DLEVEL */


/* warnings */
#if (HTA_WLEVEL >= 0)
#define HTA_WARN(L,X)    { if (L <=  Tracing::wlevel_) { std::cout << "WARN[" << L << ", " << Traits::Default::myPlace() << "]  " << X << std::endl << std::flush; } }
#else
#define HTA_WARN(L,X)    { ; }
#endif /* HTA_WLEVEL */


/* deprecation */
#ifdef HTA_REPORT_DEPRECATION
#define HTA_DEPRECATED(X)    { std::cout << "DEPRECATED: " << X << std::endl << std::flush; }
#else
#define HTA_DEPRECATED(X)    { ; }
#endif /* HTA_REPORT_DEPRECATION */

#endif /* __TRACING_H__ */
