/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: AccumArray.h,v 1.8 2006/08/01 16:41:49 bikshand Exp $
 * Authors: Ganesh Bikshandi
 */


/* ================ DEFUNCT ======================== */

#ifndef  __ACCUM_ARRAY_H__
#define  __ACCUM_ARRAY_H__

#include <cassert>
#include "MapReduce.h"

using namespace std;

template<typename T, int DIM, typename TRAIT> class HTA;

/**
 * map-reduce operator to count the number of occurences of a integers
 * inspired from MATLAB

 * TODO: Assumes the HTA is 1D or 2D with 1 X N layout -- should generalize this 
 */
template<typename T, int DIM, typename TRAIT>
struct AccumArray : public Operator::MapReduce<T, HTA<T, DIM, TRAIT>, DIM>  
{
  AccumArray(HTA<T, DIM, TRAIT> counter) : counter_(counter) { }
  
  void init (const MapContext<T, DIM>& context) {
      this->context_ = context;
  }

  void map_none (T& val) {
    assert (DIM == 1);
    counter_(this->context_.position_)[Tuple<DIM>(val)]++;
  }
  
  virtual const HTA<T, DIM, TRAIT>& result () const 
  {
    return counter_;
  }
  
  bool doReduce () { return false;}

  Operator::MapReduceKey keyType () const { return Operator::MR_NONE; }
  
  HTA<T, DIM, TRAIT>  counter_;
  
}; // struct Counter;

#endif
