/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: SerialTrait.h,v 1.34.4.3 2007/08/23 17:03:04 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __SERIAL_TRAIT_H__
#define __SERIAL_TRAIT_H__

#include "Traits.h"

#include "tbb/task_scheduler_init.h"

#ifdef HTA_INTEL_MKL
#include "IntelMKL.h"
#else
#ifdef HTA_NAS_FT
#include "NasFT.h"
#else
#ifdef HTA_IBM_ESSL
#include "IbmESSL.h"
#else
#include "MathKernel.h"
#endif
#endif
#endif

#include "debug/Alloc.h"
#include "debug/Timers.h"
#include "debug/Counters.h"

#include "serial/HTAOImpl.h"

//#include "serial/LeafHTAImpl.h"
#include "serial/OLocalAllocator.h"
#include "serial/Serial.h"
#include "serial/SerialHelper.h"

#include "util/Dummy.h"
#include "util/Finalizable.h"

// forward declarations
template<typename T, int DIM, typename TRAIT> class HTA;

namespace Traits {
  
#if (HTA_TRAIT == HTA_TRAIT_SERIAL)
  static tbb::task_scheduler_init* TBB_init;
  
  void Default::barrier ()                     { }

  int  Default::nPlaces ()                     { return Serial::nThreads; }

  int  Default::myPlace ()                     { return Serial::myThread; }

  int  Default::nullPlace ()                   { return Serial::nullThread; } 

  int  Default::rootPlace ()                   { return Serial::myThread; } 

  void Default::sync()                         { }
  
  void Default::async()                        { }

  void Default::init (int argc, char** argv)   
  { static const char *defopts = "T:";
    int i;

    int number_of_threads = tbb::task_scheduler_init::automatic;

    while ((i = getopt(argc, argv, defopts)) != -1) {
      switch(i) {
      case 'T': number_of_threads = static_cast<int>(strtol(optarg, (char **)NULL, 10));
	break;
      default:
	/* exit(1); */ ;
      }
    }

#ifdef HTA_DEBUG_COUNT
    Counters::enable();
#endif // HTA_DEBUG_COUNT   
     TBB_init = new tbb::task_scheduler_init(number_of_threads);
  }

  void Default::finalize ()                    
  { 
    Finalizable::finalizeAll();

#ifdef HTA_DEBUG_TIME
    Timers::printAll();
#endif // HTA_DEBUG_TIME

#ifdef HTA_DEBUG_COUNT
    Counters::printAll();
#endif // HTA_DEBUG_COUNT

#ifdef HTA_DEBUG_MEM
    Alloc::undeleted();
#endif // HTA_DEBUG_MEM
    TBB_init->terminate();
    delete TBB_init;

  }
#endif

  /* serial HTA implementation */
  template <typename T, int DIM>
  struct SerialDense : public Default
  {
    static const int                                                 dim = DIM;
    static const int                                                 sparse = false;

    typedef SerialDense<T, DIM>                                      Trait;
    typedef SerialDense<T, DIM>                                      TraitSubref;
    typedef T*                                                       TypeInternal;
    
    typedef Trait                                                    TraitSparse;        // not implemented but faked with dense
    typedef TraitSubref                                              TraitSparseSubref;  // not implemented but faked with dense
    typedef Dummy*                                                   TypeSparseInternal; // dummy

    typedef Trait                                                    TraitDense;
    typedef TraitSubref                                              TraitDenseSubref;
    typedef TypeInternal                                             TypeDenseInternal;

    typedef HTA<T, DIM, Trait>                                       Type;
    typedef HTA<T, DIM, TraitSubref>                                 TypeTile;
    typedef T                                                        TypeScalar;
    typedef HTAOImpl<T, DIM, Trait>                                   TypeImpl;       //JG: changed  
    typedef HTAOImpl<T, DIM, TraitSubref>                             TypeTileImpl;   //JG: changed
    typedef OLocalAllocator<T>                                        TypeAllocator;  //JG: changed
    typedef SerialHelper<Trait>                                      TypeHelper;

#ifdef HTA_INTEL_MKL
    typedef IntelMKL<T>                                              TypeMathKernel;
#else
#ifdef HTA_NAS_FT
    typedef NasFT<T>                                                 TypeMathKernel;
#else
#ifdef HTA_IBM_ESSL
    typedef IbmESSL<T>                                               TypeMathKernel;
#else
    typedef MathKernel<T>                                            TypeMathKernel;
#endif
#endif
#endif

  };
  
}//end namespace

#endif /* __SERIAL_TRAIT_H__ */

