/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: CompSparseRow.h,v 1.6 2006/06/07 21:23:52 gabrielt Exp $
 * Authors: Christoph von Praun
 */

#ifndef __COMP_SPARSE_ROW_H__
#define __COMP_SPARSE_ROW_H__

#include <cassert>

template <typename T, int DIM>
struct CompSparseRow 
{
};

/* specialization for 2 dimensions */
template <>
struct CompSparseRow<double, 2>
{
  typedef double T;
  static const int DIM = 2;
  
  CompSparseRow () {}

  CompSparseRow(T* vals, const int* row_start, const int* col_idx) :
    rowPtr_(row_start), 
    colIdx_(col_idx), 
    values_(vals)
  {
    assert (vals);
    assert (row_start);
    assert (col_idx);
  }
  
  ~CompSparseRow() 
  {
    // do not release sparse data structures
    // we assume memory management of these structures is done outside htalib
  }

  CompSparseRow<T, DIM>& operator= (const CompSparseRow<T, DIM>& rhs) 
  {
    colIdx_ = rhs.colIdx_;
    rowPtr_ = rhs.rowPtr_;
    values_ = rhs.values_;
    return *this;
  }
  
    
  CompSparseRow<T, DIM>& operator= (T* val_) 
  {
    values_ = val_;
    return *this;
  }

  bool operator== (const CompSparseRow<T, DIM>& rhs)  const 
  {
    return rhs.values_ == this->values_;
  }

  bool operator!= (const CompSparseRow<T, DIM>& rhs)  const 
  {
    return rhs.values_ != this->values_;
  }

  bool operator!= (void* _null)  const 
  {
    return values_ != _null;
  }

  T& operator[] (int i) const {
    cout<<"make sure this is correct"<<endl;
    return values_[i];
  }

  T& operator() (int i, int j) const 
  {
    for (int t = rowPtr_[i];  t < rowPtr_[i+1]; t++)
      if (colIdx_[t] == j) 
	return values_[t];
    assert (dummy_ == 0.0);
    return dummy_;
  }

  const int* idx () 
  {
    return colIdx_;
  }
  
  const int* ptr () 
  {
    return rowPtr_;
  }

  T* values() 
  {
    return values_;
  }


private:

  const int* rowPtr_;
  const int* colIdx_;
  T* values_;
  static T dummy_;
};

template <typename T, int DIM> 
ostream& operator<< (ostream& ost, const CompSparseRow<T, DIM>& t) 
{
  ost << "CompSparseRow()";
  return ost;
};

double CompSparseRow<double, 2>::dummy_ = 0.0;


#endif /* __COMP_SPARSE_ROW_H__ */
