/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: SparseHelper.h,v 1.16 2006/06/20 14:28:38 gabrielt Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __SPARSE_HELPER_H__
#define __SPARSE_HELPER_H__

#include <cassert>

#include "ConstDistribution.h"
#include "CyclicDistribution.h"
#include "MemMapping.h"
#include "LocalMemMapping.h"
#include "LeafMemMapping.h"
#include "Order.h"
#include "Tuple.h"
#include "serial/Serial.h"
#include "util/VectorOps.h"

/**
 * template class that implements functionality common to the 
 * serial leaf and non-leaf HTAs. All methods must follow a precise signature that 
 * is mandated by the callers of this template.
 */
template<typename TRAIT>
struct SparseHelper 
{
  /**
  * Allocates contiguous memory.
  *
  * @param  tiling      the tiling of the HTA
  * @return             allocator for local memory
  */
  static typename TRAIT::TypeAllocator getAllocator_ (int linear_size, 
						      int block_size,
						      typename TRAIT::TypeScalar* data = NULL) 
  {
    return typename TRAIT::TypeAllocator(1);
  }

  /**
   * @param dist_level  ignored resp. irrelevant for the serial helper. This paranmeter
   *                    is just here for compatibility with a distributed implementation.
   */
  static MemMapping<TRAIT::dim>* getMemMapping_ (Order order, 
						 const typename Tuple<TRAIT::dim>::Seq& tiling, 
						 size_t levels, size_t dist_level,
						 const Distribution<TRAIT::dim>* dist = NULL) 
  {
    int nlevels = (order == TILE) ? 0 : levels;

    MemMapping<TRAIT::dim>* ret;
    if (levels == 0) 
      ret = new LeafMemMapping<TRAIT::dim>(0, nlevels, tiling, Tuple<TRAIT::dim>::zero);
    else 
      ret = new LocalMemMapping<TRAIT::dim>(levels, 0, nlevels, tiling, order, Tuple<TRAIT::dim>::zero);
    return ret;     
  }
  
  static const Distribution<TRAIT::dim>* getDistribution_() 
  {
    if (Traits::Default::nPlaces() == 1)
      return &ConstDistribution<TRAIT::dim>::get(Traits::Default::myPlace());
    else
      return new CyclicDistribution<TRAIT::dim>(Traits::Default::nPlaces());
  }
};

#endif /* __SPARSE_HELPER_H__ */
