/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: SparseTrait.h,v 1.19 2006/10/23 11:41:35 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __SPARSE_TRAIT_H__
#define __SPARSE_TRAIT_H__

#ifdef HTALIB_INTEL_MKL
#include "IntelMKL.h"
#else
#include "MathKernel.h"
#endif

#include "serial/SerialTrait.h"
#include "sparse/HTASparseImpl.h"
//#include "sparse/LeafHTASparseImpl.h"
#include "sparse/SparseAllocator.h"
#include "sparse/SparseHelper.h"
#include "sparse/CompSparseCol.h"
#include "sparse/CompSparseRow.h"

// forward declarations
template<typename T, int DIM, typename TRAIT> class HTA;

namespace Traits {
  
 #if (HTA_TRAIT == HTA_TRAIT_SPARSE)
  // this should be defined conditionally, ony if the default trait is UPC
  // to avoid name clas with other trais that might be used ...
  // not part of the trait struct because struct is a template ...
  void Default::barrier ()                     { }
  int  Default::nPlaces ()                     { return Serial::nThreads; }
  int  Default::myPlace ()                     { return Serial::myThread; }
  int  Default::nullPlace ()                   { return Serial::nullThread; } 
  int  Default::rootPlace ()                   { return Serial::myThread; } 
  void Default::sync()                         { }
  void Default::async()                        { }
  void Default::init (int argc, char** argv)   { }
  std::ostream& Default::getStdout()              { return std::cout; }
  void Default::finalize ()                    { }
#endif

  /* sparse HTA implementation */
  template <typename T, int DIM>
  struct SerialSparse : public Default
  {
    static const int                                                       dim = DIM;
    //static const int                                                       level = LEVEL;
    static const bool                                                      sparse = true;

    typedef SerialSparse<T, DIM>                                           Trait;
    typedef SerialSparse<T, DIM>                                           TraitSubref;
    typedef CompSparseRow<T, DIM>                                          TypeInternal;

    typedef Trait                                                          TraitSparse;
    typedef TraitSubref                                                    TraitSparseSubref;   
    typedef TypeInternal                                                   TypeSparseInternal;

    typedef SerialDense<T, DIM>                                            TraitDense;
    typedef SerialDense<T, DIM>                                            TraitDenseSubref;
    typedef T*                                                             TypeDenseInternal;

    typedef HTA<T, DIM, Trait>                                             Type;
    typedef HTA<T, DIM, TraitSubref>                                       TypeTile;
    typedef T                                                              TypeScalar;
    typedef HTASparseImpl<T, DIM, Trait>                                   TypeImpl;
    typedef HTASparseImpl<T, DIM, TraitSubref>                             TypeTileImpl;
    typedef SparseAllocator<T>                                             TypeAllocator;
    typedef SparseHelper<Trait>                                            TypeHelper;

#ifdef HTA_INTEL_MKL
    typedef IntelMKL<T>                                                    TypeMathKernel;
#else
    typedef MathKernel<T>                                                  TypeMathKernel;
#endif

  };
  
};

#endif /* __SPARSE_TRAIT_H__ */

