/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Cache.h,v 1.5 2006/05/25 15:26:06 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __CACHE_H__
#define  __CACHE_H__

#include <functional>
#include <map>

using namespace std;

template<typename V>
class HashAndCompare 
{
public:

  static int hash (const V& v)  
  { 
    return 0;
  }
  
  static bool equal (const V& arg1, const V& arg2) 
  {
    assert (false);
    return false;
  }
};

/* 
 * cache for items of type V and keys of type K
 * HCF is an object that shall provided a hash and 
 * compare functions among objects of type K.
 */
template<typename V, typename HCF = HashAndCompare<V> > 
class Cache 
{
  
protected:

  Cache (int size) : size_(size)
  {}

  virtual ~Cache ()
  {}

  /** 
   * @return   a value from the cache if key can be matched
   *           otherwise NULL.
   */
  V* take_ (const V& key) 
  {
    V* ret = NULL;
    int hval = HCF::hash(key);
    typename map<int, V*, less<int> >::iterator it = data_.find(hval);
    
    if (it != data_.end()) {
      V* tmp = (*it).second;
      if (HCF::equal(key, *tmp)) {
	data_.erase(it);
	ret = tmp;
      }
    } 
    return ret;
  }

  /**
   * @return true if cache accepted the item
   */
  bool put_ (V* val) 
  {
    assert (val);

    bool ret = false;
    if (data_.size() < size_) {
      int hval = HCF::hash(*val);
      if (data_.find(hval) == data_.end()) {
	data_[hval] = val;
	ret = true;
      }
    }
    return ret;
  }

  /**
   * @removes all elements from the cache and deletes them with *delete*
   */
  void clean_ () 
  {
    for (typename map<int, V*>::iterator it = data_.begin(); it != data_.end(); ++it) {
      V* tmp = (*it).second;
      delete tmp;
    }
    data_.clear();
  }

  /**
   * @return number of elements that are currently in the cache
   */
  size_t fill () 
  {
    return data_.size();
  }

private:
  
  const size_t size_;  
  map<int, V*> data_;

};

#endif  /* __CACHE_H__ */
