/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: DynamicPool.h,v 1.2 2006/05/19 14:36:37 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __DYNAMIC_POOL_H__
#define  __DYNAMIC_POOL_H__

#include <climits>
#include <vector>

using namespace std;

/* 
 */
template<typename V> 
class DynamicPool 
{
  
protected:

  DynamicPool () : size_(INT_MAX) 
  { }

  DynamicPool (int size) : size_(size)
  { }

  virtual ~DynamicPool () {}

  /** 
   * @return   a value from the cache if key can be matched
   *           otherwise NULL.
   */
  V* take_ ()
  {
    V* ret;
    if (data_.empty()) 
      ret = NULL;
    else {
      ret = data_.back();
      data_.pop_back();
    }
    return ret;
  }

  /**
   * @return true if cache accepted the item
   */
  bool put_ (V* val) 
  {
    assert (val);
    if (data_.size() == size_)
      return false;
    else {
      data_.push_back(val);
      return true;
    }
  }

  void finalize_ () 
  {
    // iterate over all items in the buffer and delete them
    for (typename vector<V*>::iterator it = data_.begin(); it != data_.end(); ++ it) {
      V* tmp = *it;
      delete tmp;
    }
  }
  
private:
  
  const size_t size_;  
  vector<V*> data_;

};

#endif  /* __DYNAMIC_POOL_H__ */
