/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Finalizable.h,v 1.2 2006/07/08 22:46:45 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __FINALIZABLE_H__
#define  __FINALIZABLE_H__

#include <set>

using namespace std;

class Finalizable
{

public:

  static void finalizeAll () 
  {
    for (set<Finalizable*>::iterator it = items_.begin(); it != items_.end(); ++it) {
      Finalizable* tmp = (*it);
      tmp -> finalize();
      // in the mpi version, this frees an invalid pointer ...
      // delete tmp;
    }
    items_.clear();
  }

protected:
  
  Finalizable () 
  {
    items_.insert(this);
  }

  virtual ~Finalizable () 
  {
    items_.erase(this);
  }

private:

  virtual void finalize () = 0;

  static set<Finalizable*> items_;
  
};

set<Finalizable*> Finalizable::items_;

#endif  /* __FINALIZABLE_H__ */
