  /**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Sequence.h,v 1.17 2006/06/07 13:06:38 vonpraun Exp $
 * Authors: Christoph von Praun
 */

#ifndef  __SEQUENCE_H__
#define  __SEQUENCE_H__

#include "debug/Tracing.h"

using namespace std;

/* enables convenient syntax */
template <class T>
class Sequence {
public:
  
  /* empty sequence */
  Sequence() :
    cur_(0)
  {}

  /**
   * implicit conversion from a T to a Sequence<T>
   */
  Sequence (const T& first) : cur_(1)
  {
    seq_[0] = first;
  }

  Sequence (const T& first, const T& second) : cur_(2)
  {
    seq_[0] = first; 
    seq_[1] = second;
  }


  Sequence (const T& first, const T& second, const T& third) : cur_(3)
  {
    seq_[0] = first;
    seq_[1] = second;
    seq_[2] = third;
  }
  
  Sequence (const T& first, const T& second, const T& third, const T& fourth) : cur_(4)
  {
    seq_[0] = first;
    seq_[1] = second;
    seq_[2] = third;
    seq_[3] = fourth;
  }
  
  Sequence (const T& first, const T& second, const T& third, const T& fourth, const T& fifth) : cur_(5)
  {
    seq_[0] = first;
    seq_[1] = second;
    seq_[2] = third;
    seq_[3] = fourth;
    seq_[4] = fifth;
  }
  
  bool operator== (const Sequence<T>& rhs) const
  {
    if (this == &rhs)
      return true;
    
    bool ret;
    if (cur_ == rhs.cur_) {
      ret = true;
      for (int i = 0; ret && i < cur_; ++i) 
	if (seq_[i] != rhs.seq_[i])
	  ret = false;
    } else 
      ret = false;
    return ret;
  }

  Sequence<T>& operator,(const T& t) 
  {
    assert (cur_ < MAX_SEQ_);
    seq_[cur_++] = t;
    return *this;
  }
    
  const T& operator[] (int idx) const 
  {
    assert (idx < MAX_SEQ_ && idx >= 0);
    assert (idx < cur_);
    return seq_[idx];
  }
  
  T& operator[] (int idx)
  {
    assert (idx < MAX_SEQ_ && idx >= 0);
    if (idx >= cur_)
      cur_ = idx + 1;
    return seq_[idx];
  }

  const T* raw() const 
  {
    return seq_;
  }

  int length() const
  {
    return cur_;
  }
  
  /**
   * return the trailing sequence
   * starting at idx
   */
  Sequence<T> trailing (int idx) const {
    Sequence<T> ret;
    int k = 0;
    for (int i = idx; i < length(); i++, k++) {
      ret[k] = seq_[i];
    }
    return ret;
  }
 

private:
  template<class T2>
  friend ostream& operator<< (ostream& ost, const Sequence<T2>& t);

  static const int MAX_SEQ_ = 5;
  int cur_;
  T seq_[MAX_SEQ_];

};

template<typename T>
ostream& operator << (ostream& ost, const Sequence<T>& t) 
{
  for (int i = 0; i < t.cur_; i++)
    ost << t[i] << " " ;
  return ost;
}


#endif /* __SEQUENCE_H__ */
