/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: StaticPool.h,v 1.1 2006/05/16 18:43:18 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __STATIC_POOL_H__
#define  __STATIC_POOL_H__

#include <climits>
#include <vector>

using namespace std;

/** 
 * Pool of static size - pool does manage memory itself.
 * put operations must be issued in reverse order of take.
 */
template<typename T> 
class StaticPool 
{
  
protected:

  StaticPool (int size) : size_(size), cur_(0)
  { 
    assert (size > 0);
    data_ = new T[size];
  }

  StaticPool::~StaticPool() 
  {
    delete data_;
  }
  
  /** 
   * @return   pointer to i consecutive records if possible ..
   *           NULL otherwise
   */
  T* take_ (int n = 1)
  {
    assert (n > 0);

    T* ret;
    if (size_ - cur_ >= n) {
      ret = &data_[cur_];
      cur_ += n;
    } else 
      ret = NULL;
    return ret;
  }

  /**
   * @return true if cache accepted the item
   */
  bool put_ (T* val, int n = 1) 
  {
    assert (val);
    assert (n > 0 && n < size_ - cur_); 
    assert (val == &data_[cur_ - n]);

    cur_ -= n;
    return true;
  }
  
private:
  
  const size_t size_;
  size_t cur_;
  T* data_;

};

#endif  /* __STATIC_POOL_H__ */
