/*
 * Copyright (c) 2005 IBM Corporation
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 */

/*
 * Version: $Id: Timer.h,v 1.4 2006/10/23 11:43:07 fraguela Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef __TIMER_H__
#define __TIMER_H__

#include <iostream>
#include <sys/time.h>

using namespace std;

class Timer {

  friend ostream& operator << (ostream& ost, const Timer& t);

private:

  long long start_;
  long long total_;
  long long base_sec_;
  long long base_usec_;

  /* relative current time in microsecs since object creation */
  long long currentTime_() {
    long long ret = 0;
    struct timeval tv;
    gettimeofday(&tv, 0);
    long diff = tv.tv_usec - base_usec_;
    if (diff >= 0) 
      ret = (tv.tv_sec - base_sec_) * 1000000LL + (long long) diff;
    else 
      ret = (tv.tv_sec - base_sec_ - 1) * 1000000LL + (1000000LL + (long long) diff);
    return ret;
  }
  
public:
  
  Timer() : start_(0LL), total_(0LL), base_sec_(0LL), base_usec_(0LL) {
    struct timeval tv;
    gettimeofday(&tv, 0);
    base_sec_ = tv.tv_sec;
    base_usec_ = tv.tv_usec;
  }

  void start()
  {
    struct timeval tv;
    gettimeofday(&tv, 0);
    start_ = currentTime_();
  }

  void stop()
  {
    long long duration = currentTime_() - start_;
    total_ += duration;
  }
  
  /* number of milliseconds */
  double get() const 
  {
    return total_ / 1000.0;
  }

};

ostream& operator << (ostream& ost, const Timer& t) {  
  ost << t.get() << " [ms]"; 
  return ost;
}

#endif //__TIMER_H__
