  /**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: TypeHelper.h,v 1.6 2006/08/20 21:13:19 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#ifndef  __TYPE_HELPER_H__
#define  __TYPE_HELPER_H__

#include <climits> 
#include <functional>
#include "util/complex"

template <typename T>
class TypeHelper
{
public:
  static bool isNull (const T& t) {
    return false;
  }

  static T nullVal () 
  {
    assert (false);
    return 0;
  }

  template <typename OP>
  static T identityVal (const OP& op) 
  {
    assert (false);
    return 0;
  }

};

/* specialization for double */
template<>
class TypeHelper<double>
{
public:
  static bool isNull (const double t) {
    return t == 0.0;
  } 

  static double nullVal () 
  {
    return 0.0;
  }

  template <typename OP>
  static double identityVal (const OP& op) 
  {
    if (typeid(plus<double>) == typeid(OP))
      return 0.0;
    else if (typeid(multiplies<double>) == typeid(OP))
      return 1.0;
    else if (typeid(Operator::max<double>) == typeid(OP))
      return (double) INT_MIN;
    else if (typeid(Operator::min<double>) == typeid(OP))
      return (double) INT_MAX;
    else if (typeid(minus<double>) == typeid(OP))
      return 0.0;
    else if (typeid(divides<double>) == typeid(OP))
      return 1.0;
    else {
      assert (false);
      return 0.0;
    }
  }
};

/* specialization for int */
template<>
class TypeHelper<int>
{
public:
  static bool isNull (const int t) {
    return t == 0;
  } 

  static int nullVal () 
  {
    return 0;
  }

  template <typename OP>
  static int identityVal (const OP& op) 
  {
    if (typeid(plus<int>) == typeid(OP))
      return 0;
    else if (typeid(multiplies<int>) == typeid(OP))
      return 1;
    else if (typeid(Operator::max<int>) == typeid(OP))
      return INT_MIN;
    else if (typeid(Operator::min<int>) == typeid(OP))
      return INT_MAX;
    else if (typeid(minus<int>) == typeid(OP))
      return 0;
    else if (typeid(divides<int>) == typeid(OP))
      return 1;
    else {
      assert (false);
      return 0;
    }
  }
};

/* specialization for complex<double> */
template<>
class TypeHelper<complex<double> >
{
public:
  static bool isNull (const complex<double> t) {
    return t == complex<double>(0.0, 0.0);
  } 
  
  static complex<double> nullVal () 
  {
    return complex<double>(0.0, 0.0);
  }

  template <typename OP>
  static complex<double> identityVal (const OP& op) 
  {
    if (typeid(plus<complex<double> >) == typeid(OP))
      return complex<double>(0.0, 0.0);
    else if (typeid(multiplies<complex<double> >) == typeid(OP))
      return complex<double>(1.0, 1.0);
    else if (typeid(Operator::max<complex<double> >) == typeid(OP))
      return complex<double> ((double) INT_MIN, (double) INT_MIN);
    else if (typeid(Operator::min<complex<double> >) == typeid(OP))
      return complex<double> ((double) INT_MAX, (double) INT_MAX);
    else if (typeid(minus<complex<double> >) == typeid(OP))
      return complex<double>(0.0, 0.0);
    else if (typeid(divides<complex<double> >) == typeid(OP))
      return complex<double>(1.0, 1.0);
    else {
      assert (false);
      return 0.0;
    }
  }
};

#endif /* __TYPE_HELPER_H__ */
