/*
 * Copyright (c) 2005 IBM Corporation
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 */

/*
 * Version: $Id: Test_Shape_2D.cpp,v 1.10 2006/05/11 21:41:08 vonpraun Exp $
 * Authors: Ganesh Bikshandi
 *
 * Test the ord, coord, coordLocal_ functionalities of Shape for a 2D case 
 * Also test these functionalities for a "composed" shape 
 * A composed shape is a shape that is composed from another shape and 
 * its ord is w.r.t that shape
 *
 * TODO: finish ordLocal_ and test it too
 */

#include <iostream>

#include "htalib_serial.h"

#define TESTNAME "Shape_2D"
#ifdef NDEBUG
#error Testing with NDEBUG is meaningless.
#endif

int main () {
  int n = 0;

  Sequence<Triplet> x_args(Triplet(0,3), Triplet(0,5));
  Shape<2> x(x_args);
  
  for(int i = 0; i <= 3; i++) 
    for(int j = 0; j <= 5; j++) {
      cout << "n=" << n << " x.ord(" << i << "," << j << ")=" << x.ord(Tuple<2>(i,j)) << endl;
      assert(n == x.ord(Tuple<2>(i,j)));
      n++;
    }

  Sequence<Triplet> c_args(Triplet(0), Triplet(0));
  Shape<2> c(c_args);
  Sequence<Triplet> a_args(Triplet(1), Triplet(1));
  Shape<2> a(a_args);
  Sequence<Triplet> b_args(Triplet(2), Triplet(2));
  Shape<2> b(b_args);
  
  cout << "c.card=" << c.card() << endl;
  cout << "a.card=" << a.card() << endl;
  cout << "b.card=" << b.card() << endl;

  assert(c.card() == 1);
  assert(a.card() == 4);
  assert(b.card() == 9);

  cout << c << a << b << endl;

  cout << TESTNAME << " PASSED" << endl;
}
