/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_Shape_iterator.cpp,v 1.11 2006/05/16 13:16:30 vonpraun Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#include <algorithm>
#include <iostream>

#include "htalib_serial.h"

#define TESTNAME "Shape_iterator"
#ifdef NDEBUG
#error Testing with NDEBUG is meaningless.
#endif

using namespace std;

void test_plain() {
  int n = 0;
  Sequence<Triplet> t_args(Triplet(1,4,1), Triplet(2,5,1));
  Shape<2> s(t_args);

  cout << "test_plain ------------------------------------" << endl;  

  for(Shape<2>::iterator i = s.begin(); i != s.end(); ++i, ++n) {
    Tuple<2> t = *i;
    int i_ord = i.ord();
    int s_ord = s.ord(t);
    cout << "t=" << t << "it.ord=" << i_ord << " s.ord=" << s_ord << endl;
    assert (i_ord >= 0);
    assert (s_ord >= 0);
    assert (i_ord == s_ord);
  }
  assert( n == s.card() );
}



void test_strided() {
  int n = 0;
  Sequence<Triplet> t_args (Triplet(2,8,2), Triplet(3,9,2));
  Shape<2> s(t_args);

  cout << "test_strided ------------------------------------" << endl;

  for(Shape<2>::iterator i = s.begin(); i != s.end(); ++i, ++n) {
    Tuple<2> t = *i;
    int i_ord = i.ord();
    int s_ord = s.ord(t);
    cout << "t=" << t << "it.ord=" << i_ord << " s.ord=" << s_ord << endl;
    assert (i_ord >= 0);
    assert (s_ord >= 0);
    assert (i_ord == s_ord);
  }
  assert( n == s.card() );
}

void test2() {
  int n = 0;
  Sequence<Triplet> t_args(Triplet(1,4,1), Triplet(2,5,1));
  Shape<2> s(t_args);

  for(Shape<2>::iterator i = s.begin(); i != s.end(); ++i, ++n) {
    cout << "it= " << i.print_(cout) << endl;
  }
  assert( n == s.card() );
}



int main( int argc, char** argv)
{
  test2();
  test_plain();
  test_strided();
  
  cout << TESTNAME << " PASSED" << endl;
  return 0;
}
