/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_StridedOp2.cpp,v 1.1 2006/05/19 21:49:13 vonpraun Exp $
 * Authors: Ganesh Bikshandi
 */

#include "htalib_serial.h"
#include <complex>

#define T Tuple<2>
 
typedef HTA<double, 2, 1> Grid;

#define TESTNAME "HTA_StridedOp2"

int main (int argc, char** argv) {
  
  Tuple<2>::Seq tiling = ((T(4,4), T(2,2)));

  /* GB: temporaries -- allocate it outside the function */
  Grid u = HTA<double, 2, 1>::alloc(tiling, NULL, TILE);
  Grid s = HTA<double, 2, 1>::alloc(tiling, NULL, TILE); 
  
  s = 1.0;

  for (int i = 0; i < 8; i++)
    for (int j = 0; j < 8; j++)
	assert (s[T(i, j)] != 0);
  
  /* even points */
  Triplet I(0, 3, 2);
  Triplet J(0, 3, 2);

  /* odd points */
  Triplet II(1, 3, 2);
  Triplet JJ(1, 3, 2);

  /* assign even points */
  u()((I, J)) =  (Grid) s()((I, J));

  cout << "u--------------------" << endl;
  cout << u << endl;
  cout << "--------------------" << endl;

  /* assign odd points */
  u()((II, JJ)) =  (Grid) s()((II, JJ)); 
  
  cout << "u--------------------" << endl;
  cout << u << endl;
  cout << "--------------------" << endl;

  Grid tmp0 = u()((I, J));
  Grid tmp1 = u()((II, JJ));
  
  cout << "tmp0--------------------" << endl;
  cout << tmp0 << endl;
  cout << "--------------------" << endl;

  cout << "tmp1--------------------" << endl;
  cout << tmp0 << endl;
  cout << "--------------------" << endl;

  cout << tmp0(T(0,0)).shape() << endl;
  cout << tmp1(T(0,0)).shape() << endl;

  for (int i = 0; i < 8; i++) {
    for (int j = 0; j < 8; j++) {
      if (i+j % 2 == 0)
	assert (u[T(i, j)] != 0);
    }
  }

  cout << TESTNAME << " PASSED " << endl;
}
