/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_add.cpp,v 1.23 2007/08/06 19:18:56 fraguela Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include "htalib_serial.h"

#define TESTNAME "HTA_add"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_1H;
typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_0H;

using namespace std;

void run () {

  // allocate HTA  
  // Tuple<2> tiling[] = { Tuple<2>(4,4), Tuple<2>(2,2) };
  HTA_1H h = 
    // HTA<double, 2, 1>::alloc (tiling, &alloc, ROW /* TILE */);
    HTA_1H::alloc(1, (Tuple<2>(4,4), Tuple<2>(2,2)), NULL, ROW /* TILE */);
  
  // create data set
  double* a = h.raw();
  double sum = 0;
  double expected_mx = 0;
  for(int i=0; i < 64; i++){
    a[i] = i ; //sin((double)i * theta);
    sum += a[i];
    if (a[i] > expected_mx)
      expected_mx = a[i];
  }
 
  cout << "h=" << h << endl;

  // check generic sum-reduce
  double diff = abs(h.reduce(plus<double>()) - sum);
  cout << "sum = " << sum << endl;
  cout << "reduce = " << h.reduce(plus<double>()) << endl;
  assert(diff < pow((double)10,-6));
 
  // check sum reduce on sum
  { 
    HTA_1H r = h + h;
    diff = abs(r.reduce(plus<double>()) - (2 * sum));
    assert(diff < pow((double)10,-6)); 
    cout << "SumReduce OK" << endl;
  }

  // print maximum
  { 
    double mx = h.reduce(Operator::max<double>());
    cout << "max: " << mx << " (expected: " << expected_mx << ")" << endl;
    assert(mx - expected_mx <  pow((double)10,-6));
    cout << "MaxReduce OK" << endl;
  }
  
  // check widening from scalar (right arg is scalar)
  { 
    HTA_1H r = h + 2.0;
    cout << r << endl;
    diff = abs(r.reduce(plus<double>()) - (sum + (64 * 2.0)));
    cout << "diff= " << diff << endl;
    assert(diff < pow((double)10,-4));
    cout << "Widening from scalar OK (1)" << endl;
   }

  // check widening from scalar (left arg is scalar)
  { 
    HTA_1H r = 2.0 + h;
    diff = abs(r.reduce(plus<double>()) - (sum + (64 * 2.0)));
    cout << "diff= " << diff << endl;
    assert(diff < pow((double)10,-4));
    cout << "Widening from scalar OK (2)" << endl;
  }
  
  // check widening from LeafHTA (right arg smaller than left arg)
  {
   
    // allocate HTA
    HTA_0H h_sub = 
      HTA_0H::alloc(Tuple<2>(4,4));
    
    // initialize array
    double* a_sub = h_sub.raw();
    for(int i=0; i < 16; i++) {
      a_sub[i] = 1;
    }
    

    HTA_1H r = h + h_sub;
    diff = abs(r.reduce(plus<double>()) - (sum + 64));
    cout << "Widening from leaf-HTA OK" << endl;
  }
  
  // check if printing works
  cout << "HTA h=" << h << endl;
  
}

int main( int argc, char** argv) 
{

  Traits::Default::init(argc, argv);
  
  run();

  Traits::Default::finalize();

  cout << TESTNAME << " PASSED" << endl;
  return 0;
}
