/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_allocShell.cpp,v 1.5 2006/07/11 04:24:06 bikshand Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#include <iostream>
#include <math.h>
#include "htalib_serial.h"

#define TESTNAME "HTA_allocShell"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

#define ORDER ROW
#define TYPE double
typedef HTA<TYPE, 2> HTA_1;
typedef HTA<TYPE, 2> HTA_0;

using namespace std;

void test1()
{
  HTA_0 leaf = HTA_0::allocLocal (0,Tuple<2>(2,2));
  HTA_1 hta_shell = HTA_1::allocShell(1, Tuple<2>(2,2));
  
  // put some numbers into the leaf
  for (int i = 0; i < 2; ++i)
    for (int j = 0; j < 2; ++j)
      leaf [Tuple<2>(i,j)] = i * 2 + j;
  
  Shape<2>::iterator it = hta_shell.shape().begin();
  Shape<2>::iterator it_end = hta_shell.shape().end();
  
  for(; it != it_end; it++) 
    hta_shell.initTile(*it, leaf); 
  hta_shell.initMemMapping();

  cout << "hta_shell=" << endl << hta_shell << endl;
}

int main (int argc, char** argv)
{  

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  assert (Alloc::memcount == 0);
#endif  

  test1();

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  Alloc::undeleted ();
  assert (Alloc::memcount == 0);
#endif

  cout<<TESTNAME<<" PASSED"<<endl;
  
  return 0;
}
