/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_binexpr.cpp,v 1.4 2007/08/07 19:48:33 brodman2 Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include "htalib_serial.h"

#define TESTNAME "HTA_binexpr"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_1H;
typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_0H;

using namespace std;

int main( int argc, char** argv) {

  Traits::Default::init(argc, argv);

  HTA_1H h = 
    HTA_1H::alloc(1, (Tuple<2>(4,4), Tuple<2>(2,2)), NULL, ROW /* TILE */);

  HTA_1H r = 
    HTA_1H::alloc(1, (Tuple<2>(4,4), Tuple<2>(2,2)), NULL, ROW /* TILE */);

  HTA_1H s = 
    HTA_1H::alloc(1, (Tuple<2>(4,4), Tuple<2>(2,2)), NULL, ROW /* TILE */);

  // create data set
  double* a = h.raw();
  double sum = 0.0;
  double diff = 0.0;
  for(int i=0; i < 64; i++){
    a[i] = i ; 
    sum += a[i];
  }

  // check sum reduce on sum
  { 
    r = h + 2.0;
    diff = abs(r.reduce(plus<double>()) - (sum + 2.0 * 64));
    assert(diff < pow((double)10,-6)); 
    cout << "SumReduce OK" << endl;
  }

  // check sum reduce on sum
  { 
    s = h + r;
    diff = abs(s.reduce(plus<double>()) - (2.0 * sum + 2.0 * 64));
    assert(diff < pow((double)10,-6)); 
    cout << "SumReduce OK" << endl;
  }


  // check sum reduce on sum
  { 
    s = 1.3 * (h + r);
    diff = abs(s.reduce(plus<double>()) - 1.3 * (2.0 * sum + 2.0 * 64));
    assert(diff < pow((double)10,-6)); 
    cout << "SumReduce OK" << endl;
  }

  // check sum reduce on sum
  { 
    s = (h + r) * 1.3;
    diff = abs(s.reduce(plus<double>()) - 1.3 * (2.0 * sum + 2.0 * 64));
    assert(diff < pow((double)10,-6)); 
    cout << "SumReduce OK" << endl;
  }

  // check sum reduce on sum
  { 
    s = r + h + r;
    diff = abs(s.reduce(plus<double>()) - (3.0 * sum + 2.0 * 128));
    assert(diff < pow((double)10,-6)); 
    cout << "SumReduce OK" << endl;
  }

  Traits::Default::finalize();

  cout << TESTNAME << " PASSED" << endl;
  return 0;
}
