/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_dpermute.cpp,v 1.11.2.1 2007/08/14 17:56:04 brodman2 Exp $
 * Authors: Ganesh Bikshandi
 */


#include <cassert>
#include <cmath>
#include <iostream>

#include "htalib_serial.h"

#define TESTNAME "Test_HTA_dpermute"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2> HTA_2H;
typedef HTA<double, 2> HTA_1H;
typedef HTA<double, 2> HTA_0H;

using namespace std;

void print_hta (const HTA_2H& hta, int nx, int ny);

void test (Order o) {

  int nx = 4;
  int ny = 4;

  HTA_2H h = 
    HTA_2H::alloc(2,(Tuple<2>(2, 2), Tuple<2>(2,1), Tuple<2>(1,2)), NULL, o);

  // create data set
  double* a = h.raw();
  for(int i=0; i < 16; i++) {
    a[i] = i ; 
  }
 
  cout << "h=" << endl;
  print_hta(h, 4, 4);
  
  Tuple<2> perm(1, 0);
  HTA_2H h_perm = h.dpermute(perm);
 
  cout << "h_perm=" << endl;
  print_hta(h_perm, 4, 4);

  cout << endl << endl;
  // check if this is really a permutation
  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      assert (h[Tuple<2>(i,j)] == h_perm[Tuple<2>(j,i)]);
    }
  }
}

void print_hta (const HTA_2H& h, int nx, int ny) 
{

  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
	cout << h[Tuple<2>(i,j)] << "\t" ;
      }
      cout << endl;
  }
}

int main( int argc, char** argv) {
  Traits::Default::init(argc, argv);
    test(TILE);
  test(ROW); 
  cout << TESTNAME << " PASSED" << endl;
  Traits::Default::finalize();
  return 0;
}


/* For a tiled array, the outoput should be 
h=
        0       1       2
        3       4       5
        6       7       8
        9       10      11
        12      13      14
        
        15      16      17
        18      19      20
        21      22      23
        24      25      26
        27      28      29

h_perm=
        0       3       6       9       12
        1       4       7       10      13
        2       5       8       11      14
        
        15      18      21      24      27
        16      19      22      25      28
        17      20      23      26      29


for row-major arrays, the output should be

h=
        0       1       2
        6       7       8
        12      13      14
        18      19      20
        24      25      26
        
        3       4       5
        9       10      11
        15      16      17
        21      22      23
        27      28      29
        

h_perm=
        0       6       12      18      24
        1       7       13      19      25
        2       8       14      20      26
        
        3       9       15      21      27
        4       10      16      22      28
        5       11      17      23      29
*/
