/** Copyright (c) 2007 Basilio B. Fraguela Universidade da Coruna */

#include <iostream>

#include "htalib_serial.h"

#define TESTNAME "HTA_easySyntax"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef Tuple<1> T1;
typedef Tuple<2> T2;

typedef int Type;

typedef HTA<Type, 1> H1;
typedef HTA<Type, 2> H2;

static void test1()
{
  
  H1 h1 = H1::alloc(1,  (T1(4), T1(4)), NULL, ROW);
  H1 g1 = H1::alloc(1,  (T1(4), T1(4)), NULL, ROW);
  H1 r1 = H1::alloc(1,  (T1(4), T1(4)), NULL, ROW);

  Type * ph1 = h1.raw();
  Type * pg1 = g1.raw();
  Type * pr1 = r1.raw();

  for(int i=0; i < 16; ++i) {
    ph1[i] = i;
    pg1[i] = i + 10;
    pr1[i] = 0;
  }

  cout << "Tile 1: " << h1(1); //Used to be h1(Tuple<1>(1))
  cout << "sixth value: " << h1[6] << endl; //Used to be h1[Tuple<1>(6)]
  assert(h1[6] == 6);
  cout << "tenth value: " << h1[10] << endl;
  assert(h1[10] == 10);
  cout << "Tiles {1:2} : " << h1(Range(1,2));
  cout << "Data {1:2}(0:1)" << h1(Range(1,2))(Range(0,1));
  
  H1 mtile = h1(0);
  
  cout << mtile[Range(1,2)];
}

/* Output:
Tile 1: 4 5 6 7 
sixth value: 6
tenth value: 10
Tiles {1:2} : 
        4 5 6 7 

        8 9 10 11 

Data {1:2}(0:1)
        4 5 

        8 9 

1 2
 */

static void test2()
{

  H2 h2 = H2::alloc(1, (T2(2, 2), T2(2, 2)), NULL, ROW);
  H2 g2 = H2::alloc(1, (T2(2, 2), T2(2, 2)), NULL, ROW);
  H2 r2 = H2::alloc(1, (T2(2, 2), T2(2, 2)), NULL, ROW);

  Type * ph2 = h2.raw();
  Type * pg2 = g2.raw();
  Type * pr2 = r2.raw();

  for(int i=0; i < 16; ++i) {
    ph2[i] = i;
    pg2[i] = i + 10;
    pr2[i] = 0;
  }

  cout << "Tile (1,1):\n" << h2(1,1);          //Used to be h2(Tuple<2>(1,1))
  cout << "Elem [1][1]: " << h2[1][1] << endl; //Used to be h2[Tuple<2>(1,1)]
  cout << "Elem [3][3]: " << h2[3][3] << endl;
  cout << "Tiles {0,0:1}: " << h2(0, Range(0,1)); //Used to be h2((Triplet(0,0), Triplet(0,1)))
  cout << " **Notice the bug when printing: the second column is not the one expected!\n";
  cout << "Tiles {1,0:1}(:,1): " << h2(1, Range(0,1))(Range(0,1), 1); //Used to be h2((Triplet(1,1), Triplet(0,1)))((Triplet(0,1), Triplet(1,1)))
  h2(1, Range(0,1))(Range(0,1), 1) = 0;
  cout << " **but when we fill it with zeros, the right colum is zeroed:\n";
  cout << h2 << endl;
  
  H2 mtile = h2(0,0);
  cout << mtile[0][Range(0,1)] << endl;
  cout << mtile[Range(0,1)][1];
}

/* Output:
   Tile (1,1):
10      11      
14      15      
Elem [1][1]: 5
Elem [3][3]: 15
Tiles {0,0:1}: 
0       1       |       2       3       
4       5       |       6       7       
 **Notice the bug when printing: the second column is not the one expected!
Tiles {1,0:1}(:,1): 
9       |       10      
13      |       14      
 **but when we fill it with zeros, the right colum is zeroed:

0       1       |       2       3       
4       5       |       6       7       
----------------------------------------------------------------
8       0       |       10      0       
12      0       |       14      0       

 Notice that without the printing bug, h1{1,0:1}(:,1) should have printed
9       |       11      
13      |       15

0       1

1
5
 */

int main(int argc, char **argv)
{
  Traits::Default::init(argc, argv);

  cout << "---------------------Test with 1D HTAs\n";

  test1();

  cout << "---------------------Test with 2D HTAs\n";

  test2();

  Traits::Default::finalize();

  cout << endl;

  cout << TESTNAME << " PASSED" << endl;

  return 0;
}
