/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_flatten2.cpp,v 1.5 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include "htalib_serial.h"

#define TESTNAME "HTA_flatten2"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_1H;

using namespace std;

void run(Order ord) {

  HTA_1H h = 
    HTA_1H::alloc(1, (Tuple<2>(4,4), Tuple<2>(2,2)), NULL, ord);

  // initialize contiguously
  double* r = h.raw();
  for (int i=0; i < 64; ++i) 
    r[i] = i;

  cout << "h=" << h << endl;

  // then initialize according to row-major walk-through
  for(int i=0; i < 8; i++) {
    for(int j=0; j < 8; j++) {
      
      h [Tuple<2>(i, j)] = i * 8 + j ; 
    }
  }
  cout << "h=" << h << endl;
}

int main( int argc, char** argv) 
{
  cout << "tiled ----------------------------------------" << endl;
  run(TILE);
  cout << "row ----------------------------------------" << endl;
  run(ROW);
  cout << TESTNAME << " PASSED" << endl;
  return 0;
}


/* expected output:

tiled ----------------------------------------
h=
        0       1       2       3
        4       5       6       7
        8       9       10      11
        12      13      14      15
        
        16      17      18      19
        20      21      22      23
        24      25      26      27
        28      29      30      31
        
        32      33      34      35
        36      37      38      39
        40      41      42      43
        44      45      46      47
        
        48      49      50      51
        52      53      54      55
        56      57      58      59
        60      61      62      63
        

h=
        0       1       2       3
        8       9       10      11
        16      17      18      19
        24      25      26      27
        
        4       5       6       7
        12      13      14      15
        20      21      22      23
        28      29      30      31
        
        32      33      34      35
        40      41      42      43
        48      49      50      51
        56      57      58      59
        
        36      37      38      39
        44      45      46      47
        52      53      54      55
        60      61      62      63
        

HTA_flatten2 PASSED
row ----------------------------------------
h=
        0       1       2       3
        8       9       10      11
        16      17      18      19
        24      25      26      27
        
        4       5       6       7
        12      13      14      15
        20      21      22      23
        28      29      30      31
        
        32      33      34      35
        40      41      42      43
        48      49      50      51
        56      57      58      59
        
        36      37      38      39
        44      45      46      47
        52      53      54      55
        60      61      62      63
        

h=
        0       1       2       3
        8       9       10      11
        16      17      18      19
        24      25      26      27
        
        4       5       6       7
        12      13      14      15
        20      21      22      23
        28      29      30      31
        
        32      33      34      35
        40      41      42      43
        48      49      50      51
        56      57      58      59
        
        36      37      38      39
        44      45      46      47
        52      53      54      55
        60      61      62      63
        

HTA_flatten2 PASSED
HTA_flatten2 PASSED
*/

