/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_flatten_multilevel2.cpp,v 1.6 2006/07/11 04:24:06 bikshand Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include <complex>

#include "htalib_serial.h"

#define TESTNAME "Test_HTA_flatten_multilevel2"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

#define TYPE int

typedef Tuple<2> T;
typedef HTA<TYPE, 2> HTA_2H;
typedef HTA<TYPE, 2> HTA_1H;
typedef HTA<TYPE, 2> HTA_0H;

using namespace std;


int main( int argc, char** argv) {

  int total_size = 64;

  // create data set
  HTA_2H h = HTA_2H::alloc(2, (T(4, 4), T(2, 1), T(1, 2)), NULL, ROW);
  
  // cout << "Error #1 : " << &(h(T(2,2))) << " " << &(h[T(0,0)](T(2,2))) << endl;

  int* arr = h.raw();
  for (int i = 0; i < total_size; ++ i) {
    arr[i] = i;
  }

  // check if flattening works
  int ctr = 0;
  for (int i = 0; i < 8; ++ i) {
    for (int j = 0; j < 8; ++ j) {
      assert (h[Tuple<2>(i,j)] == ctr);
      ctr ++;
    }
  }

  ctr = 0;
  // basically the whole array
  // problem: the trailling () seems to spoil the mem-mapping of the result
  HTA_2H h2 = h( (Triplet(0,0), Triplet(0,1)) ) () ();

  cout << "h2=====================================" << endl;
  cout << h2 << endl;
  cout << "=====================================" << endl;
  
  for (int i = 0; i < 4; ++ i) {
    for (int j = 0; j < 8; ++ j) {
      assert (h2 [Tuple<2>(i,j)] == ctr);
      ctr ++;
    }
  }

  ctr = 0;
  HTA_2H h3 = h () ( (Triplet(0,0), Triplet(0,0)) ) ();
  
  cout << "h3=====================================" << endl;
  cout << h3 << endl;
  cout << "=====================================" << endl;

  for (int i = 0; i < 4; ++ i) {
    for (int j = 0; j < 8; ++ j) {
      int tmp = h3 [Tuple<2>(i,j)];
      assert (tmp == ctr);
      //if (ctr > 0 && ctr % 4 == 0)
      //	ctr += 5;
      //else 
	ctr++;
    }
  }

  ctr = 32;
  HTA_2H h4 = h ((Triplet(0,0), Triplet(0,0)))  ((Triplet(1,1), Triplet(0,0))) ();
  
  cout << "h4=====================================" << endl;
  cout << h4 << endl;
  cout << "=====================================" << endl;

  for (int i = 0; i < 4; ++ i) {
    for (int j = 0; j < 4; ++ j) {
      int tmp = h4 [Tuple<2>(i,j)];
      assert (tmp == ctr);
      ctr++;
      if (ctr % 4 == 0)
	ctr += 4;
    }
  }
  

  cout << "h[0,0]=" << &(h [T(0,0)])  << endl;
  cout << "h(0,0)[0,0]=" << &(h (T(0,0))[T(0,0)])  << endl;
  cout << "h(0,0)(0,0)[0,0]=" << &(h (T(0,0))(T(0,0))[T(0,0)])  << endl;
  cout << endl;

  cout << "h[0,1]=" << &(h [T(0,1)]) << endl;
  cout << "h(0,0)[0,1]=" << &(h (T(0,0))[T(0,1)]) << endl;
  cout << "h(0,0)(0,0)[0,1]=" << &(h (T(0,0)) (T(0,0)) [T(0,1)])  << endl;
  cout << endl;

  cout << "h[2,6]=" << &(h [T(2,6)])  << endl;
  cout << "h(0,1)[2,2]=" << &(h (T(0,1)) [T(2,2)])  << endl;

  cout << TESTNAME << " Passed " << endl;
}


