/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_ft.cpp,v 1.13 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include <complex>
#include <stdlib.h>

#include "htalib_serial.h"
#include "mkl_dfti.h"

#define TESTNAME "Test_HTA_ft"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<complex<double>, 3> HTA_2H;
typedef HTA<complex<double>, 3> HTA_1H;
typedef HTA<complex<double>, 3> HTA_0H;

using namespace std;


#define nx  16 
#define ny  16 
#define nz  16
void ft_hta (complex<double> a[nx][ny][nz], complex<double> b[nx][ny][nz]);
int main( int argc, char** argv) {

  int stride[2]; 
  long status;
  complex<double> in[nx][ny][nz];
  complex<double> tmp[nx][ny][nz];
  // double in[nx][nx];
  DFTI_DESCRIPTOR_HANDLE handle;
  
  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      for (int k = 0; k < nz; k++) {
      in[i][j][k] = complex<double>( ((double) rand()) / (double) RAND_MAX, ((double) rand()) / (double) RAND_MAX);
      //in[i][j][k] = complex<double>(i, j+k);
      tmp[i][j][k] = in[i][j][k];
      }
    }
  }

 /* for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      for (int k = 0; k < nz; k++) {
      cout << in[i][j][k] << " ";
      }
      cout << endl;
    }
    cout << endl;
  }
*/ 
  cout << endl;

//  #if 0
  /* FT along dim = 0 */
  status = DftiCreateDescriptor(&handle, DFTI_DOUBLE, DFTI_COMPLEX, 1, nz);
  status = DftiCommitDescriptor(handle);
  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      status = DftiComputeForward(handle, &(in[i][j][0]));
    }
  }

  cout << endl;

  /* FT along dim = 1 */
  stride[0] = 0;
  stride[1] = &(in[0][1][0]) - &(in[0][0][0]);
  status = DftiCreateDescriptor(&handle, DFTI_DOUBLE, DFTI_COMPLEX, 1, ny);
  status = DftiSetValue(handle, DFTI_INPUT_STRIDES, stride);
  status = DftiSetValue(handle, DFTI_OUTPUT_STRIDES, stride);
  status = DftiCommitDescriptor(handle);
  for (int i = 0; i < nx; i++) {
    for (int k = 0; k < nz; k++) {
      status = DftiComputeForward(handle, &(in[i][0][k]));
    }
  }
  
//  #endif  
  /* FT along dim = 2 */
  stride[0] = 0;
  stride[1] = &(in[1][0][0]) - &(in[0][0][0]);
  status = DftiCreateDescriptor(&handle, DFTI_DOUBLE, DFTI_COMPLEX, 1, nx);
  status = DftiSetValue(handle, DFTI_INPUT_STRIDES, stride);
  status = DftiSetValue(handle, DFTI_OUTPUT_STRIDES, stride);
  status = DftiCommitDescriptor(handle);
  for (int j = 0; j < ny; j++) {
    for (int k = 0; k < nz; k++) {
      status = DftiComputeForward(handle, &(in[0][j][k]));
    }
  }
  
  ft_hta(in, tmp);

  cout << TESTNAME << " PASSED" << endl;
}

void ft_hta (complex<double> a[nx][ny][nz], complex<double> b[nx][ny][nz]) {
  
  // create data set
  HTA_2H h = HTA_2H::alloc(2, (Tuple<3>(nx/4, ny/4, nz/4), Tuple<3>(4, 4, 1), Tuple<3>(1, 1, 4)), NULL, ROW);
 
  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      for (int k = 0; k < nz; k++) {
	h[Tuple<3>(i,j,k)] = b[i][j][k];
      }
    }
  }
  

  h.ft(1, 1);
  h.ft(0, 1);
  HTA_2H h_perm = h.dpermute(Tuple<2>(0, 2));
  h_perm.ft(0, 1);
  h = h_perm.dpermute(Tuple<2>(0, 2));


  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      for (int k = 0; k < nz; k++) {
	assert (real(h[Tuple<3>(i,j,k)])- real(a[i][j][k]) < 1.0e-6);
	assert (imag(h[Tuple<3>(i,j,k)])- imag(a[i][j][k]) < 1.0e-6);
//	cout << h[Tuple<3>(i,j,k)] << " " << a[i][j][k] << endl;
      }
    }
  }
}

