/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_ft_essl.cpp,v 1.2 2006/07/27 14:13:12 vonpraun Exp $
 * Authors: Ganesh Bikshandi
 */

#define _COMPLEX_

#include <cmath>
#include <complex>
#include <iostream>

#include "essl.h"
#include "htalib_serial.h"

#define TESTNAME "Test_HTA_ft_essl"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<complex<double>, 3> HTA_2H;
typedef HTA<complex<double>, 3> HTA_1H;
typedef HTA<complex<double>, 3> HTA_0H;

using namespace std;

#define NX  4 
#define NY  4 
#define NZ  4

void ft_hta (complex<double> a[NX][NY][NZ], complex<double> b[NX][NY][NZ]);
void ft_array (complex<double> a[NX][NY][NZ]);
void print_array (const complex<double> a[NX][NY][NZ]);
void print_hta (const HTA_2H& hta);

int main( int argc, char** argv) {
  
  complex<double> in[NX][NY][NZ];
  // this Array holds the initialization data
  complex<double> init[NX][NY][NZ];

  // (0) initialize array
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
      for (int k = 0; k < NZ; k++) {
      in[i][j][k] = init[i][j][k] = complex<double>( ((double) rand()) / (double) RAND_MAX, ((double) rand()) / (double) RAND_MAX);
      }
    }
  }
  
  ft_array(in);

  cout << "END fft array" << endl;
  cout << "=============================================================================" << endl;
  cout << "START fft hta" << endl;
  
  ft_hta(in, init);

  cout << TESTNAME << " PASSED" << endl;
}

/*
int    esvscft_er(int, void *, int, int, void *, int, int, _ESVI, 
                   int, int,  float, double *, _ESVI, double *, _ESVI);
*/

void ft_array (complex<double> in[NX][NY][NZ]) 
{
  int stride[2]; 

  // (1) print initialized array
  cout << "Array before fft " << endl;
  print_array(in);

  /* FT along dim = 1 */
  stride[0] = 0;
  stride[1] = &(in[0][1][0]) - &(in[0][0][0]);
  cout << "Array fft-1 stride " << stride[1] << endl;
  
  int INIT = 0;
  int status = 0;
  int n = NY; // length of the transform (dim 1) -- not all lenghts are permissible
  int m = 1;  // number of sequences to be transformed
  int naux1 = (n > 2048) ? (int) (20000.0 + (2.28 * (double) n)) : 20000; // magic formula in ESSL
  double* AUX1 = new double[naux1]; 
  int incx1 = &(in[0][1][0]) - &(in[0][0][0]);
  int incx2 = 0;
  int incy1 = incx1;
  int incy2 = incx2;
  double* AUX2 = NULL;
  float SCALE = 1.0; // forward discrete fourier transform
  complex<double>* X = NULL;
  complex<double>* Y = NULL;

  for (int i = 0; i < NX; i++) {
    for (int k = 0; k < NZ; k++) {
      // status = DftiComputeForward(handle, &(in[i][0][k]));
      X = &(in[i][0][k]); 
      Y = X;
      // check alignment on doubleword boundary
      assert ((((unsigned) Y) & 0x7u) == 0x0);
      cout << "Array fft-1 stride=" << incx1 << " size=" << n << endl;
      dcft (1   ,     X    , incx1  ,  incx2 ,   Y ,   incy1  ,  incy2 ,   n   ,  m ,      1 ,  SCALE,  AUX1 , naux1 ,  AUX2    , 0); // init aux
      dcft (0   ,     X    , incx1  ,  incx2 ,   Y ,   incy1  ,  incy2 ,   n   ,  m ,      1 ,  SCALE,  AUX1 , naux1 ,  AUX2    , 0); // init aux
    }
  }

  // (2) print array after forward FFT on dim-1
  cout << "Array after fft-1 " << endl;
  print_array(in);
}

/** compute FFT on the HTA and compare the result to the FFT computed on the plain array */
void ft_hta (complex<double> a[NX][NY][NZ], complex<double> b[NX][NY][NZ]) 
{
  
  // create data set
  HTA_2H h = HTA_2H::alloc(2,(Tuple<3>(NX/2, NY/2, NZ/2), Tuple<3>(2, 2, 1), Tuple<3>(1, 1, 2)), NULL, ROW);
  // HTA_2H h = HTA_2H::alloc((Tuple<3>(NX, NY, NZ), Tuple<3>(1, 1, 1), Tuple<3>(1, 1, 1)), NULL, ROW);
  
  // (0) initialize
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
      for (int k = 0; k < NZ; k++) {
	h[Tuple<3>(i,j,k)] = b[i][j][k];
      }
    }
  }
	
  // (1) print the initialized hta 
  // cout << "HTA before ft " << endl;
  print_hta(h);

  h.ft(1, 1);
  
  // (2) print after fft-1
  cout << "HTA after FFT dim-1 " << endl;
  print_hta(h);
  
  //h.ft(0, 1);

  // (3) print after fft-0
  // cout << "HTA after FFT dim-0 " << endl;
  // print_hta(h);
  // -- OK till here!

  //HTA_2H h_perm = h.dpermute(Tuple<2>(2, 0));
  //h_perm.ft(0, 1);
  //h = h_perm.dpermute(Tuple<2>(2, 0));

  // (4) print after fft-2
  // cout << "HTA after FFT dim-2 " << endl;
  //print_hta(h);

  // (5) validate
  //for (int i = 0; i < NX; i++) {
  //for (int j = 0; j < NY; j++) {
  //  for (int k = 0; k < NZ; k++) {
  //assert (real(h[Tuple<3>(i,j,k)])- real(a[i][j][k]) < 1.0e-6);
  //assert (imag(h[Tuple<3>(i,j,k)])- imag(a[i][j][k]) < 1.0e-6);
  //  }
  //}
  //}
}

void print_array (const complex<double> in[NX][NY][NZ]) 
{
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
      for (int k = 0; k < NZ; k++) {
	cout << in[i][j][k] << "\t";
      }
      cout << endl;
    }
    cout << endl;
  }
}

void print_hta (const HTA_2H& h) 
{
  for (int i = 0; i < NX; i++) {
    for (int j = 0; j < NY; j++) {
      for (int k = 0; k < NZ; k++) {
	cout << h[Tuple<3>(i,j,k)] << "\t" ;
      }
      cout << endl;
    }
    cout << endl;
  }
}
