
/** Copyright (c) 2007 Basilio B. Fraguela Universidade da Coruna */

#include <iostream>
#include "htalib_serial.h"

#define TESTNAME "HTA_hmap"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef Tuple<1> T1;
typedef Tuple<2> T2;

typedef int Type;

typedef HTA<Type, 1> H1;
typedef HTA<Type, 2> H2;

template<typename T, typename H>
struct Unary {
  const T k_;
  
  Unary(const T &k = T())
    : k_(k)
    {}

  void operator() (H dest, H src) {
    dest = src + k_;
  }

};


static void test1()
{

  H1 h1 = H1::alloc(1,  (T1(4), T1(4)), NULL, ROW);
  H1 r1 = H1::alloc(1,  (T1(4), T1(4)), NULL, ROW);

  Type * ph1 = h1.raw();
  Type * pr1 = r1.raw();

  for(int i=0; i < 16; ++i) {
    ph1[i] = i;
    pr1[i] = 0;
  }

  r1.hmap(Unary<Type, H1>(1), h1, 0);

  cout << r1;

  for(int i=0; i < 16; ++i)
    assert(pr1[i] == (ph1[i] + 1));

  r1.hmap(Unary<Type, H1>(10), h1, 0);

  cout << r1;

  for(int i=0; i < 16; ++i)
    assert(pr1[i] == (ph1[i] + 10));
}

/* Should print

       1 2 3 4 

        5 6 7 8 

        9 10 11 12 

        13 14 15 16 


        10 11 12 13 

        14 15 16 17 

        18 19 20 21 

        22 23 24 25 

 */

static void test2()
{

  H2 h2 = H2::alloc(1, (T2(2, 2), T2(2, 2)), NULL, ROW);
  H2 r2 = H2::alloc(1, (T2(2, 2), T2(2, 2)), NULL, ROW);

  Type * ph2 = h2.raw();
  Type * pr2 = r2.raw();

  for(int i=0; i < 16; ++i) {
    ph2[i] = i;
    pr2[i] = 0;
  }

  r2.hmap(Unary<Type, H2>(1), h2, 0);

  cout << r2;

  for(int i=0; i < 16; ++i)
    assert(pr2[i] == (ph2[i] + 1));

}
/* Should print

1       2       |       3       4
5       6       |       7       8
----------------------------------------------------------------
9       10      |       11      12
13      14      |       15      16

 */

int main(int argc, char **argv)
{
  Traits::Default::init(argc, argv);
  
  test1();

  test2();

  Traits::Default::finalize();

  cout << TESTNAME << " PASSED" << endl;
  return 0;
}
