/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_map.cpp,v 1.4 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#include "htalib_serial.h"

#define TESTNAME "Test_HTA_map"
#ifdef NDEBUG
#warning "Testing with NDEBUG=1!"
#endif

using namespace std;

int main( int argc, char** argv) {
  
  double* arr = new double[64];
  for (int i = 0; i < 64; ++i) 
    arr[i] = (double) i;

  HTA<double, 2> h_rhs  = HTA<double, 2>::allocLocal(0,(Tuple<2>(8, 8)), arr);
  cout << "1 -----------------" << endl;
  cout << h_rhs << endl;
  cout << "1a -----------------" << endl;
  

  HTA<double, 2> h_lhs  = HTA<double, 2>::alloc (0,(Tuple<2>(4, 4)));
  cout << "2 -----------------" << endl;
  
  HTA<double, 2> h_lhs2 = HTA<double, 2>::alloc (0, (Tuple<2>(2, 2)));
  cout << "3 -----------------" << endl;



  Shape<2> s ((Triplet(0, 7, 2),Triplet(0, 7, 2)));
  cout << "4 -----------------" << endl;
  
  h_lhs = h_rhs [s];
  cout << "5 -----------------" << endl;
  cout << h_lhs << endl;
  cout << "5a -----------------" << endl;

  Shape<2> s2((Triplet(2, 5, 2),Triplet(2, 5, 2)));
  cout << "6 -----------------" << endl;

  h_lhs2 = h_rhs [s2];
  cout << "7 -----------------" << endl;
  cout << h_lhs2 << endl;
  cout << "7a -----------------" << endl;

  // this is an array full of 0s.
  HTA<double, 2> big_2  = HTA<double, 2>::alloc(0, (Tuple<2>(8, 8)));
  cout << "8 -----------------" << endl;
  // not a linear iterator, but both arrays are IteratorConformable
  // this seros out some elements of h_rhs
  h_rhs[s] = big_2 [s];
  cout << "9 -----------------" << endl;
  cout << h_rhs << endl;
  cout << "9a -----------------" << endl;

  Shape<2> s3 ((Triplet(0, 6, 2),Triplet(0, 6, 2)));
  h_rhs[s3 + Tuple<2>(0, 1)] = big_2[s3];
  cout << "10 -----------------" << endl;
  // this seros out yet more elements
  cout << h_rhs << endl;
  cout << "10a -----------------" << endl;

  cout << TESTNAME << " PASSED" << endl;
  return 0;
}


/* the output of this testcase should be:

Starting program: /home/praun/workspace_hta/htalib/testcases/serial/Test_HTA_foreach 
1 -----------------
0       1       2       3       4       5       6       7
8       9       10      11      12      13      14      15
16      17      18      19      20      21      22      23
24      25      26      27      28      29      30      31
32      33      34      35      36      37      38      39
40      41      42      43      44      45      46      47
48      49      50      51      52      53      54      55
56      57      58      59      60      61      62      63

1a -----------------
2 -----------------
3 -----------------
4 -----------------
5 -----------------
0       2       4       6
16      18      20      22
32      34      36      38
48      50      52      54

5a -----------------
6 -----------------
7 -----------------
18      20
34      36

7a -----------------
8 -----------------
1
9 -----------------
0       1       0       3       0       5       0       7
8       9       10      11      12      13      14      15
0       17      0       19      0       21      0       23
24      25      26      27      28      29      30      31
0       33      0       35      0       37      0       39
40      41      42      43      44      45      46      47
0       49      0       51      0       53      0       55
56      57      58      59      60      61      62      63

9a -----------------
1
10 -----------------
0       1       0       3       0       5       0       7
0       9       0       11      0       13      0       15
0       17      0       19      0       21      0       23
0       25      0       27      0       29      0       31
0       33      0       35      0       37      0       39
0       41      0       43      0       45      0       47
0       49      0       51      0       53      0       55
0       57      0       59      0       61      0       63

10a -----------------
Test_HTA_foreach PASSED

*/
