/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_mapReduce.cpp,v 1.5 2007/08/07 19:51:53 brodman2 Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>

#include "htalib_serial.h"

#define TESTNAME "HTA_mapReduce"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

#define TYPE int
typedef HTA<TYPE, 2> HTA_0H;
typedef HTA<TYPE, 2> HTA_1H;
using namespace std;

void init(int elems, HTA_1H h) { 
  TYPE* tmp = h.raw();
  assert (tmp != NULL);
  for(int j=0; j < elems; j++) 
    tmp[j] = j;
}

template<typename T, int N, int DIM>
struct XchgData 
{
  Tuple<DIM> key[N];
  T val[N];
};

template<typename T, int N, int DIM>
struct MaxN : public Operator::MapReduce<T, XchgData<T, N, DIM>, DIM>  
{

  MaxN() : minIdx_(0) { }
  
  void reduce (const XchgData<T, N, DIM>& d) 
  { 
    for (int i = 0; i < N; i++) 
      map_tuple(d.key[i], d.val[i]);
  }
  
  void map_tuple (const Tuple<DIM>& key, T val) 
  { 
    if (val > data_.val[minIdx_]) {
      data_.val[minIdx_] = val;
      data_.key[minIdx_] = key;
      T min = val;
      for (int i = 0; i < N; i++) {
	if (data_.val[i] < min) {
	  min = data_.val[i];
	  minIdx_ = i;
	}
      }
    }
  } // map

  Operator::MapReduceKey keyType () const { return Operator::MR_TUPLE; }

  virtual const XchgData<T, N, DIM>& result () const 
  {
    return data_;
  }

  void print() 
  {
    cout << "Maxima are: ";
    for (int i = 0; i < N; ++i) 
      cout << data_.key[i] << "=" << data_.val[i] << " ";
    cout << endl;
  }
  
private:

  XchgData<T, N, DIM> data_;
  int minIdx_;
}; // struct MaxN

void level_1() 
{
  cout << "level 1 start -----------------------------------" << endl;
  // allocate HTA

  HTA_1H a =  HTA_1H::alloc (1, (Tuple<2>(2,2), Tuple<2>(2,2)), NULL, ROW); 

  init (16, a);

  cout << "a=" << endl << a << endl;

  cout << "-----------------------------------" << endl;

  /* test of library version ... */
  MaxN<TYPE, 5, 2> op;
  a.mapReduce (op);
  op.print();

  cout << "level 1 end -----------------------------------" << endl;
}

int main(int argc, char** argv) {
  Traits::Default::init(argc, argv);
  level_1();
  Traits::Default::finalize();
  cout << TESTNAME << " PASSED" << endl;
  Alloc::undeleted();
  return 0;
}


/* correct output: 
level 1 start -----------------------------------
a=

        0       1
        4       5
        
        2       3
        6       7
        
        8       9
        12      13
        
        10      11
        14      15
        

-----------------------------------
Maxima are: (3, 2)=14 (3, 1)=13 (3, 0)=12 (3, 3)=15 (2, 3)=11 
level 1 end -----------------------------------
HTA_mapReduce PASSED
*/
