/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_map_1D.cpp,v 1.3 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#include "htalib_serial.h"

#define TESTNAME "HTA_map_1D"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

using namespace std;

const int DIM = 1;
const int N = 8;

int main( int argc, char** argv) {
  
  double* arr = new double[N];
  for (int i = 0; i < N; ++i) 
    arr[i] = (double) i;

  HTA<double, DIM> h_rhs  = HTA<double, DIM>::allocLocal(0,(Tuple<DIM>(N)), arr);
  cout << "1 -----------------" << endl;
  cout << h_rhs << endl;
  cout << "1a -----------------" << endl;

  HTA<double, DIM> h_lhs  = HTA<double, DIM>::alloc (0, (Tuple<DIM>(N/2)));
  cout << "2 -----------------" << endl;
  
  HTA<double, DIM> h_lhs2 = HTA<double, DIM>::alloc (0, (Tuple<DIM>(N/4)));
  cout << "3 -----------------" << endl;

  Shape<DIM> s ((Triplet(0, N-1, 2)));
  cout << "4 -----------------" << endl;
  h_lhs = h_rhs[(s)];
  cout << "5 -----------------" << endl;
  cout << h_lhs << endl;
  cout << "5a -----------------" << endl;

  Shape<DIM> s2((Triplet(2, N/2+1, 2)));
  cout << "6 -----------------" << endl;

  h_lhs2 = h_rhs[(s2)];
  cout << "7 -----------------" << endl;
  cout << h_lhs2 << endl;
  cout << "7a -----------------" << endl;

  // this is an array full of 0s.
  HTA<double, DIM> big_2  = HTA<double, DIM>::alloc(0, (Tuple<DIM>(N)));
  cout << "8 -----------------" << endl;
  // not a linear iterator, but both arrays are IteratorConformable
  // this seros out some elements of h_rhs
  h_rhs[s] = big_2[s];
  cout << "9 -----------------" << endl;
  cout << h_rhs << endl;
  cout << "9a -----------------" << endl;

  Shape<DIM> s3 ((Triplet(0, N-2, 2)));
  h_rhs[s3 + Tuple<DIM>(1)] = big_2[s3];
  cout << "10 -----------------" << endl;
  // this zeros out yet more elements
  cout << h_rhs << endl;
  cout << "10a -----------------" << endl;

  cout << TESTNAME << " PASSED" << endl;
  return 0;
}


/* the output of this testcase should be:

1 -----------------
0 1 2 3 4 5 6 7 

1a -----------------
2 -----------------
3 -----------------
4 -----------------
5 -----------------
0 2 4 6 

5a -----------------
6 -----------------
7 -----------------
2 4 

7a -----------------
8 -----------------
9 -----------------
0 1 0 3 0 5 0 7 

9a -----------------
10 -----------------
0 0 0 0 0 0 0 0 

10a -----------------
Test_HTA_map_1D PASSED


*/
