/*********************************************************************/
/*                     I.B.M. CONFIDENTIAL                           */
/*      (c) COPYRIGHT IBM CORP. 2005 ALL RIGHTS RESERVED.            */
/*                                                                   */
/*                LICENSED MATERIALS-PROPERTY OF IBM.                */
/*                                                                   */
/*********************************************************************/

/*
 * Version: $Id: Test_HTA_matvecmul.cpp,v 1.3.2.1 2007/08/14 19:22:29 brodman2 Exp $
 * Authors: Ganesh Bikshandi
 */

/* Description:
 *   This file serves to test HTA::flatten(Shape&);
 */

#include <iostream>
#include <cmath>

#include "htalib_serial.h"

#define TESTNAME "HTA_matvecmul"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_0H;
typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_1H;
using namespace std;


void level_0() 
{
  cout << "level 0 start -----------------------------------" << endl;
  
  // allocate HTA
  HTA_0H a =  HTA_0H::alloc (0,Tuple<2>(6,6));
  HTA_0H b =  HTA_0H::alloc (0,Tuple<2>(6,1));
  HTA_0H c =  HTA_0H::alloc (0,Tuple<2>(6,1));

  // initialize matrix and vector
  double* a_x = a.raw();
  double* b_x = b.raw();
  for(int i=0; i < 36; i++) 
    a_x[i] = i;
  for(int i=0; i < 6; i++) 
    b_x[i] = i;
  
  // c is initialized with 0
  
  cout << "a=" << endl << a << endl;
  cout << "b=" << endl << b << endl;

  cout << "-----------------------------------" << endl;
  
  /* test of library version ... */
  HTA_0H::matmul (a, b, &c);
  
  cout << "c (library matmul) =" << endl << c << endl;

  cout << "level 0 end -----------------------------------" << endl;
}

void level_1() 
{
  cout << "level 1 start -----------------------------------" << endl;
  // allocate HTA

  HTA_1H a =  HTA_1H::alloc (1, (Tuple<2>(3,3), Tuple<2>(2,2)), NULL, TILE); // matrix
  HTA_1H b =  HTA_1H::alloc (1, (Tuple<2>(3,1), Tuple<2>(2,1)), NULL, TILE); // vector
  HTA_1H c =  HTA_1H::alloc (1, (Tuple<2>(3,1), Tuple<2>(2,1)), NULL, TILE); // vector

  // initialize array
  double* a_x = a.raw();
  double* b_x = b.raw();
  for(int i=0; i < 36; i++) 
    a_x[i] = i;
  for(int i=0; i < 6; i++) 
    b_x[i] = i;
  
  // c is initiaized with 0
  
  cout << "a=" << endl << a << endl;
  cout << "b=" << endl << b << endl;

  cout << "-----------------------------------" << endl;

  /* test of library version ... */
  c = HTA_1H::matmul (a, b, NULL);

  cout << "c (library matmul) =" << endl << c << endl;
  
  cout << "level 1 end -----------------------------------" << endl;
}

int main(int argc, char** argv) 
{
  Traits::Default::init(argc, argv);

  level_0();
  level_1();
  cout << TESTNAME << " PASSED" << endl;

  Traits::Default::finalize();
}
