/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_permute.cpp,v 1.8.2.1 2007/08/14 19:22:29 brodman2 Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>

#include "htalib_serial.h"

#define TESTNAME "Test_HTA_permute"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2> HTA_1H;
typedef HTA<double, 2> HTA_0H;

using namespace std;

void print_hta (const HTA_1H& hta);

void test (Order o) {

  int nx = 6;
  int ny = 5;

  HTA_1H h = 
    HTA_1H::alloc(1, (Tuple<2>(3,5), Tuple<2>(2,1)), NULL, o);

  // create data set
  double* a = h.raw();
  for(int i=0; i < 30; i++){
    a[i] = i ;
  }
 
  cout << "h=" << h << endl;
  
  // check generic sum-reduce
  Tuple<2> perm(1, 0);
  HTA_1H h_perm = h.permute(perm);
  
  cout << "h_perm=" << h_perm << endl;
  
  // check if this is really a permutation
  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
      assert (h[Tuple<2>(i,j)] == h_perm[Tuple<2>(j,i)]);
    }
  }
}

void print_hta (const HTA_1H& h) 
{
  int nx = 6;
  int ny = 5;

  for (int i = 0; i < nx; i++) {
    for (int j = 0; j < ny; j++) {
	cout << h[Tuple<2>(i,j)] << "\t" ;
      }
      cout << endl;
  }
}

int main( int argc, char** argv) {
  Traits::Default::init(argc, argv);

  test(TILE);
  test(ROW); 

  cout << TESTNAME << " PASSED" << endl;

  Traits::Default::finalize();
  return 0;
}
