/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_reduce.cpp,v 1.7 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>
#include "htalib_serial.h"

#define TESTNAME "HTA_reduce"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_1H;
typedef HTA<double, 2, Traits::SerialDense<double, 2> > HTA_0H;

using namespace std;

void reduce_0() {

  HTA_0H h = 
    HTA_0H::alloc(0, Tuple<2>(3,5));

  // create data set
  double* a = h.raw();
  for(int i=0; i < 15; i++)
    a[i] = i ; //sin((double)i * theta);
 
  cout << "h=" << endl << h << endl;
  
  // check generic sum-reduce
  HTA_0H h_red_col = h.reduce(plus<double>(), 1, 0x0);
  cout << "h_red_col=" << endl << h_red_col << endl;

  // check generic sum-reduce
  HTA_0H h_red_row = h.reduce(plus<double>(), 0, 0x0);
  cout << "h_red_row=" << endl << h_red_row << endl;
  
  // check generic sum-reduce
  HTA_0H h_red_col_all = h.reduce(plus<double>(), 1, 0x2);
  cout << "h_red_col_all=" << endl << h_red_col_all << endl;

  // check generic sum-reduce
  HTA_0H h_red_row_all = h.reduce(plus<double>(), 0, 0x1);
  cout << "h_red_row_all=" << endl << h_red_row_all << endl;

}

void reduce_1() {

  HTA_1H h = 
    HTA_1H::alloc(1, (Tuple<2>(3,5), Tuple<2>(2,1)), NULL, TILE);

  // create data set
  double* a = h.raw();
  double sum = 0;
  double expected_mx = 0;
  for(int i=0; i < 30; i++){
    a[i] = i ; //sin((double)i * theta);
    sum += a[i];
    if (a[i] > expected_mx)
      expected_mx = a[i];
  }
 
  cout << "h=" << endl << h << endl;
  
  // check generic sum-reduce
  HTA_1H h_red_col = h.reduce(plus<double>(), 1, 0x0);
  cout << "h_red_col=" << endl << h_red_col << endl;
  
  HTA_1H h_red_row = h.reduce(plus<double>(), 0, 0x0);
  cout << "h_red_row=" << endl << h_red_row << endl;
  
  HTA_1H h_red_col_all = h.reduce(plus<double>(), 1, 0x2);
  cout << "h_red_col_all=" << endl << h_red_col_all << endl;
  
  HTA_1H h_red_row_all = h.reduce(plus<double>(), 0, 0x1);
  cout << "h_red_row_all=" << endl << h_red_row_all << endl;
}


int main( int argc, char** argv) {
  reduce_0();
  reduce_1();  
  cout << TESTNAME << " PASSED" << endl;
  return 0;
}

/** output should be:
h=
0       1       2       3       4
5       6       7       8       9
10      11      12      13      14

h_red_col=
10
35
60

h_red_row=
15      18      21      24      27

Test_HTA_permute PASSED
h=

        0       1       2       3       4
        5       6       7       8       9
        10      11      12      13      14
        
        15      16      17      18      19
        20      21      22      23      24
        25      26      27      28      29
        

h_red_col=

        10
        35
        60
        
        85
        110
        135
        

h_red_row=

        75      81      87      93      99
        

Test_HTA_permute PASSED
*/
