/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_scan.cpp,v 1.3 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#include <iostream>
#include <math.h>
#include "htalib_serial.h"

#define TESTNAME "HTA_scan"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

#define ORDER ROW
#define TYPE double
typedef HTA<TYPE, 2> HTA_1;
typedef HTA<TYPE, 2> HTA_0;

using namespace std;

/*
 * This testcase check the subref and subref_multi operation with 
 * step > 1.
 */

void simple() {
  
  int OUTER_TILING = 3;
  int INNER_TILING = 4;

  Sequence<Tuple<2> > tiling(Tuple<2>(INNER_TILING, INNER_TILING), Tuple<2>(OUTER_TILING, OUTER_TILING));
  HTA_1 h = HTA_1::alloc(1, tiling, NULL, ORDER);

  cout << "h --------------------------------------" << endl;
  
  // initialize
  int n = 0;
  for (int i = 0; i < OUTER_TILING * INNER_TILING; ++ i)
    for (int j = 0; j < OUTER_TILING * INNER_TILING; ++ j) {
      h [Tuple<2>(i, j)] = n++;
    }
  
  cout << endl << h << endl ;

  cout << "h.scan --------------------------------------" << endl;

  cout << endl << h.scan()  << endl;

  cout << "h().scan --------------------------------------" << endl;

  cout << endl << h().scan() << endl;

}




int main (int argc, char** argv)
{  

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  assert (Alloc::memcount == 0);
#endif  
  
  simple();

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  Alloc::undeleted ();
  assert (Alloc::memcount == 0);
#endif

  cout<<TESTNAME<<" PASSED"<<endl;
  
  return 0;
}
