/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_subref_multi.cpp,v 1.13 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi, Christoph von Praun
 */

#include <iostream>
#include <math.h>
#include "htalib_serial.h"

#define TESTNAME "HTA_subref_multi"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

#define ORDER ROW
#define TYPE double
typedef HTA<TYPE, 2> HTA_1;
typedef HTA<TYPE, 2> HTA_0;

using namespace std;

void Test_HTA_subref_multi()
{
  TYPE epsilon = pow((double)10, -6);
  TYPE diff, sum = 0.0;
  TYPE* a = new TYPE[16];
  for(int i=0; i<16; i++) {
    a[i] = i; 
    sum += a[i]; 
  }
  
  HTA_1 h = HTA_1::allocLocal (1, (Tuple<2>(2,2), Tuple<2>(2,2)), a, ORDER); 

  assert (h(Tuple<2>(0,0))[Tuple<2>(0,0)] == a[0]);
  assert (h(Tuple<2>(0,1))[Tuple<2>(0,0)] == a[2]);
  assert (h(Tuple<2>(1,0))[Tuple<2>(0,0)] == a[8]);
  assert (h(Tuple<2>(1,1))[Tuple<2>(0,0)] == a[10]);
  
  diff = abs(h.reduce(plus<TYPE>()) - sum) ;
  assert (diff < epsilon);
  
  Sequence<Triplet> args_1 (Triplet(0,1), Triplet(0,0));
  Sequence<Triplet> args_2 (Triplet(0,1), Triplet(0,1));
  Shape<2> s1 [] = { Shape<2>(args_1), Shape<2>(args_2) };

  cout << "h=" << h << endl;
    
  Sequence<Triplet> args_3 (Triplet(0,1), Triplet(1,1));
  Sequence<Triplet> args_4 (Triplet(0,1), Triplet(0,1));

  cout << "h.tileAt (2, ({ Triplet(0,1), Triplet(0,0) }, { Triplet(0,1), Triplet(0,1) }))=" << h.tileAt (2, s1) << endl;

  Shape<2> s2 [] = { Shape<2>(args_3), Shape<2>(args_4) };
  cout << "h.tileAt (2,  ({ Triplet(0,1), Triplet(1,1) }, { Triplet(0,1), Triplet(0,1) }))=" << h.tileAt (2, s2) << endl;
  
  h.tileAt (2, s1) = h.tileAt (2, s2); 

  diff = abs (h.reduce(plus<TYPE>()) - 2 * h.tileAt(2,s2).reduce(plus<TYPE>())) ;
  assert(diff < epsilon);
}


int main (int argc, char** argv)
{  

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  assert (Alloc::memcount == 0);
#endif  

  Test_HTA_subref_multi ();

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  Alloc::undeleted ();
  assert (Alloc::memcount == 0);
#endif

  cout<<TESTNAME<<" PASSED"<<endl;
  
  return 0;
}
