/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_subref_multi3.cpp,v 1.7.2.1 2007/08/14 19:22:29 brodman2 Exp $
 * Authors: Christoph von Praun
 */

#include <iostream>
#include <math.h>
#include "htalib_serial.h"

#define TESTNAME "HTA_subref_multi3"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

#define ORDER ROW
#define TYPE double
typedef HTA<TYPE, 2> HTA_1;
typedef HTA<TYPE, 2> HTA_0;

using namespace std;

/*
 * This testcase check the subref and subref_multi operation with 
 * step > 1.
 */

void very_simple() {
  
  HTA_0 h = HTA_0::alloc(0, Tuple<2>(5,5));
  
  cout << "very simple --------------------------------------" << endl;

  // initialize
  int n = 0;
  for (int i = 0; i < 5; ++ i)
    for (int j = 0; j < 5; ++ j) {
      h [Tuple<2>(i, j)] = n++;
    }
  
  h [(Triplet(0,3,2), Triplet(0,3,2))] = -1;
  
  cout << "h=" << endl << h << endl;
  double csum = h.reduce(plus<TYPE>());
  cout << "checksum =" << csum << endl; 
  assert (csum == 272);
}


void simple() {

  HTA_1 h = HTA_1::alloc(1, (Tuple<2>(4,4), Tuple<2>(1,1)), NULL, ORDER);

  cout << "simple --------------------------------------" << endl;
  
  // initialize
  int n = 0;
  for (int i = 0; i < 4; ++ i)
    for (int j = 0; j < 4; ++ j) {
      h [Tuple<2>(i, j)] = n++;
    }
  
  h ((Triplet(0,0), Triplet(0,0))) ((Triplet(0,3,2), Triplet(0,3,2))) = -1;

  cout << "h=" << h << endl;
  double csum = h.reduce(plus<TYPE>());
  cout << "checksum =" << csum << endl;
  assert (csum == 96);
}

void run() 
{
  HTA_1 h = HTA_1::alloc(1, (Tuple<2>(6,6), Tuple<2>(2,2)), NULL, ORDER);
  
  // initialize
  int n = 0;
  for (int i = 0; i < 12; ++ i)
    for (int j = 0; j < 12; ++ j) {
      h [Tuple<2>(i, j)] = n++;
    }
      
  cout << "h=" << h << endl;
  double csum1 = h.reduce(plus<TYPE>());
  cout << "checksum =" << csum1 << endl;
  assert (csum1 == 10296);
  
  h ((Triplet(0,1), Triplet(0,1))) ((Triplet(0,5,2), Triplet(5,5))) = 
    h ((Triplet(0,1), Triplet(0,1))) ((Triplet(0,5,2), Triplet(0,0)));
  
  cout << "h=" << h << endl;
  double csum2 = h.reduce(plus<TYPE>());
  cout << "checksum =" << csum2 << endl;
  assert (csum2 == 10236);
  
  h ((Triplet(0,1), Triplet(0,1))) ((Triplet(1,4,2), Triplet(1,4,2))) = -1;

  cout << "h=" << h << endl;
  double csum3 = h.reduce(plus<TYPE>());
  cout << "checksum =" << csum3 << endl;
  assert (csum3 == 9180);
  
}


int main (int argc, char** argv)
{  
  Traits::Default::init(argc, argv);
#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  assert (Alloc::memcount == 0);
#endif  
  
  very_simple();
  simple();
  run();

#ifdef HTA_DEBUG_MEM
  cout << Alloc::memcount << endl;
  Alloc::undeleted ();
  assert (Alloc::memcount == 0);
#endif

  cout<<TESTNAME<<" PASSED"<<endl;
  Traits::Default::finalize();
  return 0;
}
