/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: stencil.cpp,v 1.14 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi
 *
 * A micro-benchmark that
 * represents 'resid' part of MG
 */

#include <cstdlib>
#include <cstring>
#include <iostream>

#include "htalib_serial.h"

using namespace std;

#define T(i,j,k) Tuple<3>(i,j,k)

#define DIMS 3
#define LEVELS 0

typedef class HTA<double, DIMS> Leaf;

int N;

void residWholeArray_newSyntax_fewerTemps ()
{ 
  Shape<3> s ((Triplet(0,N-1), Triplet(0,N-1), Triplet(0,N-1)));

  Leaf u = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  Leaf r = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));

  Shape<3> s1 ((Triplet (1,N-2), Triplet (1, N-2), Triplet(0,N-1)));
  Shape<3> s2 ((Triplet (1,N-2), Triplet (1, N-2), Triplet(1,N-2)));
  Leaf tmpu1 = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  Leaf tmpu2 = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  
  HTA<double, DIMS> tmp_u1 = tmpu1 [s1];
  tmp_u1 += u [s1 + T(0 ,-1,0)];
  tmp_u1 += u [s1 + T(0 , 1,0)];
  tmp_u1 += u [s1 + T(-1,0 ,0)];
  tmp_u1 += u [s1 + T(1 ,0 ,0)];
  
  HTA<double, DIMS> tmp_u2 = tmpu1 [s1];
  tmp_u2 += u [s1 + T(-1,-1,0)];
  tmp_u2 += u [s1 + T(-1, 1,0)];
  tmp_u2 += u [s1 + T( 1,-1,0)];
  tmp_u2 += u [s1 + T( 1, 1,0)];

  u [s2] -= u [s2] * 0.5;
  u [s2] -= ( tmpu2 [s2] + tmpu1 [s2 + T(0,0,-1)] + tmpu1 [s2 + T(0,0,1)] ) * 0.25;
  u [s2] -= ( tmpu2 [s2 + T(0,0,-1)] + tmpu2 [s2 + T(0,0,1)] ) * 0.125;  
  
  r [s2] = u [s2];
}

void residWholeArray_newSyntax ()
{ 
  Shape<3> s ((Triplet(0,N-1), Triplet(0,N-1), Triplet(0,N-1)));

  Leaf u = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  Leaf r = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));

  Shape<3> s1 ((Triplet (1,N-2), Triplet (1, N-2), Triplet(0,N-1)));
  Shape<3> s2 ((Triplet (1,N-2), Triplet (1, N-2), Triplet(1,N-2)));
  Leaf tmpu1 = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  Leaf tmpu2 = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  
  tmpu1 [s1] = u [s1 + T(0,-1,0)] + u [s1 + T(0,1,0)] + u [s1 + T(-1,0,0)] + u[s1 + T(1,0,0)];
  tmpu2 [s1] = u [s1 + T(-1,-1,0)] + u [s1 + T(-1,1,0)] + u [s1 + T(1,-1,0)] + u [s1 + T(1,1,0)];
  r [s2] = u [s2] - u [s2] *.5 + (tmpu2 [s2] + tmpu1 [s2 + T(0,0,-1)] + tmpu1 [s2 + T(0,0,1)]) * 0.25 + (tmpu2 [s2 + T(0,0,-1)] + tmpu2 [s2 + T(0,0,1)]) * 0.125;  
}

void residWholeArray ()
{ 
  Tuple<3> tiling = Tuple<3>(N,N,N);

  Leaf u = HTA<double, DIMS>::alloc (LEVELS,tiling);
  Leaf r = HTA<double, DIMS>::alloc (LEVELS,tiling);

  Sequence<Triplet> t_args1(Triplet (1,N-2), Triplet (1, N-2), Triplet(0,N-1));
  Sequence<Triplet> t_args2(Triplet (1,N-2), Triplet (1, N-2), Triplet(1,N-2));

  Shape<3> s1 (t_args1);
  Shape<3> s2 (t_args2);
  Leaf tmpu1 = HTA<double, DIMS>::alloc (LEVELS,tiling);
  Leaf tmpu2 = HTA<double, DIMS>::alloc (LEVELS,tiling);
  
  tmpu1 [s1] = u [s1 + T(0,-1,0)] + u [s1 + T(0,1,0)] + u [s1 + T(-1,0,0)] + u [s1 + T(1,0,0)];
  tmpu2 [s1] = u [s1 + T(-1,-1,0)] + u [s1 + T(-1,1,0)] + u [s1 + T(1,-1,0)] + u [s1 + T(1,1,0)];
  r [s2] = u [s2] 
    - u [s2] *.5 
    + (tmpu2 [s2] + tmpu1 [s2 + T(0,0,-1)] + tmpu1 [s2 + T(0,0,1)]) * 0.25 
    + (tmpu2 [s2 + T(0,0,-1)] + tmpu2 [s2 + T(0,0,1)]) * 0.125;  
}

void residShapeByRows ()
{ 
    
  Tuple<3> tiling = Tuple<3>(N,N,N);
  Leaf u = HTA<double, DIMS>::alloc (LEVELS,tiling);
  Leaf r = HTA<double, DIMS>::alloc (LEVELS,tiling);
  
  Tuple<3> tmp_tiling(1,1,N);
  Leaf tmpu1 = HTA<double, DIMS>::alloc (LEVELS,tmp_tiling);
  Leaf tmpu2 = HTA<double, DIMS>::alloc (LEVELS,tmp_tiling);
  
  Shape<3> tmp_shape ((Triplet (0,0), Triplet (0,0), Triplet(0,N-1)));
  Shape<3> tmp_shapein ((Triplet (0,0), Triplet (0,0), Triplet(1,N-2)));

  for (int i = 1; i < N-1; i++)
    for (int j = 1; j < N-1; j++)
    {
      Shape<3> s1 = tmp_shape + T(i, j, 0);
      
      tmpu1 [tmp_shape] = u [s1 + T(0,-1,0)] + u [s1 + T(0,1,0)] + u [s1 + T(-1,0,0)] + u [s1 + T(1,0,0)];
      tmpu2 [tmp_shape] = u [s1 + T(-1,-1,0)] + u [s1 + T(-1,1,0)] + u [s1 + T(1,-1,0)] + u [s1 + T(1,1,0)];

      Shape<3> s2 = tmp_shapein + T(i, j, 0);
      
      r [s2] = u [s2] - u [s2] *.5 + 
	      (tmpu2 [tmp_shapein] + tmpu1 [tmp_shapein + T(0,0,-1)] + tmpu1 [tmp_shapein + T(0,0,1)]) * 0.25 +
	      (tmpu2 [tmp_shapein + T(0,0,-1)] + tmpu2 [tmp_shapein + T(0,0,1)]) * 0.125;
    }

}

void residOperatorParen ()
{

  double *u1 = new double[N];
  double *u2 = new double[N];
  
  Tuple<3> tiling = Tuple<3>(N,N,N);
  Shape<3> s ((Triplet(0,N-1), Triplet(0,N-1), Triplet(0,N-1)));
  
  Leaf u = HTA<double, DIMS>::alloc (LEVELS,tiling);
  Leaf r = HTA<double, DIMS>::alloc (LEVELS,tiling);

  for (int i = 1; i < N-1; i++)
    for (int j = 1; j < N-1; j++)
      {
	for (int k = 0; k < N; k++)
	  {
	    u1 [k] = u [T(i,j-1,k)] + u [T(i,j+1,k)]
	      + u [T(i-1,j,k)] + u [T(i+1,j,k)];
	    u2 [k] = u [T(i-1,j-1,k)] + u [T(i-1,j+1,k)]
	      + u [T(i+1,j-1,k)] + u [T(i+1,j+1,k)];
	  }
	for (int k = 1; k < N-1; k++)
	  {
	    r [T(i,j,k)] =  u [T(i,j,k)] 
	      - 0.5 * u [T(i,j,k)]
	      - 0.25 * ( u2[k] + u1[k-1] + u1[k+1] )
	      - 0.125 * ( u2[k-1] + u2[k+1] );
	  }
      }
      
  delete [] u1;
  delete [] u2;

}

void residPlainArray ()
{
  
  double a[N][N][N];
  double b[N][N][N];
  double u1[N];
  double u2[N];
  memset (a, 1, sizeof(double) * (N * N * N));
  
  for (int i = 1; i < N-1; i++)
    for (int j = 1; j < N-1; j++)
      {
	for (int k = 0; k < N; k++)
	  {
	    u1[k] = a[i][j-1][k] + a[i][j+1][k]
	      + a[i-1][j][k] + a[i+1][j][k];
	    u2[k] = a[i-1][j-1][k] + a[i-1][j+1][k]
	      + a[i+1][j-1][k] + a[i+1][j+1][k];
	  }
	for (int k = 1; k < N-1; k++)
	  {
	    b[i][j][k] =  a[i][j][k] 
	      - 0.5 * a[i][j][k]
	      - 0.25 * (u2[k] + u1[k-1] + u1[k+1])
	      - 0.125 * (u2[k-1] + u2[k+1]);
	  }
      }
}


void residOperator_lconv ()
{
  
  Leaf u = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  Leaf r = HTA<double, DIMS>::alloc (LEVELS,Tuple<3>(N,N,N));
  
  double weights[] = {0.5, 0, -0.25, 0, -0.125};
  u.lconv(weights, &r);
}


#define FACTOR 20

int main (int argc, char** argv)
{ 
  int factor = FACTOR;
  Timer timer[7];

  if (argc < 2 ) {
    cout << "stencil N [FACTOR]\n applies MG resid to NxNxN tile FACTOR times. Default FACTOR=" << FACTOR << endl;
    return 0;
  }
  
  N = atoi (argv[1]);
  if (argc == 3 )
    factor = atoi (argv[2]);
  
  timer[0].start();
  //opearte on whole PrivateLeafHTA (i.e do not use flatten or operator())
  //seg-faults!
  for(int i = 0; i < factor; i++)
    residWholeArray ();
  timer[0].stop();

  timer[1].start();
  for(int i = 0; i < factor; i++)
    residShapeByRows ();
  timer[1].stop();
  
  timer[2].start();
  //uses operator () (two dispatches : HTA::operator () -> PrivateLeafHTA::Flatten())
  for(int i = 0; i < factor; i++)
    residOperatorParen ();
  timer[2].stop();
  
  timer[3].start();
  for(int i = 0; i < factor; i++)
    residPlainArray ();
  timer[3].stop();
  
  timer[4].start();
  for(int i = 0; i < factor; i++)
    residWholeArray_newSyntax();
  timer[4].stop();
  
  timer[5].start();
  for(int i = 0; i < factor; i++)
    residWholeArray_newSyntax_fewerTemps();
  timer[5].stop();
  
  timer[6].start();
  for(int i = 0; i < factor; i++)
    residOperator_lconv();
  timer[6].stop();
  
  cout << " WholeArray: " << timer[0] << endl
       << " ShapeByRows: " << timer[1] << endl
       << " PrivateLeafHTA + operator(): " << timer[2] << endl
       << " Plain Array: " << timer[3] << endl
       << " WholeArray (new syntax) " << timer[4] << endl 
       << " WholeArray (new syntax, fewer temps) " << timer[5] << endl
       << " HTA operator lconv " << timer[6] << endl;
  return 0;
}
