/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_sparse.cpp,v 1.7 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>

#include "htalib_sparse.h"

#define TESTNAME "HTA_sparse"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2> HTA_1H;
typedef HTA<double, 2> HTA_0H;
typedef HTA<double, 2, Traits::HTA_DEFAULT_TRAIT<double, 2>::TraitDense> HTA_1H_DENSE;

using namespace std;

/* 
 * this is a symmetric matrix - it should work with both 
 * column and row-compressed representations 
 */
int main (int argc, char** argv) { 
  
  int M = 4;
  int N = 4;
  int nz = 5;

  double values[nz];
  int rowind[nz];
  int colptr[M+1];

  for (int i = 0; i < nz; i++) {
    values[i] = (double) i + 1.0;
  }

  /* row indices */
  rowind[0] = 1;
  rowind[1] = 3;
  rowind[2] = 0;
  rowind[3] = 2;
  rowind[4] = 0;

  /* column ptrs */
  colptr[0] = 0;
  colptr[1] = colptr[0] + 2;
  colptr[2] = colptr[1] + 1;
  colptr[3] = colptr[2] + 1;
  colptr[4] = colptr[3] + 1;

  HTA_1H h = HTA_1H::alloc(1,(Tuple<2>(M, N), Tuple<2>(2,2)), NULL, TILE);
 
  HTA_1H_DENSE v = HTA_1H_DENSE::alloc(1,(Tuple<2>(N, 1), Tuple<2>(2, 2)), NULL, TILE);
  
  for (int i = 0; i < N * 2; ++ i)
    for (int j = 0; j < 2; ++j)
      v [Tuple<2>(i, j)] = i + 1.0;
 

  for (Shape<2>::iterator i = h.shape().begin(); i != h.shape().end(); ++i) {
    h (*i).initSparse(values, colptr, rowind);
  }


  //cout << "h -----------------------------------" << endl;
  //cout << h << endl;


  /* each tile of h looks as follows: 
   *
   */


 
  HTA_1H_DENSE result = HTA_1H::lmatmul(h,v);
  
  cout << "h -----------------------------------" << endl;
  cout << h << endl;
  
  cout << "v -----------------------------------" << endl;
  cout << v << endl;

  cout << "result -----------------------------------" << endl;
  cout << result << endl;
 
}


/* 
 * expected result:

*/

