/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 * software.  
 */

/*
 * Version: $Id: Test_HTA_sparse2.cpp,v 1.4 2006/06/07 21:29:01 gabrielt Exp $
 * Authors: Ganesh Bikshandi
 */

#include <iostream>
#include <cmath>

#include "htalib_sparse.h"

#define TESTNAME "HTA_sparse2"
#ifdef NDEBUG
#warning Testing without assertions!
#endif

typedef HTA<double, 2> HTA_1H;
typedef HTA<double, 2> HTA_0H;
typedef HTA<double, 2, Traits::HTA_DEFAULT_TRAIT<double, 2>::TraitDense> HTA_1H_DENSE;

using namespace std;

int main (int argc, char** argv) { 
  
  int M = 4;
  int N = 4;
  int nz = 6;

  double values[nz];
  int colind[nz];
  int rowptr[M+1];

  for (int i = 0; i < nz; i++) {
    values[i] = (double) i + 1.0;
  }

  /* row indices */
  colind[0] = 1;
  colind[1] = 0;
  colind[2] = 1;
  colind[3] = 1;
  colind[4] = 2;
  colind[5] = 2;

  /* column ptrs */
  rowptr[0] = 0;
  rowptr[1] = rowptr[0] + 1;
  rowptr[2] = rowptr[1] + 2;
  rowptr[3] = rowptr[2] + 2;
  rowptr[4] = rowptr[3] + 1;
  rowptr[5] = rowptr[4];

  HTA_1H h = HTA_1H::alloc(1,(Tuple<2>(M, N), Tuple<2>(2,2)), NULL, TILE);
  HTA_1H_DENSE v = HTA_1H_DENSE::alloc(1,(Tuple<2>(N, 1), Tuple<2>(2, 2)), NULL, TILE);
  
  for (int i = 0; i < N * 2; ++ i)
    for (int j = 0; j < 2; ++j)
      v [Tuple<2>(i, j)] = i + 1.0;
  
  for (Shape<2>::iterator i = h.shape().begin(); i != h.shape().end(); ++i) {
    h (*i).initSparse(values, rowptr, colind);
  }


  /* each tile of h looks as follows: 
   *
   * 0 2 0 0 
   * 1 3 4 0
   * 0 0 5 0
   * 0 0 0 6
   */
  
  HTA_1H_DENSE result = HTA_1H::lmatmul(h,v);
  
  cout << "h -----------------------------------" << endl;
  cout << h << endl;
  
  cout << "v -----------------------------------" << endl;
  cout << v << endl;

  cout << "result -----------------------------------" << endl;
  cout << result << endl;
 
}


/* expected result:
0       2       0       0       |       0       2       0       0
1       3       4       0       |       1       3       4       0
0       0       5       6       |       0       0       5       6
----------------------------------------------------------------
0       2       0       0       |       0       2       0       0
1       3       4       0       |       1       3       4       0
0       0       5       6       |       0       0       5       6

v -----------------------------------

1       |       1
2       |       2
3       |       3
4       |       4
----------------------------------------------------------------
5       |       5
6       |       6
7       |       7
8       |       8

result -----------------------------------

4       |       4
19      |       19
39      |       39
----------------------------------------------------------------
12      |       12
51      |       51
83      |       83
*/
