/**
 * IBM SOFTWARE DISCLAIMER 
 *
 * This file is part of htalib, a library for hierarchically tiled arrays.
 * Copyright (c) 2005 IBM Corporation.
 *
 * Permission to use, copy, modify and distribute this software for
 * any noncommercial purpose and without fee is hereby granted,
 * provided that this copyright and permission notice appear on all
 * copies of the software. The name of the IBM Corporation may not be
 * used in any advertising or publicity pertaining to the use of the
 * software. IBM makes no warranty or representations about the
 * suitability of the software for any purpose.  It is provided "AS
 * IS" without any express or implied warranty, including the implied
 * warranties of merchantability, fitness for a particular purpose and
 * non-infringement.  IBM shall not be liable for any direct,
 * indirect, special or consequential damages resulting from the loss
 * of use, data or projects, whether in an action of contract or tort,
 * arising out of or in connection with the use or performance of this
 */

/*
 * Version: $Id: MapReduce.h,v 1.24 2006/08/03 05:48:24 bikshand Exp $
 * Authors:Christoph von Praun
 */

#ifndef __MAP_REDUCE_H__
#define __MAP_REDUCE_H__

#include <cassert>
#include <Tuple.h>
#include <MapContext.h>

using namespace std;

namespace Operator {

  /* TODO: remove this */
  enum MapReduceKey { 
    MR_NONE,
    MR_PROC_OFF,  // processor and local offset in array
    MR_TUPLE      // coordinates at level 0
  }; 

  
  /**
   * base class for map-reduce operators 
   * subclasses should override the required methods
   */
  template<typename T, typename DATA, int DIM>
  class MapReduce
  {
    
  public:
    
    /* data type to be exchanged during the reduction */
    typedef DATA XData;

    MapReduce() {}
    virtual ~MapReduce() {}

    void reduce (const DATA& data) { } 
        
    /** 
     * Init method 
     */
    void init (const MapContext<T, DIM>& context) {}
    
    /** 
     * map functions - one of those is called by the runtime system, 
     * depending on the key type      
     */
    void map_none (const T& val) { assert (false); }
    void map_proc_off (int proc, int idx, const T& val) { assert (false); }
    void map_tuple (const Tuple<DIM>& t, const T& val) { assert (false); }

    void map_none (const T& val, const T& arg1) { assert (false); }
    void map_tuple (const Tuple<DIM>& t, const T& val,  const T& arg1) { assert (false); }
    void map_none (const T& val, const T& arg1, const T& arg2) { assert (false); }
    void map_proc_off (int proc, int idx, const T& val, const T& arg1, const T& arg2) { assert (false); }
    void map_tuple (const Tuple<DIM>& t, const T& val,  const T& arg1, const T& arg2) { assert (false); }

    void map_init (const Tuple<DIM>& t) { assert (false); }

    /**
     * The keytype determines the map function that is invoked 
     * by the runtime system.
     */
    virtual MapReduceKey keyType () const = 0;

    /**
     * @return reference to data that participates in the exchange
     */
    virtual const DATA& result () const = 0;
    
    /** 
     * Information that the runtime system might exploit 
     * for optimization.
     */
    bool isThreadSafe () { return false; }

    /** 
     * Information that the runtime system might exploit 
     * for optimization.
     */
    bool doReduce () { return true; }

  protected:
    
    /* it is the responsiblity of the derived classes to initialize this using *init* */
    MapContext<T, DIM> context_;     
  };


}
  
#endif /* __MAP_REDUCE_H__ */
